📦
8189 /agent.js.map
5134 /agent.js
2815 /node_modules/@frida/base64-js/index.js.map
1428 /node_modules/@frida/base64-js/index.js
↻ base64-js
36822 /node_modules/@frida/buffer/index.js.map
25206 /node_modules/@frida/buffer/index.js
↻ buffer
8079 /node_modules/@frida/events/events.js.map
5739 /node_modules/@frida/events/events.js
↻ events
2132 /node_modules/@frida/ieee754/index.js.map
1001 /node_modules/@frida/ieee754/index.js
↻ ieee754
17121 /node_modules/@frida/path/index.js.map
9533 /node_modules/@frida/path/index.js
↻ path
1289 /node_modules/@frida/process/index.js.map
1074 /node_modules/@frida/process/index.js
↻ process
2345 /node_modules/@frida/readable-stream/errors.js.map
2217 /node_modules/@frida/readable-stream/errors.js
2301 /node_modules/@frida/readable-stream/lib/abort_controller.js.map
1595 /node_modules/@frida/readable-stream/lib/abort_controller.js
977 /node_modules/@frida/readable-stream/lib/add-abort-signal.js.map
609 /node_modules/@frida/readable-stream/lib/add-abort-signal.js
2997 /node_modules/@frida/readable-stream/lib/buffer_list.js.map
1736 /node_modules/@frida/readable-stream/lib/buffer_list.js
2808 /node_modules/@frida/readable-stream/lib/compose.js.map
1771 /node_modules/@frida/readable-stream/lib/compose.js
4914 /node_modules/@frida/readable-stream/lib/destroy.js.map
3161 /node_modules/@frida/readable-stream/lib/destroy.js
6529 /node_modules/@frida/readable-stream/lib/duplex.js.map
4977 /node_modules/@frida/readable-stream/lib/duplex.js
3502 /node_modules/@frida/readable-stream/lib/end-of-stream.js.map
2065 /node_modules/@frida/readable-stream/lib/end-of-stream.js
14317 /node_modules/@frida/readable-stream/lib/event_target.js.map
10300 /node_modules/@frida/readable-stream/lib/event_target.js
1806 /node_modules/@frida/readable-stream/lib/from.js.map
1086 /node_modules/@frida/readable-stream/lib/from.js
1922 /node_modules/@frida/readable-stream/lib/legacy.js.map
1189 /node_modules/@frida/readable-stream/lib/legacy.js
352 /node_modules/@frida/readable-stream/lib/once.js.map
95 /node_modules/@frida/readable-stream/lib/once.js
548 /node_modules/@frida/readable-stream/lib/passthrough.js.map
238 /node_modules/@frida/readable-stream/lib/passthrough.js
4518 /node_modules/@frida/readable-stream/lib/pipeline.js.map
2667 /node_modules/@frida/readable-stream/lib/pipeline.js
788 /node_modules/@frida/readable-stream/lib/promises.js.map
409 /node_modules/@frida/readable-stream/lib/promises.js
18120 /node_modules/@frida/readable-stream/lib/readable.js.map
13334 /node_modules/@frida/readable-stream/lib/readable.js
826 /node_modules/@frida/readable-stream/lib/state.js.map
426 /node_modules/@frida/readable-stream/lib/state.js
2608 /node_modules/@frida/readable-stream/lib/transform.js.map
1644 /node_modules/@frida/readable-stream/lib/transform.js
3922 /node_modules/@frida/readable-stream/lib/utils.js.map
3422 /node_modules/@frida/readable-stream/lib/utils.js
12733 /node_modules/@frida/readable-stream/lib/writable.js.map
9107 /node_modules/@frida/readable-stream/lib/writable.js
1315 /node_modules/@frida/readable-stream/readable.js.map
1243 /node_modules/@frida/readable-stream/readable.js
↻ readable-stream
503 /node_modules/@frida/stream/index.js.map
438 /node_modules/@frida/stream/index.js
↻ stream
5123 /node_modules/@frida/string_decoder/lib/string_decoder.js.map
3481 /node_modules/@frida/string_decoder/lib/string_decoder.js
↻ string_decoder
3835 /node_modules/@frida/util/support/types.js.map
3002 /node_modules/@frida/util/support/types.js
12014 /node_modules/@frida/util/util.js.map
8557 /node_modules/@frida/util/util.js
↻ util
21875 /node_modules/frida-fs/dist/index.js.map
14522 /node_modules/frida-fs/dist/index.js
↻ fs
4240 /node_modules/frida-remote-stream/dist/index.js.map
3010 /node_modules/frida-remote-stream/dist/index.js
↻ frida-remote-stream
✄
{"version":3,"file":"agent.js","names":["Buffer","RemoteStreamController","fs","fsPath","S_IFMT","S_IFREG","S_IFDIR","S_IFCHR","S_IFBLK","S_IFIFO","S_IFLNK","S_IFSOCK","constants","pointerSize","Process","cachedUsers","Map","cachedGroups","getpwduidR","getgrgidR","entryFromStats","path","name","stats","mode","type","typeFromMode","target","targetPath","readlinkSync","targetType","targetPerms","s","statSync","permissionsFromMode","e","nlink","resolveUserID","uid","resolveGroupID","gid","size","mtimeMs","Error","toString","access","i","get","undefined","platform","pwd","SystemFunction","Module","getExportByName","buf","res","pwdCapacity","bufCapacity","Memory","alloc","add","r","value","errno","entry","readPointer","isNull","readUtf8String","set","group","groupCapacity","agent","constructor","_Agent_streamController","this","_Agent_onMessage","message","rawData","data","from","__classPrivateFieldGet","receive","stanza","payload","recv","_Agent_onStreamControllerSendRequest","packet","send","buffer","_Agent_onStreamControllerStreamRequest","stream","index","parseInt","label","details","filename","isDirectory","join","writer","createWriteStream","onStreamError","error","detachListeners","end","onWriterError","destroy","onWriterFinish","removeListener","pipe","addListener","events","on","ls","paths","length","fileGroup","entries","errors","directoryGroups","digDeeper","lstatSync","push","isSymbolicLink","names","readdirSync","curPath","rm","flags","dirs","files","force","includes","pending","slice","shift","filter","map","unshift","unlinkSync","collectError","rmdirSync","async","total","reader","createReadStream","open","transfer","Promise","resolve","reject","onReaderError","rpc","exports","bind","pull"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/","sources":["agent.ts"],"mappings":"0WAASA,MAAc,gBAChBC,MAAwD,6BACxDC,MAAQ,YACRC,MAAY,OAEnB,MAAMC,OACFA,EAAMC,QACNA,EAAOC,QACPA,EAAOC,QACPA,EAAOC,QACPA,EAAOC,QACPA,EAAOC,QACPA,EAAOC,SACPA,GACAT,EAAGU,WAEDC,YAAEA,GAAgBC,QAElBC,EAAc,IAAIC,IAClBC,EAAe,IAAID,IACzB,IAAIE,EAA0H,KAC1HC,EAAyH,KA+T7H,SAASC,EAAeC,EAAcC,EAAcC,GAChD,MAAMC,KAAEA,GAASD,EACXE,EAAOC,EAAaF,GAE1B,IAAIG,EACJ,GAAa,MAATF,EAAc,CACd,MAAMG,EAAa1B,EAAG2B,aAAaR,GACnC,IAAIS,EACAC,EACJ,IACI,MAAMC,EAAI9B,EAAG+B,SAASZ,GACtBU,EAAcG,EAAoBF,EAAER,MACpCM,EAAaJ,EAAaM,EAAER,K,CAC9B,MAAOW,GACLL,EAAa,I,CAEjBH,EAAS,CAACC,EAA4B,OAAfE,EAAuB,CAACA,EAAYC,GAAgB,K,MAE3EJ,EAAS,KAGb,MAAO,CACHL,EACAK,EACAF,EACAS,EAAoBV,GACpBD,EAAMa,MACNC,EAAcd,EAAMe,KACpBC,EAAehB,EAAMiB,KACrBjB,EAAMkB,KACNlB,EAAMmB,QAEd,CAEA,SAAShB,EAAaF,GAClB,OAAQA,EAAOpB,GACX,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAS,MAAO,IACrB,KAAKC,EAAU,MAAO,IAE1B,MAAM,IAAIgC,MAAM,mBAAmBnB,EAAKoB,SAAS,MACrD,CAEA,SAASV,EAAoBV,GACzB,IAAIqB,EAAS,GACb,IAAK,IAAIC,EAAI,GAAU,IAAPA,EAAUA,GAAK,EAEvBD,GADuB,IAArBrB,IAASsB,EAAK,GACN,IAEA,IAGVD,GAD6B,IAA3BrB,IAAUsB,EAAI,EAAM,GACZ,IAEA,IAGVD,GAD6B,IAA3BrB,IAAUsB,EAAI,EAAM,GACZ,IAEA,IAGlB,OAAOD,CACX,CAEA,SAASR,EAAcC,GACnB,IAAIhB,EAAOP,EAAYgC,IAAIT,GAC3B,QAAaU,IAAT1B,EACA,OAAOA,EAGX,GAAyB,YAArBR,QAAQmC,SACR3B,EAAOgB,EAAIM,eACR,CAOH,IAAIM,EANe,OAAfhC,IACAA,EAAa,IAAIiC,eAAeC,OAAOC,gBAAgB,KAAM,cACzD,MACA,CAAC,OAAQ,UAAW,UAAW,SAAU,aAIjD,IACIC,EAEAC,EAHAC,EAAc,IAEdC,EAAc,KAElB,OAAG,CACCP,EAAMQ,OAAOC,MAAMH,EAAcC,EAAc5C,GAC/CyC,EAAMJ,EAAIU,IAAIJ,GACdD,EAAMD,EAAIM,IAAIH,GAEd,MAAMI,EAAI3C,EAAWoB,EAAKY,EAAKI,EAAKG,EAAaF,GACjD,GAAgB,IAAZM,EAAEC,MACF,MAEJ,GAvaG,KAuaCD,EAAEE,MACF,MAAM,IAAIpB,MAAM,6BAA6BL,MAAQuB,EAAEE,SAE3DN,GAAe,C,CAGnB,MAAMO,EAAQT,EAAIU,cAId3C,EAHC0C,EAAME,SAGA5B,EAAIM,WAFJoB,EAAMC,cAAcE,gB,CAQnC,OAFApD,EAAYqD,IAAI9B,EAAKhB,GAEdA,CACX,CAEA,SAASiB,EAAeC,GACpB,IAAIlB,EAAOL,EAAa8B,IAAIP,GAC5B,QAAaQ,IAAT1B,EACA,OAAOA,EAGX,GAAyB,YAArBR,QAAQmC,SACR3B,EAAOkB,EAAII,eACR,CAOH,IAAIyB,EANc,OAAdlD,IACAA,EAAY,IAAIgC,eAAeC,OAAOC,gBAAgB,KAAM,cACxD,MACA,CAAC,OAAQ,UAAW,UAAW,SAAU,aAIjD,IACIC,EAEAC,EAHAe,EAAgB,IAEhBb,EAAc,KAElB,OAAG,CACCY,EAAQX,OAAOC,MAAMW,EAAgBb,EAAc5C,GACnDyC,EAAMe,EAAMT,IAAIU,GAChBf,EAAMD,EAAIM,IAAIH,GAEd,MAAMI,EAAI1C,EAAUqB,EAAK6B,EAAOf,EAAKG,EAAaF,GAClD,GAAgB,IAAZM,EAAEC,MACF,MAEJ,GAvdG,KAudCD,EAAEE,MACF,MAAM,IAAIpB,MAAM,8BAA8BH,MAAQqB,EAAEE,SAE5DN,GAAe,C,CAGnB,MAAMO,EAAQT,EAAIU,cAId3C,EAHC0C,EAAME,SAGA1B,EAAII,WAFJoB,EAAMC,cAAcE,gB,CAQnC,OAFAlD,EAAamD,IAAI5B,EAAKlB,GAEfA,CACX,C,wDAEA,MAAMiD,EAAQ,IAled,MAGIC,cAFAC,EAAAL,IAAAM,KAAoB,IAAIzE,GAiNxB0E,EAAAP,IAAAM,MAAa,CAACE,EAAcC,KAGxB,GAAa,WAFQD,EAAQnD,KAEN,CACnB,MAAMqD,EAAmC,OAAZD,EAAoB7E,EAAO+E,KAAKF,GAAW,KACxEG,EAAAN,KAAID,EAAA,KAAmBQ,QAAQ,CAC3BC,OAAQN,EAAQO,QAChBL,Q,CAIRM,KAAKJ,EAAAN,KAAIC,EAAA,KAAY,IAGzBU,EAAAjB,IAAAM,MAAkCY,IAC9BC,KAAK,CACD9D,KAAM,SACN0D,QAASG,EAAOJ,QACjBI,EAAOR,MAAMU,OAAsB,IAG1CC,EAAArB,IAAAM,MAAoCgB,IAChC,MAAMC,EAAQC,SAASF,EAAOG,OAExBC,EAAUJ,EAAOI,QACjBC,EAAmBD,EAAQC,SAC3BpE,EAAiBmE,EAAQnE,OAE/B,IAAIN,EAAsB,KAC1B,IACcnB,EAAG+B,SAASN,GAChBqE,gBACF3E,EAAOlB,EAAO8F,KAAKtE,EAAQoE,G,CAEjC,MAAO5D,G,CAEI,OAATd,IACAA,EAAOM,GAGX,MAAMuE,EAAShG,EAAGiG,kBAAkB9E,GAOpC,SAAS+E,EAAcC,GACnBC,IACAJ,EAAOK,MAEPhB,KAAK,CACD9D,KAAM,gBACNkE,QACAU,MAAOA,EAAMzB,SAErB,CAEA,SAAS4B,EAAcH,GACnBC,IACAZ,EAAOe,UAEPlB,KAAK,CACD9D,KAAM,gBACNkE,QACAU,MAAOA,EAAMzB,SAErB,CAEA,SAAS8B,IACLJ,IAEAf,KAAK,CACD9D,KAAM,kBACNkE,SAER,CAEA,SAASW,IACLJ,EAAOS,eAAe,SAAUD,GAChCR,EAAOS,eAAe,QAASH,GAC/Bd,EAAOiB,eAAe,QAASP,EACnC,CAzCAV,EAAOkB,KAAKV,GAEZR,EAAOmB,YAAY,QAAST,GAC5BF,EAAOW,YAAY,QAASL,GAC5BN,EAAOW,YAAY,SAAUH,EAqC7B,IAhSAtB,KAAKJ,EAAAN,KAAIC,EAAA,MACTK,EAAAN,KAAID,EAAA,KAAmBqC,OAAOC,GAAG,OAAQ/B,EAAAN,KAAIW,EAAA,MAC7CL,EAAAN,KAAID,EAAA,KAAmBqC,OAAOC,GAAG,SAAU/B,EAAAN,KAAIe,EAAA,KACnD,CAEAuB,GAAGC,GACsB,IAAjBA,EAAMC,SACND,EAAQ,CAAuB,YAArBnG,QAAQmC,SAA0B,OAAS,MAGzD,MAAMkE,EAAuB,CACzB9F,KAAM,GACN+F,QAAS,GACTC,OAAQ,IAENC,EAA+B,GAErC,IAAK,MAAMjG,KAAQ4F,EAAO,CACtB,IAAI1F,EAQAgG,EAPJ,IACIhG,EAAQrB,EAAGsH,UAAUnG,E,CACvB,MAAOc,GACLgF,EAAUE,OAAOI,KAAK,CAACpG,EAAOc,EAAYyC,UAC1C,Q,CAIJ,GAAIrD,EAAMmG,iBAAkB,CACxB,IAAI1F,EACJ,IACIA,EAAI9B,EAAG+B,SAASZ,GAChBkG,EAAYvF,EAAEgE,cACVuB,IACAhG,EAAQS,E,CAEd,MAAOG,GACLoF,GAAY,C,OAGhBA,EAAYhG,EAAMyE,cAGtB,GAAIuB,EAAW,CACX,IAAII,EACJ,IACIA,EAAQzH,EAAG0H,YAAYvG,E,CACzB,MAAOc,GACLmF,EAAgBG,KAAK,CACjBpG,OACA+F,QAAS,GACTC,OAAQ,CAAC,CAAChG,EAAOc,EAAYyC,YAEjC,Q,CAGJ,MAAMwC,EAAuB,GAC7B,IAAK,MAAM9F,KAAQqG,EAAO,CACtB,MAAME,EAAU1H,EAAO8F,KAAK5E,EAAOC,GACnC,IACI,MACM0C,EAAQ5C,EAAeyG,EAASvG,EADX,MAATA,EAAgBC,EAAQrB,EAAGsH,UAAUK,IAEvDT,EAAQK,KAAKzD,E,CACf,MAAO7B,G,EAIbmF,EAAgBG,KAAK,CACjBpG,OACA+F,UACAC,OAAQ,I,MAGZF,EAAUC,QAAQK,KAAKrG,EAAeC,EAAMA,EAAME,G,CAI1D,OAAQ4F,EAAUC,QAAQF,OAAS,GAAKC,EAAUE,OAAOH,OAAS,EAC5D,CAACC,KAAcG,GACfA,CACV,CAEAQ,GAAGb,EAAiBc,GAChB,MAAMV,EAAmB,GAEnBW,EAAiB,GACjBC,EAAkB,GAElBC,EAAQH,EAAMI,SAAS,SAG7B,GAFkBJ,EAAMI,SAAS,aAElB,CACX,MAAMC,EAAUnB,EAAMoB,QACtB,OAAa,CACT,MAAMhH,EAAO+G,EAAQE,QACrB,QAAatF,IAAT3B,EACA,MAGJ,IAAIW,EACJ,IACIA,EAAI9B,EAAG+B,SAASZ,E,CAClB,MAAOc,GACL8F,EAAMR,KAAKpG,GACX,Q,CAGAW,EAAEgE,eACFoC,EAAQX,QAAQvH,EAAG0H,YAAYvG,GAC1BkH,QAAOxC,GAAyB,MAAbA,GAAiC,OAAbA,IACvCyC,KAAIzC,GAAY5F,EAAO8F,KAAK5E,EAAM0E,MACvCiC,EAAKS,QAAQpH,IAEb4G,EAAMQ,QAAQpH,E,OAItB4G,EAAMR,QAAQR,GAGlB,IAAK,MAAM5F,KAAQ4G,EACf,IACI/H,EAAGwI,WAAWrH,E,CAChB,MAAOc,GACA+F,GACDS,EAAatH,EAAMc,E,CAK/B,IAAK,MAAMd,KAAQ2G,EACf,IACI9H,EAAG0I,UAAUvH,E,CACf,MAAOc,GACLwG,EAAatH,EAAMc,E,CAI3B,SAASwG,EAAatH,EAAcc,GAChCkF,EAAOI,KAAK,GAAGpG,MAAUc,EAAYyC,UACzC,CAEA,OAAOyC,CACX,CAEAwB,WAAW5B,GACP,IAAI6B,EAAQ,EACZ,IAAK,MAAMzH,KAAQ4F,EACf,IAEI6B,GADU5I,EAAG+B,SAASZ,GACXoB,I,CACb,MAAON,G,CAGboD,KAAK,CACD9D,KAAM,cACNqH,UAGJ,IAAInD,EAAQ,EACZ,IAAK,MAAMtE,KAAQ4F,EAAO,CACtB,MAAM8B,EAAS7I,EAAG8I,iBAAiB3H,GAC7B6E,EAAS6C,EAAOnC,KAAK5B,EAAAN,KAAID,EAAA,KAAmBwE,KAAKtD,EAAM/C,aAEvDsG,EAAW,IAAIC,SAAQ,CAACC,EAASC,KAKnC,SAASC,EAAcjD,GACnBC,IACAJ,EAAOK,MACP8C,EAAOhD,EACX,CAEA,SAASG,EAAcH,GACnBC,IACAyC,EAAOtC,UACP2C,EAAQ,KACZ,CAEA,SAAS1C,IACLJ,IACA8C,EAAQ,KACZ,CAEA,SAAS9C,IACLJ,EAAOS,eAAe,SAAUD,GAChCR,EAAOS,eAAe,QAASH,GAC/BuC,EAAOpC,eAAe,QAAS2C,EACnC,CAzBAP,EAAOlC,YAAY,QAASyC,GAC5BpD,EAAOW,YAAY,QAASL,GAC5BN,EAAOW,YAAY,SAAUH,EAuB7B,IAGJ,UACUwC,C,CACR,MAAO/G,GACLoD,KAAK,CACD9D,KAAM,gBACNkE,QACAU,MAAQlE,EAAYyC,S,CAI5Be,G,CAER,GAoRJ4D,IAAIC,QAAU,CACVxC,GAAIzC,EAAMyC,GAAGyC,KAAKlF,GAClBuD,GAAIvD,EAAMuD,GAAG2B,KAAKlF,GAClBmF,KAAMnF,EAAMmF,KAAKD,KAAKlF"}
✄
var e,t,r,n,s=this&&this.__classPrivateFieldGet||function(e,t,r,n){if("a"===r&&!n)throw new TypeError("Private accessor was defined without a getter");if("function"==typeof t?e!==t||!n:!t.has(e))throw new TypeError("Cannot read private member from an object whose class did not declare it");return"m"===r?n:"a"===r?n.call(e):n?n.value:t.get(e)};import{Buffer as o}from"buffer";import i from"frida-remote-stream";import a from"fs";import c from"path";const{S_IFMT:l,S_IFREG:u,S_IFDIR:f,S_IFCHR:d,S_IFBLK:p,S_IFIFO:h,S_IFLNK:m,S_IFSOCK:y}=a.constants,{pointerSize:S}=Process,w=new Map,g=new Map;let v=null,b=null;function L(e,t,r){const{mode:n}=r,s=I(n);let o;if("l"===s){const t=a.readlinkSync(e);let r,n;try{const t=a.statSync(e);n=_(t.mode),r=I(t.mode)}catch(e){r=null}o=[t,null!==r?[r,n]:null]}else o=null;return[t,o,s,_(n),r.nlink,F(r.uid),M(r.gid),r.size,r.mtimeMs]}function I(e){switch(e&l){case u:return"-";case f:return"d";case d:return"c";case p:return"b";case h:return"p";case m:return"l";case y:return"s"}throw new Error(`Invalid mode: 0x${e.toString(16)}`)}function _(e){let t="";for(let r=8;-1!==r;r-=3)t+=0!=(e>>>r&1)?"r":"-",t+=0!=(e>>>r-1&1)?"w":"-",t+=0!=(e>>>r-2&1)?"x":"-";return t}function F(e){let t=w.get(e);if(void 0!==t)return t;if("windows"===Process.platform)t=e.toString();else{let r;null===v&&(v=new SystemFunction(Module.getExportByName(null,"getpwuid_r"),"int",["uint","pointer","pointer","size_t","pointer"]));let n,s,o=128,i=1024;for(;;){r=Memory.alloc(o+i+S),n=r.add(o),s=n.add(i);const t=v(e,r,n,i,s);if(0===t.value)break;if(34!==t.errno)throw new Error(`Unable to resolve user ID ${e}: ${t.errno}`);i*=2}const a=s.readPointer();t=a.isNull()?e.toString():a.readPointer().readUtf8String()}return w.set(e,t),t}function M(e){let t=g.get(e);if(void 0!==t)return t;if("windows"===Process.platform)t=e.toString();else{let r;null===b&&(b=new SystemFunction(Module.getExportByName(null,"getgrgid_r"),"int",["uint","pointer","pointer","size_t","pointer"]));let n,s,o=128,i=1024;for(;;){r=Memory.alloc(o+i+S),n=r.add(o),s=n.add(i);const t=b(e,r,n,i,s);if(0===t.value)break;if(34!==t.errno)throw new Error(`Unable to resolve group ID ${e}: ${t.errno}`);i*=2}const a=s.readPointer();t=a.isNull()?e.toString():a.readPointer().readUtf8String()}return g.set(e,t),t}e=new WeakMap,t=new WeakMap,r=new WeakMap,n=new WeakMap;const k=new class{constructor(){e.set(this,new i),t.set(this,((r,n)=>{if("stream"===r.type){const t=null!==n?o.from(n):null;s(this,e,"f").receive({stanza:r.payload,data:t})}recv(s(this,t,"f"))})),r.set(this,(e=>{send({type:"stream",payload:e.stanza},e.data?.buffer)})),n.set(this,(e=>{const t=parseInt(e.label),r=e.details,n=r.filename,s=r.target;let o=null;try{a.statSync(s).isDirectory()&&(o=c.join(s,n))}catch(e){}null===o&&(o=s);const i=a.createWriteStream(o);function l(e){d(),i.end(),send({type:"push:io-error",index:t,error:e.message})}function u(r){d(),e.destroy(),send({type:"push:io-error",index:t,error:r.message})}function f(){d(),send({type:"push:io-success",index:t})}function d(){i.removeListener("finish",f),i.removeListener("error",u),e.removeListener("error",l)}e.pipe(i),e.addListener("error",l),i.addListener("error",u),i.addListener("finish",f)})),recv(s(this,t,"f")),s(this,e,"f").events.on("send",s(this,r,"f")),s(this,e,"f").events.on("stream",s(this,n,"f"))}ls(e){0===e.length&&(e=["windows"===Process.platform?"C:\\":"/"]);const t={path:"",entries:[],errors:[]},r=[];for(const n of e){let e,s;try{e=a.lstatSync(n)}catch(e){t.errors.push([n,e.message]);continue}if(e.isSymbolicLink()){let t;try{t=a.statSync(n),s=t.isDirectory(),s&&(e=t)}catch(e){s=!1}}else s=e.isDirectory();if(s){let t;try{t=a.readdirSync(n)}catch(e){r.push({path:n,entries:[],errors:[[n,e.message]]});continue}const s=[];for(const r of t){const t=c.join(n,r);try{const n=L(t,r,"."===r?e:a.lstatSync(t));s.push(n)}catch(e){}}r.push({path:n,entries:s,errors:[]})}else t.entries.push(L(n,n,e))}return t.entries.length>0||t.errors.length>0?[t,...r]:r}rm(e,t){const r=[],n=[],s=[],o=t.includes("force");if(t.includes("recursive")){const t=e.slice();for(;;){const e=t.shift();if(void 0===e)break;let r;try{r=a.statSync(e)}catch(t){s.push(e);continue}r.isDirectory()?(t.push(...a.readdirSync(e).filter((e=>"."!==e&&".."!==e)).map((t=>c.join(e,t)))),n.unshift(e)):s.unshift(e)}}else s.push(...e);for(const e of s)try{a.unlinkSync(e)}catch(t){o||i(e,t)}for(const e of n)try{a.rmdirSync(e)}catch(t){i(e,t)}function i(e,t){r.push(`${e}: ${t.message}`)}return r}async pull(t){let r=0;for(const e of t)try{r+=a.statSync(e).size}catch(e){}send({type:"pull:status",total:r});let n=0;for(const r of t){const t=a.createReadStream(r),o=t.pipe(s(this,e,"f").open(n.toString())),i=new Promise(((e,r)=>{function n(e){a(),o.end(),r(e)}function s(r){a(),t.destroy(),e(null)}function i(){a(),e(null)}function a(){o.removeListener("finish",i),o.removeListener("error",s),t.removeListener("error",n)}t.addListener("error",n),o.addListener("error",s),o.addListener("finish",i)}));try{await i}catch(e){send({type:"pull:io-error",index:n,error:e.message})}n++}}};rpc.exports={ls:k.ls.bind(k),rm:k.rm.bind(k),pull:k.pull.bind(k)};
✄
{"version":3,"file":"index.js","names":["lookup","revLookup","code","i","len","length","charCodeAt","getLens","b64","Error","validLen","indexOf","byteLength","lens","placeHoldersLen","toByteArray","arr","Uint8Array","_byteLength","curByte","tmp","encodeChunk","uint8","start","end","output","push","num","join","fromByteArray","extraBytes","parts","maxChunkLength","len2"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/base64-js/","sources":[""],"mappings":"AAAA,MAAMA,EAAS,GACTC,EAAY,GAEZC,EAAO,mEACb,IAAK,IAAIC,EAAI,EAAGC,EAAMF,EAAKG,OAAQF,EAAIC,IAAOD,EAC5CH,EAAOG,GAAKD,EAAKC,GACjBF,EAAUC,EAAKI,WAAWH,IAAMA,EAQlC,SAASI,EAASC,GAChB,MAAMJ,EAAMI,EAAIH,OAEhB,GAAID,EAAM,EAAI,EACZ,MAAM,IAAIK,MAAM,kDAKlB,IAAIC,EAAWF,EAAIG,QAAQ,MACT,IAAdD,IAAiBA,EAAWN,GAMhC,MAAO,CAACM,EAJgBA,IAAaN,EACjC,EACA,EAAKM,EAAW,EAGtB,CApBAT,EAAU,IAAIK,WAAW,IAAM,GAC/BL,EAAU,IAAIK,WAAW,IAAM,UAsBxB,SAASM,WAAYJ,GAC1B,MAAMK,EAAON,EAAQC,GACfE,EAAWG,EAAK,GAChBC,EAAkBD,EAAK,GAC7B,OAAuC,GAA9BH,EAAWI,GAAuB,EAAKA,CAClD,QAMO,SAASC,YAAaP,GAC3B,MAAMK,EAAON,EAAQC,GACfE,EAAWG,EAAK,GAChBC,EAAkBD,EAAK,GAEvBG,EAAM,IAAIC,WATlB,SAAsBT,EAAKE,EAAUI,GACnC,OAAuC,GAA9BJ,EAAWI,GAAuB,EAAKA,CAClD,CAO6BI,CAAYV,EAAKE,EAAUI,IAEtD,IAAIK,EAAU,EAGd,MAAMf,EAAMU,EAAkB,EAC1BJ,EAAW,EACXA,EAEJ,IAAIP,EACJ,IAAKA,EAAI,EAAGA,EAAIC,EAAKD,GAAK,EAAG,CAC3B,MAAMiB,EACHnB,EAAUO,EAAIF,WAAWH,KAAO,GAChCF,EAAUO,EAAIF,WAAWH,EAAI,KAAO,GACpCF,EAAUO,EAAIF,WAAWH,EAAI,KAAO,EACrCF,EAAUO,EAAIF,WAAWH,EAAI,IAC/Ba,EAAIG,KAAcC,GAAO,GAAM,IAC/BJ,EAAIG,KAAcC,GAAO,EAAK,IAC9BJ,EAAIG,KAAmB,IAANC,CACnB,CAEA,GAAwB,IAApBN,EAAuB,CACzB,MAAMM,EACHnB,EAAUO,EAAIF,WAAWH,KAAO,EAChCF,EAAUO,EAAIF,WAAWH,EAAI,KAAO,EACvCa,EAAIG,KAAmB,IAANC,CACnB,CAEA,GAAwB,IAApBN,EAAuB,CACzB,MAAMM,EACHnB,EAAUO,EAAIF,WAAWH,KAAO,GAChCF,EAAUO,EAAIF,WAAWH,EAAI,KAAO,EACpCF,EAAUO,EAAIF,WAAWH,EAAI,KAAO,EACvCa,EAAIG,KAAcC,GAAO,EAAK,IAC9BJ,EAAIG,KAAmB,IAANC,CACnB,CAEA,OAAOJ,CACT,CASA,SAASK,EAAaC,EAAOC,EAAOC,GAClC,MAAMC,EAAS,GACf,IAAK,IAAItB,EAAIoB,EAAOpB,EAAIqB,EAAKrB,GAAK,EAAG,CACnC,MAAMiB,GACFE,EAAMnB,IAAM,GAAM,WAClBmB,EAAMnB,EAAI,IAAM,EAAK,QACP,IAAfmB,EAAMnB,EAAI,IACbsB,EAAOC,KAbF1B,GADiB2B,EAcMP,IAbT,GAAK,IACxBpB,EAAO2B,GAAO,GAAK,IACnB3B,EAAO2B,GAAO,EAAI,IAClB3B,EAAa,GAAN2B,GAWT,CAfF,IAA0BA,EAgBxB,OAAOF,EAAOG,KAAK,GACrB,QAEO,SAASC,cAAeP,GAC7B,MAAMlB,EAAMkB,EAAMjB,OACZyB,EAAa1B,EAAM,EACnB2B,EAAQ,GACRC,EAAiB,MAGvB,IAAK,IAAI7B,EAAI,EAAG8B,EAAO7B,EAAM0B,EAAY3B,EAAI8B,EAAM9B,GAAK6B,EACtDD,EAAML,KAAKL,EAAYC,EAAOnB,EAAIA,EAAI6B,EAAkBC,EAAOA,EAAQ9B,EAAI6B,IAI7E,GAAmB,IAAfF,EAAkB,CACpB,MAAMV,EAAME,EAAMlB,EAAM,GACxB2B,EAAML,KACJ1B,EAAOoB,GAAO,GACdpB,EAAQoB,GAAO,EAAK,IACpB,KAEJ,MAAO,GAAmB,IAAfU,EAAkB,CAC3B,MAAMV,GAAOE,EAAMlB,EAAM,IAAM,GAAKkB,EAAMlB,EAAM,GAChD2B,EAAML,KACJ1B,EAAOoB,GAAO,IACdpB,EAAQoB,GAAO,EAAK,IACpBpB,EAAQoB,GAAO,EAAK,IACpB,IAEJ,CAEA,OAAOW,EAAMH,KAAK,GACpB"}
✄
const t=[],o=[],n="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";for(let c=0,h=n.length;c<h;++c)t[c]=n[c],o[n.charCodeAt(c)]=c;function r(t){const o=t.length;if(o%4>0)throw new Error("Invalid string. Length must be a multiple of 4");let n=t.indexOf("=");-1===n&&(n=o);return[n,n===o?0:4-n%4]}o["-".charCodeAt(0)]=62,o["_".charCodeAt(0)]=63;export function byteLength(t){const o=r(t),n=o[0],e=o[1];return 3*(n+e)/4-e}export function toByteArray(t){const n=r(t),e=n[0],c=n[1],h=new Uint8Array(function(t,o,n){return 3*(o+n)/4-n}(0,e,c));let s=0;const a=c>0?e-4:e;let f;for(f=0;f<a;f+=4){const n=o[t.charCodeAt(f)]<<18|o[t.charCodeAt(f+1)]<<12|o[t.charCodeAt(f+2)]<<6|o[t.charCodeAt(f+3)];h[s++]=n>>16&255,h[s++]=n>>8&255,h[s++]=255&n}if(2===c){const n=o[t.charCodeAt(f)]<<2|o[t.charCodeAt(f+1)]>>4;h[s++]=255&n}if(1===c){const n=o[t.charCodeAt(f)]<<10|o[t.charCodeAt(f+1)]<<4|o[t.charCodeAt(f+2)]>>2;h[s++]=n>>8&255,h[s++]=255&n}return h}function e(o,n,r){const e=[];for(let h=n;h<r;h+=3){const n=(o[h]<<16&16711680)+(o[h+1]<<8&65280)+(255&o[h+2]);e.push(t[(c=n)>>18&63]+t[c>>12&63]+t[c>>6&63]+t[63&c])}var c;return e.join("")}export function fromByteArray(o){const n=o.length,r=n%3,c=[],h=16383;for(let t=0,s=n-r;t<s;t+=h)c.push(e(o,t,t+h>s?s:t+h));if(1===r){const r=o[n-1];c.push(t[r>>2]+t[r<<4&63]+"==")}else if(2===r){const r=(o[n-2]<<8)+o[n-1];c.push(t[r>>10]+t[r>>4&63]+t[r<<2&63]+"=")}return c.join("")}
✄
{"version":3,"file":"index.js","names":["base64","ieee754","config","INSPECT_MAX_BYTES","K_MAX_LENGTH","createBuffer","length","RangeError","buf","Uint8Array","Object","setPrototypeOf","Buffer","prototype","TYPED_ARRAY_SUPPORT","defineProperty","enumerable","get","isBuffer","this","buffer","byteOffset","arg","encodingOrOffset","TypeError","allocUnsafe","from","value","string","encoding","isEncoding","byteLength","actual","write","slice","fromString","ArrayBuffer","isView","arrayView","copy","fromArrayBuffer","fromArrayLike","fromArrayView","SharedArrayBuffer","valueOf","b","obj","len","checked","undefined","Number","isNaN","type","Array","isArray","data","fromObject","Symbol","toPrimitive","assertSize","size","array","i","toString","poolSize","alloc","fill","allocUnsafeSlow","SlowBuffer","mustMatch","arguments","loweredCase","utf8ToBytes","base64ToBytes","toLowerCase","slowToString","start","end","hexSlice","utf8Slice","asciiSlice","latin1Slice","base64Slice","utf16leSlice","swap","n","m","bidirectionalIndexOf","val","dir","arrayIndexOf","indexOf","call","lastIndexOf","arr","indexSize","arrLength","valLength","String","read","readUInt16BE","foundIndex","found","j","hexWrite","offset","remaining","strLen","parsed","parseInt","substr","utf8Write","blitBuffer","asciiWrite","str","byteArray","push","charCodeAt","asciiToBytes","base64Write","ucs2Write","units","c","hi","lo","utf16leToBytes","fromByteArray","Math","min","res","firstByte","codePoint","bytesPerSequence","secondByte","thirdByte","fourthByte","tempCodePoint","codePoints","fromCharCode","apply","decodeCodePointsArray","_isBuffer","compare","a","x","y","concat","list","pos","set","swap16","swap32","swap64","toLocaleString","equals","inspect","max","replace","trim","for","target","thisStart","thisEnd","thisCopy","targetCopy","includes","isFinite","Error","toJSON","_arr","ret","out","hexSliceLookupTable","bytes","checkOffset","ext","checkInt","wrtBigUInt64LE","checkIntBI","BigInt","wrtBigUInt64BE","checkIEEE754","writeFloat","littleEndian","noAssert","writeDouble","newBuf","subarray","readUintLE","readUIntLE","mul","readUintBE","readUIntBE","readUint8","readUInt8","readUint16LE","readUInt16LE","readUint16BE","readUint32LE","readUInt32LE","readUint32BE","readUInt32BE","readBigUInt64LE","validateNumber","first","last","boundsError","readBigUInt64BE","readIntLE","pow","readIntBE","readInt8","readInt16LE","readInt16BE","readInt32LE","readInt32BE","readBigInt64LE","readBigInt64BE","readFloatLE","readFloatBE","readDoubleLE","readDoubleBE","writeUintLE","writeUIntLE","writeUintBE","writeUIntBE","writeUint8","writeUInt8","writeUint16LE","writeUInt16LE","writeUint16BE","writeUInt16BE","writeUint32LE","writeUInt32LE","writeUint32BE","writeUInt32BE","writeBigUInt64LE","writeBigUInt64BE","writeIntLE","limit","sub","writeIntBE","writeInt8","writeInt16LE","writeInt16BE","writeInt32LE","writeInt32BE","writeBigInt64LE","writeBigInt64BE","writeFloatLE","writeFloatBE","writeDoubleLE","writeDoubleBE","targetStart","copyWithin","code","errors","E","sym","getMessage","Base","constructor","super","writable","configurable","name","stack","message","addNumericalSeparator","range","ERR_OUT_OF_RANGE","checkBounds","ERR_INVALID_ARG_TYPE","floor","ERR_BUFFER_OUT_OF_BOUNDS","input","msg","received","isInteger","abs","INVALID_BASE64_RE","Infinity","leadSurrogate","toByteArray","split","base64clean","src","dst","alphabet","table","i16","kMaxLength"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/buffer/","sources":[""],"mappings":";;;;;;UAQYA,MAAY,sBACZC,MAAa,iBAElB,MAAMC,OAAS,CACpBC,kBAAmB,IAGrB,MAAMC,EAAe,kBACZA,iBAoBT,SAASC,EAAcC,GACrB,GAAIA,EAtBe,WAuBjB,MAAM,IAAIC,WAAW,cAAgBD,EAAS,kCAGhD,MAAME,EAAM,IAAIC,WAAWH,GAE3B,OADAI,OAAOC,eAAeH,EAAKI,OAAOC,WAC3BL,CACT,CA1BAI,OAAOE,qBAAsB,EAE7BJ,OAAOK,eAAeH,OAAOC,UAAW,SAAU,CAChDG,YAAY,EACZC,IAAK,WACH,GAAKL,OAAOM,SAASC,MACrB,OAAOA,KAAKC,MACd,IAGFV,OAAOK,eAAeH,OAAOC,UAAW,SAAU,CAChDG,YAAY,EACZC,IAAK,WACH,GAAKL,OAAOM,SAASC,MACrB,OAAOA,KAAKE,UACd,WAuBK,SAAST,OAAQU,EAAKC,EAAkBjB,GAE7C,GAAmB,iBAARgB,EAAkB,CAC3B,GAAgC,iBAArBC,EACT,MAAM,IAAIC,UACR,sEAGJ,OAAOC,EAAYH,EACrB,CACA,OAAOI,EAAKJ,EAAKC,EAAkBjB,EACrC,CAIA,SAASoB,EAAMC,EAAOJ,EAAkBjB,GACtC,GAAqB,iBAAVqB,EACT,OAoHJ,SAAqBC,EAAQC,GACH,iBAAbA,GAAsC,KAAbA,IAClCA,EAAW,QAGb,IAAKjB,OAAOkB,WAAWD,GACrB,MAAM,IAAIL,UAAU,qBAAuBK,GAG7C,MAAMvB,EAAwC,EAA/ByB,EAAWH,EAAQC,GAClC,IAAIrB,EAAMH,EAAaC,GAEvB,MAAM0B,EAASxB,EAAIyB,MAAML,EAAQC,GAE7BG,IAAW1B,IAIbE,EAAMA,EAAI0B,MAAM,EAAGF,IAGrB,OAAOxB,CACT,CA1IW2B,CAAWR,EAAOJ,GAG3B,GAAIa,YAAYC,OAAOV,GACrB,OAiJJ,SAAwBW,GACtB,GAAIA,aAAqB7B,WAAY,CACnC,MAAM8B,EAAO,IAAI9B,WAAW6B,GAC5B,OAAOE,EAAgBD,EAAKnB,OAAQmB,EAAKlB,WAAYkB,EAAKR,WAC5D,CACA,OAAOU,EAAcH,EACvB,CAvJWI,CAAcf,GAGvB,GAAa,MAATA,EACF,MAAM,IAAIH,UACR,yHACiDG,GAIrD,GAAIA,aAAiBS,aAChBT,GAASA,EAAMP,kBAAkBgB,YACpC,OAAOI,EAAgBb,EAAOJ,EAAkBjB,GAGlD,GAAIqB,aAAiBgB,mBAChBhB,GAASA,EAAMP,kBAAkBuB,kBACpC,OAAOH,EAAgBb,EAAOJ,EAAkBjB,GAGlD,GAAqB,iBAAVqB,EACT,MAAM,IAAIH,UACR,yEAIJ,MAAMoB,EAAUjB,EAAMiB,SAAWjB,EAAMiB,UACvC,GAAe,MAAXA,GAAmBA,IAAYjB,EACjC,OAAOf,OAAOc,KAAKkB,EAASrB,EAAkBjB,GAGhD,MAAMuC,EAkJR,SAAqBC,GACnB,GAAIlC,OAAOM,SAAS4B,GAAM,CACxB,MAAMC,EAA4B,EAAtBC,EAAQF,EAAIxC,QAClBE,EAAMH,EAAa0C,GAEzB,OAAmB,IAAfvC,EAAIF,QAIRwC,EAAIP,KAAK/B,EAAK,EAAG,EAAGuC,GAHXvC,CAKX,CAEA,QAAmByC,IAAfH,EAAIxC,OACN,MAA0B,iBAAfwC,EAAIxC,QAAuB4C,OAAOC,MAAML,EAAIxC,QAC9CD,EAAa,GAEfoC,EAAcK,GAGvB,GAAiB,WAAbA,EAAIM,MAAqBC,MAAMC,QAAQR,EAAIS,MAC7C,OAAOd,EAAcK,EAAIS,KAE7B,CAzKYC,CAAW7B,GACrB,GAAIkB,EAAG,OAAOA,EAEd,GAAsB,oBAAXY,QAAgD,MAAtBA,OAAOC,aACH,mBAA9B/B,EAAM8B,OAAOC,aACtB,OAAO9C,OAAOc,KAAKC,EAAM8B,OAAOC,aAAa,UAAWnC,EAAkBjB,GAG5E,MAAM,IAAIkB,UACR,yHACiDG,EAErD,CAmBA,SAASgC,EAAYC,GACnB,GAAoB,iBAATA,EACT,MAAM,IAAIpC,UAAU,0CACf,GAAIoC,EAAO,EAChB,MAAM,IAAIrD,WAAW,cAAgBqD,EAAO,iCAEhD,CA0BA,SAASnC,EAAamC,GAEpB,OADAD,EAAWC,GACJvD,EAAauD,EAAO,EAAI,EAAoB,EAAhBZ,EAAQY,GAC7C,CAuCA,SAASnB,EAAeoB,GACtB,MAAMvD,EAASuD,EAAMvD,OAAS,EAAI,EAA4B,EAAxB0C,EAAQa,EAAMvD,QAC9CE,EAAMH,EAAaC,GACzB,IAAK,IAAIwD,EAAI,EAAGA,EAAIxD,EAAQwD,GAAK,EAC/BtD,EAAIsD,GAAgB,IAAXD,EAAMC,GAEjB,OAAOtD,CACT,CAUA,SAASgC,EAAiBqB,EAAOxC,EAAYf,GAC3C,GAAIe,EAAa,GAAKwC,EAAM9B,WAAaV,EACvC,MAAM,IAAId,WAAW,wCAGvB,GAAIsD,EAAM9B,WAAaV,GAAcf,GAAU,GAC7C,MAAM,IAAIC,WAAW,wCAGvB,IAAIC,EAYJ,OAVEA,OADiByC,IAAf5B,QAAuC4B,IAAX3C,EACxB,IAAIG,WAAWoD,QACDZ,IAAX3C,EACH,IAAIG,WAAWoD,EAAOxC,GAEtB,IAAIZ,WAAWoD,EAAOxC,EAAYf,GAI1CI,OAAOC,eAAeH,EAAKI,OAAOC,WAE3BL,CACT,CA2BA,SAASwC,EAAS1C,GAGhB,GAAIA,GA3Qe,WA4QjB,MAAM,IAAIC,WAAW,0DA5QJ,YA6Q8BwD,SAAS,IAAM,UAEhE,OAAgB,EAATzD,CACT,CA1NAM,OAAOoD,SAAW,KA6DlBpD,OAAOc,KAAO,SAAUC,EAAOJ,EAAkBjB,GAC/C,OAAOoB,EAAKC,EAAOJ,EAAkBjB,EACvC,EAIAI,OAAOC,eAAeC,OAAOC,UAAWJ,WAAWI,WACnDH,OAAOC,eAAeC,OAAQH,YA8B9BG,OAAOqD,MAAQ,SAAUL,EAAMM,EAAMrC,GACnC,OArBF,SAAgB+B,EAAMM,EAAMrC,GAE1B,OADA8B,EAAWC,GACPA,GAAQ,EACHvD,EAAauD,QAETX,IAATiB,EAIyB,iBAAbrC,EACVxB,EAAauD,GAAMM,KAAKA,EAAMrC,GAC9BxB,EAAauD,GAAMM,KAAKA,GAEvB7D,EAAauD,EACtB,CAOSK,CAAML,EAAMM,EAAMrC,EAC3B,EAUAjB,OAAOa,YAAc,SAAUmC,GAC7B,OAAOnC,EAAYmC,EACrB,EAIAhD,OAAOuD,gBAAkB,SAAUP,GACjC,OAAOnC,EAAYmC,EACrB,SAsGO,SAASQ,WAAY9D,GAI1B,OAHKA,GAAUA,IACbA,EAAS,GAEJM,OAAOqD,OAAO3D,EACvB,CAiGA,SAASyB,EAAYH,EAAQC,GAC3B,GAAIjB,OAAOM,SAASU,GAClB,OAAOA,EAAOtB,OAEhB,GAAI8B,YAAYC,OAAOT,IAAWA,aAAkBQ,YAClD,OAAOR,EAAOG,WAEhB,GAAsB,iBAAXH,EACT,MAAM,IAAIJ,UACR,kGAC0BI,GAI9B,MAAMmB,EAAMnB,EAAOtB,OACb+D,EAAaC,UAAUhE,OAAS,IAAsB,IAAjBgE,UAAU,GACrD,IAAKD,GAAqB,IAARtB,EAAW,OAAO,EAGpC,IAAIwB,GAAc,EAClB,OACE,OAAQ1C,GACN,IAAK,QACL,IAAK,SACL,IAAK,SACH,OAAOkB,EACT,IAAK,OACL,IAAK,QACH,OAAOyB,EAAY5C,GAAQtB,OAC7B,IAAK,OACL,IAAK,QACL,IAAK,UACL,IAAK,WACH,OAAa,EAANyC,EACT,IAAK,MACH,OAAOA,IAAQ,EACjB,IAAK,SACH,OAAO0B,EAAc7C,GAAQtB,OAC/B,QACE,GAAIiE,EACF,OAAOF,GAAa,EAAIG,EAAY5C,GAAQtB,OAE9CuB,GAAY,GAAKA,GAAU6C,cAC3BH,GAAc,EAGtB,CAGA,SAASI,EAAc9C,EAAU+C,EAAOC,GACtC,IAAIN,GAAc,EAclB,SALctB,IAAV2B,GAAuBA,EAAQ,KACjCA,EAAQ,GAINA,EAAQzD,KAAKb,OACf,MAAO,GAOT,SAJY2C,IAAR4B,GAAqBA,EAAM1D,KAAKb,UAClCuE,EAAM1D,KAAKb,QAGTuE,GAAO,EACT,MAAO,GAOT,IAHAA,KAAS,KACTD,KAAW,GAGT,MAAO,GAKT,IAFK/C,IAAUA,EAAW,UAGxB,OAAQA,GACN,IAAK,MACH,OAAOiD,EAAS3D,KAAMyD,EAAOC,GAE/B,IAAK,OACL,IAAK,QACH,OAAOE,EAAU5D,KAAMyD,EAAOC,GAEhC,IAAK,QACH,OAAOG,EAAW7D,KAAMyD,EAAOC,GAEjC,IAAK,SACL,IAAK,SACH,OAAOI,EAAY9D,KAAMyD,EAAOC,GAElC,IAAK,SACH,OAAOK,EAAY/D,KAAMyD,EAAOC,GAElC,IAAK,OACL,IAAK,QACL,IAAK,UACL,IAAK,WACH,OAAOM,EAAahE,KAAMyD,EAAOC,GAEnC,QACE,GAAIN,EAAa,MAAM,IAAI/C,UAAU,qBAAuBK,GAC5DA,GAAYA,EAAW,IAAI6C,cAC3BH,GAAc,EAGtB,CAUA,SAASa,EAAMvC,EAAGwC,EAAGC,GACnB,MAAMxB,EAAIjB,EAAEwC,GACZxC,EAAEwC,GAAKxC,EAAEyC,GACTzC,EAAEyC,GAAKxB,CACT,CAyIA,SAASyB,EAAsBnE,EAAQoE,EAAKnE,EAAYQ,EAAU4D,GAEhE,GAAsB,IAAlBrE,EAAOd,OAAc,OAAQ,EAmBjC,GAhB0B,iBAAfe,GACTQ,EAAWR,EACXA,EAAa,GACJA,EAAa,WACtBA,EAAa,WACJA,GAAc,aACvBA,GAAc,YAEhBA,GAAcA,EACV6B,OAAOC,MAAM9B,KAEfA,EAAaoE,EAAM,EAAKrE,EAAOd,OAAS,GAItCe,EAAa,IAAGA,EAAaD,EAAOd,OAASe,GAC7CA,GAAcD,EAAOd,OAAQ,CAC/B,GAAImF,EAAK,OAAQ,EACZpE,EAAaD,EAAOd,OAAS,CACpC,MAAO,GAAIe,EAAa,EAAG,CACzB,IAAIoE,EACC,OAAQ,EADJpE,EAAa,CAExB,CAQA,GALmB,iBAARmE,IACTA,EAAM5E,OAAOc,KAAK8D,EAAK3D,IAIrBjB,OAAOM,SAASsE,GAElB,OAAmB,IAAfA,EAAIlF,QACE,EAEHoF,EAAatE,EAAQoE,EAAKnE,EAAYQ,EAAU4D,GAClD,GAAmB,iBAARD,EAEhB,OADAA,GAAY,IACgC,mBAAjC/E,WAAWI,UAAU8E,QAC1BF,EACKhF,WAAWI,UAAU8E,QAAQC,KAAKxE,EAAQoE,EAAKnE,GAE/CZ,WAAWI,UAAUgF,YAAYD,KAAKxE,EAAQoE,EAAKnE,GAGvDqE,EAAatE,EAAQ,CAACoE,GAAMnE,EAAYQ,EAAU4D,GAG3D,MAAM,IAAIjE,UAAU,uCACtB,CAEA,SAASkE,EAAcI,EAAKN,EAAKnE,EAAYQ,EAAU4D,GACrD,IA0BI3B,EA1BAiC,EAAY,EACZC,EAAYF,EAAIxF,OAChB2F,EAAYT,EAAIlF,OAEpB,QAAiB2C,IAAbpB,IAEe,UADjBA,EAAWqE,OAAOrE,GAAU6C,gBACY,UAAb7C,GACV,YAAbA,GAAuC,aAAbA,GAAyB,CACrD,GAAIiE,EAAIxF,OAAS,GAAKkF,EAAIlF,OAAS,EACjC,OAAQ,EAEVyF,EAAY,EACZC,GAAa,EACbC,GAAa,EACb5E,GAAc,CAChB,CAGF,SAAS8E,EAAM3F,EAAKsD,GAClB,OAAkB,IAAdiC,EACKvF,EAAIsD,GAEJtD,EAAI4F,aAAatC,EAAIiC,EAEhC,CAGA,GAAIN,EAAK,CACP,IAAIY,GAAc,EAClB,IAAKvC,EAAIzC,EAAYyC,EAAIkC,EAAWlC,IAClC,GAAIqC,EAAKL,EAAKhC,KAAOqC,EAAKX,GAAqB,IAAhBa,EAAoB,EAAIvC,EAAIuC,IAEzD,IADoB,IAAhBA,IAAmBA,EAAavC,GAChCA,EAAIuC,EAAa,IAAMJ,EAAW,OAAOI,EAAaN,OAEtC,IAAhBM,IAAmBvC,GAAKA,EAAIuC,GAChCA,GAAc,CAGpB,MAEE,IADIhF,EAAa4E,EAAYD,IAAW3E,EAAa2E,EAAYC,GAC5DnC,EAAIzC,EAAYyC,GAAK,EAAGA,IAAK,CAChC,IAAIwC,GAAQ,EACZ,IAAK,IAAIC,EAAI,EAAGA,EAAIN,EAAWM,IAC7B,GAAIJ,EAAKL,EAAKhC,EAAIyC,KAAOJ,EAAKX,EAAKe,GAAI,CACrCD,GAAQ,EACR,KACF,CAEF,GAAIA,EAAO,OAAOxC,CACpB,CAGF,OAAQ,CACV,CAcA,SAAS0C,EAAUhG,EAAKoB,EAAQ6E,EAAQnG,GACtCmG,EAASvD,OAAOuD,IAAW,EAC3B,MAAMC,EAAYlG,EAAIF,OAASmG,EAC1BnG,GAGHA,EAAS4C,OAAO5C,IACHoG,IACXpG,EAASoG,GAJXpG,EAASoG,EAQX,MAAMC,EAAS/E,EAAOtB,OAKtB,IAAIwD,EACJ,IAJIxD,EAASqG,EAAS,IACpBrG,EAASqG,EAAS,GAGf7C,EAAI,EAAGA,EAAIxD,IAAUwD,EAAG,CAC3B,MAAM8C,EAASC,SAASjF,EAAOkF,OAAW,EAAJhD,EAAO,GAAI,IACjD,GAAIZ,OAAOC,MAAMyD,GAAS,OAAO9C,EACjCtD,EAAIiG,EAAS3C,GAAK8C,CACpB,CACA,OAAO9C,CACT,CAEA,SAASiD,EAAWvG,EAAKoB,EAAQ6E,EAAQnG,GACvC,OAAO0G,EAAWxC,EAAY5C,EAAQpB,EAAIF,OAASmG,GAASjG,EAAKiG,EAAQnG,EAC3E,CAEA,SAAS2G,EAAYzG,EAAKoB,EAAQ6E,EAAQnG,GACxC,OAAO0G,EAwpCT,SAAuBE,GACrB,MAAMC,EAAY,GAClB,IAAK,IAAIrD,EAAI,EAAGA,EAAIoD,EAAI5G,SAAUwD,EAEhCqD,EAAUC,KAAyB,IAApBF,EAAIG,WAAWvD,IAEhC,OAAOqD,CACT,CA/pCoBG,CAAa1F,GAASpB,EAAKiG,EAAQnG,EACvD,CAEA,SAASiH,EAAa/G,EAAKoB,EAAQ6E,EAAQnG,GACzC,OAAO0G,EAAWvC,EAAc7C,GAASpB,EAAKiG,EAAQnG,EACxD,CAEA,SAASkH,EAAWhH,EAAKoB,EAAQ6E,EAAQnG,GACvC,OAAO0G,EAypCT,SAAyBE,EAAKO,GAC5B,IAAIC,EAAGC,EAAIC,EACX,MAAMT,EAAY,GAClB,IAAK,IAAIrD,EAAI,EAAGA,EAAIoD,EAAI5G,WACjBmH,GAAS,GAAK,KADa3D,EAGhC4D,EAAIR,EAAIG,WAAWvD,GACnB6D,EAAKD,GAAK,EACVE,EAAKF,EAAI,IACTP,EAAUC,KAAKQ,GACfT,EAAUC,KAAKO,GAGjB,OAAOR,CACT,CAvqCoBU,CAAejG,EAAQpB,EAAIF,OAASmG,GAASjG,EAAKiG,EAAQnG,EAC9E,CA8EA,SAAS4E,EAAa1E,EAAKoE,EAAOC,GAChC,OAAc,IAAVD,GAAeC,IAAQrE,EAAIF,OACtBN,EAAO8H,cAActH,GAErBR,EAAO8H,cAActH,EAAI0B,MAAM0C,EAAOC,GAEjD,CAEA,SAASE,EAAWvE,EAAKoE,EAAOC,GAC9BA,EAAMkD,KAAKC,IAAIxH,EAAIF,OAAQuE,GAC3B,MAAMoD,EAAM,GAEZ,IAAInE,EAAIc,EACR,KAAOd,EAAIe,GAAK,CACd,MAAMqD,EAAY1H,EAAIsD,GACtB,IAAIqE,EAAY,KACZC,EAAoBF,EAAY,IAChC,EACCA,EAAY,IACT,EACCA,EAAY,IACT,EACA,EAEZ,GAAIpE,EAAIsE,GAAoBvD,EAAK,CAC/B,IAAIwD,EAAYC,EAAWC,EAAYC,EAEvC,OAAQJ,GACN,KAAK,EACCF,EAAY,MACdC,EAAYD,GAEd,MACF,KAAK,EACHG,EAAa7H,EAAIsD,EAAI,GACO,MAAV,IAAbuE,KACHG,GAA6B,GAAZN,IAAqB,EAAoB,GAAbG,EACzCG,EAAgB,MAClBL,EAAYK,IAGhB,MACF,KAAK,EACHH,EAAa7H,EAAIsD,EAAI,GACrBwE,EAAY9H,EAAIsD,EAAI,GACQ,MAAV,IAAbuE,IAAsD,MAAV,IAAZC,KACnCE,GAA6B,GAAZN,IAAoB,IAAoB,GAAbG,IAAsB,EAAmB,GAAZC,EACrEE,EAAgB,OAAUA,EAAgB,OAAUA,EAAgB,SACtEL,EAAYK,IAGhB,MACF,KAAK,EACHH,EAAa7H,EAAIsD,EAAI,GACrBwE,EAAY9H,EAAIsD,EAAI,GACpByE,EAAa/H,EAAIsD,EAAI,GACO,MAAV,IAAbuE,IAAsD,MAAV,IAAZC,IAAsD,MAAV,IAAbC,KAClEC,GAA6B,GAAZN,IAAoB,IAAqB,GAAbG,IAAsB,IAAmB,GAAZC,IAAqB,EAAoB,GAAbC,EAClGC,EAAgB,OAAUA,EAAgB,UAC5CL,EAAYK,IAItB,CAEkB,OAAdL,GAGFA,EAAY,MACZC,EAAmB,GACVD,EAAY,QAErBA,GAAa,MACbF,EAAIb,KAAKe,IAAc,GAAK,KAAQ,OACpCA,EAAY,MAAqB,KAAZA,GAGvBF,EAAIb,KAAKe,GACTrE,GAAKsE,CACP,CAEA,OAQF,SAAgCK,GAC9B,MAAM1F,EAAM0F,EAAWnI,OACvB,GAAIyC,GAJuB,KAKzB,OAAOmD,OAAOwC,aAAaC,MAAMzC,OAAQuC,GAI3C,IAAIR,EAAM,GACNnE,EAAI,EACR,KAAOA,EAAIf,GACTkF,GAAO/B,OAAOwC,aAAaC,MACzBzC,OACAuC,EAAWvG,MAAM4B,EAAGA,GAdG,OAiB3B,OAAOmE,CACT,CAxBSW,CAAsBX,EAC/B,CA/qBArH,OAAOM,SAAW,SAAmB2B,GACnC,OAAY,MAALA,IAA6B,IAAhBA,EAAEgG,WACpBhG,IAAMjC,OAAOC,SACjB,EAEAD,OAAOkI,QAAU,SAAkBC,EAAGlG,GAGpC,GAFIkG,aAAatI,aAAYsI,EAAInI,OAAOc,KAAKqH,EAAGA,EAAEtC,OAAQsC,EAAEhH,aACxDc,aAAapC,aAAYoC,EAAIjC,OAAOc,KAAKmB,EAAGA,EAAE4D,OAAQ5D,EAAEd,cACvDnB,OAAOM,SAAS6H,KAAOnI,OAAOM,SAAS2B,GAC1C,MAAM,IAAIrB,UACR,yEAIJ,GAAIuH,IAAMlG,EAAG,OAAO,EAEpB,IAAImG,EAAID,EAAEzI,OACN2I,EAAIpG,EAAEvC,OAEV,IAAK,IAAIwD,EAAI,EAAGf,EAAMgF,KAAKC,IAAIgB,EAAGC,GAAInF,EAAIf,IAAOe,EAC/C,GAAIiF,EAAEjF,KAAOjB,EAAEiB,GAAI,CACjBkF,EAAID,EAAEjF,GACNmF,EAAIpG,EAAEiB,GACN,KACF,CAGF,OAAIkF,EAAIC,GAAW,EACfA,EAAID,EAAU,EACX,CACT,EAEApI,OAAOkB,WAAa,SAAqBD,GACvC,OAAQqE,OAAOrE,GAAU6C,eACvB,IAAK,MACL,IAAK,OACL,IAAK,QACL,IAAK,QACL,IAAK,SACL,IAAK,SACL,IAAK,SACL,IAAK,OACL,IAAK,QACL,IAAK,UACL,IAAK,WACH,OAAO,EACT,QACE,OAAO,EAEb,EAEA9D,OAAOsI,OAAS,SAAiBC,EAAM7I,GACrC,IAAK+C,MAAMC,QAAQ6F,GACjB,MAAM,IAAI3H,UAAU,+CAGtB,GAAoB,IAAhB2H,EAAK7I,OACP,OAAOM,OAAOqD,MAAM,GAGtB,IAAIH,EACJ,QAAeb,IAAX3C,EAEF,IADAA,EAAS,EACJwD,EAAI,EAAGA,EAAIqF,EAAK7I,SAAUwD,EAC7BxD,GAAU6I,EAAKrF,GAAGxD,OAItB,MAAMc,EAASR,OAAOa,YAAYnB,GAClC,IAAI8I,EAAM,EACV,IAAKtF,EAAI,EAAGA,EAAIqF,EAAK7I,SAAUwD,EAAG,CAChC,IAAItD,EAAM2I,EAAKrF,GACf,GAAItD,aAAeC,WACb2I,EAAM5I,EAAIF,OAASc,EAAOd,QACvBM,OAAOM,SAASV,KACnBA,EAAMI,OAAOc,KAAKlB,EAAIY,OAAQZ,EAAIa,WAAYb,EAAIuB,aAEpDvB,EAAI+B,KAAKnB,EAAQgI,IAEjB3I,WAAWI,UAAUwI,IAAIzD,KACvBxE,EACAZ,EACA4I,OAGC,KAAKxI,OAAOM,SAASV,GAC1B,MAAM,IAAIgB,UAAU,+CAEpBhB,EAAI+B,KAAKnB,EAAQgI,EACnB,CACAA,GAAO5I,EAAIF,MACb,CACA,OAAOc,CACT,EAiDAR,OAAOmB,WAAaA,EA8EpBnB,OAAOC,UAAUgI,WAAY,EAQ7BjI,OAAOC,UAAUyI,OAAS,WACxB,MAAMvG,EAAM5B,KAAKb,OACjB,GAAIyC,EAAM,GAAM,EACd,MAAM,IAAIxC,WAAW,6CAEvB,IAAK,IAAIuD,EAAI,EAAGA,EAAIf,EAAKe,GAAK,EAC5BsB,EAAKjE,KAAM2C,EAAGA,EAAI,GAEpB,OAAO3C,IACT,EAEAP,OAAOC,UAAU0I,OAAS,WACxB,MAAMxG,EAAM5B,KAAKb,OACjB,GAAIyC,EAAM,GAAM,EACd,MAAM,IAAIxC,WAAW,6CAEvB,IAAK,IAAIuD,EAAI,EAAGA,EAAIf,EAAKe,GAAK,EAC5BsB,EAAKjE,KAAM2C,EAAGA,EAAI,GAClBsB,EAAKjE,KAAM2C,EAAI,EAAGA,EAAI,GAExB,OAAO3C,IACT,EAEAP,OAAOC,UAAU2I,OAAS,WACxB,MAAMzG,EAAM5B,KAAKb,OACjB,GAAIyC,EAAM,GAAM,EACd,MAAM,IAAIxC,WAAW,6CAEvB,IAAK,IAAIuD,EAAI,EAAGA,EAAIf,EAAKe,GAAK,EAC5BsB,EAAKjE,KAAM2C,EAAGA,EAAI,GAClBsB,EAAKjE,KAAM2C,EAAI,EAAGA,EAAI,GACtBsB,EAAKjE,KAAM2C,EAAI,EAAGA,EAAI,GACtBsB,EAAKjE,KAAM2C,EAAI,EAAGA,EAAI,GAExB,OAAO3C,IACT,EAEAP,OAAOC,UAAUkD,SAAW,WAC1B,MAAMzD,EAASa,KAAKb,OACpB,OAAe,IAAXA,EAAqB,GACA,IAArBgE,UAAUhE,OAAqByE,EAAU5D,KAAM,EAAGb,GAC/CqE,EAAagE,MAAMxH,KAAMmD,UAClC,EAEA1D,OAAOC,UAAU4I,eAAiB7I,OAAOC,UAAUkD,SAEnDnD,OAAOC,UAAU6I,OAAS,SAAiB7G,GACzC,IAAKjC,OAAOM,SAAS2B,GAAI,MAAM,IAAIrB,UAAU,6BAC7C,OAAIL,OAAS0B,GACsB,IAA5BjC,OAAOkI,QAAQ3H,KAAM0B,EAC9B,EAEAjC,OAAOC,UAAU8I,QAAU,WACzB,IAAIzC,EAAM,GACV,MAAM0C,EAAM1J,OAAOC,kBAGnB,OAFA+G,EAAM/F,KAAK4C,SAAS,MAAO,EAAG6F,GAAKC,QAAQ,UAAW,OAAOC,OACzD3I,KAAKb,OAASsJ,IAAK1C,GAAO,SACvB,WAAaA,EAAM,GAC5B,EACAtG,OAAOC,UAAU4C,OAAOsG,IAAI,+BAAiCnJ,OAAOC,UAAU8I,QAE9E/I,OAAOC,UAAUiI,QAAU,SAAkBkB,EAAQpF,EAAOC,EAAKoF,EAAWC,GAI1E,GAHIF,aAAkBvJ,aACpBuJ,EAASpJ,OAAOc,KAAKsI,EAAQA,EAAOvD,OAAQuD,EAAOjI,cAEhDnB,OAAOM,SAAS8I,GACnB,MAAM,IAAIxI,UACR,wFAC2BwI,GAiB/B,QAbc/G,IAAV2B,IACFA,EAAQ,QAEE3B,IAAR4B,IACFA,EAAMmF,EAASA,EAAO1J,OAAS,QAEf2C,IAAdgH,IACFA,EAAY,QAEEhH,IAAZiH,IACFA,EAAU/I,KAAKb,QAGbsE,EAAQ,GAAKC,EAAMmF,EAAO1J,QAAU2J,EAAY,GAAKC,EAAU/I,KAAKb,OACtE,MAAM,IAAIC,WAAW,sBAGvB,GAAI0J,GAAaC,GAAWtF,GAASC,EACnC,OAAO,EAET,GAAIoF,GAAaC,EACf,OAAQ,EAEV,GAAItF,GAASC,EACX,OAAO,EAQT,GAAI1D,OAAS6I,EAAQ,OAAO,EAE5B,IAAIhB,GAJJkB,KAAa,IADbD,KAAe,GAMXhB,GAPJpE,KAAS,IADTD,KAAW,GASX,MAAM7B,EAAMgF,KAAKC,IAAIgB,EAAGC,GAElBkB,EAAWhJ,KAAKe,MAAM+H,EAAWC,GACjCE,EAAaJ,EAAO9H,MAAM0C,EAAOC,GAEvC,IAAK,IAAIf,EAAI,EAAGA,EAAIf,IAAOe,EACzB,GAAIqG,EAASrG,KAAOsG,EAAWtG,GAAI,CACjCkF,EAAImB,EAASrG,GACbmF,EAAImB,EAAWtG,GACf,KACF,CAGF,OAAIkF,EAAIC,GAAW,EACfA,EAAID,EAAU,EACX,CACT,EA2HApI,OAAOC,UAAUwJ,SAAW,SAAmB7E,EAAKnE,EAAYQ,GAC9D,OAAoD,IAA7CV,KAAKwE,QAAQH,EAAKnE,EAAYQ,EACvC,EAEAjB,OAAOC,UAAU8E,QAAU,SAAkBH,EAAKnE,EAAYQ,GAC5D,OAAO0D,EAAqBpE,KAAMqE,EAAKnE,EAAYQ,GAAU,EAC/D,EAEAjB,OAAOC,UAAUgF,YAAc,SAAsBL,EAAKnE,EAAYQ,GACpE,OAAO0D,EAAqBpE,KAAMqE,EAAKnE,EAAYQ,GAAU,EAC/D,EA4CAjB,OAAOC,UAAUoB,MAAQ,SAAgBL,EAAQ6E,EAAQnG,EAAQuB,GAE/D,QAAeoB,IAAXwD,EACF5E,EAAW,OACXvB,EAASa,KAAKb,OACdmG,EAAS,OAEJ,QAAexD,IAAX3C,GAA0C,iBAAXmG,EACxC5E,EAAW4E,EACXnG,EAASa,KAAKb,OACdmG,EAAS,MAEJ,KAAI6D,SAAS7D,GAUlB,MAAM,IAAI8D,MACR,2EAVF9D,KAAoB,EAChB6D,SAAShK,IACXA,KAAoB,OACH2C,IAAbpB,IAAwBA,EAAW,UAEvCA,EAAWvB,EACXA,OAAS2C,EAMb,CAEA,MAAMyD,EAAYvF,KAAKb,OAASmG,EAGhC,SAFexD,IAAX3C,GAAwBA,EAASoG,KAAWpG,EAASoG,GAEpD9E,EAAOtB,OAAS,IAAMA,EAAS,GAAKmG,EAAS,IAAOA,EAAStF,KAAKb,OACrE,MAAM,IAAIC,WAAW,0CAGlBsB,IAAUA,EAAW,QAE1B,IAAI0C,GAAc,EAClB,OACE,OAAQ1C,GACN,IAAK,MACH,OAAO2E,EAASrF,KAAMS,EAAQ6E,EAAQnG,GAExC,IAAK,OACL,IAAK,QACH,OAAOyG,EAAU5F,KAAMS,EAAQ6E,EAAQnG,GAEzC,IAAK,QACL,IAAK,SACL,IAAK,SACH,OAAO2G,EAAW9F,KAAMS,EAAQ6E,EAAQnG,GAE1C,IAAK,SAEH,OAAOiH,EAAYpG,KAAMS,EAAQ6E,EAAQnG,GAE3C,IAAK,OACL,IAAK,QACL,IAAK,UACL,IAAK,WACH,OAAOkH,EAAUrG,KAAMS,EAAQ6E,EAAQnG,GAEzC,QACE,GAAIiE,EAAa,MAAM,IAAI/C,UAAU,qBAAuBK,GAC5DA,GAAY,GAAKA,GAAU6C,cAC3BH,GAAc,EAGtB,EAEA3D,OAAOC,UAAU2J,OAAS,WACxB,MAAO,CACLpH,KAAM,SACNG,KAAMF,MAAMxC,UAAUqB,MAAM0D,KAAKzE,KAAKsJ,MAAQtJ,KAAM,GAExD,EA6GA,SAAS6D,EAAYxE,EAAKoE,EAAOC,GAC/B,IAAI6F,EAAM,GACV7F,EAAMkD,KAAKC,IAAIxH,EAAIF,OAAQuE,GAE3B,IAAK,IAAIf,EAAIc,EAAOd,EAAIe,IAAOf,EAC7B4G,GAAOxE,OAAOwC,aAAsB,IAATlI,EAAIsD,IAEjC,OAAO4G,CACT,CAEA,SAASzF,EAAazE,EAAKoE,EAAOC,GAChC,IAAI6F,EAAM,GACV7F,EAAMkD,KAAKC,IAAIxH,EAAIF,OAAQuE,GAE3B,IAAK,IAAIf,EAAIc,EAAOd,EAAIe,IAAOf,EAC7B4G,GAAOxE,OAAOwC,aAAalI,EAAIsD,IAEjC,OAAO4G,CACT,CAEA,SAAS5F,EAAUtE,EAAKoE,EAAOC,GAC7B,MAAM9B,EAAMvC,EAAIF,SAEXsE,GAASA,EAAQ,KAAGA,EAAQ,KAC5BC,GAAOA,EAAM,GAAKA,EAAM9B,KAAK8B,EAAM9B,GAExC,IAAI4H,EAAM,GACV,IAAK,IAAI7G,EAAIc,EAAOd,EAAIe,IAAOf,EAC7B6G,GAAOC,EAAoBpK,EAAIsD,IAEjC,OAAO6G,CACT,CAEA,SAASxF,EAAc3E,EAAKoE,EAAOC,GACjC,MAAMgG,EAAQrK,EAAI0B,MAAM0C,EAAOC,GAC/B,IAAIoD,EAAM,GAEV,IAAK,IAAInE,EAAI,EAAGA,EAAI+G,EAAMvK,OAAS,EAAGwD,GAAK,EACzCmE,GAAO/B,OAAOwC,aAAamC,EAAM/G,GAAqB,IAAf+G,EAAM/G,EAAI,IAEnD,OAAOmE,CACT,CAiCA,SAAS6C,EAAarE,EAAQsE,EAAKzK,GACjC,GAAKmG,EAAS,GAAO,GAAKA,EAAS,EAAG,MAAM,IAAIlG,WAAW,sBAC3D,GAAIkG,EAASsE,EAAMzK,EAAQ,MAAM,IAAIC,WAAW,wCAClD,CAyQA,SAASyK,EAAUxK,EAAKmB,EAAO8E,EAAQsE,EAAKnB,EAAK5B,GAC/C,IAAKpH,OAAOM,SAASV,GAAM,MAAM,IAAIgB,UAAU,+CAC/C,GAAIG,EAAQiI,GAAOjI,EAAQqG,EAAK,MAAM,IAAIzH,WAAW,qCACrD,GAAIkG,EAASsE,EAAMvK,EAAIF,OAAQ,MAAM,IAAIC,WAAW,qBACtD,CA+FA,SAAS0K,EAAgBzK,EAAKmB,EAAO8E,EAAQuB,EAAK4B,GAChDsB,EAAWvJ,EAAOqG,EAAK4B,EAAKpJ,EAAKiG,EAAQ,GAEzC,IAAImB,EAAK1E,OAAOvB,EAAQwJ,OAAO,aAC/B3K,EAAIiG,KAAYmB,EAChBA,IAAW,EACXpH,EAAIiG,KAAYmB,EAChBA,IAAW,EACXpH,EAAIiG,KAAYmB,EAChBA,IAAW,EACXpH,EAAIiG,KAAYmB,EAChB,IAAID,EAAKzE,OAAOvB,GAASwJ,OAAO,IAAMA,OAAO,aAQ7C,OAPA3K,EAAIiG,KAAYkB,EAChBA,IAAW,EACXnH,EAAIiG,KAAYkB,EAChBA,IAAW,EACXnH,EAAIiG,KAAYkB,EAChBA,IAAW,EACXnH,EAAIiG,KAAYkB,EACTlB,CACT,CAEA,SAAS2E,EAAgB5K,EAAKmB,EAAO8E,EAAQuB,EAAK4B,GAChDsB,EAAWvJ,EAAOqG,EAAK4B,EAAKpJ,EAAKiG,EAAQ,GAEzC,IAAImB,EAAK1E,OAAOvB,EAAQwJ,OAAO,aAC/B3K,EAAIiG,EAAS,GAAKmB,EAClBA,IAAW,EACXpH,EAAIiG,EAAS,GAAKmB,EAClBA,IAAW,EACXpH,EAAIiG,EAAS,GAAKmB,EAClBA,IAAW,EACXpH,EAAIiG,EAAS,GAAKmB,EAClB,IAAID,EAAKzE,OAAOvB,GAASwJ,OAAO,IAAMA,OAAO,aAQ7C,OAPA3K,EAAIiG,EAAS,GAAKkB,EAClBA,IAAW,EACXnH,EAAIiG,EAAS,GAAKkB,EAClBA,IAAW,EACXnH,EAAIiG,EAAS,GAAKkB,EAClBA,IAAW,EACXnH,EAAIiG,GAAUkB,EACPlB,EAAS,CAClB,CAkHA,SAAS4E,EAAc7K,EAAKmB,EAAO8E,EAAQsE,EAAKnB,EAAK5B,GACnD,GAAIvB,EAASsE,EAAMvK,EAAIF,OAAQ,MAAM,IAAIC,WAAW,sBACpD,GAAIkG,EAAS,EAAG,MAAM,IAAIlG,WAAW,qBACvC,CAEA,SAAS+K,EAAY9K,EAAKmB,EAAO8E,EAAQ8E,EAAcC,GAOrD,OANA7J,GAASA,EACT8E,KAAoB,EACf+E,GACHH,EAAa7K,EAAKmB,EAAO8E,EAAQ,GAEnCxG,EAAQgC,MAAMzB,EAAKmB,EAAO8E,EAAQ8E,EAAc,GAAI,GAC7C9E,EAAS,CAClB,CAUA,SAASgF,EAAajL,EAAKmB,EAAO8E,EAAQ8E,EAAcC,GAOtD,OANA7J,GAASA,EACT8E,KAAoB,EACf+E,GACHH,EAAa7K,EAAKmB,EAAO8E,EAAQ,GAEnCxG,EAAQgC,MAAMzB,EAAKmB,EAAO8E,EAAQ8E,EAAc,GAAI,GAC7C9E,EAAS,CAClB,CAzkBA7F,OAAOC,UAAUqB,MAAQ,SAAgB0C,EAAOC,GAC9C,MAAM9B,EAAM5B,KAAKb,QACjBsE,IAAUA,GAGE,GACVA,GAAS7B,GACG,IAAG6B,EAAQ,GACdA,EAAQ7B,IACjB6B,EAAQ7B,IANV8B,OAAc5B,IAAR4B,EAAoB9B,IAAQ8B,GASxB,GACRA,GAAO9B,GACG,IAAG8B,EAAM,GACVA,EAAM9B,IACf8B,EAAM9B,GAGJ8B,EAAMD,IAAOC,EAAMD,GAEvB,MAAM8G,EAASvK,KAAKwK,SAAS/G,EAAOC,GAIpC,OAFAnE,OAAOC,eAAe+K,EAAQ9K,OAAOC,WAE9B6K,CACT,EAUA9K,OAAOC,UAAU+K,WACjBhL,OAAOC,UAAUgL,WAAa,SAAqBpF,EAAQ1E,EAAYyJ,GACrE/E,KAAoB,EACpB1E,KAA4B,EACvByJ,GAAUV,EAAYrE,EAAQ1E,EAAYZ,KAAKb,QAEpD,IAAIkF,EAAMrE,KAAKsF,GACXqF,EAAM,EACNhI,EAAI,EACR,OAASA,EAAI/B,IAAe+J,GAAO,MACjCtG,GAAOrE,KAAKsF,EAAS3C,GAAKgI,EAG5B,OAAOtG,CACT,EAEA5E,OAAOC,UAAUkL,WACjBnL,OAAOC,UAAUmL,WAAa,SAAqBvF,EAAQ1E,EAAYyJ,GACrE/E,KAAoB,EACpB1E,KAA4B,EACvByJ,GACHV,EAAYrE,EAAQ1E,EAAYZ,KAAKb,QAGvC,IAAIkF,EAAMrE,KAAKsF,IAAW1E,GACtB+J,EAAM,EACV,KAAO/J,EAAa,IAAM+J,GAAO,MAC/BtG,GAAOrE,KAAKsF,IAAW1E,GAAc+J,EAGvC,OAAOtG,CACT,EAEA5E,OAAOC,UAAUoL,UACjBrL,OAAOC,UAAUqL,UAAY,SAAoBzF,EAAQ+E,GAGvD,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCa,KAAKsF,EACd,EAEA7F,OAAOC,UAAUsL,aACjBvL,OAAOC,UAAUuL,aAAe,SAAuB3F,EAAQ+E,GAG7D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCa,KAAKsF,GAAWtF,KAAKsF,EAAS,IAAM,CAC7C,EAEA7F,OAAOC,UAAUwL,aACjBzL,OAAOC,UAAUuF,aAAe,SAAuBK,EAAQ+E,GAG7D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACnCa,KAAKsF,IAAW,EAAKtF,KAAKsF,EAAS,EAC7C,EAEA7F,OAAOC,UAAUyL,aACjB1L,OAAOC,UAAU0L,aAAe,SAAuB9F,EAAQ+E,GAI7D,OAHA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,SAElCa,KAAKsF,GACTtF,KAAKsF,EAAS,IAAM,EACpBtF,KAAKsF,EAAS,IAAM,IACD,SAAnBtF,KAAKsF,EAAS,EACrB,EAEA7F,OAAOC,UAAU2L,aACjB5L,OAAOC,UAAU4L,aAAe,SAAuBhG,EAAQ+E,GAI7D,OAHA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QAEpB,SAAfa,KAAKsF,IACTtF,KAAKsF,EAAS,IAAM,GACrBtF,KAAKsF,EAAS,IAAM,EACrBtF,KAAKsF,EAAS,GAClB,EAEA7F,OAAOC,UAAU6L,gBAAkB,SAA0BjG,GAE3DkG,EADAlG,KAAoB,EACG,UACvB,MAAMmG,EAAQzL,KAAKsF,GACboG,EAAO1L,KAAKsF,EAAS,QACbxD,IAAV2J,QAAgC3J,IAAT4J,GACzBC,EAAYrG,EAAQtF,KAAKb,OAAS,GAGpC,MAAMsH,EAAKgF,EACQ,IAAjBzL,OAAOsF,GACU,MAAjBtF,OAAOsF,GACPtF,OAAOsF,GAAU,GAAK,GAElBkB,EAAKxG,OAAOsF,GACC,IAAjBtF,OAAOsF,GACU,MAAjBtF,OAAOsF,GACPoG,EAAO,GAAK,GAEd,OAAO1B,OAAOvD,IAAOuD,OAAOxD,IAAOwD,OAAO,IAC5C,EAEAvK,OAAOC,UAAUkM,gBAAkB,SAA0BtG,GAE3DkG,EADAlG,KAAoB,EACG,UACvB,MAAMmG,EAAQzL,KAAKsF,GACboG,EAAO1L,KAAKsF,EAAS,QACbxD,IAAV2J,QAAgC3J,IAAT4J,GACzBC,EAAYrG,EAAQtF,KAAKb,OAAS,GAGpC,MAAMqH,EAAKiF,EAAQ,GAAK,GACL,MAAjBzL,OAAOsF,GACU,IAAjBtF,OAAOsF,GACPtF,OAAOsF,GAEHmB,EAAKzG,OAAOsF,GAAU,GAAK,GACd,MAAjBtF,OAAOsF,GACU,IAAjBtF,OAAOsF,GACPoG,EAEF,OAAQ1B,OAAOxD,IAAOwD,OAAO,KAAOA,OAAOvD,EAC7C,EAEAhH,OAAOC,UAAUmM,UAAY,SAAoBvG,EAAQ1E,EAAYyJ,GACnE/E,KAAoB,EACpB1E,KAA4B,EACvByJ,GAAUV,EAAYrE,EAAQ1E,EAAYZ,KAAKb,QAEpD,IAAIkF,EAAMrE,KAAKsF,GACXqF,EAAM,EACNhI,EAAI,EACR,OAASA,EAAI/B,IAAe+J,GAAO,MACjCtG,GAAOrE,KAAKsF,EAAS3C,GAAKgI,EAM5B,OAJAA,GAAO,IAEHtG,GAAOsG,IAAKtG,GAAOuC,KAAKkF,IAAI,EAAG,EAAIlL,IAEhCyD,CACT,EAEA5E,OAAOC,UAAUqM,UAAY,SAAoBzG,EAAQ1E,EAAYyJ,GACnE/E,KAAoB,EACpB1E,KAA4B,EACvByJ,GAAUV,EAAYrE,EAAQ1E,EAAYZ,KAAKb,QAEpD,IAAIwD,EAAI/B,EACJ+J,EAAM,EACNtG,EAAMrE,KAAKsF,IAAW3C,GAC1B,KAAOA,EAAI,IAAMgI,GAAO,MACtBtG,GAAOrE,KAAKsF,IAAW3C,GAAKgI,EAM9B,OAJAA,GAAO,IAEHtG,GAAOsG,IAAKtG,GAAOuC,KAAKkF,IAAI,EAAG,EAAIlL,IAEhCyD,CACT,EAEA5E,OAAOC,UAAUsM,SAAW,SAAmB1G,EAAQ+E,GAGrD,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACtB,IAAfa,KAAKsF,IAC0B,GAA5B,IAAOtF,KAAKsF,GAAU,GADKtF,KAAKsF,EAE3C,EAEA7F,OAAOC,UAAUuM,YAAc,SAAsB3G,EAAQ+E,GAC3D/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QAC3C,MAAMkF,EAAMrE,KAAKsF,GAAWtF,KAAKsF,EAAS,IAAM,EAChD,OAAc,MAANjB,EAAsB,WAANA,EAAmBA,CAC7C,EAEA5E,OAAOC,UAAUwM,YAAc,SAAsB5G,EAAQ+E,GAC3D/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QAC3C,MAAMkF,EAAMrE,KAAKsF,EAAS,GAAMtF,KAAKsF,IAAW,EAChD,OAAc,MAANjB,EAAsB,WAANA,EAAmBA,CAC7C,EAEA5E,OAAOC,UAAUyM,YAAc,SAAsB7G,EAAQ+E,GAI3D,OAHA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QAEnCa,KAAKsF,GACVtF,KAAKsF,EAAS,IAAM,EACpBtF,KAAKsF,EAAS,IAAM,GACpBtF,KAAKsF,EAAS,IAAM,EACzB,EAEA7F,OAAOC,UAAU0M,YAAc,SAAsB9G,EAAQ+E,GAI3D,OAHA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QAEnCa,KAAKsF,IAAW,GACrBtF,KAAKsF,EAAS,IAAM,GACpBtF,KAAKsF,EAAS,IAAM,EACpBtF,KAAKsF,EAAS,EACnB,EAEA7F,OAAOC,UAAU2M,eAAiB,SAAyB/G,GAEzDkG,EADAlG,KAAoB,EACG,UACvB,MAAMmG,EAAQzL,KAAKsF,GACboG,EAAO1L,KAAKsF,EAAS,QACbxD,IAAV2J,QAAgC3J,IAAT4J,GACzBC,EAAYrG,EAAQtF,KAAKb,OAAS,GAGpC,MAAMkF,EAAMrE,KAAKsF,EAAS,GACL,IAAnBtF,KAAKsF,EAAS,GACK,MAAnBtF,KAAKsF,EAAS,IACboG,GAAQ,IAEX,OAAQ1B,OAAO3F,IAAQ2F,OAAO,KAC5BA,OAAOyB,EACU,IAAjBzL,OAAOsF,GACU,MAAjBtF,OAAOsF,GACPtF,OAAOsF,GAAU,GAAK,GAC1B,EAEA7F,OAAOC,UAAU4M,eAAiB,SAAyBhH,GAEzDkG,EADAlG,KAAoB,EACG,UACvB,MAAMmG,EAAQzL,KAAKsF,GACboG,EAAO1L,KAAKsF,EAAS,QACbxD,IAAV2J,QAAgC3J,IAAT4J,GACzBC,EAAYrG,EAAQtF,KAAKb,OAAS,GAGpC,MAAMkF,GAAOoH,GAAS,IACH,MAAjBzL,OAAOsF,GACU,IAAjBtF,OAAOsF,GACPtF,OAAOsF,GAET,OAAQ0E,OAAO3F,IAAQ2F,OAAO,KAC5BA,OAAOhK,OAAOsF,GAAU,GAAK,GACZ,MAAjBtF,OAAOsF,GACU,IAAjBtF,OAAOsF,GACPoG,EACJ,EAEAjM,OAAOC,UAAU6M,YAAc,SAAsBjH,EAAQ+E,GAG3D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCL,EAAQkG,KAAKhF,KAAMsF,GAAQ,EAAM,GAAI,EAC9C,EAEA7F,OAAOC,UAAU8M,YAAc,SAAsBlH,EAAQ+E,GAG3D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCL,EAAQkG,KAAKhF,KAAMsF,GAAQ,EAAO,GAAI,EAC/C,EAEA7F,OAAOC,UAAU+M,aAAe,SAAuBnH,EAAQ+E,GAG7D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCL,EAAQkG,KAAKhF,KAAMsF,GAAQ,EAAM,GAAI,EAC9C,EAEA7F,OAAOC,UAAUgN,aAAe,SAAuBpH,EAAQ+E,GAG7D,OAFA/E,KAAoB,EACf+E,GAAUV,EAAYrE,EAAQ,EAAGtF,KAAKb,QACpCL,EAAQkG,KAAKhF,KAAMsF,GAAQ,EAAO,GAAI,EAC/C,EAQA7F,OAAOC,UAAUiN,YACjBlN,OAAOC,UAAUkN,YAAc,SAAsBpM,EAAO8E,EAAQ1E,EAAYyJ,GAI9E,GAHA7J,GAASA,EACT8E,KAAoB,EACpB1E,KAA4B,GACvByJ,EAAU,CAEbR,EAAS7J,KAAMQ,EAAO8E,EAAQ1E,EADbgG,KAAKkF,IAAI,EAAG,EAAIlL,GAAc,EACK,EACtD,CAEA,IAAI+J,EAAM,EACNhI,EAAI,EAER,IADA3C,KAAKsF,GAAkB,IAAR9E,IACNmC,EAAI/B,IAAe+J,GAAO,MACjC3K,KAAKsF,EAAS3C,GAAMnC,EAAQmK,EAAO,IAGrC,OAAOrF,EAAS1E,CAClB,EAEAnB,OAAOC,UAAUmN,YACjBpN,OAAOC,UAAUoN,YAAc,SAAsBtM,EAAO8E,EAAQ1E,EAAYyJ,GAI9E,GAHA7J,GAASA,EACT8E,KAAoB,EACpB1E,KAA4B,GACvByJ,EAAU,CAEbR,EAAS7J,KAAMQ,EAAO8E,EAAQ1E,EADbgG,KAAKkF,IAAI,EAAG,EAAIlL,GAAc,EACK,EACtD,CAEA,IAAI+B,EAAI/B,EAAa,EACjB+J,EAAM,EAEV,IADA3K,KAAKsF,EAAS3C,GAAa,IAARnC,IACVmC,GAAK,IAAMgI,GAAO,MACzB3K,KAAKsF,EAAS3C,GAAMnC,EAAQmK,EAAO,IAGrC,OAAOrF,EAAS1E,CAClB,EAEAnB,OAAOC,UAAUqN,WACjBtN,OAAOC,UAAUsN,WAAa,SAAqBxM,EAAO8E,EAAQ+E,GAKhE,OAJA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,IAAM,GACtDtF,KAAKsF,GAAmB,IAAR9E,EACT8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUuN,cACjBxN,OAAOC,UAAUwN,cAAgB,SAAwB1M,EAAO8E,EAAQ+E,GAMtE,OALA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,MAAQ,GACxDtF,KAAKsF,GAAmB,IAAR9E,EAChBR,KAAKsF,EAAS,GAAM9E,IAAU,EACvB8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUyN,cACjB1N,OAAOC,UAAU0N,cAAgB,SAAwB5M,EAAO8E,EAAQ+E,GAMtE,OALA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,MAAQ,GACxDtF,KAAKsF,GAAW9E,IAAU,EAC1BR,KAAKsF,EAAS,GAAc,IAAR9E,EACb8E,EAAS,CAClB,EAEA7F,OAAOC,UAAU2N,cACjB5N,OAAOC,UAAU4N,cAAgB,SAAwB9M,EAAO8E,EAAQ+E,GAQtE,OAPA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,WAAY,GAC5DtF,KAAKsF,EAAS,GAAM9E,IAAU,GAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,GAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,EAC9BR,KAAKsF,GAAmB,IAAR9E,EACT8E,EAAS,CAClB,EAEA7F,OAAOC,UAAU6N,cACjB9N,OAAOC,UAAU8N,cAAgB,SAAwBhN,EAAO8E,EAAQ+E,GAQtE,OAPA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,WAAY,GAC5DtF,KAAKsF,GAAW9E,IAAU,GAC1BR,KAAKsF,EAAS,GAAM9E,IAAU,GAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,EAC9BR,KAAKsF,EAAS,GAAc,IAAR9E,EACb8E,EAAS,CAClB,EA8CA7F,OAAOC,UAAU+N,iBAAmB,SAA2BjN,EAAO8E,EAAS,GAC7E,OAAOwE,EAAe9J,KAAMQ,EAAO8E,EAAQ0E,OAAO,GAAIA,OAAO,sBAC/D,EAEAvK,OAAOC,UAAUgO,iBAAmB,SAA2BlN,EAAO8E,EAAS,GAC7E,OAAO2E,EAAejK,KAAMQ,EAAO8E,EAAQ0E,OAAO,GAAIA,OAAO,sBAC/D,EAEAvK,OAAOC,UAAUiO,WAAa,SAAqBnN,EAAO8E,EAAQ1E,EAAYyJ,GAG5E,GAFA7J,GAASA,EACT8E,KAAoB,GACf+E,EAAU,CACb,MAAMuD,EAAQhH,KAAKkF,IAAI,EAAI,EAAIlL,EAAc,GAE7CiJ,EAAS7J,KAAMQ,EAAO8E,EAAQ1E,EAAYgN,EAAQ,GAAIA,EACxD,CAEA,IAAIjL,EAAI,EACJgI,EAAM,EACNkD,EAAM,EAEV,IADA7N,KAAKsF,GAAkB,IAAR9E,IACNmC,EAAI/B,IAAe+J,GAAO,MAC7BnK,EAAQ,GAAa,IAARqN,GAAsC,IAAzB7N,KAAKsF,EAAS3C,EAAI,KAC9CkL,EAAM,GAER7N,KAAKsF,EAAS3C,IAAOnC,EAAQmK,GAAQ,GAAKkD,EAAM,IAGlD,OAAOvI,EAAS1E,CAClB,EAEAnB,OAAOC,UAAUoO,WAAa,SAAqBtN,EAAO8E,EAAQ1E,EAAYyJ,GAG5E,GAFA7J,GAASA,EACT8E,KAAoB,GACf+E,EAAU,CACb,MAAMuD,EAAQhH,KAAKkF,IAAI,EAAI,EAAIlL,EAAc,GAE7CiJ,EAAS7J,KAAMQ,EAAO8E,EAAQ1E,EAAYgN,EAAQ,GAAIA,EACxD,CAEA,IAAIjL,EAAI/B,EAAa,EACjB+J,EAAM,EACNkD,EAAM,EAEV,IADA7N,KAAKsF,EAAS3C,GAAa,IAARnC,IACVmC,GAAK,IAAMgI,GAAO,MACrBnK,EAAQ,GAAa,IAARqN,GAAsC,IAAzB7N,KAAKsF,EAAS3C,EAAI,KAC9CkL,EAAM,GAER7N,KAAKsF,EAAS3C,IAAOnC,EAAQmK,GAAQ,GAAKkD,EAAM,IAGlD,OAAOvI,EAAS1E,CAClB,EAEAnB,OAAOC,UAAUqO,UAAY,SAAoBvN,EAAO8E,EAAQ+E,GAM9D,OALA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,KAAO,KACnD9E,EAAQ,IAAGA,EAAQ,IAAOA,EAAQ,GACtCR,KAAKsF,GAAmB,IAAR9E,EACT8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUsO,aAAe,SAAuBxN,EAAO8E,EAAQ+E,GAMpE,OALA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,OAAS,OACzDtF,KAAKsF,GAAmB,IAAR9E,EAChBR,KAAKsF,EAAS,GAAM9E,IAAU,EACvB8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUuO,aAAe,SAAuBzN,EAAO8E,EAAQ+E,GAMpE,OALA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,OAAS,OACzDtF,KAAKsF,GAAW9E,IAAU,EAC1BR,KAAKsF,EAAS,GAAc,IAAR9E,EACb8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUwO,aAAe,SAAuB1N,EAAO8E,EAAQ+E,GAQpE,OAPA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,YAAa,YAC7DtF,KAAKsF,GAAmB,IAAR9E,EAChBR,KAAKsF,EAAS,GAAM9E,IAAU,EAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,GAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,GACvB8E,EAAS,CAClB,EAEA7F,OAAOC,UAAUyO,aAAe,SAAuB3N,EAAO8E,EAAQ+E,GASpE,OARA7J,GAASA,EACT8E,KAAoB,EACf+E,GAAUR,EAAS7J,KAAMQ,EAAO8E,EAAQ,EAAG,YAAa,YACzD9E,EAAQ,IAAGA,EAAQ,WAAaA,EAAQ,GAC5CR,KAAKsF,GAAW9E,IAAU,GAC1BR,KAAKsF,EAAS,GAAM9E,IAAU,GAC9BR,KAAKsF,EAAS,GAAM9E,IAAU,EAC9BR,KAAKsF,EAAS,GAAc,IAAR9E,EACb8E,EAAS,CAClB,EAEA7F,OAAOC,UAAU0O,gBAAkB,SAA0B5N,EAAO8E,EAAS,GAC3E,OAAOwE,EAAe9J,KAAMQ,EAAO8E,GAAS0E,OAAO,sBAAuBA,OAAO,sBACnF,EAEAvK,OAAOC,UAAU2O,gBAAkB,SAA0B7N,EAAO8E,EAAS,GAC3E,OAAO2E,EAAejK,KAAMQ,EAAO8E,GAAS0E,OAAO,sBAAuBA,OAAO,sBACnF,EAiBAvK,OAAOC,UAAU4O,aAAe,SAAuB9N,EAAO8E,EAAQ+E,GACpE,OAAOF,EAAWnK,KAAMQ,EAAO8E,GAAQ,EAAM+E,EAC/C,EAEA5K,OAAOC,UAAU6O,aAAe,SAAuB/N,EAAO8E,EAAQ+E,GACpE,OAAOF,EAAWnK,KAAMQ,EAAO8E,GAAQ,EAAO+E,EAChD,EAYA5K,OAAOC,UAAU8O,cAAgB,SAAwBhO,EAAO8E,EAAQ+E,GACtE,OAAOC,EAAYtK,KAAMQ,EAAO8E,GAAQ,EAAM+E,EAChD,EAEA5K,OAAOC,UAAU+O,cAAgB,SAAwBjO,EAAO8E,EAAQ+E,GACtE,OAAOC,EAAYtK,KAAMQ,EAAO8E,GAAQ,EAAO+E,EACjD,EAGA5K,OAAOC,UAAU0B,KAAO,SAAeyH,EAAQ6F,EAAajL,EAAOC,GACjE,IAAKjE,OAAOM,SAAS8I,GAAS,MAAM,IAAIxI,UAAU,+BAQlD,GAPKoD,IAAOA,EAAQ,GACfC,GAAe,IAARA,IAAWA,EAAM1D,KAAKb,QAC9BuP,GAAe7F,EAAO1J,SAAQuP,EAAc7F,EAAO1J,QAClDuP,IAAaA,EAAc,GAC5BhL,EAAM,GAAKA,EAAMD,IAAOC,EAAMD,GAG9BC,IAAQD,EAAO,OAAO,EAC1B,GAAsB,IAAlBoF,EAAO1J,QAAgC,IAAhBa,KAAKb,OAAc,OAAO,EAGrD,GAAIuP,EAAc,EAChB,MAAM,IAAItP,WAAW,6BAEvB,GAAIqE,EAAQ,GAAKA,GAASzD,KAAKb,OAAQ,MAAM,IAAIC,WAAW,sBAC5D,GAAIsE,EAAM,EAAG,MAAM,IAAItE,WAAW,2BAG9BsE,EAAM1D,KAAKb,SAAQuE,EAAM1D,KAAKb,QAC9B0J,EAAO1J,OAASuP,EAAchL,EAAMD,IACtCC,EAAMmF,EAAO1J,OAASuP,EAAcjL,GAGtC,MAAM7B,EAAM8B,EAAMD,EAYlB,OAVIzD,OAAS6I,EACX7I,KAAK2O,WAAWD,EAAajL,EAAOC,GAEpCpE,WAAWI,UAAUwI,IAAIzD,KACvBoE,EACA7I,KAAKwK,SAAS/G,EAAOC,GACrBgL,GAIG9M,CACT,EAMAnC,OAAOC,UAAUqD,KAAO,SAAesB,EAAKZ,EAAOC,EAAKhD,GAEtD,GAAmB,iBAAR2D,EAAkB,CAS3B,GARqB,iBAAVZ,GACT/C,EAAW+C,EACXA,EAAQ,EACRC,EAAM1D,KAAKb,QACa,iBAARuE,IAChBhD,EAAWgD,EACXA,EAAM1D,KAAKb,aAEI2C,IAAbpB,GAA8C,iBAAbA,EACnC,MAAM,IAAIL,UAAU,6BAEtB,GAAwB,iBAAbK,IAA0BjB,OAAOkB,WAAWD,GACrD,MAAM,IAAIL,UAAU,qBAAuBK,GAE7C,GAAmB,IAAf2D,EAAIlF,OAAc,CACpB,MAAMyP,EAAOvK,EAAI6B,WAAW,IACV,SAAbxF,GAAuBkO,EAAO,KAClB,WAAblO,KAEF2D,EAAMuK,EAEV,CACF,KAA0B,iBAARvK,EAChBA,GAAY,IACY,kBAARA,IAChBA,EAAMtC,OAAOsC,IAIf,GAAIZ,EAAQ,GAAKzD,KAAKb,OAASsE,GAASzD,KAAKb,OAASuE,EACpD,MAAM,IAAItE,WAAW,sBAGvB,GAAIsE,GAAOD,EACT,OAAOzD,KAQT,IAAI2C,EACJ,GANAc,KAAkB,EAClBC,OAAc5B,IAAR4B,EAAoB1D,KAAKb,OAASuE,IAAQ,EAE3CW,IAAKA,EAAM,GAGG,iBAARA,EACT,IAAK1B,EAAIc,EAAOd,EAAIe,IAAOf,EACzB3C,KAAK2C,GAAK0B,MAEP,CACL,MAAMqF,EAAQjK,OAAOM,SAASsE,GAC1BA,EACA5E,OAAOc,KAAK8D,EAAK3D,GACfkB,EAAM8H,EAAMvK,OAClB,GAAY,IAARyC,EACF,MAAM,IAAIvB,UAAU,cAAgBgE,EAClC,qCAEJ,IAAK1B,EAAI,EAAGA,EAAIe,EAAMD,IAASd,EAC7B3C,KAAK2C,EAAIc,GAASiG,EAAM/G,EAAIf,EAEhC,CAEA,OAAO5B,IACT,EAMA,MAAM6O,EAAS,CAAC,EAChB,SAASC,EAAGC,EAAKC,EAAYC,GAC3BJ,EAAOE,GAAO,cAAwBE,EACpCC,cACEC,QAEA5P,OAAOK,eAAeI,KAAM,UAAW,CACrCQ,MAAOwO,EAAWxH,MAAMxH,KAAMmD,WAC9BiM,UAAU,EACVC,cAAc,IAIhBrP,KAAKsP,KAAO,GAAGtP,KAAKsP,SAASP,KAG7B/O,KAAKuP,aAEEvP,KAAKsP,IACd,CAEIV,WACF,OAAOG,CACT,CAEIH,SAAMpO,GACRjB,OAAOK,eAAeI,KAAM,OAAQ,CAClCqP,cAAc,EACdxP,YAAY,EACZW,QACA4O,UAAU,GAEd,CAEAxM,WACE,MAAO,GAAG5C,KAAKsP,SAASP,OAAS/O,KAAKwP,SACxC,EAEJ,CA+BA,SAASC,EAAuBpL,GAC9B,IAAIyC,EAAM,GACNnE,EAAI0B,EAAIlF,OACZ,MAAMsE,EAAmB,MAAXY,EAAI,GAAa,EAAI,EACnC,KAAO1B,GAAKc,EAAQ,EAAGd,GAAK,EAC1BmE,EAAM,IAAIzC,EAAItD,MAAM4B,EAAI,EAAGA,KAAKmE,IAElC,MAAO,GAAGzC,EAAItD,MAAM,EAAG4B,KAAKmE,GAC9B,CAYA,SAASiD,EAAYvJ,EAAOqG,EAAK4B,EAAKpJ,EAAKiG,EAAQ1E,GACjD,GAAIJ,EAAQiI,GAAOjI,EAAQqG,EAAK,CAC9B,MAAM3C,EAAmB,iBAAR2C,EAAmB,IAAM,GAC1C,IAAI6I,EAWJ,MARIA,EAFA9O,EAAa,EACH,IAARiG,GAAaA,IAAQmD,OAAO,GACtB,OAAO9F,YAAYA,QAA2B,GAAlBtD,EAAa,KAASsD,IAElD,SAASA,QAA2B,GAAlBtD,EAAa,GAAS,IAAIsD,iBACtB,GAAlBtD,EAAa,GAAS,IAAIsD,IAGhC,MAAM2C,IAAM3C,YAAYuE,IAAMvE,IAElC,IAAI2K,EAAOc,iBAAiB,QAASD,EAAOlP,EACpD,EAtBF,SAAsBnB,EAAKiG,EAAQ1E,GACjC4K,EAAelG,EAAQ,eACHxD,IAAhBzC,EAAIiG,SAAsDxD,IAA7BzC,EAAIiG,EAAS1E,IAC5C+K,EAAYrG,EAAQjG,EAAIF,QAAUyB,EAAa,GAEnD,CAkBEgP,CAAYvQ,EAAKiG,EAAQ1E,EAC3B,CAEA,SAAS4K,EAAgBhL,EAAO8O,GAC9B,GAAqB,iBAAV9O,EACT,MAAM,IAAIqO,EAAOgB,qBAAqBP,EAAM,SAAU9O,EAE1D,CAEA,SAASmL,EAAanL,EAAOrB,EAAQ8C,GACnC,GAAI2E,KAAKkJ,MAAMtP,KAAWA,EAExB,MADAgL,EAAehL,EAAOyB,GAChB,IAAI4M,EAAOc,iBAAiB1N,GAAQ,SAAU,aAAczB,GAGpE,GAAIrB,EAAS,EACX,MAAM,IAAI0P,EAAOkB,yBAGnB,MAAM,IAAIlB,EAAOc,iBAAiB1N,GAAQ,SACR,MAAMA,EAAO,EAAI,YAAY9C,IAC7BqB,EACpC,CAvFAsO,EAAE,4BACA,SAAUQ,GACR,OAAIA,EACK,GAAGA,gCAGL,gDACT,GAAGlQ,YACL0P,EAAE,wBACA,SAAUQ,EAAMzO,GACd,MAAO,QAAQyO,4DAA+DzO,GAChF,GAAGR,WACLyO,EAAE,oBACA,SAAU/I,EAAK2J,EAAOM,GACpB,IAAIC,EAAM,iBAAiBlK,sBACvBmK,EAAWF,EAWf,OAVIjO,OAAOoO,UAAUH,IAAUpJ,KAAKwJ,IAAIJ,GAAS,GAAK,GACpDE,EAAWT,EAAsB1K,OAAOiL,IACd,iBAAVA,IAChBE,EAAWnL,OAAOiL,IACdA,EAAQhG,OAAO,IAAMA,OAAO,KAAOgG,IAAUhG,OAAO,IAAMA,OAAO,QACnEkG,EAAWT,EAAsBS,IAEnCA,GAAY,KAEdD,GAAO,eAAeP,eAAmBQ,IAClCD,CACT,GAAG7Q,YAiEL,MAAMiR,EAAoB,oBAgB1B,SAAShN,EAAa5C,EAAQ6F,GAE5B,IAAIU,EADJV,EAAQA,GAASgK,IAEjB,MAAMnR,EAASsB,EAAOtB,OACtB,IAAIoR,EAAgB,KACpB,MAAM7G,EAAQ,GAEd,IAAK,IAAI/G,EAAI,EAAGA,EAAIxD,IAAUwD,EAAG,CAI/B,GAHAqE,EAAYvG,EAAOyF,WAAWvD,GAG1BqE,EAAY,OAAUA,EAAY,MAAQ,CAE5C,IAAKuJ,EAAe,CAElB,GAAIvJ,EAAY,MAAQ,EAEjBV,GAAS,IAAM,GAAGoD,EAAMzD,KAAK,IAAM,IAAM,KAC9C,QACF,CAAO,GAAItD,EAAI,IAAMxD,EAAQ,EAEtBmH,GAAS,IAAM,GAAGoD,EAAMzD,KAAK,IAAM,IAAM,KAC9C,QACF,CAGAsK,EAAgBvJ,EAEhB,QACF,CAGA,GAAIA,EAAY,MAAQ,EACjBV,GAAS,IAAM,GAAGoD,EAAMzD,KAAK,IAAM,IAAM,KAC9CsK,EAAgBvJ,EAChB,QACF,CAGAA,EAAkE,OAArDuJ,EAAgB,OAAU,GAAKvJ,EAAY,MAC1D,MAAWuJ,IAEJjK,GAAS,IAAM,GAAGoD,EAAMzD,KAAK,IAAM,IAAM,KAMhD,GAHAsK,EAAgB,KAGZvJ,EAAY,IAAM,CACpB,IAAKV,GAAS,GAAK,EAAG,MACtBoD,EAAMzD,KAAKe,EACb,MAAO,GAAIA,EAAY,KAAO,CAC5B,IAAKV,GAAS,GAAK,EAAG,MACtBoD,EAAMzD,KACJe,GAAa,EAAM,IACP,GAAZA,EAAmB,IAEvB,MAAO,GAAIA,EAAY,MAAS,CAC9B,IAAKV,GAAS,GAAK,EAAG,MACtBoD,EAAMzD,KACJe,GAAa,GAAM,IACnBA,GAAa,EAAM,GAAO,IACd,GAAZA,EAAmB,IAEvB,KAAO,MAAIA,EAAY,SASrB,MAAM,IAAIoC,MAAM,sBARhB,IAAK9C,GAAS,GAAK,EAAG,MACtBoD,EAAMzD,KACJe,GAAa,GAAO,IACpBA,GAAa,GAAM,GAAO,IAC1BA,GAAa,EAAM,GAAO,IACd,GAAZA,EAAmB,IAIvB,CACF,CAEA,OAAO0C,CACT,CA2BA,SAASpG,EAAeyC,GACtB,OAAOlH,EAAO2R,YAxHhB,SAAsBzK,GAMpB,IAFAA,GAFAA,EAAMA,EAAI0K,MAAM,KAAK,IAEX9H,OAAOD,QAAQ2H,EAAmB,KAEpClR,OAAS,EAAG,MAAO,GAE3B,KAAO4G,EAAI5G,OAAS,GAAM,GACxB4G,GAAY,IAEd,OAAOA,CACT,CA4G4B2K,CAAY3K,GACxC,CAEA,SAASF,EAAY8K,EAAKC,EAAKtL,EAAQnG,GACrC,IAAIwD,EACJ,IAAKA,EAAI,EAAGA,EAAIxD,KACTwD,EAAI2C,GAAUsL,EAAIzR,QAAYwD,GAAKgO,EAAIxR,UADpBwD,EAExBiO,EAAIjO,EAAI2C,GAAUqL,EAAIhO,GAExB,OAAOA,CACT,CAIA,MAAM8G,EAAsB,WAC1B,MAAMoH,EAAW,mBACXC,EAAQ,IAAI5O,MAAM,KACxB,IAAK,IAAIS,EAAI,EAAGA,EAAI,KAAMA,EAAG,CAC3B,MAAMoO,EAAU,GAAJpO,EACZ,IAAK,IAAIyC,EAAI,EAAGA,EAAI,KAAMA,EACxB0L,EAAMC,EAAM3L,GAAKyL,EAASlO,GAAKkO,EAASzL,EAE5C,CACA,OAAO0L,CACR,CAV2B,iBAYb,CACb/R,OACAiS,WA7+DmB,WA8+DnBvR,OACAwD"}
✄
/*!
 * The buffer module from node.js, for Frida.
 *
 * @author   Feross Aboukhadijeh <https://feross.org>
 * @license  MIT
 */
import*as t from"base64-js";import*as e from"ieee754";export const config={INSPECT_MAX_BYTES:50};const r=2147483647;export{r as kMaxLength};function n(t){if(t>2147483647)throw new RangeError('The value "'+t+'" is invalid for option "size"');const e=new Uint8Array(t);return Object.setPrototypeOf(e,Buffer.prototype),e}Buffer.TYPED_ARRAY_SUPPORT=!0,Object.defineProperty(Buffer.prototype,"parent",{enumerable:!0,get:function(){if(Buffer.isBuffer(this))return this.buffer}}),Object.defineProperty(Buffer.prototype,"offset",{enumerable:!0,get:function(){if(Buffer.isBuffer(this))return this.byteOffset}});export function Buffer(t,e,r){if("number"==typeof t){if("string"==typeof e)throw new TypeError('The "string" argument must be of type string. Received type number');return o(t)}return f(t,e,r)}function f(t,e,r){if("string"==typeof t)return function(t,e){"string"==typeof e&&""!==e||(e="utf8");if(!Buffer.isEncoding(e))throw new TypeError("Unknown encoding: "+e);const r=0|a(t,e);let f=n(r);const i=f.write(t,e);i!==r&&(f=f.slice(0,i));return f}(t,e);if(ArrayBuffer.isView(t))return function(t){if(t instanceof Uint8Array){const e=new Uint8Array(t);return s(e.buffer,e.byteOffset,e.byteLength)}return u(t)}(t);if(null==t)throw new TypeError("The first argument must be one of type string, Buffer, ArrayBuffer, Array, or Array-like Object. Received type "+typeof t);if(t instanceof ArrayBuffer||t&&t.buffer instanceof ArrayBuffer)return s(t,e,r);if(t instanceof SharedArrayBuffer||t&&t.buffer instanceof SharedArrayBuffer)return s(t,e,r);if("number"==typeof t)throw new TypeError('The "value" argument must not be of type number. Received type number');const f=t.valueOf&&t.valueOf();if(null!=f&&f!==t)return Buffer.from(f,e,r);const i=function(t){if(Buffer.isBuffer(t)){const e=0|h(t.length),r=n(e);return 0===r.length||t.copy(r,0,0,e),r}if(void 0!==t.length)return"number"!=typeof t.length||Number.isNaN(t.length)?n(0):u(t);if("Buffer"===t.type&&Array.isArray(t.data))return u(t.data)}(t);if(i)return i;if("undefined"!=typeof Symbol&&null!=Symbol.toPrimitive&&"function"==typeof t[Symbol.toPrimitive])return Buffer.from(t[Symbol.toPrimitive]("string"),e,r);throw new TypeError("The first argument must be one of type string, Buffer, ArrayBuffer, Array, or Array-like Object. Received type "+typeof t)}function i(t){if("number"!=typeof t)throw new TypeError('"size" argument must be of type number');if(t<0)throw new RangeError('The value "'+t+'" is invalid for option "size"')}function o(t){return i(t),n(t<0?0:0|h(t))}function u(t){const e=t.length<0?0:0|h(t.length),r=n(e);for(let n=0;n<e;n+=1)r[n]=255&t[n];return r}function s(t,e,r){if(e<0||t.byteLength<e)throw new RangeError('"offset" is outside of buffer bounds');if(t.byteLength<e+(r||0))throw new RangeError('"length" is outside of buffer bounds');let n;return n=void 0===e&&void 0===r?new Uint8Array(t):void 0===r?new Uint8Array(t,e):new Uint8Array(t,e,r),Object.setPrototypeOf(n,Buffer.prototype),n}function h(t){if(t>=2147483647)throw new RangeError("Attempt to allocate Buffer larger than maximum size: 0x"+2147483647..toString(16)+" bytes");return 0|t}Buffer.poolSize=8192,Buffer.from=function(t,e,r){return f(t,e,r)},Object.setPrototypeOf(Buffer.prototype,Uint8Array.prototype),Object.setPrototypeOf(Buffer,Uint8Array),Buffer.alloc=function(t,e,r){return function(t,e,r){return i(t),t<=0?n(t):void 0!==e?"string"==typeof r?n(t).fill(e,r):n(t).fill(e):n(t)}(t,e,r)},Buffer.allocUnsafe=function(t){return o(t)},Buffer.allocUnsafeSlow=function(t){return o(t)};export function SlowBuffer(t){return+t!=t&&(t=0),Buffer.alloc(+t)}function a(t,e){if(Buffer.isBuffer(t))return t.length;if(ArrayBuffer.isView(t)||t instanceof ArrayBuffer)return t.byteLength;if("string"!=typeof t)throw new TypeError('The "string" argument must be one of type string, Buffer, or ArrayBuffer. Received type '+typeof t);const r=t.length,n=arguments.length>2&&!0===arguments[2];if(!n&&0===r)return 0;let f=!1;for(;;)switch(e){case"ascii":case"latin1":case"binary":return r;case"utf8":case"utf-8":return j(t).length;case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return 2*r;case"hex":return r>>>1;case"base64":return z(t).length;default:if(f)return n?-1:j(t).length;e=(""+e).toLowerCase(),f=!0}}function c(t,e,r){let n=!1;if((void 0===e||e<0)&&(e=0),e>this.length)return"";if((void 0===r||r>this.length)&&(r=this.length),r<=0)return"";if((r>>>=0)<=(e>>>=0))return"";for(t||(t="utf8");;)switch(t){case"hex":return A(this,e,r);case"utf8":case"utf-8":return m(this,e,r);case"ascii":return I(this,e,r);case"latin1":case"binary":return U(this,e,r);case"base64":return E(this,e,r);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return v(this,e,r);default:if(n)throw new TypeError("Unknown encoding: "+t);t=(t+"").toLowerCase(),n=!0}}function l(t,e,r){const n=t[e];t[e]=t[r],t[r]=n}function p(t,e,r,n,f){if(0===t.length)return-1;if("string"==typeof r?(n=r,r=0):r>2147483647?r=2147483647:r<-2147483648&&(r=-2147483648),r=+r,Number.isNaN(r)&&(r=f?0:t.length-1),r<0&&(r=t.length+r),r>=t.length){if(f)return-1;r=t.length-1}else if(r<0){if(!f)return-1;r=0}if("string"==typeof e&&(e=Buffer.from(e,n)),Buffer.isBuffer(e))return 0===e.length?-1:g(t,e,r,n,f);if("number"==typeof e)return e&=255,"function"==typeof Uint8Array.prototype.indexOf?f?Uint8Array.prototype.indexOf.call(t,e,r):Uint8Array.prototype.lastIndexOf.call(t,e,r):g(t,[e],r,n,f);throw new TypeError("val must be string, number or Buffer")}function g(t,e,r,n,f){let i,o=1,u=t.length,s=e.length;if(void 0!==n&&("ucs2"===(n=String(n).toLowerCase())||"ucs-2"===n||"utf16le"===n||"utf-16le"===n)){if(t.length<2||e.length<2)return-1;o=2,u/=2,s/=2,r/=2}function h(t,e){return 1===o?t[e]:t.readUInt16BE(e*o)}if(f){let n=-1;for(i=r;i<u;i++)if(h(t,i)===h(e,-1===n?0:i-n)){if(-1===n&&(n=i),i-n+1===s)return n*o}else-1!==n&&(i-=i-n),n=-1}else for(r+s>u&&(r=u-s),i=r;i>=0;i--){let r=!0;for(let n=0;n<s;n++)if(h(t,i+n)!==h(e,n)){r=!1;break}if(r)return i}return-1}function y(t,e,r,n){r=Number(r)||0;const f=t.length-r;n?(n=Number(n))>f&&(n=f):n=f;const i=e.length;let o;for(n>i/2&&(n=i/2),o=0;o<n;++o){const n=parseInt(e.substr(2*o,2),16);if(Number.isNaN(n))return o;t[r+o]=n}return o}function B(t,e,r,n){return D(j(e,t.length-r),t,r,n)}function w(t,e,r,n){return D(function(t){const e=[];for(let r=0;r<t.length;++r)e.push(255&t.charCodeAt(r));return e}(e),t,r,n)}function d(t,e,r,n){return D(z(e),t,r,n)}function b(t,e,r,n){return D(function(t,e){let r,n,f;const i=[];for(let o=0;o<t.length&&!((e-=2)<0);++o)r=t.charCodeAt(o),n=r>>8,f=r%256,i.push(f),i.push(n);return i}(e,t.length-r),t,r,n)}function E(e,r,n){return 0===r&&n===e.length?t.fromByteArray(e):t.fromByteArray(e.slice(r,n))}function m(t,e,r){r=Math.min(t.length,r);const n=[];let f=e;for(;f<r;){const e=t[f];let i=null,o=e>239?4:e>223?3:e>191?2:1;if(f+o<=r){let r,n,u,s;switch(o){case 1:e<128&&(i=e);break;case 2:r=t[f+1],128==(192&r)&&(s=(31&e)<<6|63&r,s>127&&(i=s));break;case 3:r=t[f+1],n=t[f+2],128==(192&r)&&128==(192&n)&&(s=(15&e)<<12|(63&r)<<6|63&n,s>2047&&(s<55296||s>57343)&&(i=s));break;case 4:r=t[f+1],n=t[f+2],u=t[f+3],128==(192&r)&&128==(192&n)&&128==(192&u)&&(s=(15&e)<<18|(63&r)<<12|(63&n)<<6|63&u,s>65535&&s<1114112&&(i=s))}}null===i?(i=65533,o=1):i>65535&&(i-=65536,n.push(i>>>10&1023|55296),i=56320|1023&i),n.push(i),f+=o}return function(t){const e=t.length;if(e<=4096)return String.fromCharCode.apply(String,t);let r="",n=0;for(;n<e;)r+=String.fromCharCode.apply(String,t.slice(n,n+=4096));return r}(n)}Buffer.isBuffer=function(t){return null!=t&&!0===t._isBuffer&&t!==Buffer.prototype},Buffer.compare=function(t,e){if(t instanceof Uint8Array&&(t=Buffer.from(t,t.offset,t.byteLength)),e instanceof Uint8Array&&(e=Buffer.from(e,e.offset,e.byteLength)),!Buffer.isBuffer(t)||!Buffer.isBuffer(e))throw new TypeError('The "buf1", "buf2" arguments must be one of type Buffer or Uint8Array');if(t===e)return 0;let r=t.length,n=e.length;for(let f=0,i=Math.min(r,n);f<i;++f)if(t[f]!==e[f]){r=t[f],n=e[f];break}return r<n?-1:n<r?1:0},Buffer.isEncoding=function(t){switch(String(t).toLowerCase()){case"hex":case"utf8":case"utf-8":case"ascii":case"latin1":case"binary":case"base64":case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return!0;default:return!1}},Buffer.concat=function(t,e){if(!Array.isArray(t))throw new TypeError('"list" argument must be an Array of Buffers');if(0===t.length)return Buffer.alloc(0);let r;if(void 0===e)for(e=0,r=0;r<t.length;++r)e+=t[r].length;const n=Buffer.allocUnsafe(e);let f=0;for(r=0;r<t.length;++r){let e=t[r];if(e instanceof Uint8Array)f+e.length>n.length?(Buffer.isBuffer(e)||(e=Buffer.from(e.buffer,e.byteOffset,e.byteLength)),e.copy(n,f)):Uint8Array.prototype.set.call(n,e,f);else{if(!Buffer.isBuffer(e))throw new TypeError('"list" argument must be an Array of Buffers');e.copy(n,f)}f+=e.length}return n},Buffer.byteLength=a,Buffer.prototype._isBuffer=!0,Buffer.prototype.swap16=function(){const t=this.length;if(t%2!=0)throw new RangeError("Buffer size must be a multiple of 16-bits");for(let e=0;e<t;e+=2)l(this,e,e+1);return this},Buffer.prototype.swap32=function(){const t=this.length;if(t%4!=0)throw new RangeError("Buffer size must be a multiple of 32-bits");for(let e=0;e<t;e+=4)l(this,e,e+3),l(this,e+1,e+2);return this},Buffer.prototype.swap64=function(){const t=this.length;if(t%8!=0)throw new RangeError("Buffer size must be a multiple of 64-bits");for(let e=0;e<t;e+=8)l(this,e,e+7),l(this,e+1,e+6),l(this,e+2,e+5),l(this,e+3,e+4);return this},Buffer.prototype.toString=function(){const t=this.length;return 0===t?"":0===arguments.length?m(this,0,t):c.apply(this,arguments)},Buffer.prototype.toLocaleString=Buffer.prototype.toString,Buffer.prototype.equals=function(t){if(!Buffer.isBuffer(t))throw new TypeError("Argument must be a Buffer");return this===t||0===Buffer.compare(this,t)},Buffer.prototype.inspect=function(){let t="";const e=config.INSPECT_MAX_BYTES;return t=this.toString("hex",0,e).replace(/(.{2})/g,"$1 ").trim(),this.length>e&&(t+=" ... "),"<Buffer "+t+">"},Buffer.prototype[Symbol.for("nodejs.util.inspect.custom")]=Buffer.prototype.inspect,Buffer.prototype.compare=function(t,e,r,n,f){if(t instanceof Uint8Array&&(t=Buffer.from(t,t.offset,t.byteLength)),!Buffer.isBuffer(t))throw new TypeError('The "target" argument must be one of type Buffer or Uint8Array. Received type '+typeof t);if(void 0===e&&(e=0),void 0===r&&(r=t?t.length:0),void 0===n&&(n=0),void 0===f&&(f=this.length),e<0||r>t.length||n<0||f>this.length)throw new RangeError("out of range index");if(n>=f&&e>=r)return 0;if(n>=f)return-1;if(e>=r)return 1;if(this===t)return 0;let i=(f>>>=0)-(n>>>=0),o=(r>>>=0)-(e>>>=0);const u=Math.min(i,o),s=this.slice(n,f),h=t.slice(e,r);for(let t=0;t<u;++t)if(s[t]!==h[t]){i=s[t],o=h[t];break}return i<o?-1:o<i?1:0},Buffer.prototype.includes=function(t,e,r){return-1!==this.indexOf(t,e,r)},Buffer.prototype.indexOf=function(t,e,r){return p(this,t,e,r,!0)},Buffer.prototype.lastIndexOf=function(t,e,r){return p(this,t,e,r,!1)},Buffer.prototype.write=function(t,e,r,n){if(void 0===e)n="utf8",r=this.length,e=0;else if(void 0===r&&"string"==typeof e)n=e,r=this.length,e=0;else{if(!isFinite(e))throw new Error("Buffer.write(string, encoding, offset[, length]) is no longer supported");e>>>=0,isFinite(r)?(r>>>=0,void 0===n&&(n="utf8")):(n=r,r=void 0)}const f=this.length-e;if((void 0===r||r>f)&&(r=f),t.length>0&&(r<0||e<0)||e>this.length)throw new RangeError("Attempt to write outside buffer bounds");n||(n="utf8");let i=!1;for(;;)switch(n){case"hex":return y(this,t,e,r);case"utf8":case"utf-8":return B(this,t,e,r);case"ascii":case"latin1":case"binary":return w(this,t,e,r);case"base64":return d(this,t,e,r);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return b(this,t,e,r);default:if(i)throw new TypeError("Unknown encoding: "+n);n=(""+n).toLowerCase(),i=!0}},Buffer.prototype.toJSON=function(){return{type:"Buffer",data:Array.prototype.slice.call(this._arr||this,0)}};function I(t,e,r){let n="";r=Math.min(t.length,r);for(let f=e;f<r;++f)n+=String.fromCharCode(127&t[f]);return n}function U(t,e,r){let n="";r=Math.min(t.length,r);for(let f=e;f<r;++f)n+=String.fromCharCode(t[f]);return n}function A(t,e,r){const n=t.length;(!e||e<0)&&(e=0),(!r||r<0||r>n)&&(r=n);let f="";for(let n=e;n<r;++n)f+=G[t[n]];return f}function v(t,e,r){const n=t.slice(e,r);let f="";for(let t=0;t<n.length-1;t+=2)f+=String.fromCharCode(n[t]+256*n[t+1]);return f}function R(t,e,r){if(t%1!=0||t<0)throw new RangeError("offset is not uint");if(t+e>r)throw new RangeError("Trying to access beyond buffer length")}function T(t,e,r,n,f,i){if(!Buffer.isBuffer(t))throw new TypeError('"buffer" argument must be a Buffer instance');if(e>f||e<i)throw new RangeError('"value" argument is out of bounds');if(r+n>t.length)throw new RangeError("Index out of range")}function L(t,e,r,n,f){k(e,n,f,t,r,7);let i=Number(e&BigInt(4294967295));t[r++]=i,i>>=8,t[r++]=i,i>>=8,t[r++]=i,i>>=8,t[r++]=i;let o=Number(e>>BigInt(32)&BigInt(4294967295));return t[r++]=o,o>>=8,t[r++]=o,o>>=8,t[r++]=o,o>>=8,t[r++]=o,r}function O(t,e,r,n,f){k(e,n,f,t,r,7);let i=Number(e&BigInt(4294967295));t[r+7]=i,i>>=8,t[r+6]=i,i>>=8,t[r+5]=i,i>>=8,t[r+4]=i;let o=Number(e>>BigInt(32)&BigInt(4294967295));return t[r+3]=o,o>>=8,t[r+2]=o,o>>=8,t[r+1]=o,o>>=8,t[r]=o,r+8}function S(t,e,r,n,f,i){if(r+n>t.length)throw new RangeError("Index out of range");if(r<0)throw new RangeError("Index out of range")}function _(t,r,n,f,i){return r=+r,n>>>=0,i||S(t,0,n,4),e.write(t,r,n,f,23,4),n+4}function x(t,r,n,f,i){return r=+r,n>>>=0,i||S(t,0,n,8),e.write(t,r,n,f,52,8),n+8}Buffer.prototype.slice=function(t,e){const r=this.length;(t=~~t)<0?(t+=r)<0&&(t=0):t>r&&(t=r),(e=void 0===e?r:~~e)<0?(e+=r)<0&&(e=0):e>r&&(e=r),e<t&&(e=t);const n=this.subarray(t,e);return Object.setPrototypeOf(n,Buffer.prototype),n},Buffer.prototype.readUintLE=Buffer.prototype.readUIntLE=function(t,e,r){t>>>=0,e>>>=0,r||R(t,e,this.length);let n=this[t],f=1,i=0;for(;++i<e&&(f*=256);)n+=this[t+i]*f;return n},Buffer.prototype.readUintBE=Buffer.prototype.readUIntBE=function(t,e,r){t>>>=0,e>>>=0,r||R(t,e,this.length);let n=this[t+--e],f=1;for(;e>0&&(f*=256);)n+=this[t+--e]*f;return n},Buffer.prototype.readUint8=Buffer.prototype.readUInt8=function(t,e){return t>>>=0,e||R(t,1,this.length),this[t]},Buffer.prototype.readUint16LE=Buffer.prototype.readUInt16LE=function(t,e){return t>>>=0,e||R(t,2,this.length),this[t]|this[t+1]<<8},Buffer.prototype.readUint16BE=Buffer.prototype.readUInt16BE=function(t,e){return t>>>=0,e||R(t,2,this.length),this[t]<<8|this[t+1]},Buffer.prototype.readUint32LE=Buffer.prototype.readUInt32LE=function(t,e){return t>>>=0,e||R(t,4,this.length),(this[t]|this[t+1]<<8|this[t+2]<<16)+16777216*this[t+3]},Buffer.prototype.readUint32BE=Buffer.prototype.readUInt32BE=function(t,e){return t>>>=0,e||R(t,4,this.length),16777216*this[t]+(this[t+1]<<16|this[t+2]<<8|this[t+3])},Buffer.prototype.readBigUInt64LE=function(t){P(t>>>=0,"offset");const e=this[t],r=this[t+7];void 0!==e&&void 0!==r||M(t,this.length-8);const n=e+256*this[++t]+65536*this[++t]+this[++t]*2**24,f=this[++t]+256*this[++t]+65536*this[++t]+r*2**24;return BigInt(n)+(BigInt(f)<<BigInt(32))},Buffer.prototype.readBigUInt64BE=function(t){P(t>>>=0,"offset");const e=this[t],r=this[t+7];void 0!==e&&void 0!==r||M(t,this.length-8);const n=e*2**24+65536*this[++t]+256*this[++t]+this[++t],f=this[++t]*2**24+65536*this[++t]+256*this[++t]+r;return(BigInt(n)<<BigInt(32))+BigInt(f)},Buffer.prototype.readIntLE=function(t,e,r){t>>>=0,e>>>=0,r||R(t,e,this.length);let n=this[t],f=1,i=0;for(;++i<e&&(f*=256);)n+=this[t+i]*f;return f*=128,n>=f&&(n-=Math.pow(2,8*e)),n},Buffer.prototype.readIntBE=function(t,e,r){t>>>=0,e>>>=0,r||R(t,e,this.length);let n=e,f=1,i=this[t+--n];for(;n>0&&(f*=256);)i+=this[t+--n]*f;return f*=128,i>=f&&(i-=Math.pow(2,8*e)),i},Buffer.prototype.readInt8=function(t,e){return t>>>=0,e||R(t,1,this.length),128&this[t]?-1*(255-this[t]+1):this[t]},Buffer.prototype.readInt16LE=function(t,e){t>>>=0,e||R(t,2,this.length);const r=this[t]|this[t+1]<<8;return 32768&r?4294901760|r:r},Buffer.prototype.readInt16BE=function(t,e){t>>>=0,e||R(t,2,this.length);const r=this[t+1]|this[t]<<8;return 32768&r?4294901760|r:r},Buffer.prototype.readInt32LE=function(t,e){return t>>>=0,e||R(t,4,this.length),this[t]|this[t+1]<<8|this[t+2]<<16|this[t+3]<<24},Buffer.prototype.readInt32BE=function(t,e){return t>>>=0,e||R(t,4,this.length),this[t]<<24|this[t+1]<<16|this[t+2]<<8|this[t+3]},Buffer.prototype.readBigInt64LE=function(t){P(t>>>=0,"offset");const e=this[t],r=this[t+7];void 0!==e&&void 0!==r||M(t,this.length-8);const n=this[t+4]+256*this[t+5]+65536*this[t+6]+(r<<24);return(BigInt(n)<<BigInt(32))+BigInt(e+256*this[++t]+65536*this[++t]+this[++t]*2**24)},Buffer.prototype.readBigInt64BE=function(t){P(t>>>=0,"offset");const e=this[t],r=this[t+7];void 0!==e&&void 0!==r||M(t,this.length-8);const n=(e<<24)+65536*this[++t]+256*this[++t]+this[++t];return(BigInt(n)<<BigInt(32))+BigInt(this[++t]*2**24+65536*this[++t]+256*this[++t]+r)},Buffer.prototype.readFloatLE=function(t,r){return t>>>=0,r||R(t,4,this.length),e.read(this,t,!0,23,4)},Buffer.prototype.readFloatBE=function(t,r){return t>>>=0,r||R(t,4,this.length),e.read(this,t,!1,23,4)},Buffer.prototype.readDoubleLE=function(t,r){return t>>>=0,r||R(t,8,this.length),e.read(this,t,!0,52,8)},Buffer.prototype.readDoubleBE=function(t,r){return t>>>=0,r||R(t,8,this.length),e.read(this,t,!1,52,8)},Buffer.prototype.writeUintLE=Buffer.prototype.writeUIntLE=function(t,e,r,n){if(t=+t,e>>>=0,r>>>=0,!n){T(this,t,e,r,Math.pow(2,8*r)-1,0)}let f=1,i=0;for(this[e]=255&t;++i<r&&(f*=256);)this[e+i]=t/f&255;return e+r},Buffer.prototype.writeUintBE=Buffer.prototype.writeUIntBE=function(t,e,r,n){if(t=+t,e>>>=0,r>>>=0,!n){T(this,t,e,r,Math.pow(2,8*r)-1,0)}let f=r-1,i=1;for(this[e+f]=255&t;--f>=0&&(i*=256);)this[e+f]=t/i&255;return e+r},Buffer.prototype.writeUint8=Buffer.prototype.writeUInt8=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,1,255,0),this[e]=255&t,e+1},Buffer.prototype.writeUint16LE=Buffer.prototype.writeUInt16LE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,2,65535,0),this[e]=255&t,this[e+1]=t>>>8,e+2},Buffer.prototype.writeUint16BE=Buffer.prototype.writeUInt16BE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,2,65535,0),this[e]=t>>>8,this[e+1]=255&t,e+2},Buffer.prototype.writeUint32LE=Buffer.prototype.writeUInt32LE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,4,4294967295,0),this[e+3]=t>>>24,this[e+2]=t>>>16,this[e+1]=t>>>8,this[e]=255&t,e+4},Buffer.prototype.writeUint32BE=Buffer.prototype.writeUInt32BE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,4,4294967295,0),this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t,e+4},Buffer.prototype.writeBigUInt64LE=function(t,e=0){return L(this,t,e,BigInt(0),BigInt("0xffffffffffffffff"))},Buffer.prototype.writeBigUInt64BE=function(t,e=0){return O(this,t,e,BigInt(0),BigInt("0xffffffffffffffff"))},Buffer.prototype.writeIntLE=function(t,e,r,n){if(t=+t,e>>>=0,!n){const n=Math.pow(2,8*r-1);T(this,t,e,r,n-1,-n)}let f=0,i=1,o=0;for(this[e]=255&t;++f<r&&(i*=256);)t<0&&0===o&&0!==this[e+f-1]&&(o=1),this[e+f]=(t/i>>0)-o&255;return e+r},Buffer.prototype.writeIntBE=function(t,e,r,n){if(t=+t,e>>>=0,!n){const n=Math.pow(2,8*r-1);T(this,t,e,r,n-1,-n)}let f=r-1,i=1,o=0;for(this[e+f]=255&t;--f>=0&&(i*=256);)t<0&&0===o&&0!==this[e+f+1]&&(o=1),this[e+f]=(t/i>>0)-o&255;return e+r},Buffer.prototype.writeInt8=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,1,127,-128),t<0&&(t=255+t+1),this[e]=255&t,e+1},Buffer.prototype.writeInt16LE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,2,32767,-32768),this[e]=255&t,this[e+1]=t>>>8,e+2},Buffer.prototype.writeInt16BE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,2,32767,-32768),this[e]=t>>>8,this[e+1]=255&t,e+2},Buffer.prototype.writeInt32LE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,4,2147483647,-2147483648),this[e]=255&t,this[e+1]=t>>>8,this[e+2]=t>>>16,this[e+3]=t>>>24,e+4},Buffer.prototype.writeInt32BE=function(t,e,r){return t=+t,e>>>=0,r||T(this,t,e,4,2147483647,-2147483648),t<0&&(t=4294967295+t+1),this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t,e+4},Buffer.prototype.writeBigInt64LE=function(t,e=0){return L(this,t,e,-BigInt("0x8000000000000000"),BigInt("0x7fffffffffffffff"))},Buffer.prototype.writeBigInt64BE=function(t,e=0){return O(this,t,e,-BigInt("0x8000000000000000"),BigInt("0x7fffffffffffffff"))},Buffer.prototype.writeFloatLE=function(t,e,r){return _(this,t,e,!0,r)},Buffer.prototype.writeFloatBE=function(t,e,r){return _(this,t,e,!1,r)},Buffer.prototype.writeDoubleLE=function(t,e,r){return x(this,t,e,!0,r)},Buffer.prototype.writeDoubleBE=function(t,e,r){return x(this,t,e,!1,r)},Buffer.prototype.copy=function(t,e,r,n){if(!Buffer.isBuffer(t))throw new TypeError("argument should be a Buffer");if(r||(r=0),n||0===n||(n=this.length),e>=t.length&&(e=t.length),e||(e=0),n>0&&n<r&&(n=r),n===r)return 0;if(0===t.length||0===this.length)return 0;if(e<0)throw new RangeError("targetStart out of bounds");if(r<0||r>=this.length)throw new RangeError("Index out of range");if(n<0)throw new RangeError("sourceEnd out of bounds");n>this.length&&(n=this.length),t.length-e<n-r&&(n=t.length-e+r);const f=n-r;return this===t?this.copyWithin(e,r,n):Uint8Array.prototype.set.call(t,this.subarray(r,n),e),f},Buffer.prototype.fill=function(t,e,r,n){if("string"==typeof t){if("string"==typeof e?(n=e,e=0,r=this.length):"string"==typeof r&&(n=r,r=this.length),void 0!==n&&"string"!=typeof n)throw new TypeError("encoding must be a string");if("string"==typeof n&&!Buffer.isEncoding(n))throw new TypeError("Unknown encoding: "+n);if(1===t.length){const e=t.charCodeAt(0);("utf8"===n&&e<128||"latin1"===n)&&(t=e)}}else"number"==typeof t?t&=255:"boolean"==typeof t&&(t=Number(t));if(e<0||this.length<e||this.length<r)throw new RangeError("Out of range index");if(r<=e)return this;let f;if(e>>>=0,r=void 0===r?this.length:r>>>0,t||(t=0),"number"==typeof t)for(f=e;f<r;++f)this[f]=t;else{const i=Buffer.isBuffer(t)?t:Buffer.from(t,n),o=i.length;if(0===o)throw new TypeError('The value "'+t+'" is invalid for argument "value"');for(f=0;f<r-e;++f)this[f+e]=i[f%o]}return this};const $={};function N(t,e,r){$[t]=class extends r{constructor(){super(),Object.defineProperty(this,"message",{value:e.apply(this,arguments),writable:!0,configurable:!0}),this.name=`${this.name} [${t}]`,this.stack,delete this.name}get code(){return t}set code(t){Object.defineProperty(this,"code",{configurable:!0,enumerable:!0,value:t,writable:!0})}toString(){return`${this.name} [${t}]: ${this.message}`}}}function C(t){let e="",r=t.length;const n="-"===t[0]?1:0;for(;r>=n+4;r-=3)e=`_${t.slice(r-3,r)}${e}`;return`${t.slice(0,r)}${e}`}function k(t,e,r,n,f,i){if(t>r||t<e){const n="bigint"==typeof e?"n":"";let f;throw f=i>3?0===e||e===BigInt(0)?`>= 0${n} and < 2${n} ** ${8*(i+1)}${n}`:`>= -(2${n} ** ${8*(i+1)-1}${n}) and < 2 ** ${8*(i+1)-1}${n}`:`>= ${e}${n} and <= ${r}${n}`,new $.ERR_OUT_OF_RANGE("value",f,t)}!function(t,e,r){P(e,"offset"),void 0!==t[e]&&void 0!==t[e+r]||M(e,t.length-(r+1))}(n,f,i)}function P(t,e){if("number"!=typeof t)throw new $.ERR_INVALID_ARG_TYPE(e,"number",t)}function M(t,e,r){if(Math.floor(t)!==t)throw P(t,r),new $.ERR_OUT_OF_RANGE(r||"offset","an integer",t);if(e<0)throw new $.ERR_BUFFER_OUT_OF_BOUNDS;throw new $.ERR_OUT_OF_RANGE(r||"offset",`>= ${r?1:0} and <= ${e}`,t)}N("ERR_BUFFER_OUT_OF_BOUNDS",(function(t){return t?`${t} is outside of buffer bounds`:"Attempt to access memory outside buffer bounds"}),RangeError),N("ERR_INVALID_ARG_TYPE",(function(t,e){return`The "${t}" argument must be of type number. Received type ${typeof e}`}),TypeError),N("ERR_OUT_OF_RANGE",(function(t,e,r){let n=`The value of "${t}" is out of range.`,f=r;return Number.isInteger(r)&&Math.abs(r)>2**32?f=C(String(r)):"bigint"==typeof r&&(f=String(r),(r>BigInt(2)**BigInt(32)||r<-(BigInt(2)**BigInt(32)))&&(f=C(f)),f+="n"),n+=` It must be ${e}. Received ${f}`,n}),RangeError);const F=/[^+/0-9A-Za-z-_]/g;function j(t,e){let r;e=e||1/0;const n=t.length;let f=null;const i=[];for(let o=0;o<n;++o){if(r=t.charCodeAt(o),r>55295&&r<57344){if(!f){if(r>56319){(e-=3)>-1&&i.push(239,191,189);continue}if(o+1===n){(e-=3)>-1&&i.push(239,191,189);continue}f=r;continue}if(r<56320){(e-=3)>-1&&i.push(239,191,189),f=r;continue}r=65536+(f-55296<<10|r-56320)}else f&&(e-=3)>-1&&i.push(239,191,189);if(f=null,r<128){if((e-=1)<0)break;i.push(r)}else if(r<2048){if((e-=2)<0)break;i.push(r>>6|192,63&r|128)}else if(r<65536){if((e-=3)<0)break;i.push(r>>12|224,r>>6&63|128,63&r|128)}else{if(!(r<1114112))throw new Error("Invalid code point");if((e-=4)<0)break;i.push(r>>18|240,r>>12&63|128,r>>6&63|128,63&r|128)}}return i}function z(e){return t.toByteArray(function(t){if((t=(t=t.split("=")[0]).trim().replace(F,"")).length<2)return"";for(;t.length%4!=0;)t+="=";return t}(e))}function D(t,e,r,n){let f;for(f=0;f<n&&!(f+r>=e.length||f>=t.length);++f)e[f+r]=t[f];return f}const G=function(){const t="0123456789abcdef",e=new Array(256);for(let r=0;r<16;++r){const n=16*r;for(let f=0;f<16;++f)e[n+f]=t[r]+t[f]}return e}();export default{config,kMaxLength:2147483647,Buffer,SlowBuffer};
✄
{"version":3,"file":"events.js","names":["EventEmitter","once","init","call","this","prototype","_events","undefined","_eventsCount","_maxListeners","defaultMaxListeners","checkListener","listener","TypeError","_getMaxListeners","that","_addListener","target","type","prepend","existing","events","Object","create","newListener","emit","unshift","push","m","length","warned","w","Error","String","name","emitter","count","warning","console","warn","onceWrapper","fired","removeListener","wrapFn","arguments","apply","_onceWrap","state","wrapped","bind","_listeners","unwrap","evlistener","arr","ret","Array","i","unwrapListeners","arrayClone","listenerCount","n","copy","Promise","resolve","reject","errorListener","err","resolver","slice","eventTargetAgnosticAddListener","handler","flags","on","addErrorHandlerIfEventEmitter","addEventListener","wrapListener","arg","removeEventListener","defineProperty","enumerable","get","set","Number","isNaN","RangeError","getPrototypeOf","setMaxListeners","getMaxListeners","args","doError","error","er","message","context","Reflect","len","listeners","addListener","prependListener","prependOnceListener","list","originalListener","position","shift","index","pop","spliceOne","off","removeAllListeners","keys","key","rawListeners","eventNames","ownKeys"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/events/","sources":[""],"mappings":"eAqBeA,SACNA,kBAAcC,WAMvB,SAASD,IACPA,EAAaE,KAAKC,KAAKC,KACzB,CAGAJ,EAAaA,aAAeA,EAE5BA,EAAaK,UAAUC,aAAUC,EACjCP,EAAaK,UAAUG,aAAe,EACtCR,EAAaK,UAAUI,mBAAgBF,EAIvC,IAAIG,EAAsB,GAE1B,SAASC,EAAcC,GACrB,GAAwB,mBAAbA,EACT,MAAM,IAAIC,UAAU,0EAA4ED,EAEpG,CAoCA,SAASE,EAAiBC,GACxB,YAA2BR,IAAvBQ,EAAKN,cACAT,EAAaU,oBACfK,EAAKN,aACd,CAkDA,SAASO,EAAaC,EAAQC,EAAMN,EAAUO,GAC5C,IAAIC,EAEJT,EAAcC,GAEd,IAAIS,EAASJ,EAAOX,QAkBpB,QAjBeC,IAAXc,GACFA,EAASJ,EAAOX,QAAUgB,OAAOC,OAAO,MACxCN,EAAOT,aAAe,SAIKD,IAAvBc,EAAOG,cACTP,EAAOQ,KAAK,cAAeP,EACfN,EAASA,SAAWA,EAASA,SAAWA,GAIpDS,EAASJ,EAAOX,SAElBc,EAAWC,EAAOH,SAGHX,IAAba,EAEFA,EAAWC,EAAOH,GAAQN,IACxBK,EAAOT,iBACJ,CACmB,mBAAbY,EAETA,EAAWC,EAAOH,GAChBC,EAAU,CAACP,EAAUQ,GAAY,CAACA,EAAUR,GAErCO,EACTC,EAASM,QAAQd,GAEjBQ,EAASO,KAAKf,GAIhB,MAAMgB,EAAId,EAAiBG,GAC3B,GAAIW,EAAI,GAAKR,EAASS,OAASD,IAAMR,EAASU,OAAQ,CACpDV,EAASU,QAAS,EAGlB,MAAMC,EAAI,IAAIC,MAAM,+CACAZ,EAASS,OAAS,IAAMI,OAAOf,GAD/B,qEAIpBa,EAAEG,KAAO,8BACTH,EAAEI,QAAUlB,EACZc,EAAEb,KAAOA,EACTa,EAAEK,MAAQhB,EAASS,OArKGQ,EAsKHN,EArKvBO,QAAQC,KAAKF,EAsKX,CACF,CAxKF,IAA4BA,EA0K1B,OAAOpB,CACT,CAaA,SAASuB,IACP,IAAKpC,KAAKqC,MAGR,OAFArC,KAAKa,OAAOyB,eAAetC,KAAKc,KAAMd,KAAKuC,QAC3CvC,KAAKqC,OAAQ,EACY,IAArBG,UAAUf,OACLzB,KAAKQ,SAAST,KAAKC,KAAKa,QAC1Bb,KAAKQ,SAASiC,MAAMzC,KAAKa,OAAQ2B,UAE5C,CAEA,SAASE,EAAU7B,EAAQC,EAAMN,GAC/B,MAAMmC,EAAQ,CAAEN,OAAO,EAAOE,YAAQpC,EAAWU,OAAQA,EAAQC,KAAMA,EAAMN,SAAUA,GACjFoC,EAAUR,EAAYS,KAAKF,GAGjC,OAFAC,EAAQpC,SAAWA,EACnBmC,EAAMJ,OAASK,EACRA,CACT,CAqHA,SAASE,EAAWjC,EAAQC,EAAMiC,GAChC,MAAM9B,EAASJ,EAAOX,QAEtB,QAAeC,IAAXc,EACF,MAAO,GAET,MAAM+B,EAAa/B,EAAOH,GAC1B,YAAmBX,IAAf6C,EACK,GAEiB,mBAAfA,EACFD,EAAS,CAACC,EAAWxC,UAAYwC,GAAc,CAACA,GAElDD,EAsDT,SAAyBE,GACvB,MAAMC,EAAM,IAAIC,MAAMF,EAAIxB,QAC1B,IAAK,IAAI2B,EAAI,EAAGA,EAAIF,EAAIzB,SAAU2B,EAChCF,EAAIE,GAAKH,EAAIG,GAAG5C,UAAYyC,EAAIG,GAElC,OAAOF,CACT,CA3DIG,CAAgBL,GAAcM,EAAWN,EAAYA,EAAWvB,OACpE,CAmBA,SAAS8B,EAAczC,GACrB,MAAMG,EAASjB,KAAKE,QAEpB,QAAeC,IAAXc,EAAsB,CACxB,MAAM+B,EAAa/B,EAAOH,GAE1B,GAA0B,mBAAfkC,EACT,OAAO,EACF,QAAmB7C,IAAf6C,EACT,OAAOA,EAAWvB,MAEtB,CAEA,OAAO,CACT,CAMA,SAAS6B,EAAWL,EAAKO,GACvB,MAAMC,EAAO,IAAIN,MAAMK,GACvB,IAAK,IAAIJ,EAAI,EAAGA,EAAII,IAAKJ,EACvBK,EAAKL,GAAKH,EAAIG,GAChB,OAAOK,CACT,CAgBA,SAAS5D,EAAKkC,EAASD,GACrB,OAAO,IAAI4B,SAAQ,SAAUC,EAASC,GACpC,SAASC,EAAcC,GACrB/B,EAAQO,eAAeR,EAAMiC,GAC7BH,EAAOE,EACT,CAEA,SAASC,IAC+B,mBAA3BhC,EAAQO,gBACjBP,EAAQO,eAAe,QAASuB,GAElCF,EAAQ,GAAGK,MAAMjE,KAAKyC,WACxB,CAEAyB,EAA+BlC,EAASD,EAAMiC,EAAU,CAAElE,MAAM,IACnD,UAATiC,GAMR,SAAuCC,EAASmC,EAASC,GAC7B,mBAAfpC,EAAQqC,IACjBH,EAA+BlC,EAAS,QAASmC,EAASC,EAE9D,CATME,CAA8BtC,EAAS8B,EAAe,CAAEhE,MAAM,GAElE,GACF,CAQA,SAASoE,EAA+BlC,EAASD,EAAMtB,EAAU2D,GAC/D,GAA0B,mBAAfpC,EAAQqC,GACbD,EAAMtE,KACRkC,EAAQlC,KAAKiC,EAAMtB,GAEnBuB,EAAQqC,GAAGtC,EAAMtB,OAEd,IAAwC,mBAA7BuB,EAAQuC,iBAYxB,MAAM,IAAI7D,UAAU,6EAA+EsB,GATnGA,EAAQuC,iBAAiBxC,GAAM,SAASyC,EAAaC,GAG/CL,EAAMtE,MACRkC,EAAQ0C,oBAAoB3C,EAAMyC,GAEpC/D,EAASgE,EACX,GAGF,CACF,CA/ZAtD,OAAOwD,eAAe9E,EAAc,sBAAuB,CACzD+E,YAAY,EACZC,IAAK,WACH,OAAOtE,CACT,EACAuE,IAAK,SAASL,GACZ,GAAmB,iBAARA,GAAoBA,EAAM,GAAKM,OAAOC,MAAMP,GACrD,MAAM,IAAIQ,WAAW,kGAAoGR,EAAM,KAEjIlE,EAAsBkE,CACxB,IAGF5E,EAAaE,KAAO,gBAEGK,IAAjBH,KAAKE,SACLF,KAAKE,UAAYgB,OAAO+D,eAAejF,MAAME,UAC/CF,KAAKE,QAAUgB,OAAOC,OAAO,MAC7BnB,KAAKI,aAAe,GAGtBJ,KAAKK,cAAgBL,KAAKK,oBAAiBF,CAC7C,EAIAP,EAAaK,UAAUiF,gBAAkB,SAAyB1B,GAChE,GAAiB,iBAANA,GAAkBA,EAAI,GAAKsB,OAAOC,MAAMvB,GACjD,MAAM,IAAIwB,WAAW,gFAAkFxB,EAAI,KAG7G,OADAxD,KAAKK,cAAgBmD,EACdxD,IACT,EAQAJ,EAAaK,UAAUkF,gBAAkB,WACvC,OAAOzE,EAAiBV,KAC1B,EAEAJ,EAAaK,UAAUoB,KAAO,SAAcP,GAC1C,MAAMsE,EAAO,GACb,IAAK,IAAIhC,EAAI,EAAGA,EAAIZ,UAAUf,OAAQ2B,IAAKgC,EAAK7D,KAAKiB,UAAUY,IAC/D,IAAIiC,EAAoB,UAATvE,EAEf,MAAMG,EAASjB,KAAKE,QACpB,QAAeC,IAAXc,EACFoE,EAAWA,QAA4BlF,IAAjBc,EAAOqE,WAC1B,IAAKD,EACR,OAAO,EAGT,GAAIA,EAAS,CACX,IAAIE,EAGJ,GAFIH,EAAK3D,OAAS,IAChB8D,EAAKH,EAAK,IACRG,aAAc3D,MAGhB,MAAM2D,EAGR,MAAMzB,EAAM,IAAIlC,MAAM,oBAAsB2D,EAAK,KAAOA,EAAGC,QAAU,IAAM,KAE3E,MADA1B,EAAI2B,QAAUF,EACRzB,CACR,CAEA,MAAMI,EAAUjD,EAAOH,GAEvB,QAAgBX,IAAZ+D,EACF,OAAO,EAET,GAAuB,mBAAZA,EACTwB,QAAQjD,MAAMyB,EAASlE,KAAMoF,OACxB,CACL,MAAMO,EAAMzB,EAAQzC,OACdmE,EAAYtC,EAAWY,EAASyB,GACtC,IAAK,IAAIvC,EAAI,EAAGA,EAAIuC,IAAOvC,EACzBsC,QAAQjD,MAAMmD,EAAUxC,GAAIpD,KAAMoF,EACtC,CAEA,OAAO,CACT,EA8DAxF,EAAaK,UAAU4F,YAAc,SAAqB/E,EAAMN,GAC9D,OAAOI,EAAaZ,KAAMc,EAAMN,GAAU,EAC5C,EAEAZ,EAAaK,UAAUmE,GAAKxE,EAAaK,UAAU4F,YAEnDjG,EAAaK,UAAU6F,gBACnB,SAAyBhF,EAAMN,GAC7B,OAAOI,EAAaZ,KAAMc,EAAMN,GAAU,EAC5C,EAoBJZ,EAAaK,UAAUJ,KAAO,SAAciB,EAAMN,GAGhD,OAFAD,EAAcC,GACdR,KAAKoE,GAAGtD,EAAM4B,EAAU1C,KAAMc,EAAMN,IAC7BR,IACT,EAEAJ,EAAaK,UAAU8F,oBACnB,SAA6BjF,EAAMN,GAGjC,OAFAD,EAAcC,GACdR,KAAK8F,gBAAgBhF,EAAM4B,EAAU1C,KAAMc,EAAMN,IAC1CR,IACT,EAGJJ,EAAaK,UAAUqC,eACnB,SAAwBxB,EAAMN,GAC5BD,EAAcC,GAEd,MAAMS,EAASjB,KAAKE,QACpB,QAAeC,IAAXc,EACF,OAAOjB,KAET,MAAMgG,EAAO/E,EAAOH,GACpB,QAAaX,IAAT6F,EACF,OAAOhG,KAET,GAAIgG,IAASxF,GAAYwF,EAAKxF,WAAaA,EACb,KAAtBR,KAAKI,aACTJ,KAAKE,QAAUgB,OAAOC,OAAO,cAEtBF,EAAOH,GACVG,EAAOqB,gBACTtC,KAAKqB,KAAK,iBAAkBP,EAAMkF,EAAKxF,UAAYA,SAElD,GAAoB,mBAATwF,EAAqB,CACrC,IAAIC,EACAC,GAAY,EAEhB,IAAK,IAAI9C,EAAI4C,EAAKvE,OAAS,EAAG2B,GAAK,EAAGA,IACpC,GAAI4C,EAAK5C,KAAO5C,GAAYwF,EAAK5C,GAAG5C,WAAaA,EAAU,CACzDyF,EAAmBD,EAAK5C,GAAG5C,SAC3B0F,EAAW9C,EACX,KACF,CAGF,GAAI8C,EAAW,EACb,OAAOlG,KAEQ,IAAbkG,EACFF,EAAKG,QA8Hf,SAAmBH,EAAMI,GACvB,KAAOA,EAAQ,EAAIJ,EAAKvE,OAAQ2E,IAC9BJ,EAAKI,GAASJ,EAAKI,EAAQ,GAC7BJ,EAAKK,KACP,CAhIUC,CAAUN,EAAME,GAGE,IAAhBF,EAAKvE,SACPR,EAAOH,GAAQkF,EAAK,SAEQ7F,IAA1Bc,EAAOqB,gBACTtC,KAAKqB,KAAK,iBAAkBP,EAAMmF,GAAoBzF,EAC1D,CAEA,OAAOR,IACT,EAEJJ,EAAaK,UAAUsG,IAAM3G,EAAaK,UAAUqC,eAEpD1C,EAAaK,UAAUuG,mBACnB,SAA4B1F,GAC1B,MAAMG,EAASjB,KAAKE,QACpB,QAAeC,IAAXc,EACF,OAAOjB,KAGT,QAA8BG,IAA1Bc,EAAOqB,eAUT,OATyB,IAArBE,UAAUf,QACZzB,KAAKE,QAAUgB,OAAOC,OAAO,MAC7BnB,KAAKI,aAAe,QACMD,IAAjBc,EAAOH,KACY,KAAtBd,KAAKI,aACTJ,KAAKE,QAAUgB,OAAOC,OAAO,aAEtBF,EAAOH,IAEXd,KAIT,GAAyB,IAArBwC,UAAUf,OAAc,CAC1B,MAAMgF,EAAOvF,OAAOuF,KAAKxF,GACzB,IAAK,IAAImC,EAAI,EAAGA,EAAIqD,EAAKhF,SAAU2B,EAAG,CACpC,MAAMsD,EAAMD,EAAKrD,GACL,mBAARsD,GACJ1G,KAAKwG,mBAAmBE,EAC1B,CAIA,OAHA1G,KAAKwG,mBAAmB,kBACxBxG,KAAKE,QAAUgB,OAAOC,OAAO,MAC7BnB,KAAKI,aAAe,EACbJ,IACT,CAEA,MAAM4F,EAAY3E,EAAOH,GAEzB,GAAyB,mBAAd8E,EACT5F,KAAKsC,eAAexB,EAAM8E,QACrB,QAAkBzF,IAAdyF,EAET,IAAK,IAAIxC,EAAIwC,EAAUnE,OAAS,EAAG2B,GAAK,EAAGA,IACzCpD,KAAKsC,eAAexB,EAAM8E,EAAUxC,IAIxC,OAAOpD,IACT,EAmBJJ,EAAaK,UAAU2F,UAAY,SAAmB9E,GACpD,OAAOgC,EAAW9C,KAAMc,GAAM,EAChC,EAEAlB,EAAaK,UAAU0G,aAAe,SAAsB7F,GAC1D,OAAOgC,EAAW9C,KAAMc,GAAM,EAChC,EAEAlB,EAAa2D,cAAgB,SAASxB,EAASjB,GAC7C,MAAqC,mBAA1BiB,EAAQwB,cACVxB,EAAQwB,cAAczC,GAEtByC,EAAcxD,KAAKgC,EAASjB,EAEvC,EAEAlB,EAAaK,UAAUsD,cAAgBA,EAiBvC3D,EAAaK,UAAU2G,WAAa,WAClC,OAAO5G,KAAKI,aAAe,EAAIsF,QAAQmB,QAAQ7G,KAAKE,SAAW,EACjE"}
✄
export default e;export{e as EventEmitter,l as once};function e(){e.init.call(this)}e.EventEmitter=e,e.prototype._events=void 0,e.prototype._eventsCount=0,e.prototype._maxListeners=void 0;let t=10;function n(e){if("function"!=typeof e)throw new TypeError('The "listener" argument must be of type Function. Received type '+typeof e)}function r(t){return void 0===t._maxListeners?e.defaultMaxListeners:t._maxListeners}function i(e,t,i,o){let s;n(i);let u=e._events;if(void 0===u?(u=e._events=Object.create(null),e._eventsCount=0):(void 0!==u.newListener&&(e.emit("newListener",t,i.listener?i.listener:i),u=e._events),s=u[t]),void 0===s)s=u[t]=i,++e._eventsCount;else{"function"==typeof s?s=u[t]=o?[i,s]:[s,i]:o?s.unshift(i):s.push(i);const n=r(e);if(n>0&&s.length>n&&!s.warned){s.warned=!0;const n=new Error("Possible EventEmitter memory leak detected. "+s.length+" "+String(t)+" listeners added. Use emitter.setMaxListeners() to increase limit");n.name="MaxListenersExceededWarning",n.emitter=e,n.type=t,n.count=s.length,f=n,console.warn(f)}}var f;return e}function o(){if(!this.fired)return this.target.removeListener(this.type,this.wrapFn),this.fired=!0,0===arguments.length?this.listener.call(this.target):this.listener.apply(this.target,arguments)}function s(e,t,n){const r={fired:!1,wrapFn:void 0,target:e,type:t,listener:n},i=o.bind(r);return i.listener=n,r.wrapFn=i,i}function u(e,t,n){const r=e._events;if(void 0===r)return[];const i=r[t];return void 0===i?[]:"function"==typeof i?n?[i.listener||i]:[i]:n?function(e){const t=new Array(e.length);for(let n=0;n<t.length;++n)t[n]=e[n].listener||e[n];return t}(i):c(i,i.length)}function f(e){const t=this._events;if(void 0!==t){const n=t[e];if("function"==typeof n)return 1;if(void 0!==n)return n.length}return 0}function c(e,t){const n=new Array(t);for(let r=0;r<t;++r)n[r]=e[r];return n}function l(e,t){return new Promise((function(n,r){function i(n){e.removeListener(t,o),r(n)}function o(){"function"==typeof e.removeListener&&e.removeListener("error",i),n([].slice.call(arguments))}h(e,t,o,{once:!0}),"error"!==t&&function(e,t,n){"function"==typeof e.on&&h(e,"error",t,n)}(e,i,{once:!0})}))}function h(e,t,n,r){if("function"==typeof e.on)r.once?e.once(t,n):e.on(t,n);else{if("function"!=typeof e.addEventListener)throw new TypeError('The "emitter" argument must be of type EventEmitter. Received type '+typeof e);e.addEventListener(t,(function i(o){r.once&&e.removeEventListener(t,i),n(o)}))}}Object.defineProperty(e,"defaultMaxListeners",{enumerable:!0,get:function(){return t},set:function(e){if("number"!=typeof e||e<0||Number.isNaN(e))throw new RangeError('The value of "defaultMaxListeners" is out of range. It must be a non-negative number. Received '+e+".");t=e}}),e.init=function(){void 0!==this._events&&this._events!==Object.getPrototypeOf(this)._events||(this._events=Object.create(null),this._eventsCount=0),this._maxListeners=this._maxListeners||void 0},e.prototype.setMaxListeners=function(e){if("number"!=typeof e||e<0||Number.isNaN(e))throw new RangeError('The value of "n" is out of range. It must be a non-negative number. Received '+e+".");return this._maxListeners=e,this},e.prototype.getMaxListeners=function(){return r(this)},e.prototype.emit=function(e){const t=[];for(let e=1;e<arguments.length;e++)t.push(arguments[e]);let n="error"===e;const r=this._events;if(void 0!==r)n=n&&void 0===r.error;else if(!n)return!1;if(n){let e;if(t.length>0&&(e=t[0]),e instanceof Error)throw e;const n=new Error("Unhandled error."+(e?" ("+e.message+")":""));throw n.context=e,n}const i=r[e];if(void 0===i)return!1;if("function"==typeof i)Reflect.apply(i,this,t);else{const e=i.length,n=c(i,e);for(let r=0;r<e;++r)Reflect.apply(n[r],this,t)}return!0},e.prototype.addListener=function(e,t){return i(this,e,t,!1)},e.prototype.on=e.prototype.addListener,e.prototype.prependListener=function(e,t){return i(this,e,t,!0)},e.prototype.once=function(e,t){return n(t),this.on(e,s(this,e,t)),this},e.prototype.prependOnceListener=function(e,t){return n(t),this.prependListener(e,s(this,e,t)),this},e.prototype.removeListener=function(e,t){n(t);const r=this._events;if(void 0===r)return this;const i=r[e];if(void 0===i)return this;if(i===t||i.listener===t)0==--this._eventsCount?this._events=Object.create(null):(delete r[e],r.removeListener&&this.emit("removeListener",e,i.listener||t));else if("function"!=typeof i){let n,o=-1;for(let e=i.length-1;e>=0;e--)if(i[e]===t||i[e].listener===t){n=i[e].listener,o=e;break}if(o<0)return this;0===o?i.shift():function(e,t){for(;t+1<e.length;t++)e[t]=e[t+1];e.pop()}(i,o),1===i.length&&(r[e]=i[0]),void 0!==r.removeListener&&this.emit("removeListener",e,n||t)}return this},e.prototype.off=e.prototype.removeListener,e.prototype.removeAllListeners=function(e){const t=this._events;if(void 0===t)return this;if(void 0===t.removeListener)return 0===arguments.length?(this._events=Object.create(null),this._eventsCount=0):void 0!==t[e]&&(0==--this._eventsCount?this._events=Object.create(null):delete t[e]),this;if(0===arguments.length){const e=Object.keys(t);for(let t=0;t<e.length;++t){const n=e[t];"removeListener"!==n&&this.removeAllListeners(n)}return this.removeAllListeners("removeListener"),this._events=Object.create(null),this._eventsCount=0,this}const n=t[e];if("function"==typeof n)this.removeListener(e,n);else if(void 0!==n)for(let t=n.length-1;t>=0;t--)this.removeListener(e,n[t]);return this},e.prototype.listeners=function(e){return u(this,e,!0)},e.prototype.rawListeners=function(e){return u(this,e,!1)},e.listenerCount=function(e,t){return"function"==typeof e.listenerCount?e.listenerCount(t):f.call(e,t)},e.prototype.listenerCount=f,e.prototype.eventNames=function(){return this._eventsCount>0?Reflect.ownKeys(this._events):[]};
✄
{"version":3,"file":"index.js","names":["read","buffer","offset","isLE","mLen","nBytes","e","m","eLen","eMax","eBias","nBits","i","d","s","NaN","Infinity","Math","pow","write","value","c","rt","abs","isNaN","floor","log","LN2"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/ieee754/","sources":[""],"mappings":";OAEO,SAASA,KAAMC,EAAQC,EAAQC,EAAMC,EAAMC,GAChD,IAAIC,EAAGC,EACP,MAAMC,EAAiB,EAATH,EAAcD,EAAO,EAC7BK,GAAQ,GAAKD,GAAQ,EACrBE,EAAQD,GAAQ,EACtB,IAAIE,GAAS,EACTC,EAAIT,EAAQE,EAAS,EAAK,EAC9B,MAAMQ,EAAIV,GAAQ,EAAI,EACtB,IAAIW,EAAIb,EAAOC,EAASU,GAOxB,IALAA,GAAKC,EAELP,EAAIQ,GAAM,IAAOH,GAAU,EAC3BG,KAAQH,EACRA,GAASH,EACFG,EAAQ,GACbL,EAAS,IAAJA,EAAWL,EAAOC,EAASU,GAChCA,GAAKC,EACLF,GAAS,EAMX,IAHAJ,EAAID,GAAM,IAAOK,GAAU,EAC3BL,KAAQK,EACRA,GAASP,EACFO,EAAQ,GACbJ,EAAS,IAAJA,EAAWN,EAAOC,EAASU,GAChCA,GAAKC,EACLF,GAAS,EAGX,GAAU,IAANL,EACFA,EAAI,EAAII,MACH,IAAIJ,IAAMG,EACf,OAAOF,EAAIQ,IAAsBC,KAAdF,GAAK,EAAI,GAE5BP,GAAQU,KAAKC,IAAI,EAAGd,GACpBE,GAAQI,CACV,CACA,OAAQI,GAAK,EAAI,GAAKP,EAAIU,KAAKC,IAAI,EAAGZ,EAAIF,EAC5C,QAEO,SAASe,MAAOlB,EAAQmB,EAAOlB,EAAQC,EAAMC,EAAMC,GACxD,IAAIC,EAAGC,EAAGc,EACNb,EAAiB,EAATH,EAAcD,EAAO,EACjC,MAAMK,GAAQ,GAAKD,GAAQ,EACrBE,EAAQD,GAAQ,EAChBa,EAAe,KAATlB,EAAca,KAAKC,IAAI,GAAI,IAAMD,KAAKC,IAAI,GAAI,IAAM,EAChE,IAAIN,EAAIT,EAAO,EAAKE,EAAS,EAC7B,MAAMQ,EAAIV,EAAO,GAAK,EAChBW,EAAIM,EAAQ,GAAgB,IAAVA,GAAe,EAAIA,EAAQ,EAAK,EAAI,EAmC5D,IAjCAA,EAAQH,KAAKM,IAAIH,GAEbI,MAAMJ,IAAUA,IAAUJ,KAC5BT,EAAIiB,MAAMJ,GAAS,EAAI,EACvBd,EAAIG,IAEJH,EAAIW,KAAKQ,MAAMR,KAAKS,IAAIN,GAASH,KAAKU,KAClCP,GAASC,EAAIJ,KAAKC,IAAI,GAAIZ,IAAM,IAClCA,IACAe,GAAK,IAGLD,GADEd,EAAII,GAAS,EACNY,EAAKD,EAELC,EAAKL,KAAKC,IAAI,EAAG,EAAIR,IAEpBW,GAAK,IACff,IACAe,GAAK,GAGHf,EAAII,GAASD,GACfF,EAAI,EACJD,EAAIG,GACKH,EAAII,GAAS,GACtBH,GAAMa,EAAQC,EAAK,GAAKJ,KAAKC,IAAI,EAAGd,GACpCE,GAAQI,IAERH,EAAIa,EAAQH,KAAKC,IAAI,EAAGR,EAAQ,GAAKO,KAAKC,IAAI,EAAGd,GACjDE,EAAI,IAIDF,GAAQ,GACbH,EAAOC,EAASU,GAAS,IAAJL,EACrBK,GAAKC,EACLN,GAAK,IACLH,GAAQ,EAKV,IAFAE,EAAKA,GAAKF,EAAQG,EAClBC,GAAQJ,EACDI,EAAO,GACZP,EAAOC,EAASU,GAAS,IAAJN,EACrBM,GAAKC,EACLP,GAAK,IACLE,GAAQ,EAGVP,EAAOC,EAASU,EAAIC,IAAU,IAAJC,CAC5B"}
✄
/*! ieee754. BSD-3-Clause License. Feross Aboukhadijeh <https://feross.org/opensource> */
export function read(t,o,a,e,h){let r,M;const p=8*h-e-1,n=(1<<p)-1,w=n>>1;let f=-7,l=a?h-1:0;const s=a?-1:1;let i=t[o+l];for(l+=s,r=i&(1<<-f)-1,i>>=-f,f+=p;f>0;)r=256*r+t[o+l],l+=s,f-=8;for(M=r&(1<<-f)-1,r>>=-f,f+=e;f>0;)M=256*M+t[o+l],l+=s,f-=8;if(0===r)r=1-w;else{if(r===n)return M?NaN:1/0*(i?-1:1);M+=Math.pow(2,e),r-=w}return(i?-1:1)*M*Math.pow(2,r-e)}export function write(t,o,a,e,h,r){let M,p,n,w=8*r-h-1;const f=(1<<w)-1,l=f>>1,s=23===h?Math.pow(2,-24)-Math.pow(2,-77):0;let i=e?0:r-1;const N=e?1:-1,c=o<0||0===o&&1/o<0?1:0;for(o=Math.abs(o),isNaN(o)||o===1/0?(p=isNaN(o)?1:0,M=f):(M=Math.floor(Math.log(o)/Math.LN2),o*(n=Math.pow(2,-M))<1&&(M--,n*=2),(o+=M+l>=1?s/n:s*Math.pow(2,1-l))*n>=2&&(M++,n/=2),M+l>=f?(p=0,M=f):M+l>=1?(p=(o*n-1)*Math.pow(2,h),M+=l):(p=o*Math.pow(2,l-1)*Math.pow(2,h),M=0));h>=8;)t[a+i]=255&p,i+=N,p/=256,h-=8;for(M=M<<h|p,w+=h;w>0;)t[a+i]=255&M,i+=N,M/=256,w-=8;t[a+i-N]|=128*c}
✄
{"version":3,"file":"index.js","names":["process","platformIsWin32","platform","isPathSeparator","code","isPosixPathSeparator","isWindowsDeviceRoot","normalizeString","path","allowAboveRoot","separator","res","lastSegmentLength","lastSlash","dots","i","length","charCodeAt","lastSlashIndex","lastIndexOf","slice","_format","sep","pathObject","dir","root","base","name","ext","_win32","resolve","args","resolvedDevice","resolvedTail","resolvedAbsolute","cwd","env","undefined","toLowerCase","len","rootEnd","device","isAbsolute","j","last","firstPart","normalize","tail","join","joined","arg","needsReplace","slashCount","firstLen","relative","from","to","fromOrig","toOrig","fromStart","fromEnd","fromLen","toStart","toEnd","toLen","lastCommonSep","fromCode","out","toNamespacedPath","resolvedPath","dirname","offset","end","matchedSlash","basename","start","extIdx","firstNonSlashEnd","extname","startDot","startPart","preDotState","format","bind","parse","ret","delimiter","win32","posix","posixCwd","regexp","replace","indexOf","_posix","trailingSeparator","hasRoot","impl"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/path/","sources":[""],"mappings":"OAqBOA,MAAa,UAEpB,MAWMC,EAAwC,UAArBD,EAAQE,SAEjC,SAASC,EAAgBC,GACvB,OARsB,KAQfA,GAPe,KAOgBA,CACxC,CAEA,SAASC,EAAqBD,GAC5B,OAZsB,KAYfA,CACT,CAEA,SAASE,EAAoBF,GAC3B,OAAQA,GArBc,IAqBcA,GAnBd,IAoBdA,GArBc,IAqBcA,GAnBd,GAoBxB,CAGA,SAASG,EAAgBC,EAAMC,EAAgBC,EAAWP,GACxD,IAAIQ,EAAM,GACNC,EAAoB,EACpBC,GAAa,EACbC,EAAO,EACPV,EAAO,EACX,IAAK,IAAIW,EAAI,EAAGA,GAAKP,EAAKQ,SAAUD,EAAG,CACrC,GAAIA,EAAIP,EAAKQ,OACXZ,EAAOI,EAAKS,WAAWF,OACpB,IAAIZ,EAAgBC,GACvB,MAEAA,EAjCkB,EAiCO,CAE3B,GAAID,EAAgBC,GAAO,CACzB,GAAIS,IAAcE,EAAI,GAAc,IAATD,QAEpB,GAAa,IAATA,EAAY,CACrB,GAAIH,EAAIK,OAAS,GAA2B,IAAtBJ,GAxCN,KAyCZD,EAAIM,WAAWN,EAAIK,OAAS,IAzChB,KA0CZL,EAAIM,WAAWN,EAAIK,OAAS,GAAiB,CAC/C,GAAIL,EAAIK,OAAS,EAAG,CAClB,MAAME,EAAiBP,EAAIQ,YAAYT,IACf,IAApBQ,GACFP,EAAM,GACNC,EAAoB,IAEpBD,EAAMA,EAAIS,MAAM,EAAGF,GACnBN,EACED,EAAIK,OAAS,EAAIL,EAAIQ,YAAYT,IAErCG,EAAYE,EACZD,EAAO,EACP,QACF,CAAO,GAAmB,IAAfH,EAAIK,OAAc,CAC3BL,EAAM,GACNC,EAAoB,EACpBC,EAAYE,EACZD,EAAO,EACP,QACF,CACF,CACIL,IACFE,GAAOA,EAAIK,OAAS,EAAI,GAAGN,MAAgB,KAC3CE,EAAoB,EAExB,MACMD,EAAIK,OAAS,EACfL,GAAO,GAAGD,IAAYF,EAAKY,MAAMP,EAAY,EAAGE,KAEhDJ,EAAMH,EAAKY,MAAMP,EAAY,EAAGE,GAClCH,EAAoBG,EAAIF,EAAY,EAEtCA,EAAYE,EACZD,EAAO,CACT,MA7EoB,KA6ETV,IAA+B,IAAVU,IAC5BA,EAEFA,GAAQ,CAEZ,CACA,OAAOH,CACT,CAaA,SAASU,EAAQC,EAAKC,GACpB,MAAMC,EAAMD,EAAWC,KAAOD,EAAWE,KACnCC,EAAOH,EAAWG,MACtB,GAAGH,EAAWI,MAAQ,KAAKJ,EAAWK,KAAO,KAC/C,OAAKJ,EAGEA,IAAQD,EAAWE,KAAO,GAAGD,IAAME,IAAS,GAAGF,IAAMF,IAAMI,IAFzDA,CAGX,CAEA,MAAMG,EAAS,CAMbC,WAAWC,GACT,IAAIC,EAAiB,GACjBC,EAAe,GACfC,GAAmB,EAEvB,IAAK,IAAInB,EAAIgB,EAAKf,OAAS,EAAGD,IAAM,EAAGA,IAAK,CAC1C,IAAIP,EACJ,GAAIO,GAAK,GAIP,GAHAP,EAAOuB,EAAKhB,GAGQ,IAAhBP,EAAKQ,OACP,cAEiC,IAA1BgB,EAAehB,OACxBR,EAAOR,EAAQmC,OAOf3B,EAAOR,EAAQoC,IAAI,IAAIJ,MAAqBhC,EAAQmC,YAIvCE,IAAT7B,GACCA,EAAKY,MAAM,EAAG,GAAGkB,gBAAkBN,EAAeM,eA1IvC,KA2IZ9B,EAAKS,WAAW,MAClBT,EAAO,GAAGwB,QAId,MAAMO,EAAM/B,EAAKQ,OACjB,IAAIwB,EAAU,EACVC,EAAS,GACTC,GAAa,EACjB,MAAMtC,EAAOI,EAAKS,WAAW,GAG7B,GAAY,IAARsB,EACEpC,EAAgBC,KAElBoC,EAAU,EACVE,GAAa,QAEV,GAAIvC,EAAgBC,GAOzB,GAFAsC,GAAa,EAETvC,EAAgBK,EAAKS,WAAW,IAAK,CAEvC,IAAI0B,EAAI,EACJC,EAAOD,EAEX,KAAOA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CACzB,MAAMC,EAAYrC,EAAKY,MAAMwB,EAAMD,GAInC,IAFAC,EAAOD,EAEAA,EAAIJ,GACJpC,EAAgBK,EAAKS,WAAW0B,KACrCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEEA,IAAMJ,GAAOI,IAAMC,IAErBH,EACE,OAAOI,MAAcrC,EAAKY,MAAMwB,EAAMD,KACxCH,EAAUG,EAEd,CACF,CACF,MACEH,EAAU,OAEHlC,EAAoBF,IAxMb,KAyMNI,EAAKS,WAAW,KAE1BwB,EAASjC,EAAKY,MAAM,EAAG,GACvBoB,EAAU,EACND,EAAM,GAAKpC,EAAgBK,EAAKS,WAAW,MAG7CyB,GAAa,EACbF,EAAU,IAId,GAAIC,EAAOzB,OAAS,EAClB,GAAIgB,EAAehB,OAAS,GAC1B,GAAIyB,EAAOH,gBAAkBN,EAAeM,cAE1C,cAEFN,EAAiBS,EAIrB,GAAIP,GACF,GAAIF,EAAehB,OAAS,EAC1B,WAKF,GAHAiB,EACE,GAAGzB,EAAKY,MAAMoB,OAAaP,IAC7BC,EAAmBQ,EACfA,GAAcV,EAAehB,OAAS,EACxC,KAGN,CAUA,OAHAiB,EAAe1B,EAAgB0B,GAAeC,EAAkB,KACjC/B,GAExB+B,EACL,GAAGF,MAAmBC,IACtB,GAAGD,IAAiBC,KAAkB,GAC1C,EAMAa,UAAUtC,GACR,MAAM+B,EAAM/B,EAAKQ,OACjB,GAAY,IAARuB,EACF,MAAO,IACT,IACIE,EADAD,EAAU,EAEVE,GAAa,EACjB,MAAMtC,EAAOI,EAAKS,WAAW,GAG7B,GAAY,IAARsB,EAGF,OAAOlC,EAAqBD,GAAQ,KAAOI,EAE7C,GAAIL,EAAgBC,GAOlB,GAFAsC,GAAa,EAETvC,EAAgBK,EAAKS,WAAW,IAAK,CAEvC,IAAI0B,EAAI,EACJC,EAAOD,EAEX,KAAOA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CACzB,MAAMC,EAAYrC,EAAKY,MAAMwB,EAAMD,GAInC,IAFAC,EAAOD,EAEAA,EAAIJ,GACJpC,EAAgBK,EAAKS,WAAW0B,KACrCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,IAAMJ,EAIR,MAAO,OAAOM,MAAcrC,EAAKY,MAAMwB,OAErCD,IAAMC,IAERH,EACE,OAAOI,MAAcrC,EAAKY,MAAMwB,EAAMD,KACxCH,EAAUG,EAEd,CACF,CACF,MACEH,EAAU,OAEHlC,EAAoBF,IA9TX,KA+TTI,EAAKS,WAAW,KAEzBwB,EAASjC,EAAKY,MAAM,EAAG,GACvBoB,EAAU,EACND,EAAM,GAAKpC,EAAgBK,EAAKS,WAAW,MAG7CyB,GAAa,EACbF,EAAU,IAId,IAAIO,EAAOP,EAAUD,EACnBhC,EAAgBC,EAAKY,MAAMoB,IACVE,EAAY,KAAMvC,GACnC,GAMF,OALoB,IAAhB4C,EAAK/B,QAAiB0B,IACxBK,EAAO,KACLA,EAAK/B,OAAS,GACdb,EAAgBK,EAAKS,WAAWsB,EAAM,MACxCQ,GAAQ,WACKV,IAAXI,EACKC,EAAa,KAAKK,IAASA,EAE7BL,EAAa,GAAGD,MAAWM,IAAS,GAAGN,IAASM,GACzD,EAMAL,WAAWlC,GACT,MAAM+B,EAAM/B,EAAKQ,OACjB,GAAY,IAARuB,EACF,OAAO,EAET,MAAMnC,EAAOI,EAAKS,WAAW,GAC7B,OAAOd,EAAgBC,IAEpBmC,EAAM,GACPjC,EAAoBF,IAvWF,KAwWlBI,EAAKS,WAAW,IAChBd,EAAgBK,EAAKS,WAAW,GACpC,EAMA+B,QAAQjB,GACN,GAAoB,IAAhBA,EAAKf,OACP,MAAO,IAET,IAAIiC,EACAJ,EACJ,IAAK,IAAI9B,EAAI,EAAGA,EAAIgB,EAAKf,SAAUD,EAAG,CACpC,MAAMmC,EAAMnB,EAAKhB,GACbmC,EAAIlC,OAAS,SACAqB,IAAXY,EACFA,EAASJ,EAAYK,EAErBD,GAAU,KAAKC,IAErB,CAEA,QAAeb,IAAXY,EACF,MAAO,IAeT,IAAIE,GAAe,EACfC,EAAa,EACjB,GAAIjD,EAAgB0C,EAAU5B,WAAW,IAAK,GAC1CmC,EACF,MAAMC,EAAWR,EAAU7B,OACvBqC,EAAW,GACXlD,EAAgB0C,EAAU5B,WAAW,QACrCmC,EACEC,EAAW,IACTlD,EAAgB0C,EAAU5B,WAAW,MACrCmC,EAGFD,GAAe,GAIvB,CACA,GAAIA,EAAc,CAEhB,KAAOC,EAAaH,EAAOjC,QACpBb,EAAgB8C,EAAOhC,WAAWmC,KACvCA,IAIEA,GAAc,IAChBH,EAAS,KAAKA,EAAO7B,MAAMgC,KAC/B,CAEA,OAAOvB,EAAOiB,UAAUG,EAC1B,EAWAK,SAASC,EAAMC,GACb,GAAID,IAASC,EACX,MAAO,GAET,MAAMC,EAAW5B,EAAOC,QAAQyB,GAC1BG,EAAS7B,EAAOC,QAAQ0B,GAE9B,GAAIC,IAAaC,EACf,MAAO,GAKT,IAHAH,EAAOE,EAASnB,kBAChBkB,EAAKE,EAAOpB,eAGV,MAAO,GAGT,IAAIqB,EAAY,EAChB,KAAOA,EAAYJ,EAAKvC,QA7cJ,KA8cbuC,EAAKtC,WAAW0C,IACrBA,IAGF,IAAIC,EAAUL,EAAKvC,OACnB,KACE4C,EAAU,EAAID,GApdI,KAqdlBJ,EAAKtC,WAAW2C,EAAU,IAE1BA,IAEF,MAAMC,EAAUD,EAAUD,EAG1B,IAAIG,EAAU,EACd,KAAOA,EAAUN,EAAGxC,QA7dA,KA8dbwC,EAAGvC,WAAW6C,IACnBA,IAGF,IAAIC,EAAQP,EAAGxC,OACf,KAAO+C,EAAQ,EAAID,GAneC,KAoebN,EAAGvC,WAAW8C,EAAQ,IAC3BA,IAEF,MAAMC,EAAQD,EAAQD,EAGhB9C,EAAS6C,EAAUG,EAAQH,EAAUG,EAC3C,IAAIC,GAAiB,EACjBlD,EAAI,EACR,KAAOA,EAAIC,EAAQD,IAAK,CACtB,MAAMmD,EAAWX,EAAKtC,WAAW0C,EAAY5C,GAC7C,GAAImD,IAAaV,EAAGvC,WAAW6C,EAAU/C,GACvC,MAhfgB,KAifTmD,IACPD,EAAgBlD,EACpB,CAIA,GAAIA,IAAMC,GACR,IAAuB,IAAnBiD,EACF,OAAOP,MACJ,CACL,GAAIM,EAAQhD,EAAQ,CAClB,GA5fgB,KA4fZwC,EAAGvC,WAAW6C,EAAU/C,GAI1B,OAAO2C,EAAOtC,MAAM0C,EAAU/C,EAAI,GAEpC,GAAU,IAANA,EAGF,OAAO2C,EAAOtC,MAAM0C,EAAU/C,EAElC,CACI8C,EAAU7C,IAxgBI,KAygBZuC,EAAKtC,WAAW0C,EAAY5C,GAI9BkD,EAAgBlD,EACD,IAANA,IAGTkD,EAAgB,KAGG,IAAnBA,IACFA,EAAgB,EACpB,CAEA,IAAIE,EAAM,GAGV,IAAKpD,EAAI4C,EAAYM,EAAgB,EAAGlD,GAAK6C,IAAW7C,EAClDA,IAAM6C,GA5hBQ,KA6hBdL,EAAKtC,WAAWF,KAClBoD,GAAsB,IAAfA,EAAInD,OAAe,KAAO,QAQrC,OAJA8C,GAAWG,EAIPE,EAAInD,OAAS,EACR,GAAGmD,IAAMT,EAAOtC,MAAM0C,EAASC,MAviBpB,KAyiBhBL,EAAOzC,WAAW6C,MAClBA,EACGJ,EAAOtC,MAAM0C,EAASC,GAC/B,EAMAK,iBAAiB5D,GAEf,GAAoB,iBAATA,GAAqC,IAAhBA,EAAKQ,OACnC,OAAOR,EAET,MAAM6D,EAAexC,EAAOC,QAAQtB,GAEpC,GAAI6D,EAAarD,QAAU,EACzB,OAAOR,EAET,GA5jBoB,KA4jBhB6D,EAAapD,WAAW,IAE1B,GA9jBkB,KA8jBdoD,EAAapD,WAAW,GAA4B,CACtD,MAAMb,EAAOiE,EAAapD,WAAW,GACrC,GA9jBgB,KA8jBZb,GAlkBY,KAkkBmBA,EAEjC,MAAO,eAAeiE,EAAajD,MAAM,IAE7C,OACK,GACLd,EAAoB+D,EAAapD,WAAW,KArkB1B,KAskBlBoD,EAAapD,WAAW,IAvkBN,KAwkBlBoD,EAAapD,WAAW,GAGxB,MAAO,UAAUoD,IAGnB,OAAO7D,CACT,EAMA8D,QAAQ9D,GACN,MAAM+B,EAAM/B,EAAKQ,OACjB,GAAY,IAARuB,EACF,MAAO,IACT,IAAIC,GAAW,EACX+B,EAAS,EACb,MAAMnE,EAAOI,EAAKS,WAAW,GAE7B,GAAY,IAARsB,EAGF,OAAOpC,EAAgBC,GAAQI,EAAO,IAIxC,GAAIL,EAAgBC,IAKlB,GAFAoC,EAAU+B,EAAS,EAEfpE,EAAgBK,EAAKS,WAAW,IAAK,CAEvC,IAAI0B,EAAI,EACJC,EAAOD,EAEX,KAAOA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,GACJpC,EAAgBK,EAAKS,WAAW0B,KACrCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,IAAMJ,EAER,OAAO/B,EAELmC,IAAMC,IAKRJ,EAAU+B,EAAS5B,EAAI,EAE3B,CACF,CACF,OAESrC,EAAoBF,IAhpBX,KAipBTI,EAAKS,WAAW,KACzBuB,EACED,EAAM,GAAKpC,EAAgBK,EAAKS,WAAW,IAAM,EAAI,EACvDsD,EAAS/B,GAGX,IAAIgC,GAAO,EACPC,GAAe,EACnB,IAAK,IAAI1D,EAAIwB,EAAM,EAAGxB,GAAKwD,IAAUxD,EACnC,GAAIZ,EAAgBK,EAAKS,WAAWF,KAClC,IAAK0D,EAAc,CACjBD,EAAMzD,EACN,KACF,OAGA0D,GAAe,EAInB,IAAa,IAATD,EAAY,CACd,IAAiB,IAAbhC,EACF,MAAO,IAETgC,EAAMhC,CACR,CACA,OAAOhC,EAAKY,MAAM,EAAGoD,EACvB,EAOAE,SAASlE,EAAMoB,GACb,IAAI+C,EAAQ,EACRH,GAAO,EACPC,GAAe,EAWnB,GANIjE,EAAKQ,QAAU,GACfV,EAAoBE,EAAKS,WAAW,KA5rBpB,KA6rBhBT,EAAKS,WAAW,KAClB0D,EAAQ,QAGEtC,IAART,GAAqBA,EAAIZ,OAAS,GAAKY,EAAIZ,QAAUR,EAAKQ,OAAQ,CACpE,GAAIY,IAAQpB,EACV,MAAO,GACT,IAAIoE,EAAShD,EAAIZ,OAAS,EACtB6D,GAAoB,EACxB,IAAK,IAAI9D,EAAIP,EAAKQ,OAAS,EAAGD,GAAK4D,IAAS5D,EAAG,CAC7C,MAAMX,EAAOI,EAAKS,WAAWF,GAC7B,GAAIZ,EAAgBC,IAGlB,IAAKqE,EAAc,CACjBE,EAAQ5D,EAAI,EACZ,KACF,OAE0B,IAAtB8D,IAGFJ,GAAe,EACfI,EAAmB9D,EAAI,GAErB6D,GAAU,IAERxE,IAASwB,EAAIX,WAAW2D,IACR,KAAZA,IAGJJ,EAAMzD,IAKR6D,GAAU,EACVJ,EAAMK,GAId,CAMA,OAJIF,IAAUH,EACZA,EAAMK,GACU,IAATL,IACPA,EAAMhE,EAAKQ,QACNR,EAAKY,MAAMuD,EAAOH,EAC3B,CACA,IAAK,IAAIzD,EAAIP,EAAKQ,OAAS,EAAGD,GAAK4D,IAAS5D,EAC1C,GAAIZ,EAAgBK,EAAKS,WAAWF,KAGlC,IAAK0D,EAAc,CACjBE,EAAQ5D,EAAI,EACZ,KACF,OACkB,IAATyD,IAGTC,GAAe,EACfD,EAAMzD,EAAI,GAId,OAAa,IAATyD,EACK,GACFhE,EAAKY,MAAMuD,EAAOH,EAC3B,EAMAM,QAAQtE,GACN,IAAImE,EAAQ,EACRI,GAAY,EACZC,EAAY,EACZR,GAAO,EACPC,GAAe,EAGfQ,EAAc,EAMdzE,EAAKQ,QAAU,GArxBC,KAsxBhBR,EAAKS,WAAW,IAChBX,EAAoBE,EAAKS,WAAW,MACtC0D,EAAQK,EAAY,GAGtB,IAAK,IAAIjE,EAAIP,EAAKQ,OAAS,EAAGD,GAAK4D,IAAS5D,EAAG,CAC7C,MAAMX,EAAOI,EAAKS,WAAWF,GAC7B,GAAIZ,EAAgBC,IAGlB,IAAKqE,EAAc,CACjBO,EAAYjE,EAAI,EAChB,KACF,OAGW,IAATyD,IAGFC,GAAe,EACfD,EAAMzD,EAAI,GA7yBM,KA+yBdX,GAEgB,IAAd2E,EACFA,EAAWhE,EACY,IAAhBkE,IACPA,EAAc,IACO,IAAdF,IAGTE,GAAe,EAEnB,CAEA,OAAkB,IAAdF,IACS,IAATP,GAEgB,IAAhBS,GAEiB,IAAhBA,GACAF,IAAaP,EAAM,GACnBO,IAAaC,EAAY,EACrB,GAEFxE,EAAKY,MAAM2D,EAAUP,EAC9B,EAEAU,OAAQ7D,EAAQ8D,KAAK,KAAM,MAY3BC,MAAM5E,GACJ,MAAM6E,EAAM,CAAE5D,KAAM,GAAID,IAAK,GAAIE,KAAM,GAAIE,IAAK,GAAID,KAAM,IAC1D,GAAoB,IAAhBnB,EAAKQ,OACP,OAAOqE,EAET,MAAM9C,EAAM/B,EAAKQ,OACjB,IAAIwB,EAAU,EACVpC,EAAOI,EAAKS,WAAW,GAE3B,GAAY,IAARsB,EACF,OAAIpC,EAAgBC,IAGlBiF,EAAI5D,KAAO4D,EAAI7D,IAAMhB,EACd6E,IAETA,EAAI3D,KAAO2D,EAAI1D,KAAOnB,EACf6E,GAGT,GAAIlF,EAAgBC,IAIlB,GADAoC,EAAU,EACNrC,EAAgBK,EAAKS,WAAW,IAAK,CAEvC,IAAI0B,EAAI,EACJC,EAAOD,EAEX,KAAOA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,GACJpC,EAAgBK,EAAKS,WAAW0B,KACrCA,IAEF,GAAIA,EAAIJ,GAAOI,IAAMC,EAAM,CAIzB,IAFAA,EAAOD,EAEAA,EAAIJ,IACHpC,EAAgBK,EAAKS,WAAW0B,KACtCA,IAEEA,IAAMJ,EAERC,EAAUG,EACDA,IAAMC,IAEfJ,EAAUG,EAAI,EAElB,CACF,CACF,OACK,GAAIrC,EAAoBF,IA74BX,KA84BTI,EAAKS,WAAW,GAAmB,CAE5C,GAAIsB,GAAO,EAIT,OADA8C,EAAI5D,KAAO4D,EAAI7D,IAAMhB,EACd6E,EAGT,GADA7C,EAAU,EACNrC,EAAgBK,EAAKS,WAAW,IAAK,CACvC,GAAY,IAARsB,EAIF,OADA8C,EAAI5D,KAAO4D,EAAI7D,IAAMhB,EACd6E,EAET7C,EAAU,CACZ,CACF,CACIA,EAAU,IACZ6C,EAAI5D,KAAOjB,EAAKY,MAAM,EAAGoB,IAE3B,IAAIuC,GAAY,EACZC,EAAYxC,EACZgC,GAAO,EACPC,GAAe,EACf1D,EAAIP,EAAKQ,OAAS,EAIlBiE,EAAc,EAGlB,KAAOlE,GAAKyB,IAAWzB,EAErB,GADAX,EAAOI,EAAKS,WAAWF,GACnBZ,EAAgBC,IAGlB,IAAKqE,EAAc,CACjBO,EAAYjE,EAAI,EAChB,KACF,OAGW,IAATyD,IAGFC,GAAe,EACfD,EAAMzD,EAAI,GAj8BM,KAm8BdX,GAEgB,IAAd2E,EACFA,EAAWhE,EACY,IAAhBkE,IACPA,EAAc,IACO,IAAdF,IAGTE,GAAe,GA4BnB,OAxBa,IAATT,KACgB,IAAdO,GAEgB,IAAhBE,GAEiB,IAAhBA,GACAF,IAAaP,EAAM,GACnBO,IAAaC,EAAY,EAC5BK,EAAI3D,KAAO2D,EAAI1D,KAAOnB,EAAKY,MAAM4D,EAAWR,IAE5Ca,EAAI1D,KAAOnB,EAAKY,MAAM4D,EAAWD,GACjCM,EAAI3D,KAAOlB,EAAKY,MAAM4D,EAAWR,GACjCa,EAAIzD,IAAMpB,EAAKY,MAAM2D,EAAUP,KAQjCa,EAAI7D,IADFwD,EAAY,GAAKA,IAAcxC,EACvBhC,EAAKY,MAAM,EAAG4D,EAAY,GAE1BK,EAAI5D,KAET4D,CACT,EAEA/D,IAAK,KACLgE,UAAW,IACXC,MAAO,KACPC,MAAO,MAGHC,EAAW,MACf,GAAIxF,EAAiB,CAGnB,MAAMyF,EAAS,MACf,MAAO,KACL,MAAMvD,EAAMnC,EAAQmC,MAAMwD,QAAQD,EAAQ,KAC1C,OAAOvD,EAAIf,MAAMe,EAAIyD,QAAQ,KAAK,CAEtC,CAGA,MAAO,IAAM5F,EAAQmC,KACtB,EAbgB,GAeX0D,EAAS,CAMb/D,WAAWC,GACT,IAAIsC,EAAe,GACfnC,GAAmB,EAEvB,IAAK,IAAInB,EAAIgB,EAAKf,OAAS,EAAGD,IAAM,IAAMmB,EAAkBnB,IAAK,CAC/D,MAAMP,EAAOO,GAAK,EAAIgB,EAAKhB,GAAK0E,IAGZ,IAAhBjF,EAAKQ,SAITqD,EAAe,GAAG7D,KAAQ6D,IAC1BnC,EAlhCkB,KAmhChB1B,EAAKS,WAAW,GACpB,CASA,OAHAoD,EAAe9D,EAAgB8D,GAAenC,EAAkB,IACjC7B,GAE3B6B,EACK,IAAImC,IAENA,EAAarD,OAAS,EAAIqD,EAAe,GAClD,EAMAvB,UAAUtC,GACR,GAAoB,IAAhBA,EAAKQ,OACP,MAAO,IAET,MAAM0B,EA3iCc,KA4iClBlC,EAAKS,WAAW,GACZ6E,EA7iCc,KA8iClBtF,EAAKS,WAAWT,EAAKQ,OAAS,GAKhC,OAAoB,KAFpBR,EAAOD,EAAgBC,GAAOkC,EAAY,IAAKrC,IAEtCW,OACH0B,EACK,IACFoD,EAAoB,KAAO,KAEhCA,IACFtF,GAAQ,KAEHkC,EAAa,IAAIlC,IAASA,EACnC,EAMAkC,WAAWlC,GACFA,EAAKQ,OAAS,GAnkCD,KAokCbR,EAAKS,WAAW,GAOzB+B,QAAQjB,GACN,GAAoB,IAAhBA,EAAKf,OACP,MAAO,IACT,IAAIiC,EACJ,IAAK,IAAIlC,EAAI,EAAGA,EAAIgB,EAAKf,SAAUD,EAAG,CACpC,MAAMmC,EAAMnB,EAAKhB,GACbmC,EAAIlC,OAAS,SACAqB,IAAXY,EACFA,EAASC,EAETD,GAAU,IAAIC,IAEpB,CACA,YAAeb,IAAXY,EACK,IACF4C,EAAO/C,UAAUG,EAC1B,EAOAK,SAASC,EAAMC,GACb,GAAID,IAASC,EACX,MAAO,GAMT,IAHAD,EAAOsC,EAAO/D,QAAQyB,OACtBC,EAAKqC,EAAO/D,QAAQ0B,IAGlB,MAAO,GAET,MACMI,EAAUL,EAAKvC,OACf6C,EAAUD,EAFE,EAIZI,EAAQR,EAAGxC,OADD,EAIVA,EAAU6C,EAAUG,EAAQH,EAAUG,EAC5C,IAAIC,GAAiB,EACjBlD,EAAI,EACR,KAAOA,EAAIC,EAAQD,IAAK,CACtB,MAAMmD,EAAWX,EAAKtC,WAXN,EAW6BF,GAC7C,GAAImD,IAAaV,EAAGvC,WATN,EAS2BF,GACvC,MA1nCgB,KA2nCTmD,IACPD,EAAgBlD,EACpB,CACA,GAAIA,IAAMC,EACR,GAAIgD,EAAQhD,EAAQ,CAClB,GAhoCgB,KAgoCZwC,EAAGvC,WAhBK,EAgBgBF,GAG1B,OAAOyC,EAAGpC,MAnBA,EAmBgBL,EAAI,GAEhC,GAAU,IAANA,EAGF,OAAOyC,EAAGpC,MAxBA,EAwBgBL,EAE9B,MAAW8C,EAAU7C,IA1oCH,KA2oCZuC,EAAKtC,WA9BK,EA8BkBF,GAI9BkD,EAAgBlD,EACD,IAANA,IAGTkD,EAAgB,IAKtB,IAAIE,EAAM,GAGV,IAAKpD,EA9Ca,EA8CGkD,EAAgB,EAAGlD,GAAK6C,IAAW7C,EAClDA,IAAM6C,GA5pCQ,KA6pCdL,EAAKtC,WAAWF,KAClBoD,GAAsB,IAAfA,EAAInD,OAAe,KAAO,OAMrC,MAAO,GAAGmD,IAAMX,EAAGpC,MApDH,EAoDmB6C,IACrC,EAMAG,iBAAiB5D,GAERA,EAOT8D,QAAQ9D,GACN,GAAoB,IAAhBA,EAAKQ,OACP,MAAO,IACT,MAAM+E,EAvrCc,KAurCJvF,EAAKS,WAAW,GAChC,IAAIuD,GAAO,EACPC,GAAe,EACnB,IAAK,IAAI1D,EAAIP,EAAKQ,OAAS,EAAGD,GAAK,IAAKA,EACtC,GA3rCkB,KA2rCdP,EAAKS,WAAWF,IAClB,IAAK0D,EAAc,CACjBD,EAAMzD,EACN,KACF,OAGA0D,GAAe,EAInB,OAAa,IAATD,EACKuB,EAAU,IAAM,IACrBA,GAAmB,IAARvB,EACN,KACFhE,EAAKY,MAAM,EAAGoD,EACvB,EAOAE,SAASlE,EAAMoB,GACb,IAAI+C,EAAQ,EACRH,GAAO,EACPC,GAAe,EAEnB,QAAYpC,IAART,GAAqBA,EAAIZ,OAAS,GAAKY,EAAIZ,QAAUR,EAAKQ,OAAQ,CACpE,GAAIY,IAAQpB,EACV,MAAO,GACT,IAAIoE,EAAShD,EAAIZ,OAAS,EACtB6D,GAAoB,EACxB,IAAK,IAAI9D,EAAIP,EAAKQ,OAAS,EAAGD,GAAK,IAAKA,EAAG,CACzC,MAAMX,EAAOI,EAAKS,WAAWF,GAC7B,GA9tCgB,KA8tCZX,GAGF,IAAKqE,EAAc,CACjBE,EAAQ5D,EAAI,EACZ,KACF,OAE0B,IAAtB8D,IAGFJ,GAAe,EACfI,EAAmB9D,EAAI,GAErB6D,GAAU,IAERxE,IAASwB,EAAIX,WAAW2D,IACR,KAAZA,IAGJJ,EAAMzD,IAKR6D,GAAU,EACVJ,EAAMK,GAId,CAMA,OAJIF,IAAUH,EACZA,EAAMK,GACU,IAATL,IACPA,EAAMhE,EAAKQ,QACNR,EAAKY,MAAMuD,EAAOH,EAC3B,CACA,IAAK,IAAIzD,EAAIP,EAAKQ,OAAS,EAAGD,GAAK,IAAKA,EACtC,GArwCkB,KAqwCdP,EAAKS,WAAWF,IAGlB,IAAK0D,EAAc,CACjBE,EAAQ5D,EAAI,EACZ,KACF,OACkB,IAATyD,IAGTC,GAAe,EACfD,EAAMzD,EAAI,GAId,OAAa,IAATyD,EACK,GACFhE,EAAKY,MAAMuD,EAAOH,EAC3B,EAMAM,QAAQtE,GACN,IAAIuE,GAAY,EACZC,EAAY,EACZR,GAAO,EACPC,GAAe,EAGfQ,EAAc,EAClB,IAAK,IAAIlE,EAAIP,EAAKQ,OAAS,EAAGD,GAAK,IAAKA,EAAG,CACzC,MAAMX,EAAOI,EAAKS,WAAWF,GAC7B,GAvyCkB,KAuyCdX,GASS,IAAToE,IAGFC,GAAe,EACfD,EAAMzD,EAAI,GArzCM,KAuzCdX,GAEgB,IAAd2E,EACFA,EAAWhE,EACY,IAAhBkE,IACPA,EAAc,IACO,IAAdF,IAGTE,GAAe,QArBf,IAAKR,EAAc,CACjBO,EAAYjE,EAAI,EAChB,KACF,CAoBJ,CAEA,OAAkB,IAAdgE,IACS,IAATP,GAEgB,IAAhBS,GAEiB,IAAhBA,GACAF,IAAaP,EAAM,GACnBO,IAAaC,EAAY,EACrB,GAEFxE,EAAKY,MAAM2D,EAAUP,EAC9B,EAEAU,OAAQ7D,EAAQ8D,KAAK,KAAM,KAY3BC,MAAM5E,GACJ,MAAM6E,EAAM,CAAE5D,KAAM,GAAID,IAAK,GAAIE,KAAM,GAAIE,IAAK,GAAID,KAAM,IAC1D,GAAoB,IAAhBnB,EAAKQ,OACP,OAAOqE,EACT,MAAM3C,EAh2Cc,KAi2ClBlC,EAAKS,WAAW,GAClB,IAAI0D,EACAjC,GACF2C,EAAI5D,KAAO,IACXkD,EAAQ,GAERA,EAAQ,EAEV,IAAII,GAAY,EACZC,EAAY,EACZR,GAAO,EACPC,GAAe,EACf1D,EAAIP,EAAKQ,OAAS,EAIlBiE,EAAc,EAGlB,KAAOlE,GAAK4D,IAAS5D,EAAG,CACtB,MAAMX,EAAOI,EAAKS,WAAWF,GAC7B,GAt3CkB,KAs3CdX,GASS,IAAToE,IAGFC,GAAe,EACfD,EAAMzD,EAAI,GAp4CM,KAs4CdX,GAEgB,IAAd2E,EACFA,EAAWhE,EACY,IAAhBkE,IACPA,EAAc,IACO,IAAdF,IAGTE,GAAe,QArBf,IAAKR,EAAc,CACjBO,EAAYjE,EAAI,EAChB,KACF,CAoBJ,CAEA,IAAa,IAATyD,EAAY,CACd,MAAMG,EAAsB,IAAdK,GAAmBtC,EAAa,EAAIsC,GAChC,IAAdD,GAEgB,IAAhBE,GAEiB,IAAhBA,GACDF,IAAaP,EAAM,GACnBO,IAAaC,EAAY,EAC3BK,EAAI3D,KAAO2D,EAAI1D,KAAOnB,EAAKY,MAAMuD,EAAOH,IAExCa,EAAI1D,KAAOnB,EAAKY,MAAMuD,EAAOI,GAC7BM,EAAI3D,KAAOlB,EAAKY,MAAMuD,EAAOH,GAC7Ba,EAAIzD,IAAMpB,EAAKY,MAAM2D,EAAUP,GAEnC,CAOA,OALIQ,EAAY,EACdK,EAAI7D,IAAMhB,EAAKY,MAAM,EAAG4D,EAAY,GAC7BtC,IACP2C,EAAI7D,IAAM,KAEL6D,CACT,EAEA/D,IAAK,IACLgE,UAAW,IACXC,MAAO,KACPC,MAAO,MAGTK,EAAON,MAAQ1D,EAAO0D,MAAQ1D,EAC9BgE,EAAOL,MAAQ3D,EAAO2D,MAAQK,EAE9B,MAAMG,EAAO/F,EAAkB4B,EAASgE,iBACzBG,SACR,MAAMlE,QACJgB,UACEJ,WACCM,KACNM,SACIc,iBACQE,QACTI,SACCI,QACDI,OACDE,MACD9D,IACFgE,UACMC,MACJC,OAEHQ"}
✄
import e from"process";const t="win32"===e.platform;function r(e){return 47===e||92===e}function o(e){return 47===e}function n(e){return e>=65&&e<=90||e>=97&&e<=122}function l(e,t,r,o){let n="",l=0,i=-1,c=0,a=0;for(let s=0;s<=e.length;++s){if(s<e.length)a=e.charCodeAt(s);else{if(o(a))break;a=47}if(o(a)){if(i===s-1||1===c);else if(2===c){if(n.length<2||2!==l||46!==n.charCodeAt(n.length-1)||46!==n.charCodeAt(n.length-2)){if(n.length>2){const e=n.lastIndexOf(r);-1===e?(n="",l=0):(n=n.slice(0,e),l=n.length-1-n.lastIndexOf(r)),i=s,c=0;continue}if(0!==n.length){n="",l=0,i=s,c=0;continue}}t&&(n+=n.length>0?`${r}..`:"..",l=2)}else n.length>0?n+=`${r}${e.slice(i+1,s)}`:n=e.slice(i+1,s),l=s-i-1;i=s,c=0}else 46===a&&-1!==c?++c:c=-1}return n}function i(e,t){const r=t.dir||t.root,o=t.base||`${t.name||""}${t.ext||""}`;return r?r===t.root?`${r}${o}`:`${r}${e}${o}`:o}const c={resolve(...t){let o="",i="",c=!1;for(let l=t.length-1;l>=-1;l--){let a;if(l>=0){if(a=t[l],0===a.length)continue}else 0===o.length?a=e.cwd():(a=e.env[`=${o}`]||e.cwd(),(void 0===a||a.slice(0,2).toLowerCase()!==o.toLowerCase()&&92===a.charCodeAt(2))&&(a=`${o}\\`));const s=a.length;let h=0,f="",d=!1;const C=a.charCodeAt(0);if(1===s)r(C)&&(h=1,d=!0);else if(r(C))if(d=!0,r(a.charCodeAt(1))){let e=2,t=e;for(;e<s&&!r(a.charCodeAt(e));)e++;if(e<s&&e!==t){const o=a.slice(t,e);for(t=e;e<s&&r(a.charCodeAt(e));)e++;if(e<s&&e!==t){for(t=e;e<s&&!r(a.charCodeAt(e));)e++;e!==s&&e===t||(f=`\\\\${o}\\${a.slice(t,e)}`,h=e)}}}else h=1;else n(C)&&58===a.charCodeAt(1)&&(f=a.slice(0,2),h=2,s>2&&r(a.charCodeAt(2))&&(d=!0,h=3));if(f.length>0)if(o.length>0){if(f.toLowerCase()!==o.toLowerCase())continue}else o=f;if(c){if(o.length>0)break}else if(i=`${a.slice(h)}\\${i}`,c=d,d&&o.length>0)break}return i=l(i,!c,"\\",r),c?`${o}\\${i}`:`${o}${i}`||"."},normalize(e){const t=e.length;if(0===t)return".";let i,c=0,a=!1;const s=e.charCodeAt(0);if(1===t)return o(s)?"\\":e;if(r(s))if(a=!0,r(e.charCodeAt(1))){let o=2,n=o;for(;o<t&&!r(e.charCodeAt(o));)o++;if(o<t&&o!==n){const l=e.slice(n,o);for(n=o;o<t&&r(e.charCodeAt(o));)o++;if(o<t&&o!==n){for(n=o;o<t&&!r(e.charCodeAt(o));)o++;if(o===t)return`\\\\${l}\\${e.slice(n)}\\`;o!==n&&(i=`\\\\${l}\\${e.slice(n,o)}`,c=o)}}}else c=1;else n(s)&&58===e.charCodeAt(1)&&(i=e.slice(0,2),c=2,t>2&&r(e.charCodeAt(2))&&(a=!0,c=3));let h=c<t?l(e.slice(c),!a,"\\",r):"";return 0!==h.length||a||(h="."),h.length>0&&r(e.charCodeAt(t-1))&&(h+="\\"),void 0===i?a?`\\${h}`:h:a?`${i}\\${h}`:`${i}${h}`},isAbsolute(e){const t=e.length;if(0===t)return!1;const o=e.charCodeAt(0);return r(o)||t>2&&n(o)&&58===e.charCodeAt(1)&&r(e.charCodeAt(2))},join(...e){if(0===e.length)return".";let t,o;for(let r=0;r<e.length;++r){const n=e[r];n.length>0&&(void 0===t?t=o=n:t+=`\\${n}`)}if(void 0===t)return".";let n=!0,l=0;if(r(o.charCodeAt(0))){++l;const e=o.length;e>1&&r(o.charCodeAt(1))&&(++l,e>2&&(r(o.charCodeAt(2))?++l:n=!1))}if(n){for(;l<t.length&&r(t.charCodeAt(l));)l++;l>=2&&(t=`\\${t.slice(l)}`)}return c.normalize(t)},relative(e,t){if(e===t)return"";const r=c.resolve(e),o=c.resolve(t);if(r===o)return"";if((e=r.toLowerCase())===(t=o.toLowerCase()))return"";let n=0;for(;n<e.length&&92===e.charCodeAt(n);)n++;let l=e.length;for(;l-1>n&&92===e.charCodeAt(l-1);)l--;const i=l-n;let a=0;for(;a<t.length&&92===t.charCodeAt(a);)a++;let s=t.length;for(;s-1>a&&92===t.charCodeAt(s-1);)s--;const h=s-a,f=i<h?i:h;let d=-1,C=0;for(;C<f;C++){const r=e.charCodeAt(n+C);if(r!==t.charCodeAt(a+C))break;92===r&&(d=C)}if(C!==f){if(-1===d)return o}else{if(h>f){if(92===t.charCodeAt(a+C))return o.slice(a+C+1);if(2===C)return o.slice(a+C)}i>f&&(92===e.charCodeAt(n+C)?d=C:2===C&&(d=3)),-1===d&&(d=0)}let A="";for(C=n+d+1;C<=l;++C)C!==l&&92!==e.charCodeAt(C)||(A+=0===A.length?"..":"\\..");return a+=d,A.length>0?`${A}${o.slice(a,s)}`:(92===o.charCodeAt(a)&&++a,o.slice(a,s))},toNamespacedPath(e){if("string"!=typeof e||0===e.length)return e;const t=c.resolve(e);if(t.length<=2)return e;if(92===t.charCodeAt(0)){if(92===t.charCodeAt(1)){const e=t.charCodeAt(2);if(63!==e&&46!==e)return`\\\\?\\UNC\\${t.slice(2)}`}}else if(n(t.charCodeAt(0))&&58===t.charCodeAt(1)&&92===t.charCodeAt(2))return`\\\\?\\${t}`;return e},dirname(e){const t=e.length;if(0===t)return".";let o=-1,l=0;const i=e.charCodeAt(0);if(1===t)return r(i)?e:".";if(r(i)){if(o=l=1,r(e.charCodeAt(1))){let n=2,i=n;for(;n<t&&!r(e.charCodeAt(n));)n++;if(n<t&&n!==i){for(i=n;n<t&&r(e.charCodeAt(n));)n++;if(n<t&&n!==i){for(i=n;n<t&&!r(e.charCodeAt(n));)n++;if(n===t)return e;n!==i&&(o=l=n+1)}}}}else n(i)&&58===e.charCodeAt(1)&&(o=t>2&&r(e.charCodeAt(2))?3:2,l=o);let c=-1,a=!0;for(let o=t-1;o>=l;--o)if(r(e.charCodeAt(o))){if(!a){c=o;break}}else a=!1;if(-1===c){if(-1===o)return".";c=o}return e.slice(0,c)},basename(e,t){let o=0,l=-1,i=!0;if(e.length>=2&&n(e.charCodeAt(0))&&58===e.charCodeAt(1)&&(o=2),void 0!==t&&t.length>0&&t.length<=e.length){if(t===e)return"";let n=t.length-1,c=-1;for(let a=e.length-1;a>=o;--a){const s=e.charCodeAt(a);if(r(s)){if(!i){o=a+1;break}}else-1===c&&(i=!1,c=a+1),n>=0&&(s===t.charCodeAt(n)?-1==--n&&(l=a):(n=-1,l=c))}return o===l?l=c:-1===l&&(l=e.length),e.slice(o,l)}for(let t=e.length-1;t>=o;--t)if(r(e.charCodeAt(t))){if(!i){o=t+1;break}}else-1===l&&(i=!1,l=t+1);return-1===l?"":e.slice(o,l)},extname(e){let t=0,o=-1,l=0,i=-1,c=!0,a=0;e.length>=2&&58===e.charCodeAt(1)&&n(e.charCodeAt(0))&&(t=l=2);for(let n=e.length-1;n>=t;--n){const t=e.charCodeAt(n);if(r(t)){if(!c){l=n+1;break}}else-1===i&&(c=!1,i=n+1),46===t?-1===o?o=n:1!==a&&(a=1):-1!==o&&(a=-1)}return-1===o||-1===i||0===a||1===a&&o===i-1&&o===l+1?"":e.slice(o,i)},format:i.bind(null,"\\"),parse(e){const t={root:"",dir:"",base:"",ext:"",name:""};if(0===e.length)return t;const o=e.length;let l=0,i=e.charCodeAt(0);if(1===o)return r(i)?(t.root=t.dir=e,t):(t.base=t.name=e,t);if(r(i)){if(l=1,r(e.charCodeAt(1))){let t=2,n=t;for(;t<o&&!r(e.charCodeAt(t));)t++;if(t<o&&t!==n){for(n=t;t<o&&r(e.charCodeAt(t));)t++;if(t<o&&t!==n){for(n=t;t<o&&!r(e.charCodeAt(t));)t++;t===o?l=t:t!==n&&(l=t+1)}}}}else if(n(i)&&58===e.charCodeAt(1)){if(o<=2)return t.root=t.dir=e,t;if(l=2,r(e.charCodeAt(2))){if(3===o)return t.root=t.dir=e,t;l=3}}l>0&&(t.root=e.slice(0,l));let c=-1,a=l,s=-1,h=!0,f=e.length-1,d=0;for(;f>=l;--f)if(i=e.charCodeAt(f),r(i)){if(!h){a=f+1;break}}else-1===s&&(h=!1,s=f+1),46===i?-1===c?c=f:1!==d&&(d=1):-1!==c&&(d=-1);return-1!==s&&(-1===c||0===d||1===d&&c===s-1&&c===a+1?t.base=t.name=e.slice(a,s):(t.name=e.slice(a,c),t.base=e.slice(a,s),t.ext=e.slice(c,s))),t.dir=a>0&&a!==l?e.slice(0,a-1):t.root,t},sep:"\\",delimiter:";",win32:null,posix:null},a=(()=>{if(t){const t=/\\/g;return()=>{const r=e.cwd().replace(t,"/");return r.slice(r.indexOf("/"))}}return()=>e.cwd()})(),s={resolve(...e){let t="",r=!1;for(let o=e.length-1;o>=-1&&!r;o--){const n=o>=0?e[o]:a();0!==n.length&&(t=`${n}/${t}`,r=47===n.charCodeAt(0))}return t=l(t,!r,"/",o),r?`/${t}`:t.length>0?t:"."},normalize(e){if(0===e.length)return".";const t=47===e.charCodeAt(0),r=47===e.charCodeAt(e.length-1);return 0===(e=l(e,!t,"/",o)).length?t?"/":r?"./":".":(r&&(e+="/"),t?`/${e}`:e)},isAbsolute:e=>e.length>0&&47===e.charCodeAt(0),join(...e){if(0===e.length)return".";let t;for(let r=0;r<e.length;++r){const o=e[r];o.length>0&&(void 0===t?t=o:t+=`/${o}`)}return void 0===t?".":s.normalize(t)},relative(e,t){if(e===t)return"";if((e=s.resolve(e))===(t=s.resolve(t)))return"";const r=e.length,o=r-1,n=t.length-1,l=o<n?o:n;let i=-1,c=0;for(;c<l;c++){const r=e.charCodeAt(1+c);if(r!==t.charCodeAt(1+c))break;47===r&&(i=c)}if(c===l)if(n>l){if(47===t.charCodeAt(1+c))return t.slice(1+c+1);if(0===c)return t.slice(1+c)}else o>l&&(47===e.charCodeAt(1+c)?i=c:0===c&&(i=0));let a="";for(c=1+i+1;c<=r;++c)c!==r&&47!==e.charCodeAt(c)||(a+=0===a.length?"..":"/..");return`${a}${t.slice(1+i)}`},toNamespacedPath:e=>e,dirname(e){if(0===e.length)return".";const t=47===e.charCodeAt(0);let r=-1,o=!0;for(let t=e.length-1;t>=1;--t)if(47===e.charCodeAt(t)){if(!o){r=t;break}}else o=!1;return-1===r?t?"/":".":t&&1===r?"//":e.slice(0,r)},basename(e,t){let r=0,o=-1,n=!0;if(void 0!==t&&t.length>0&&t.length<=e.length){if(t===e)return"";let l=t.length-1,i=-1;for(let c=e.length-1;c>=0;--c){const a=e.charCodeAt(c);if(47===a){if(!n){r=c+1;break}}else-1===i&&(n=!1,i=c+1),l>=0&&(a===t.charCodeAt(l)?-1==--l&&(o=c):(l=-1,o=i))}return r===o?o=i:-1===o&&(o=e.length),e.slice(r,o)}for(let t=e.length-1;t>=0;--t)if(47===e.charCodeAt(t)){if(!n){r=t+1;break}}else-1===o&&(n=!1,o=t+1);return-1===o?"":e.slice(r,o)},extname(e){let t=-1,r=0,o=-1,n=!0,l=0;for(let i=e.length-1;i>=0;--i){const c=e.charCodeAt(i);if(47!==c)-1===o&&(n=!1,o=i+1),46===c?-1===t?t=i:1!==l&&(l=1):-1!==t&&(l=-1);else if(!n){r=i+1;break}}return-1===t||-1===o||0===l||1===l&&t===o-1&&t===r+1?"":e.slice(t,o)},format:i.bind(null,"/"),parse(e){const t={root:"",dir:"",base:"",ext:"",name:""};if(0===e.length)return t;const r=47===e.charCodeAt(0);let o;r?(t.root="/",o=1):o=0;let n=-1,l=0,i=-1,c=!0,a=e.length-1,s=0;for(;a>=o;--a){const t=e.charCodeAt(a);if(47!==t)-1===i&&(c=!1,i=a+1),46===t?-1===n?n=a:1!==s&&(s=1):-1!==n&&(s=-1);else if(!c){l=a+1;break}}if(-1!==i){const o=0===l&&r?1:l;-1===n||0===s||1===s&&n===i-1&&n===l+1?t.base=t.name=e.slice(o,i):(t.name=e.slice(o,n),t.base=e.slice(o,i),t.ext=e.slice(n,i))}return l>0?t.dir=e.slice(0,l-1):r&&(t.dir="/"),t},sep:"/",delimiter:":",win32:null,posix:null};s.win32=c.win32=c,s.posix=c.posix=s;const h=t?c:s;export default h;export const{resolve,normalize,isAbsolute,join,relative,toNamespacedPath,dirname,basename,extname,format,parse,sep,delimiter,win32,posix}=h;
✄
{"version":3,"file":"index.js","names":["nextTick","Script","title","browser","platform","Process","detectPlatform","pid","id","env","argv","version","Frida","versions","noop","on","addListener","once","off","removeListener","removeAllListeners","emit","prependListener","prependOnceListener","listeners","name","binding","Error","cwd","chdir","dir","umask"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/process/","sources":[""],"mappings":"OAAO,MAAMA,SAAWC,OAAOD,gBAExB,MAAME,MAAQ,eACd,MAAMC,SAAU,SAChB,MAAMC,SA2Db,WACI,MAAMA,EAAWC,QAAQD,SACzB,MAAqB,YAAbA,EAA0B,QAAUA,CAChD,CA9DwBE,UACjB,MAAMC,IAAMF,QAAQG,UACpB,MAAMC,IAAM,CAAC,SACb,MAAMC,KAAO,UACb,MAAMC,QAAUC,MAAMD,eACtB,MAAME,SAAW,CAAC,EAEzB,SAASC,IAAQ,QAEV,MAAMC,GAAKD,SACX,MAAME,YAAcF,SACpB,MAAMG,KAAOH,SACb,MAAMI,IAAMJ,SACZ,MAAMK,eAAiBL,SACvB,MAAMM,mBAAqBN,SAC3B,MAAMO,KAAOP,SACb,MAAMQ,gBAAkBR,SACxB,MAAMS,oBAAsBT,SAE5B,MAAMU,UAAY,SAAUC,GAAQ,MAAO,EAAI,SAE/C,SAASC,QAAQD,GACpB,MAAM,IAAIE,MAAM,mCACpB,QAEO,SAASC,MACZ,MAA6B,YAArBvB,QAAQD,SAA0B,OAAS,GACvD,QACO,SAASyB,MAAMC,GAClB,MAAM,IAAIH,MAAM,iCACpB,QACO,SAASI,QAAU,OAAO,CAAG,eAErB,CACX/B,SACAE,MACAC,QArCmB,MAsCnBC,SACAG,IACAE,IACAC,KACAC,QACAE,SACAE,GACAC,YACAC,KACAC,IACAC,eACAC,mBACAC,KACAC,gBACAC,oBACAC,UACAE,QACAE,IACAC,MACAE"}
✄
export const nextTick=Script.nextTick;export const title="Frida";export const browser=!1;export const platform=function(){const e=Process.platform;return"windows"===e?"win32":e}();export const pid=Process.id;export const env={};export const argv=[];export const version=Frida.version;export const versions={};function e(){}export const on=e;export const addListener=e;export const once=e;export const off=e;export const removeListener=e;export const removeAllListeners=e;export const emit=e;export const prependListener=e;export const prependOnceListener=e;export const listeners=function(e){return[]};export function binding(e){throw new Error("process.binding is not supported")}export function cwd(){return"windows"===Process.platform?"C:\\":"/"}export function chdir(e){throw new Error("process.chdir is not supported")}export function umask(){return 0}export default{nextTick,title,browser:false,platform,pid,env,argv,version,versions,on,addListener,once,off,removeListener,removeAllListeners,emit,prependListener,prependOnceListener,listeners,binding,cwd,chdir,umask};
✄
{"version":3,"file":"errors.js","names":["format","messages","Map","codes","aggregateTwoErrors","innerError","outerError","Array","isArray","errors","push","err","AggregateError","message","code","makeNodeErrorWithCode","Base","key","args","error","self","msg","get","Reflect","apply","match","length","unshift","getMessage","Object","defineProperties","value","enumerable","writable","configurable","toString","this","name","E","sym","val","def","otherClasses","set","forEach","clazz","AbortError","Error","constructor","super","TypeError","RangeError"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/","sources":[""],"mappings":"iBAASA,MAAc,OAEvB,MAAMC,EAAW,IAAIC,WACd,MAAMC,MAAQ,CAAC,SAEf,SAASC,mBAAmBC,EAAYC,GAC7C,GAAID,GAAcC,GAAcD,IAAeC,EAAY,CACzD,GAAIC,MAAMC,QAAQF,EAAWG,QAG3B,OADAH,EAAWG,OAAOC,KAAKL,GAChBC,EAGT,MAAMK,EAAM,IAAIC,eAAe,CAC7BN,EACAD,GACCC,EAAWO,SAEd,OADAF,EAAIG,KAAOR,EAAWQ,KACfH,CACT,CACA,OAAON,GAAcC,CACvB,CAEA,SAASS,EAAsBC,EAAMC,GACnC,OAAO,YAAsBC,GAC3B,MAAMC,EAAQ,IAAIH,EACZH,EAkCV,SAAoBI,EAAKC,EAAME,GAC7B,MAAMC,EAAMpB,EAASqB,IAAIL,GAEzB,GAAmB,mBAARI,EACT,OAAOE,QAAQC,MAAMH,EAAKD,EAAMF,IAGVG,EAAII,MAAM,gBAAkB,IAAIC,OACxD,OAAoB,IAAhBR,EAAKQ,OACAL,GAETH,EAAKS,QAAQN,GACNE,QAAQC,MAAMxB,EAAQ,KAAMkB,GACrC,CA/CoBU,CAAWX,EAAKC,EAAMC,GAkBtC,OAjBAU,OAAOC,iBAAiBX,EAAO,CAC7BN,QAAS,CACPkB,MAAOlB,EACPmB,YAAY,EACZC,UAAU,EACVC,cAAc,GAEhBC,SAAU,CACRJ,QACE,MAAO,GAAGK,KAAKC,SAASpB,OAASmB,KAAKvB,SACxC,EACAmB,YAAY,EACZC,UAAU,EACVC,cAAc,KAGlBf,EAAML,KAAOG,EACNE,CACT,CACF,CAEA,SAASmB,EAAEC,EAAKC,EAAKC,KAAQC,GAC3BzC,EAAS0C,IAAIJ,EAAKC,GAClBC,EAAM1B,EAAsB0B,EAAKF,GAEL,IAAxBG,EAAahB,QACfgB,EAAaE,SAASC,IACpBJ,EAAII,EAAMR,MAAQtB,EAAsB8B,EAAON,EAAI,IAGvDpC,MAAMoC,GAAOE,CACf,QAiBO,MAAMK,mBAAmBC,MAC9BC,cACEC,MAAM,6BACNb,KAAKtB,KAAO,YACZsB,KAAKC,KAAO,YACd,EAGFC,EAAE,sBAAuB,6CAA8CS,OACvET,EAAE,0BAA2B,sBAAuBY,WACpDZ,EAAE,uBAAwB,wBAAyBY,WACnDZ,EAAE,wBAAyB,yBAA0BY,UAAWC,YAChEb,EAAE,2BAA4B,uBAAwBY,UAAWC,YACjEb,EAAE,mBAAoB,qCAAsCY,WAC5DZ,EAAE,6BAA8B,mCAAoCS,OACpET,EAAE,mBAAoB,mBAAoBY,WAC1CZ,EAAE,wBAAyB,iCAAkCS,OAC7DT,EAAE,mBAAoB,eAAgBa,YACtCb,EAAE,8BACA,6CACAS,OACFT,EAAE,yBAA0B,4BAA6BS,OACzDT,EAAE,uBAAwB,8CAA+CS,OACzET,EAAE,yBAA0B,sCAAuCY,WACnEZ,EAAE,6BAA8B,kBAAmBS,OACnDT,EAAE,4BAA6B,0BAA2BS,OAC1DT,EAAE,qCACA,mCAAoCS,OACtCT,EAAE,6BAA8B,kBAAmBS,OACnDT,EAAE,uBAAwB,uBAAwBY"}
✄
import{format as r}from"util";const e=new Map;export const codes={};export function aggregateTwoErrors(r,e){if(r&&e&&r!==e){if(Array.isArray(e.errors))return e.errors.push(r),e;const t=new AggregateError([e,r],e.message);return t.code=e.code,t}return r||e}function t(t,E){return function(...o){const n=new t,a=function(t,E,o){const n=e.get(t);if("function"==typeof n)return Reflect.apply(n,o,E);(n.match(/%[dfijoOs]/g)||[]).length;return 0===E.length?n:(E.unshift(n),Reflect.apply(r,null,E))}(E,o,n);return Object.defineProperties(n,{message:{value:a,enumerable:!1,writable:!0,configurable:!0},toString:{value(){return`${this.name} [${E}]: ${this.message}`},enumerable:!1,writable:!0,configurable:!0}}),n.code=E,n}}function E(r,E,o,...n){e.set(r,E),o=t(o,r),0!==n.length&&n.forEach((e=>{o[e.name]=t(e,r)})),codes[r]=o}export class AbortError extends Error{constructor(){super("The operation was aborted"),this.code="ABORT_ERR",this.name="AbortError"}}E("ERR_EVENT_RECURSION",'The event "%s" is already being dispatched',Error),E("ERR_ILLEGAL_CONSTRUCTOR","Illegal constructor",TypeError),E("ERR_INVALID_ARG_TYPE","Invalid argument type",TypeError),E("ERR_INVALID_ARG_VALUE","Invalid argument value",TypeError,RangeError),E("ERR_INVALID_RETURN_VALUE","Invalid return value",TypeError,RangeError),E("ERR_INVALID_THIS",'Value of "this" must be of type %s',TypeError),E("ERR_METHOD_NOT_IMPLEMENTED","The %s method is not implemented",Error),E("ERR_MISSING_ARGS","Missing argument",TypeError),E("ERR_MULTIPLE_CALLBACK","Callback called multiple times",Error),E("ERR_OUT_OF_RANGE","Out of range",RangeError),E("ERR_STREAM_ALREADY_FINISHED","Cannot call %s after a stream was finished",Error),E("ERR_STREAM_CANNOT_PIPE","Cannot pipe, not readable",Error),E("ERR_STREAM_DESTROYED","Cannot call %s after a stream was destroyed",Error),E("ERR_STREAM_NULL_VALUES","May not write null values to stream",TypeError),E("ERR_STREAM_PREMATURE_CLOSE","Premature close",Error),E("ERR_STREAM_PUSH_AFTER_EOF","stream.push() after EOF",Error),E("ERR_STREAM_UNSHIFT_AFTER_END_EVENT","stream.unshift() after end event",Error),E("ERR_STREAM_WRITE_AFTER_END","write after end",Error),E("ERR_UNKNOWN_ENCODING","Unknown encoding: %s",TypeError);
✄
{"version":3,"file":"abort_controller.js","names":["errorCodes","defineEventHandler","EventTarget","Event","kTrustEvent","inspect","ERR_ILLEGAL_CONSTRUCTOR","ERR_INVALID_THIS","kAborted","Symbol","customInspect","self","obj","depth","options","opts","Object","assign","constructor","name","AbortSignal","aborted","undefined","validateAbortSignal","this","custom","static","createAbortSignal","signal","setPrototypeOf","prototype","defineProperties","enumerable","defineProperty","toStringTag","writable","configurable","value","kSignal","validateAbortController","AbortController","abort","event","dispatchEvent","abortSignal"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"gBAGkBA,MAAkB,4CAElCC,iBACAC,WACAC,iBACAC,MACK,sCAEEC,MAAe,OAExB,MAAMC,wBACJA,EAAuBC,iBACvBA,GACEP,SAEG,MAAMQ,SAAWC,OAAO,YAE/B,SAASC,EAAcC,EAAMC,EAAKC,EAAOC,GACvC,GAAID,EAAQ,EACV,OAAOF,EAET,MAAMI,EAAOC,OAAOC,OAAO,CAAC,EAAGH,EAAS,CACtCD,MAAyB,OAAlBC,EAAQD,MAAiB,KAAOC,EAAQD,MAAQ,IAGzD,MAAO,GAAGF,EAAKO,YAAYC,QAAQd,EAAQO,EAAKG,IAClD,QAOO,MAAMK,oBAAoBlB,EAC/BgB,cACE,MAAM,IAAIZ,CACZ,CAEIe,cAEF,OAZJ,SAA6BT,GAC3B,QAAwBU,IAApBV,IAAMJ,UACR,MAAM,IAAID,EAAiB,cAC/B,CAQIgB,CAAoBC,QACXA,KAAKhB,SAChB,CAEA,CAACH,EAAQoB,QAAQZ,EAAOC,GACtB,OAAOJ,EAAcc,KAAM,CACzBH,QAASG,KAAKH,SACbR,EAAOC,EACZ,CAEAY,eACE,OAAOC,GAAkB,EAC3B,EAgBF,SAASA,EAAkBN,GAAU,GACnC,MAAMO,EAAS,IAAI1B,EAGnB,OAFAc,OAAOa,eAAeD,EAAQR,YAAYU,WAC1CF,EAAOpB,UAAYa,EACZO,CACT,CAlBAZ,OAAOe,iBAAiBX,YAAYU,UAAW,CAC7CT,QAAS,CAAEW,YAAY,KAGzBhB,OAAOiB,eAAeb,YAAYU,UAAWrB,OAAOyB,YAAa,CAC/DC,UAAU,EACVH,YAAY,EACZI,cAAc,EACdC,MAAO,gBAGTpC,EAAmBmB,YAAYU,UAAW,SAqB1C,MAAMQ,EAAU7B,OAAO,UAEvB,SAAS8B,EAAwB3B,GAC/B,QAAuBU,IAAnBV,IAAM0B,GACR,MAAM,IAAI/B,EAAiB,kBAC/B,QAEO,MAAMiC,gBACXtB,cACEM,KAAKc,GAAWX,GAClB,CAEIC,aAEF,OADAW,EAAwBf,MACjBA,KAAKc,EACd,CAEAG,QACEF,EAAwBf,MA9B5B,SAAqBI,GACnB,GAAIA,EAAOpB,UAAW,OACtBoB,EAAOpB,WAAY,EACnB,MAAMkC,EAAQ,IAAIvC,EAAM,QAAS,CAC/BC,CAACA,IAAc,IAEjBwB,EAAOe,cAAcD,EACvB,CAwBIE,CAAYpB,KAAKc,GACnB,CAEA,CAACjC,EAAQoB,QAAQZ,EAAOC,GACtB,OAAOJ,EAAcc,KAAM,CACzBI,OAAQJ,KAAKI,QACZf,EAAOC,EACZ,EAGFE,OAAOe,iBAAiBS,gBAAgBV,UAAW,CACjDF,OAAQ,CAAEI,YAAY,GACtBS,MAAO,CAAET,YAAY,KAGvBhB,OAAOiB,eAAeO,gBAAgBV,UAAWrB,OAAOyB,YAAa,CACnEC,UAAU,EACVH,YAAY,EACZI,cAAc,EACdC,MAAO"}
✄
import{codes as t}from"../errors.js";import{defineEventHandler as r,EventTarget as o,Event as e,kTrustEvent as n}from"./event_target.js";import{inspect as i}from"util";const{ERR_ILLEGAL_CONSTRUCTOR:b,ERR_INVALID_THIS:l}=t;export const kAborted=Symbol("kAborted");function s(t,r,o,e){if(o<0)return t;const n=Object.assign({},e,{depth:null===e.depth?null:e.depth-1});return`${t.constructor.name} ${i(r,n)}`}export class AbortSignal extends o{constructor(){throw new b}get aborted(){return function(t){if(void 0===t?.[kAborted])throw new l("AbortSignal")}(this),!!this[kAborted]}[i.custom](t,r){return s(this,{aborted:this.aborted},t,r)}static abort(){return a(!0)}}function a(t=!1){const r=new o;return Object.setPrototypeOf(r,AbortSignal.prototype),r[kAborted]=t,r}Object.defineProperties(AbortSignal.prototype,{aborted:{enumerable:!0}}),Object.defineProperty(AbortSignal.prototype,Symbol.toStringTag,{writable:!1,enumerable:!1,configurable:!0,value:"AbortSignal"}),r(AbortSignal.prototype,"abort");const c=Symbol("signal");function u(t){if(void 0===t?.[c])throw new l("AbortController")}export class AbortController{constructor(){this[c]=a()}get signal(){return u(this),this[c]}abort(){u(this),function(t){if(t[kAborted])return;t[kAborted]=!0;const r=new e("abort",{[n]:!0});t.dispatchEvent(r)}(this[c])}[i.custom](t,r){return s(this,{signal:this.signal},t,r)}}Object.defineProperties(AbortController.prototype,{signal:{enumerable:!0},abort:{enumerable:!0}}),Object.defineProperty(AbortController.prototype,Symbol.toStringTag,{writable:!1,enumerable:!1,configurable:!0,value:"AbortController"});
✄
{"version":3,"file":"add-abort-signal.js","names":["eos","AbortError","errorCodes","ERR_INVALID_ARG_TYPE","addAbortSignal","signal","stream","name","validateAbortSignal","obj","pipe","module","exports","addAbortSignalNoValidate","onAbort","destroy","aborted","addEventListener","removeEventListener"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OAAOA,MAAS,0CAEdC,WACSC,MACJ,eAEP,MAAMC,qBAAEA,GAAyBD,SAgB1B,SAASE,eAAeC,EAAQC,GAErC,GAb0B,EAACD,EAAQE,KACnC,GAAsB,iBAAXF,KACJ,YAAaA,GAClB,MAAM,IAAIF,EAAqBI,EAAM,cAAeF,EACtD,EAQAG,CAAoBH,EAAQ,YALRI,EAMFH,IALmB,mBAAbG,EAAIC,KAM1B,MAAM,IAAIP,EAAqB,SAAU,gBAAiBG,GAP9D,IAAsBG,EASpB,OAAOE,OAAOC,QAAQC,yBAAyBR,EAAQC,EACzD,QAEO,SAASO,yBAAyBR,EAAQC,GAC/C,GAAsB,iBAAXD,KAAyB,YAAaA,GAC/C,OAAOC,EAET,MAAMQ,EAAU,KACdR,EAAOS,QAAQ,IAAId,EAAa,EAQlC,OANII,EAAOW,QACTF,KAEAT,EAAOY,iBAAiB,QAASH,GACjCd,EAAIM,GAAQ,IAAMD,EAAOa,oBAAoB,QAASJ,MAEjDR,CACT"}
✄
import t from"./end-of-stream.js";import{AbortError as e,codes as o}from"../errors.js";const{ERR_INVALID_ARG_TYPE:r}=o;export function addAbortSignal(t,e){if(((t,e)=>{if("object"!=typeof t||!("aborted"in t))throw new r(e,"AbortSignal",t)})(t,"signal"),!(o=e)||"function"!=typeof o.pipe)throw new r("stream","stream.Stream",e);var o;return module.exports.addAbortSignalNoValidate(t,e)}export function addAbortSignalNoValidate(o,r){if("object"!=typeof o||!("aborted"in o))return r;const n=()=>{r.destroy(new e)};return o.aborted?n():(o.addEventListener("abort",n),t(r,(()=>o.removeEventListener("abort",n)))),r}
✄
{"version":3,"file":"buffer_list.js","names":["Buffer","inspect","BufferList","constructor","this","head","tail","length","push","v","entry","data","next","unshift","shift","ret","clear","join","s","p","concat","n","alloc","allocUnsafe","i","set","consume","hasStrings","slice","_getString","_getBuffer","first","Symbol","iterator","c","str","retLen","buf","Uint8Array","buffer","byteOffset","custom","_","options","depth","customInspect"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"iBAASA,MAAc,2BACdC,MAAe,sBAET,MAAMC,EACnBC,cACEC,KAAKC,KAAO,KACZD,KAAKE,KAAO,KACZF,KAAKG,OAAS,CAChB,CAEAC,KAAKC,GACH,MAAMC,EAAQ,CAAEC,KAAMF,EAAGG,KAAM,MAC3BR,KAAKG,OAAS,EAChBH,KAAKE,KAAKM,KAAOF,EAEjBN,KAAKC,KAAOK,EACdN,KAAKE,KAAOI,IACVN,KAAKG,MACT,CAEAM,QAAQJ,GACN,MAAMC,EAAQ,CAAEC,KAAMF,EAAGG,KAAMR,KAAKC,MAChB,IAAhBD,KAAKG,SACPH,KAAKE,KAAOI,GACdN,KAAKC,KAAOK,IACVN,KAAKG,MACT,CAEAO,QACE,GAAoB,IAAhBV,KAAKG,OACP,OACF,MAAMQ,EAAMX,KAAKC,KAAKM,KAMtB,OALoB,IAAhBP,KAAKG,OACPH,KAAKC,KAAOD,KAAKE,KAAO,KAExBF,KAAKC,KAAOD,KAAKC,KAAKO,OACtBR,KAAKG,OACAQ,CACT,CAEAC,QACEZ,KAAKC,KAAOD,KAAKE,KAAO,KACxBF,KAAKG,OAAS,CAChB,CAEAU,KAAKC,GACH,GAAoB,IAAhBd,KAAKG,OACP,MAAO,GACT,IAAIY,EAAIf,KAAKC,KACTU,EAAM,GAAKI,EAAER,KACjB,KAAOQ,EAAIA,EAAEP,MACXG,GAAOG,EAAIC,EAAER,KACf,OAAOI,CACT,CAEAK,OAAOC,GACL,GAAoB,IAAhBjB,KAAKG,OACP,OAAOP,EAAOsB,MAAM,GACtB,MAAMP,EAAMf,EAAOuB,YAAYF,IAAM,GACrC,IAAIF,EAAIf,KAAKC,KACTmB,EAAI,EACR,KAAOL,GACLJ,EAAIU,IAAIN,EAAER,KAAMa,GAChBA,GAAKL,EAAER,KAAKJ,OACZY,EAAIA,EAAEP,KAER,OAAOG,CACT,CAGAW,QAAQL,EAAGM,GACT,MAAMhB,EAAOP,KAAKC,KAAKM,KACvB,GAAIU,EAAIV,EAAKJ,OAAQ,CAEnB,MAAMqB,EAAQjB,EAAKiB,MAAM,EAAGP,GAE5B,OADAjB,KAAKC,KAAKM,KAAOA,EAAKiB,MAAMP,GACrBO,CACT,CACA,OAAIP,IAAMV,EAAKJ,OAENH,KAAKU,QAGPa,EAAavB,KAAKyB,WAAWR,GAAKjB,KAAK0B,WAAWT,EAC3D,CAEAU,QACE,OAAO3B,KAAKC,KAAKM,IACnB,CAEA,EAAEqB,OAAOC,YACP,IAAK,IAAId,EAAIf,KAAKC,KAAMc,EAAGA,EAAIA,EAAEP,WACzBO,EAAER,IAEZ,CAGAkB,WAAWR,GACT,IAAIN,EAAM,GACNI,EAAIf,KAAKC,KACT6B,EAAI,EACR,EAAG,CACD,MAAMC,EAAMhB,EAAER,KACd,KAAIU,EAAIc,EAAI5B,QAGL,CACDc,IAAMc,EAAI5B,QACZQ,GAAOoB,IACLD,EACEf,EAAEP,KACJR,KAAKC,KAAOc,EAAEP,KAEdR,KAAKC,KAAOD,KAAKE,KAAO,OAE1BS,GAAOoB,EAAIP,MAAM,EAAGP,GACpBjB,KAAKC,KAAOc,EACZA,EAAER,KAAOwB,EAAIP,MAAMP,IAErB,KACF,CAhBEN,GAAOoB,EACPd,GAAKc,EAAI5B,SAgBT2B,CACJ,OAASf,EAAIA,EAAEP,MAEf,OADAR,KAAKG,QAAU2B,EACRnB,CACT,CAGAe,WAAWT,GACT,MAAMN,EAAMf,EAAOuB,YAAYF,GACzBe,EAASf,EACf,IAAIF,EAAIf,KAAKC,KACT6B,EAAI,EACR,EAAG,CACD,MAAMG,EAAMlB,EAAER,KACd,KAAIU,EAAIgB,EAAI9B,QAGL,CACDc,IAAMgB,EAAI9B,QACZQ,EAAIU,IAAIY,EAAKD,EAASf,KACpBa,EACEf,EAAEP,KACJR,KAAKC,KAAOc,EAAEP,KAEdR,KAAKC,KAAOD,KAAKE,KAAO,OAE1BS,EAAIU,IAAI,IAAIa,WAAWD,EAAIE,OAAQF,EAAIG,WAAYnB,GAC3Ce,EAASf,GACjBjB,KAAKC,KAAOc,EACZA,EAAER,KAAO0B,EAAIT,MAAMP,IAErB,KACF,CAjBEN,EAAIU,IAAIY,EAAKD,EAASf,GACtBA,GAAKgB,EAAI9B,SAiBT2B,CACJ,OAASf,EAAIA,EAAEP,MAEf,OADAR,KAAKG,QAAU2B,EACRnB,CACT,CAGA,CAACd,EAAQwC,QAAQC,EAAGC,GAClB,OAAO1C,EAAQG,KAAM,IAChBuC,EAEHC,MAAO,EAEPC,eAAe,GAEnB"}
✄
import{Buffer as t}from"buffer";import{inspect as e}from"util";export default class h{constructor(){this.head=null,this.tail=null,this.length=0}push(t){const e={data:t,next:null};this.length>0?this.tail.next=e:this.head=e,this.tail=e,++this.length}unshift(t){const e={data:t,next:this.head};0===this.length&&(this.tail=e),this.head=e,++this.length}shift(){if(0===this.length)return;const t=this.head.data;return 1===this.length?this.head=this.tail=null:this.head=this.head.next,--this.length,t}clear(){this.head=this.tail=null,this.length=0}join(t){if(0===this.length)return"";let e=this.head,h=""+e.data;for(;e=e.next;)h+=t+e.data;return h}concat(e){if(0===this.length)return t.alloc(0);const h=t.allocUnsafe(e>>>0);let i=this.head,s=0;for(;i;)h.set(i.data,s),s+=i.data.length,i=i.next;return h}consume(t,e){const h=this.head.data;if(t<h.length){const e=h.slice(0,t);return this.head.data=h.slice(t),e}return t===h.length?this.shift():e?this._getString(t):this._getBuffer(t)}first(){return this.head.data}*[Symbol.iterator](){for(let t=this.head;t;t=t.next)yield t.data}_getString(t){let e="",h=this.head,i=0;do{const s=h.data;if(!(t>s.length)){t===s.length?(e+=s,++i,h.next?this.head=h.next:this.head=this.tail=null):(e+=s.slice(0,t),this.head=h,h.data=s.slice(t));break}e+=s,t-=s.length,++i}while(h=h.next);return this.length-=i,e}_getBuffer(e){const h=t.allocUnsafe(e),i=e;let s=this.head,n=0;do{const t=s.data;if(!(e>t.length)){e===t.length?(h.set(t,i-e),++n,s.next?this.head=s.next:this.head=this.tail=null):(h.set(new Uint8Array(t.buffer,t.byteOffset,e),i-e),this.head=s,s.data=t.slice(e));break}h.set(t,i-e),e-=t.length,++n}while(s=s.next);return this.length-=n,h}[e.custom](t,h){return e(this,{...h,depth:0,customInspect:!1})}}
✄
{"version":3,"file":"compose.js","names":["destroyer","Duplex","AbortError","errorCodes","pipeline","isNodeStream","isReadable","isWritable","ERR_INVALID_ARG_VALUE","ERR_MISSING_ARGS","ComposeDuplex","constructor","options","super","readable","this","_readableState","ended","endEmitted","writable","_writableState","ending","finished","compose","streams","length","from","orgStreams","idx","n","ondrain","onfinish","onreadable","onclose","d","head","tail","err","cb","destroy","writableObjectMode","readableObjectMode","_write","chunk","encoding","callback","write","_final","end","on","push","_read","buf","read","_destroy"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"oBAASA,MAAiB,sBACnBC,MAAY,mCAEjBC,WACSC,MACJ,kCACEC,MAAgB,uCAEvBC,gBACAC,gBACAC,MACK,aAEP,MAAMC,sBACJA,EAAqBC,iBACrBA,GACEN,EAGJ,MAAMO,UAAsBT,EAC1BU,YAAYC,GACVC,MAAMD,IAIoB,IAAtBA,GAASE,WACXC,KAAKC,eAAeF,UAAW,EAC/BC,KAAKC,eAAeC,OAAQ,EAC5BF,KAAKC,eAAeE,YAAa,IAGT,IAAtBN,GAASO,WACXJ,KAAKK,eAAeD,UAAW,EAC/BJ,KAAKK,eAAeC,QAAS,EAC7BN,KAAKK,eAAeH,OAAQ,EAC5BF,KAAKK,eAAeE,UAAW,EAEnC,iBAGa,SAASC,KAAWC,GACjC,GAAuB,IAAnBA,EAAQC,OACV,MAAM,IAAIhB,EAAiB,WAG7B,GAAuB,IAAnBe,EAAQC,OACV,OAAOxB,EAAOyB,KAAKF,EAAQ,IAG7B,MAAMG,EAAa,IAAIH,GAMvB,GAJ0B,mBAAfA,EAAQ,KACjBA,EAAQ,GAAKvB,EAAOyB,KAAKF,EAAQ,KAGQ,mBAAhCA,EAAQA,EAAQC,OAAS,GAAmB,CACrD,MAAMG,EAAMJ,EAAQC,OAAS,EAC7BD,EAAQI,GAAO3B,EAAOyB,KAAKF,EAAQI,GACrC,CAEA,IAAK,IAAIC,EAAI,EAAGA,EAAIL,EAAQC,SAAUI,EACpC,GAAKxB,EAAamB,EAAQK,IAA1B,CAIA,GAAIA,EAAIL,EAAQC,OAAS,IAAMnB,EAAWkB,EAAQK,IAChD,MAAM,IAAIrB,EACR,WAAWqB,KACXF,EAAWE,GACX,oBAGJ,GAAIA,EAAI,IAAMtB,EAAWiB,EAAQK,IAC/B,MAAM,IAAIrB,EACR,WAAWqB,KACXF,EAAWE,GACX,mBAZJ,CAiBF,IAAIC,EACAC,EACAC,EACAC,EACAC,EAeJ,MAAMC,EAAOX,EAAQ,GACfY,EAAOhC,EAASoB,GAdtB,SAAoBa,GAClB,MAAMC,EAAKL,EACXA,EAAU,KAENK,EACFA,EAAGD,GACMA,EACTH,EAAEK,QAAQF,GACAvB,GAAaK,GACvBe,EAAEK,SAEN,IAKMpB,IAAaZ,EAAW4B,GACxBrB,IAAaR,EAAW8B,GA0F9B,OArFAF,EAAI,IAAIxB,EAAc,CAEpB8B,qBAAsBL,GAAMK,mBAC5BC,qBAAsBL,GAAMI,mBAC5BrB,WACAL,aAGEK,IACFe,EAAEQ,OAAS,SAASC,EAAOC,EAAUC,GAC/BV,EAAKW,MAAMH,EAAOC,GACpBC,IAEAf,EAAUe,CAEd,EAEAX,EAAEa,OAAS,SAASF,GAClBV,EAAKa,MACLjB,EAAWc,CACb,EAEAV,EAAKc,GAAG,SAAS,WACf,GAAInB,EAAS,CACX,MAAMQ,EAAKR,EACXA,EAAU,KACVQ,GACF,CACF,IAEAF,EAAKa,GAAG,UAAU,WAChB,GAAIlB,EAAU,CACZ,MAAMO,EAAKP,EACXA,EAAW,KACXO,GACF,CACF,KAGExB,IACFsB,EAAKa,GAAG,YAAY,WAClB,GAAIjB,EAAY,CACd,MAAMM,EAAKN,EACXA,EAAa,KACbM,GACF,CACF,IAEAF,EAAKa,GAAG,OAAO,WACbf,EAAEgB,KAAK,KACT,IAEAhB,EAAEiB,MAAQ,WACR,OAAa,CACX,MAAMC,EAAMhB,EAAKiB,OAEjB,GAAY,OAARD,EAEF,YADApB,EAAaE,EAAEiB,OAIjB,IAAKjB,EAAEgB,KAAKE,GACV,MAEJ,CACF,GAGFlB,EAAEoB,SAAW,SAASjB,EAAKQ,GACpBR,GAAmB,OAAZJ,IACVI,EAAM,IAAInC,GAGZ8B,EAAa,KACbF,EAAU,KACVC,EAAW,KAEK,OAAZE,EACFY,EAASR,IAETJ,EAAUY,EACV7C,EAAUoC,EAAMC,GAEpB,EAEOH,CACT"}
✄
import{destroyer as t}from"./destroy.js";import e from"./duplex.js";import{AbortError as n,codes as r}from"../errors.js";import{pipeline as i}from"./pipeline.js";import{isNodeStream as o,isReadable as l,isWritable as a}from"./utils.js";const{ERR_INVALID_ARG_VALUE:s,ERR_MISSING_ARGS:f}=r;class d extends e{constructor(t){super(t),!1===t?.readable&&(this._readableState.readable=!1,this._readableState.ended=!0,this._readableState.endEmitted=!0),!1===t?.writable&&(this._writableState.writable=!1,this._writableState.ending=!0,this._writableState.ended=!0,this._writableState.finished=!0)}}export default function u(...r){if(0===r.length)throw new f("streams");if(1===r.length)return e.from(r[0]);const u=[...r];if("function"==typeof r[0]&&(r[0]=e.from(r[0])),"function"==typeof r[r.length-1]){const t=r.length-1;r[t]=e.from(r[t])}for(let t=0;t<r.length;++t)if(o(r[t])){if(t<r.length-1&&!l(r[t]))throw new s(`streams[${t}]`,u[t],"must be readable");if(t>0&&!a(r[t]))throw new s(`streams[${t}]`,u[t],"must be writable")}let c,b,w,h,m;const _=r[0],p=i(r,(function(t){const e=h;h=null,e?e(t):t?m.destroy(t):j||S||m.destroy()})),S=!!a(_),j=!!l(p);return m=new d({writableObjectMode:!!_?.writableObjectMode,readableObjectMode:!!p?.writableObjectMode,writable:S,readable:j}),S&&(m._write=function(t,e,n){_.write(t,e)?n():c=n},m._final=function(t){_.end(),b=t},_.on("drain",(function(){if(c){const t=c;c=null,t()}})),p.on("finish",(function(){if(b){const t=b;b=null,t()}}))),j&&(p.on("readable",(function(){if(w){const t=w;w=null,t()}})),p.on("end",(function(){m.push(null)})),m._read=function(){for(;;){const t=p.read();if(null===t)return void(w=m._read);if(!m.push(t))return}}),m._destroy=function(e,r){e||null===h||(e=new n),w=null,c=null,b=null,null===h?r(e):(h=r,t(p,e))},m}
✄
{"version":3,"file":"destroy.js","names":["aggregateTwoErrors","errorCodes","AbortError","kDestroyed","isDestroyed","isFinished","isServerRequest","process","ERR_MULTIPLE_CALLBACK","kDestroy","Symbol","kConstruct","checkError","err","w","r","stack","errored","destroy","cb","this","_readableState","_writableState","s","destroyed","constructed","_destroy","once","er","self","called","onDestroy","closed","nextTick","emitErrorCloseNT","emitCloseNT","result","then","call","emitErrorNT","closeEmitted","emitClose","emit","errorEmitted","undestroy","reading","ended","readable","endEmitted","finalCalled","prefinished","writable","ending","finished","errorOrDestroy","stream","sync","autoDestroy","construct","_construct","listenerCount","constructNT","onConstruct","emitConstructNT","isRequest","setHeader","abort","emitCloseLegacy","emitErrorCloseLegacy","destroyer","socket","req","close"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"6BACEA,WACSC,gBACTC,MACK,oCAELC,iBACAC,gBACAC,qBACAC,MACK,oBAEAC,MAAa,UAEpB,MAAMC,sBACJA,GACEP,EAEEQ,EAAWC,OAAO,YAClBC,EAAaD,OAAO,cAE1B,SAASE,EAAWC,EAAKC,EAAGC,GACtBF,IAEFA,EAAIG,MAEAF,IAAMA,EAAEG,UACVH,EAAEG,QAAUJ,GAEVE,IAAMA,EAAEE,UACVF,EAAEE,QAAUJ,GAGlB,QAIO,SAASK,QAAQL,EAAKM,GAC3B,MAAMJ,EAAIK,KAAKC,eACTP,EAAIM,KAAKE,eAETC,EAAIT,GAAKC,EAEf,OAAKD,GAAKA,EAAEU,WAAeT,GAAKA,EAAES,WACd,mBAAPL,GACTA,IAGKC,OAMTR,EAAWC,EAAKC,EAAGC,GAEfD,IACFA,EAAEU,WAAY,GAEZT,IACFA,EAAES,WAAY,GAIXD,EAAEE,YAKLC,EAASN,KAAMP,EAAKM,GAJpBC,KAAKO,KAAKlB,GAAU,SAASmB,GAC3BF,EAASN,KAAMpB,EAAmB4B,EAAIf,GAAMM,EAC9C,IAKKC,KACT,CAEA,SAASM,EAASG,EAAMhB,EAAKM,GAC3B,IAAIW,GAAS,EAEb,SAASC,EAAUlB,GACjB,GAAIiB,EACF,OAEFA,GAAS,EAET,MAAMf,EAAIc,EAAKR,eACTP,EAAIe,EAAKP,eAEfV,EAAWC,EAAKC,EAAGC,GAEfD,IACFA,EAAEkB,QAAS,GAETjB,IACFA,EAAEiB,QAAS,GAGK,mBAAPb,GACTA,EAAGN,GAGDA,EACFN,EAAQ0B,SAASC,EAAkBL,EAAMhB,GAEzCN,EAAQ0B,SAASE,EAAaN,EAElC,CACA,IACE,MAAMO,EAASP,EAAKH,SAASb,GAAO,KAAMkB,GAC1C,GAAc,MAAVK,EAAgB,CAClB,MAAMC,EAAOD,EAAOC,KACA,mBAATA,GACTA,EAAKC,KACHF,GACA,WACE7B,EAAQ0B,SAASF,EAAW,KAC9B,IACA,SAASlB,GACPN,EAAQ0B,SAASF,EAAWlB,EAC9B,GAEN,CAGF,CAFE,MAAOA,GACPkB,EAAUlB,EACZ,CACF,CAEA,SAASqB,EAAiBL,EAAMhB,GAC9B0B,EAAYV,EAAMhB,GAClBsB,EAAYN,EACd,CAEA,SAASM,EAAYN,GACnB,MAAMd,EAAIc,EAAKR,eACTP,EAAIe,EAAKP,eAEXR,IACFA,EAAE0B,cAAe,GAEfzB,IACFA,EAAEyB,cAAe,IAGd1B,GAAKA,EAAE2B,WAAe1B,GAAKA,EAAE0B,YAChCZ,EAAKa,KAAK,QAEd,CAEA,SAASH,EAAYV,EAAMhB,GACzB,MAAME,EAAIc,EAAKR,eACTP,EAAIe,EAAKP,eAEVR,GAAKA,EAAE6B,cAAkB5B,GAAKA,EAAE4B,eAIjC7B,IACFA,EAAE6B,cAAe,GAEf5B,IACFA,EAAE4B,cAAe,GAGnBd,EAAKa,KAAK,QAAS7B,GACrB,QAEO,SAAS+B,YACd,MAAM7B,EAAIK,KAAKC,eACTP,EAAIM,KAAKE,eAEXP,IACFA,EAAEU,aAAc,EAChBV,EAAEiB,QAAS,EACXjB,EAAEyB,cAAe,EACjBzB,EAAES,WAAY,EACdT,EAAEE,QAAU,KACZF,EAAE4B,cAAe,EACjB5B,EAAE8B,SAAU,EACZ9B,EAAE+B,OAAuB,IAAf/B,EAAEgC,SACZhC,EAAEiC,YAA4B,IAAfjC,EAAEgC,UAGfjC,IACFA,EAAEW,aAAc,EAChBX,EAAEU,WAAY,EACdV,EAAEkB,QAAS,EACXlB,EAAE0B,cAAe,EACjB1B,EAAEG,QAAU,KACZH,EAAE6B,cAAe,EACjB7B,EAAEmC,aAAc,EAChBnC,EAAEoC,aAAc,EAChBpC,EAAEgC,OAAuB,IAAfhC,EAAEqC,SACZrC,EAAEsC,QAAwB,IAAftC,EAAEqC,SACbrC,EAAEuC,UAA0B,IAAfvC,EAAEqC,SAEnB,QAEO,SAASG,eAAeC,EAAQ1C,EAAK2C,GAO1C,MAAMzC,EAAIwC,EAAOlC,eACXP,EAAIyC,EAAOjC,eAEjB,GAAKR,GAAKA,EAAEU,WAAeT,GAAKA,EAAES,UAChC,OAAOJ,KAGJL,GAAKA,EAAE0C,aAAiB3C,GAAKA,EAAE2C,YAClCF,EAAOrC,QAAQL,GACRA,IAEPA,EAAIG,MAEAF,IAAMA,EAAEG,UACVH,EAAEG,QAAUJ,GAEVE,IAAMA,EAAEE,UACVF,EAAEE,QAAUJ,GAEV2C,EACFjD,EAAQ0B,SAASM,EAAagB,EAAQ1C,GAEtC0B,EAAYgB,EAAQ1C,GAG1B,QAEO,SAAS6C,UAAUH,EAAQpC,GAChC,GAAiC,mBAAtBoC,EAAOI,WAChB,OAGF,MAAM5C,EAAIwC,EAAOlC,eACXP,EAAIyC,EAAOjC,eAEbP,IACFA,EAAEU,aAAc,GAEdX,IACFA,EAAEW,aAAc,GAGlB8B,EAAO5B,KAAKhB,EAAYQ,GAEpBoC,EAAOK,cAAcjD,GAAc,GAKvCJ,EAAQ0B,SAAS4B,EAAaN,EAChC,CAEA,SAASM,EAAYN,GACnB,IAAIzB,GAAS,EAEb,SAASgC,EAAYjD,GACnB,GAAIiB,EAEF,YADAwB,eAAeC,EAAQ1C,GAAO,IAAIL,GAGpCsB,GAAS,EAET,MAAMf,EAAIwC,EAAOlC,eACXP,EAAIyC,EAAOjC,eACXC,EAAIT,GAAKC,EAEXA,IACFA,EAAEU,aAAc,GAEdX,IACFA,EAAEW,aAAc,GAGdF,EAAEC,UACJ+B,EAAOb,KAAKjC,EAAUI,GACbA,EACTyC,eAAeC,EAAQ1C,GAAK,GAE5BN,EAAQ0B,SAAS8B,EAAiBR,EAEtC,CAEA,IACE,MAAMnB,EAASmB,EAAOI,WAAWG,GACjC,GAAc,MAAV1B,EAAgB,CAClB,MAAMC,EAAOD,EAAOC,KACA,mBAATA,GACTA,EAAKC,KACHF,GACA,WACE7B,EAAQ0B,SAAS6B,EAAa,KAChC,IACA,SAASjD,GACPN,EAAQ0B,SAAS6B,EAAajD,EAChC,GAEN,CAGF,CAFE,MAAOA,GACPiD,EAAYjD,EACd,CACF,CAEA,SAASkD,EAAgBR,GACvBA,EAAOb,KAAK/B,EACd,CAEA,SAASqD,EAAUT,GACjB,OAAOA,GAAUA,EAAOU,WAAqC,mBAAjBV,EAAOW,KACrD,CAEA,SAASC,EAAgBZ,GACvBA,EAAOb,KAAK,QACd,CAEA,SAAS0B,EAAqBb,EAAQ1C,GACpC0C,EAAOb,KAAK,QAAS7B,GACrBN,EAAQ0B,SAASkC,EAAiBZ,EACpC,QAGO,SAASc,UAAUd,EAAQ1C,GAC3B0C,IAAUnD,EAAYmD,KAItB1C,GAAQR,EAAWkD,KACtB1C,EAAM,IAAIX,GAIRI,EAAgBiD,IAClBA,EAAOe,OAAS,KAChBf,EAAOrC,QAAQL,IACNmD,EAAUT,GACnBA,EAAOW,QACEF,EAAUT,EAAOgB,KAC1BhB,EAAOgB,IAAIL,QACwB,mBAAnBX,EAAOrC,QACvBqC,EAAOrC,QAAQL,GACkB,mBAAjB0C,EAAOiB,MAEvBjB,EAAOiB,QACE3D,EACTN,EAAQ0B,SAASmC,EAAsBb,GAEvChD,EAAQ0B,SAASkC,EAAiBZ,GAG/BA,EAAO/B,YACV+B,EAAOpD,IAAc,GAEzB"}
✄
import{aggregateTwoErrors as t,codes as e,AbortError as r}from"../errors.js";import{kDestroyed as o,isDestroyed as n,isFinished as c,isServerRequest as i}from"./utils.js";import s from"process";const{ERR_MULTIPLE_CALLBACK:d}=e,l=Symbol("kDestroy"),a=Symbol("kConstruct");function u(t,e,r){t&&(t.stack,e&&!e.errored&&(e.errored=t),r&&!r.errored&&(r.errored=t))}export function destroy(e,r){const o=this._readableState,n=this._writableState,c=n||o;return n&&n.destroyed||o&&o.destroyed?("function"==typeof r&&r(),this):(u(e,n,o),n&&(n.destroyed=!0),o&&(o.destroyed=!0),c.constructed?f(this,e,r):this.once(l,(function(o){f(this,t(o,e),r)})),this)}function f(t,e,r){let o=!1;function n(e){if(o)return;o=!0;const n=t._readableState,c=t._writableState;u(e,c,n),c&&(c.closed=!0),n&&(n.closed=!0),"function"==typeof r&&r(e),e?s.nextTick(y,t,e):s.nextTick(m,t)}try{const r=t._destroy(e||null,n);if(null!=r){const t=r.then;"function"==typeof t&&t.call(r,(function(){s.nextTick(n,null)}),(function(t){s.nextTick(n,t)}))}}catch(e){n(e)}}function y(t,e){b(t,e),m(t)}function m(t){const e=t._readableState,r=t._writableState;r&&(r.closeEmitted=!0),e&&(e.closeEmitted=!0),(r&&r.emitClose||e&&e.emitClose)&&t.emit("close")}function b(t,e){const r=t._readableState,o=t._writableState;o&&o.errorEmitted||r&&r.errorEmitted||(o&&(o.errorEmitted=!0),r&&(r.errorEmitted=!0),t.emit("error",e))}export function undestroy(){const t=this._readableState,e=this._writableState;t&&(t.constructed=!0,t.closed=!1,t.closeEmitted=!1,t.destroyed=!1,t.errored=null,t.errorEmitted=!1,t.reading=!1,t.ended=!1===t.readable,t.endEmitted=!1===t.readable),e&&(e.constructed=!0,e.destroyed=!1,e.closed=!1,e.closeEmitted=!1,e.errored=null,e.errorEmitted=!1,e.finalCalled=!1,e.prefinished=!1,e.ended=!1===e.writable,e.ending=!1===e.writable,e.finished=!1===e.writable)}export function errorOrDestroy(t,e,r){const o=t._readableState,n=t._writableState;if(n&&n.destroyed||o&&o.destroyed)return this;o&&o.autoDestroy||n&&n.autoDestroy?t.destroy(e):e&&(e.stack,n&&!n.errored&&(n.errored=e),o&&!o.errored&&(o.errored=e),r?s.nextTick(b,t,e):b(t,e))}export function construct(t,e){if("function"!=typeof t._construct)return;const r=t._readableState,o=t._writableState;r&&(r.constructed=!1),o&&(o.constructed=!1),t.once(a,e),t.listenerCount(a)>1||s.nextTick(_,t)}function _(t){let e=!1;function r(r){if(e)return void errorOrDestroy(t,r??new d);e=!0;const o=t._readableState,n=t._writableState,c=n||o;o&&(o.constructed=!0),n&&(n.constructed=!0),c.destroyed?t.emit(l,r):r?errorOrDestroy(t,r,!0):s.nextTick(p,t)}try{const e=t._construct(r);if(null!=e){const t=e.then;"function"==typeof t&&t.call(e,(function(){s.nextTick(r,null)}),(function(t){s.nextTick(r,t)}))}}catch(t){r(t)}}function p(t){t.emit(a)}function S(t){return t&&t.setHeader&&"function"==typeof t.abort}function k(t){t.emit("close")}function x(t,e){t.emit("error",e),s.nextTick(k,t)}export function destroyer(t,e){t&&!n(t)&&(e||c(t)||(e=new r),i(t)?(t.socket=null,t.destroy(e)):S(t)?t.abort():S(t.req)?t.req.abort():"function"==typeof t.destroy?t.destroy(e):"function"==typeof t.close?t.close():e?s.nextTick(x,t):s.nextTick(k,t),t.destroyed||(t[o]=!0))}
✄
{"version":3,"file":"duplex.js","names":["AbortController","destroyer","eos","AbortError","errorCodes","from","Readable","isReadable","isWritable","isIterable","isNodeStream","isReadableNodeStream","isWritableNodeStream","isDuplexNodeStream","Writable","process","ERR_INVALID_ARG_TYPE","ERR_INVALID_RETURN_VALUE","Object","setPrototypeOf","Duplex","prototype","method","keys","options","this","call","allowHalfOpen","readable","_readableState","ended","endEmitted","writable","_writableState","ending","finished","defineProperties","getOwnPropertyDescriptor","writableHighWaterMark","writableObjectMode","writableBuffer","writableLength","writableFinished","writableCorked","writableEnded","writableNeedDrain","destroyed","get","undefined","set","value","body","duplexify","Duplexify","constructor","super","name","_duplexify","write","final","destroy","fn","promise","resolve","createDeferredPromise","ac","signal","async","chunk","done","cb","nextTick","aborted","encoding","err","abort","fromAsyncGen","objectMode","then","d","val","push","read","pair","r","wrap","w","ondrain","onfinish","onreadable","onclose","onfinished","readableObjectMode","_write","callback","_final","end","on","_read","buf","_destroy","reject","Promise","res","rej"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"0BA0BSA,MAAuB,4CACvBC,MAAiB,sBACnBC,MAAS,0CAEdC,WACSC,MACJ,sBACAC,MAAU,mBACVC,MAAc,qCAEnBC,gBACAC,gBACAC,kBACAC,0BACAC,0BACAC,wBACAC,MACK,oBACAC,MAAc,uBAEdC,MAAa,UAEpB,MAAMC,qBACJA,EAAoBC,yBACpBA,GACEb,EAEJc,OAAOC,eAAeC,EAAOC,UAAWf,EAASe,WACjDH,OAAOC,eAAeC,EAAQd,GAI5B,IAAK,MAAMgB,KAAUJ,OAAOK,KAAKT,EAASO,WACnCD,EAAOC,UAAUC,KACpBF,EAAOC,UAAUC,GAAUR,EAASO,UAAUC,mBAIrC,SAASF,EAAOI,GAC7B,KAAMC,gBAAgBL,GACpB,OAAO,IAAIA,EAAOI,GAEpBlB,EAASoB,KAAKD,KAAMD,GACpBV,EAASY,KAAKD,KAAMD,GAEhBA,GACFC,KAAKE,eAA0C,IAA1BH,EAAQG,eAEJ,IAArBH,EAAQI,WACVH,KAAKI,eAAeD,UAAW,EAC/BH,KAAKI,eAAeC,OAAQ,EAC5BL,KAAKI,eAAeE,YAAa,IAGV,IAArBP,EAAQQ,WACVP,KAAKQ,eAAeD,UAAW,EAC/BP,KAAKQ,eAAeC,QAAS,EAC7BT,KAAKQ,eAAeH,OAAQ,EAC5BL,KAAKQ,eAAeE,UAAW,IAGjCV,KAAKE,eAAgB,CAEzB,CAEAT,OAAOkB,iBAAiBhB,EAAOC,UAAW,CACxCW,SACEd,OAAOmB,yBAAyBvB,EAASO,UAAW,YACtDiB,sBACEpB,OAAOmB,yBAAyBvB,EAASO,UAAW,yBACtDkB,mBACErB,OAAOmB,yBAAyBvB,EAASO,UAAW,sBACtDmB,eACEtB,OAAOmB,yBAAyBvB,EAASO,UAAW,kBACtDoB,eACEvB,OAAOmB,yBAAyBvB,EAASO,UAAW,kBACtDqB,iBACExB,OAAOmB,yBAAyBvB,EAASO,UAAW,oBACtDsB,eACEzB,OAAOmB,yBAAyBvB,EAASO,UAAW,kBACtDuB,cACE1B,OAAOmB,yBAAyBvB,EAASO,UAAW,iBACtDwB,kBACE3B,OAAOmB,yBAAyBvB,EAASO,UAAW,qBAEtDyB,UAAW,CACTC,MACE,YAA4BC,IAAxBvB,KAAKI,qBACiBmB,IAAxBvB,KAAKQ,iBAGAR,KAAKI,eAAeiB,WAAarB,KAAKQ,eAAea,UAC9D,EACAG,IAAIC,GAGEzB,KAAKI,gBAAkBJ,KAAKQ,iBAC9BR,KAAKI,eAAeiB,UAAYI,EAChCzB,KAAKQ,eAAea,UAAYI,EAEpC,KAIJ9B,EAAOf,KAAO,SAAS8C,GACrB,OAAOC,EAAUD,EAAM,OACzB,EAGA,MAAME,UAAkBjC,EACtBkC,YAAY9B,GACV+B,MAAM/B,IAIoB,IAAtBA,GAASI,WACXH,KAAKI,eAAeD,UAAW,EAC/BH,KAAKI,eAAeC,OAAQ,EAC5BL,KAAKI,eAAeE,YAAa,IAGT,IAAtBP,GAASQ,WACXP,KAAKQ,eAAeD,UAAW,EAC/BP,KAAKQ,eAAeC,QAAS,EAC7BT,KAAKQ,eAAeH,OAAQ,EAC5BL,KAAKQ,eAAeE,UAAW,EAEnC,EAGF,SAASiB,EAAUD,EAAMK,GACvB,GAAI3C,EAAmBsC,GACrB,OAAOA,EAGT,GAAIxC,EAAqBwC,GACvB,OAAOM,EAAW,CAAE7B,SAAUuB,IAGhC,GAAIvC,EAAqBuC,GACvB,OAAOM,EAAW,CAAEzB,SAAUmB,IAGhC,GAAIzC,EAAayC,GACf,OAAOM,EAAW,CAAEzB,UAAU,EAAOJ,UAAU,IAGjD,GAAoB,mBAATuB,EAAqB,CAC9B,MAAMD,MAAEA,EAAKQ,MAAEA,EAAKC,MAAEA,EAAKC,QAAEA,GAyGjC,SAAsBC,GACpB,IAAIC,QAAEA,EAAOC,QAAEA,GAAYC,IAC3B,MAAMC,EAAK,IAAIjE,EACTkE,EAASD,EAAGC,OAYlB,MAAO,CACLhB,MAZYW,EAAGM,kBACf,OAAa,CACX,MAAMC,MAAEA,EAAKC,KAAEA,EAAIC,GAAEA,SAAaR,EAElC,GADA/C,EAAQwD,SAASD,GACbD,EAAM,OACV,GAAIH,EAAOM,QAAS,MAAM,IAAIrE,QACxBiE,IACHN,UAASC,WAAYC,IAC1B,CACF,CATiBG,GASZ,CAAED,WAILR,MAAMU,EAAOK,EAAUH,GACrBP,EAAQ,CAAEK,QAAOC,MAAM,EAAOC,MAChC,EACAX,MAAMW,GACJP,EAAQ,CAAEM,MAAM,EAAMC,MACxB,EACAV,QAAQc,EAAKJ,GACXL,EAAGU,QACHL,EAAGI,EACL,EAEJ,CArI6CE,CAAazB,GAEtD,GAAI1C,EAAWyC,GACb,OAAO7C,EAAKgD,EAAWH,EAAO,CAE5B2B,YAAY,EACZnB,QACAC,QACAC,YAIJ,MAAMkB,EAAO5B,GAAO4B,KACpB,GAAoB,mBAATA,EAAqB,CAC9B,IAAIC,EAEJ,MAAMjB,EAAUgB,EAAKpD,KAAKwB,GACvB8B,IACC,GAAW,MAAPA,EACF,MAAM,IAAI/D,EAAyB,QAAS,OAAQ+D,EACtD,IAEDN,IACCzE,EAAU8E,EAAGL,EAAI,IAIrB,OAAOK,EAAI,IAAI1B,EAAU,CAEvBwB,YAAY,EACZjD,UAAU,EACV8B,QACAC,MAAMW,GACJX,GAAMQ,UACJ,UACQL,EACN/C,EAAQwD,SAASD,EAAI,KAGvB,CAFE,MAAOI,GACP3D,EAAQwD,SAASD,EAAII,EACvB,IAEJ,EACAd,WAEJ,CAEA,MAAM,IAAI3C,EACR,2CAA4CuC,EAAMN,EACtD,CAEA,GAAIzC,EAAW0C,GACb,OAAO9C,EAAKgD,EAAWF,EAAM,CAE3B0B,YAAY,EACZ7C,UAAU,IAId,GAC4B,iBAAnBmB,GAAMnB,UACa,iBAAnBmB,GAAMvB,SACb,CAWA,OAAO6B,EAAW,CAAE7B,SAVHuB,GAAMvB,SACrBjB,EAAqBwC,GAAMvB,UAAYuB,GAAMvB,SAC3CwB,EAAUD,EAAKvB,eACjBoB,EAO4BhB,SALbmB,GAAMnB,SACrBpB,EAAqBuC,GAAMnB,UAAYmB,GAAMnB,SAC3CoB,EAAUD,EAAKnB,eACjBgB,GAGJ,CAEA,MAAM8B,EAAO3B,GAAM2B,KACnB,GAAoB,mBAATA,EAAqB,CAC9B,IAAIC,EAcJ,OAZAD,EAAKpD,KAAKyB,GACP6B,IACY,MAAPA,GACFD,EAAEE,KAAKD,GAETD,EAAEE,KAAK,KAAK,IAEbP,IACCzE,EAAU8E,EAAGL,EAAI,IAIdK,EAAI,IAAI1B,EAAU,CACvBwB,YAAY,EACZ7C,UAAU,EACVkD,OAAQ,GAEZ,CAEA,MAAM,IAAIlE,EACRwC,EACA,CAAC,OAAQ,iBAAkB,iBAAkB,SAAU,WACtD,gBAAiB,WAAY,8BAA+B,WAC7DL,EACJ,CAgCA,SAASM,EAAW0B,GAClB,MAAMC,EAAID,EAAKvD,UAA0C,mBAAvBuD,EAAKvD,SAASsD,KAC9C5E,EAAS+E,KAAKF,EAAKvD,UAAYuD,EAAKvD,SAChC0D,EAAIH,EAAKnD,SAEf,IAGIuD,EACAC,EACAC,EACAC,EACAX,EAPAnD,IAAarB,EAAW6E,GACxBpD,IAAaxB,EAAW8E,GAQ5B,SAASK,EAAWjB,GAClB,MAAMJ,EAAKoB,EACXA,EAAU,KAENpB,EACFA,EAAGI,GACMA,EACTK,EAAEnB,QAAQc,GACA9C,GAAaI,GACvB+C,EAAEnB,SAEN,CA2GA,OAtGAmB,EAAI,IAAI1B,EAAU,CAEhBuC,qBAAsBR,GAAGQ,mBACzBrD,qBAAsB+C,GAAG/C,mBACzBX,WACAI,aAGEA,IACF9B,EAAIoF,GAAIZ,IACN1C,GAAW,EACP0C,GACFzE,EAAUmF,EAAGV,GAEfiB,EAAWjB,EAAI,IAGjBK,EAAEc,OAAS,SAASzB,EAAOK,EAAUqB,GAC/BR,EAAE5B,MAAMU,EAAOK,GACjBqB,IAEAP,EAAUO,CAEd,EAEAf,EAAEgB,OAAS,SAASD,GAClBR,EAAEU,MACFR,EAAWM,CACb,EAEAR,EAAEW,GAAG,SAAS,WACZ,GAAIV,EAAS,CACX,MAAMjB,EAAKiB,EACXA,EAAU,KACVjB,GACF,CACF,IAEAgB,EAAEW,GAAG,UAAU,WACb,GAAIT,EAAU,CACZ,MAAMlB,EAAKkB,EACXA,EAAW,KACXlB,GACF,CACF,KAGE1C,IACF1B,EAAIkF,GAAIV,IACN9C,GAAW,EACP8C,GACFzE,EAAUmF,EAAGV,GAEfiB,EAAWjB,EAAI,IAGjBU,EAAEa,GAAG,YAAY,WACf,GAAIR,EAAY,CACd,MAAMnB,EAAKmB,EACXA,EAAa,KACbnB,GACF,CACF,IAEAc,EAAEa,GAAG,OAAO,WACVlB,EAAEE,KAAK,KACT,IAEAF,EAAEmB,MAAQ,WACR,OAAa,CACX,MAAMC,EAAMf,EAAEF,OAEd,GAAY,OAARiB,EAEF,YADAV,EAAaV,EAAEmB,OAIjB,IAAKnB,EAAEE,KAAKkB,GACV,MAEJ,CACF,GAGFpB,EAAEqB,SAAW,SAAS1B,EAAKoB,GACpBpB,GAAmB,OAAZgB,IACVhB,EAAM,IAAIvE,GAGZsF,EAAa,KACbF,EAAU,KACVC,EAAW,KAEK,OAAZE,EACFI,EAASpB,IAETgB,EAAUI,EACV7F,EAAUqF,EAAGZ,GACbzE,EAAUmF,EAAGV,GAEjB,EAEOK,CACT,CAEA,SAASf,IACP,IAAID,EACAsC,EAMJ,MAAO,CAAEvC,QALO,IAAIwC,SAAQ,CAACC,EAAKC,KAChCzC,EAAUwC,EACVF,EAASG,CAAG,IAGIzC,UAASsC,SAC7B"}
✄
import{AbortController as e}from"./abort_controller.js";import{destroyer as t}from"./destroy.js";import r from"./end-of-stream.js";import{AbortError as a,codes as o}from"../errors.js";import i from"./from.js";import n from"./readable.js";import{isReadable as l,isWritable as b,isIterable as d,isNodeStream as s,isReadableNodeStream as c,isWritableNodeStream as w,isDuplexNodeStream as p}from"./utils.js";import f from"./writable.js";import u from"process";const{ERR_INVALID_ARG_TYPE:y,ERR_INVALID_RETURN_VALUE:h}=o;Object.setPrototypeOf(j.prototype,n.prototype),Object.setPrototypeOf(j,n);for(const g of Object.keys(f.prototype))j.prototype[g]||(j.prototype[g]=f.prototype[g]);export default function j(e){if(!(this instanceof j))return new j(e);n.call(this,e),f.call(this,e),e?(this.allowHalfOpen=!1!==e.allowHalfOpen,!1===e.readable&&(this._readableState.readable=!1,this._readableState.ended=!0,this._readableState.endEmitted=!0),!1===e.writable&&(this._writableState.writable=!1,this._writableState.ending=!0,this._writableState.ended=!0,this._writableState.finished=!0)):this.allowHalfOpen=!0}Object.defineProperties(j.prototype,{writable:Object.getOwnPropertyDescriptor(f.prototype,"writable"),writableHighWaterMark:Object.getOwnPropertyDescriptor(f.prototype,"writableHighWaterMark"),writableObjectMode:Object.getOwnPropertyDescriptor(f.prototype,"writableObjectMode"),writableBuffer:Object.getOwnPropertyDescriptor(f.prototype,"writableBuffer"),writableLength:Object.getOwnPropertyDescriptor(f.prototype,"writableLength"),writableFinished:Object.getOwnPropertyDescriptor(f.prototype,"writableFinished"),writableCorked:Object.getOwnPropertyDescriptor(f.prototype,"writableCorked"),writableEnded:Object.getOwnPropertyDescriptor(f.prototype,"writableEnded"),writableNeedDrain:Object.getOwnPropertyDescriptor(f.prototype,"writableNeedDrain"),destroyed:{get(){return void 0!==this._readableState&&void 0!==this._writableState&&(this._readableState.destroyed&&this._writableState.destroyed)},set(e){this._readableState&&this._writableState&&(this._readableState.destroyed=e,this._writableState.destroyed=e)}}}),j.from=function(e){return O(e,"body")};class _ extends j{constructor(e){super(e),!1===e?.readable&&(this._readableState.readable=!1,this._readableState.ended=!0,this._readableState.endEmitted=!0),!1===e?.writable&&(this._writableState.writable=!1,this._writableState.ending=!0,this._writableState.ended=!0,this._writableState.finished=!0)}}function O(r,o){if(p(r))return r;if(c(r))return m({readable:r});if(w(r))return m({writable:r});if(s(r))return m({writable:!1,readable:!1});if("function"==typeof r){const{value:n,write:l,final:b,destroy:s}=function(t){let{promise:r,resolve:o}=S();const i=new e,n=i.signal;return{value:t(async function*(){for(;;){const{chunk:e,done:t,cb:i}=await r;if(u.nextTick(i),t)return;if(n.aborted)throw new a;yield e,({promise:r,resolve:o}=S())}}(),{signal:n}),write(e,t,r){o({chunk:e,done:!1,cb:r})},final(e){o({done:!0,cb:e})},destroy(e,t){i.abort(),t(e)}}}(r);if(d(n))return i(_,n,{objectMode:!0,write:l,final:b,destroy:s});const c=n?.then;if("function"==typeof c){let e;const r=c.call(n,(e=>{if(null!=e)throw new h("nully","body",e)}),(r=>{t(e,r)}));return e=new _({objectMode:!0,readable:!1,write:l,final(e){b((async()=>{try{await r,u.nextTick(e,null)}catch(t){u.nextTick(e,t)}}))},destroy:s})}throw new h("Iterable, AsyncIterable or AsyncFunction",o,n)}if(d(r))return i(_,r,{objectMode:!0,writable:!1});if("object"==typeof r?.writable||"object"==typeof r?.readable){return m({readable:r?.readable?c(r?.readable)?r?.readable:O(r.readable):void 0,writable:r?.writable?w(r?.writable)?r?.writable:O(r.writable):void 0})}const n=r?.then;if("function"==typeof n){let e;return n.call(r,(t=>{null!=t&&e.push(t),e.push(null)}),(r=>{t(e,r)})),e=new _({objectMode:!0,writable:!1,read(){}})}throw new y(o,["Blob","ReadableStream","WritableStream","Stream","Iterable","AsyncIterable","Function","{ readable, writable } pair","Promise"],r)}function m(e){const o=e.readable&&"function"!=typeof e.readable.read?n.wrap(e.readable):e.readable,i=e.writable;let d,s,c,w,p,f=!!l(o),u=!!b(i);function y(e){const t=w;w=null,t?t(e):e?p.destroy(e):f||u||p.destroy()}return p=new _({readableObjectMode:!!o?.readableObjectMode,writableObjectMode:!!i?.writableObjectMode,readable:f,writable:u}),u&&(r(i,(e=>{u=!1,e&&t(o,e),y(e)})),p._write=function(e,t,r){i.write(e,t)?r():d=r},p._final=function(e){i.end(),s=e},i.on("drain",(function(){if(d){const e=d;d=null,e()}})),i.on("finish",(function(){if(s){const e=s;s=null,e()}}))),f&&(r(o,(e=>{f=!1,e&&t(o,e),y(e)})),o.on("readable",(function(){if(c){const e=c;c=null,e()}})),o.on("end",(function(){p.push(null)})),p._read=function(){for(;;){const e=o.read();if(null===e)return void(c=p._read);if(!p.push(e))return}}),p._destroy=function(e,r){e||null===w||(e=new a),c=null,d=null,s=null,null===w?r(e):(w=r,t(i,e),t(o,e))},p}function S(){let e,t;return{promise:new Promise(((r,a)=>{e=r,t=a})),resolve:e,reject:t}}
✄
{"version":3,"file":"end-of-stream.js","names":["AbortError","errorCodes","once","isClosed","isReadable","isReadableNodeStream","isReadableFinished","isWritable","isWritableNodeStream","isWritableFinished","isNodeStream","_willEmitClose","process","ERR_STREAM_PREMATURE_CLOSE","isRequest","stream","setHeader","abort","nop","eos","options","callback","arguments","length","readable","writable","wState","_writableState","rState","_readableState","onlegacyfinish","onfinish","willEmitClose","writableFinished","destroyed","readableFinished","call","onend","onerror","err","closed","onclose","errored","onrequest","req","on","aborted","error","nextTick","errorEmitted","cleanup","removeListener","signal","endCallback","originalCallback","args","removeEventListener","apply","addEventListener"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"qBAIEA,WACSC,MACJ,sBACAC,MAAU,+BAEfC,gBACAC,0BACAC,wBACAC,gBACAC,0BACAC,wBACAC,kBACAC,mBACiBC,MACZ,oBAEAC,MAAa,UAEpB,MAAMC,2BACJA,GACEZ,EAEJ,SAASa,EAAUC,GACjB,OAAOA,EAAOC,WAAqC,mBAAjBD,EAAOE,KAC3C,CAEA,MAAMC,EAAM,sBAEG,SAASC,EAAIJ,EAAQK,EAASC,GAClB,IAArBC,UAAUC,QACZF,EAAWD,EACXA,EAAU,CAAC,GACS,MAAXA,IACTA,EAAU,CAAC,GAGbC,EAAWnB,EAAKmB,GAEhB,MAAMG,EAAWJ,EAAQI,WACD,IAArBJ,EAAQI,UAAsBnB,EAAqBU,GAChDU,EAAWL,EAAQK,WACD,IAArBL,EAAQK,UAAsBjB,EAAqBO,GAElDL,EAAaK,GAOjB,MAAMW,EAASX,EAAOY,eAChBC,EAASb,EAAOc,eAEhBC,EAAiB,KAChBf,EAAOU,UAAUM,GAAU,EAMlC,IAAIC,EACFrB,EAAeI,IACfV,EAAqBU,KAAYS,GACjChB,EAAqBO,KAAYU,EAG/BQ,EAAmBxB,EAAmBM,GAAQ,GAClD,MAAMgB,EAAW,KACfE,GAAmB,EAIflB,EAAOmB,YAAWF,GAAgB,KAElCA,GAAmBjB,EAAOS,WAAYA,KACrCA,IAAYW,GAAkBd,EAASe,KAAKrB,GAAO,EAG1D,IAAIoB,EAAmB7B,EAAmBS,GAAQ,GAClD,MAAMsB,EAAQ,KACZF,GAAmB,EAIfpB,EAAOmB,YAAWF,GAAgB,KAElCA,GAAmBjB,EAAOU,WAAYA,KACrCA,IAAYQ,GAAkBZ,EAASe,KAAKrB,GAAO,EAGpDuB,EAAWC,IACflB,EAASe,KAAKrB,EAAQwB,EAAI,EAG5B,IAAIC,EAASrC,EAASY,GAEtB,MAAM0B,EAAU,KACdD,GAAS,EAET,MAAME,EAAUhB,GAAQgB,SAAWd,GAAQc,QAE3C,OAAIA,GAA8B,kBAAZA,EACbrB,EAASe,KAAKrB,EAAQ2B,KAG3BlB,GAAaW,GACV7B,EAAmBS,GAAQ,OAI9BU,GAAaQ,GACVxB,EAAmBM,GAAQ,SAKlCM,EAASe,KAAKrB,GATHM,EAASe,KAAKrB,EACA,IAAIF,EAQR,EAGjB8B,EAAY,KAChB5B,EAAO6B,IAAIC,GAAG,SAAUd,EAAS,EAG/BjB,EAAUC,IACZA,EAAO8B,GAAG,WAAYd,GACjBC,GACHjB,EAAO8B,GAAG,QAASJ,GAEjB1B,EAAO6B,IAAKD,IACX5B,EAAO8B,GAAG,UAAWF,IACjBlB,IAAaC,IACtBX,EAAO8B,GAAG,MAAOf,GACjBf,EAAO8B,GAAG,QAASf,IAIhBE,GAA2C,kBAAnBjB,EAAO+B,SAClC/B,EAAO8B,GAAG,UAAWJ,GAGvB1B,EAAO8B,GAAG,MAAOR,GACjBtB,EAAO8B,GAAG,SAAUd,IACE,IAAlBX,EAAQ2B,OAAiBhC,EAAO8B,GAAG,QAASP,GAChDvB,EAAO8B,GAAG,QAASJ,GAEfD,EACF5B,EAAQoC,SAASP,GACRf,GAAQuB,cAAgBrB,GAAQqB,aACpCjB,GACHpB,EAAQoC,SAASP,IAGlBjB,GACCQ,IAAiB5B,EAAWW,KAC7BkB,GAAqB1B,EAAWQ,MAIhCU,GACCO,IAAiBzB,EAAWQ,KAC7BoB,GAAqB/B,EAAWW,IAGvBa,GAAUb,EAAO6B,KAAO7B,EAAO+B,SACzClC,EAAQoC,SAASP,GARjB7B,EAAQoC,SAASP,GAWnB,MAAMS,EAAU,KACd7B,EAAWH,EACXH,EAAOoC,eAAe,UAAWV,GACjC1B,EAAOoC,eAAe,WAAYpB,GAClChB,EAAOoC,eAAe,QAASV,GAC/B1B,EAAOoC,eAAe,UAAWR,GAC7B5B,EAAO6B,KAAK7B,EAAO6B,IAAIO,eAAe,SAAUpB,GACpDhB,EAAOoC,eAAe,MAAOrB,GAC7Bf,EAAOoC,eAAe,QAASrB,GAC/Bf,EAAOoC,eAAe,SAAUpB,GAChChB,EAAOoC,eAAe,MAAOd,GAC7BtB,EAAOoC,eAAe,QAASb,GAC/BvB,EAAOoC,eAAe,QAASV,EAAQ,EAGzC,GAAIrB,EAAQgC,SAAWZ,EAAQ,CAC7B,MAAMvB,EAAQ,KAEZ,MAAMoC,EAAchC,EACpB6B,IACAG,EAAYjB,KAAKrB,EAAQ,IAAIf,EAAa,EAE5C,GAAIoB,EAAQgC,OAAON,QACjBlC,EAAQoC,SAAS/B,OACZ,CACL,MAAMqC,EAAmBjC,EACzBA,EAAWnB,GAAK,IAAIqD,KAClBnC,EAAQgC,OAAOI,oBAAoB,QAASvC,GAC5CqC,EAAiBG,MAAM1C,EAAQwC,EAAK,IAEtCnC,EAAQgC,OAAOM,iBAAiB,QAASzC,EAC3C,CACF,CAEA,OAAOiC,CACT"}
✄
import{AbortError as e,codes as r}from"../errors.js";import o from"./once.js";import{isClosed as t,isReadable as n,isReadableNodeStream as s,isReadableFinished as i,isWritable as l,isWritableNodeStream as a,isWritableFinished as c,isNodeStream as d,willEmitClose as m}from"./utils.js";import b from"process";const{ERR_STREAM_PREMATURE_CLOSE:f}=r;function v(e){return e.setHeader&&"function"==typeof e.abort}const L=()=>{};export default function p(r,p,u){2===arguments.length?(u=p,p={}):null==p&&(p={}),u=o(u);const E=p.readable||!1!==p.readable&&s(r),w=p.writable||!1!==p.writable&&a(r);d(r);const q=r._writableState,T=r._readableState,x=()=>{r.writable||h()};let y=m(r)&&s(r)===E&&a(r)===w,g=c(r,!1);const h=()=>{g=!0,r.destroyed&&(y=!1),(!y||r.readable&&!E)&&(E&&!k||u.call(r))};let k=i(r,!1);const R=()=>{k=!0,r.destroyed&&(y=!1),(!y||r.writable&&!w)&&(w&&!g||u.call(r))},_=e=>{u.call(r,e)};let S=t(r);const j=()=>{S=!0;const e=q?.errored||T?.errored;return e&&"boolean"!=typeof e?u.call(r,e):(!E||k||i(r,!1))&&(!w||g||c(r,!1))?void u.call(r):u.call(r,new f)},A=()=>{r.req.on("finish",h)};v(r)?(r.on("complete",h),y||r.on("abort",j),r.req?A():r.on("request",A)):w&&!q&&(r.on("end",x),r.on("close",x)),y||"boolean"!=typeof r.aborted||r.on("aborted",j),r.on("end",R),r.on("finish",h),!1!==p.error&&r.on("error",_),r.on("close",j),S?b.nextTick(j):q?.errorEmitted||T?.errorEmitted?y||b.nextTick(j):(E||y&&!n(r)||!g&&l(r))&&(w||y&&!l(r)||!k&&n(r))?T&&r.req&&r.aborted&&b.nextTick(j):b.nextTick(j);const C=()=>{u=L,r.removeListener("aborted",j),r.removeListener("complete",h),r.removeListener("abort",j),r.removeListener("request",A),r.req&&r.req.removeListener("finish",h),r.removeListener("end",x),r.removeListener("close",x),r.removeListener("finish",h),r.removeListener("end",R),r.removeListener("error",_),r.removeListener("close",j)};if(p.signal&&!S){const t=()=>{const o=u;C(),o.call(r,new e)};if(p.signal.aborted)b.nextTick(t);else{const e=u;u=o(((...o)=>{p.signal.removeEventListener("abort",t),e.apply(r,o)})),p.signal.addEventListener("abort",t)}}return C}
✄
{"version":3,"file":"event_target.js","names":["errorCodes","EventEmitter","process","inspect","ERR_INVALID_ARG_TYPE","ERR_EVENT_RECURSION","ERR_MISSING_ARGS","ERR_INVALID_THIS","kIsEventTarget","Symbol","for","kIsNodeEventTarget","kMaxEventTargetListeners","kMaxEventTargetListenersWarned","kEvents","kIsBeingDispatched","kStop","kTarget","kHandlers","kWeakHandler","kHybridDispatch","kCreateEvent","kNewListener","kRemoveListener","kIsNodeStyleListener","kTrustEvent","kType","kDefaultPrevented","kCancelable","kTimestamp","kBubbles","kComposed","kPropagationStopped","isTrustedSet","WeakSet","isTrusted","Object","getOwnPropertyDescriptor","has","this","get","isEvent","value","Event","constructor","type","options","arguments","length","cancelable","bubbles","composed","Date","now","add","defineProperty","enumerable","configurable","custom","depth","name","opts","assign","Number","isInteger","defaultPrevented","timeStamp","stopImmediatePropagation","preventDefault","target","currentTarget","srcElement","composedPath","returnValue","eventPhase","AT_TARGET","NONE","cancelBubble","stopPropagation","static","kEnumerableProperty","create","defineProperties","prototype","toStringTag","writable","NodeCustomEvent","super","detail","weakListenersState","objectToWeakListenerMap","weakListeners","FinalizationRegistry","listener","remove","WeakMap","registry","map","Listener","previous","once","capture","passive","isNodeStyleListener","weak","next","undefined","removed","Boolean","callback","WeakRef","register","set","handleEvent","bind","same","deref","unregister","initEventTarget","self","Map","defaultMaxListeners","EventTarget","size","w","Error","count","emitWarning","addEventListener","isEventTarget","signal","validateEventListenerOptions","shouldAddListener","String","aborted","removeEventListener","root","handler","delete","dispatchEvent","event","nodeValue","createEvent","arg","result","call","addCatch","err","emitUncaughtException","initNodeEventTarget","NodeEventTarget","setMaxListeners","n","isNodeEventTarget","getMaxListeners","eventNames","Array","from","keys","listenerCount","off","removeListener","on","addListener","emit","hadListeners","removeAllListeners","clear","obj","promise","then","nextTick","defineEventHandler","emitter","wrappedHandler","eventHandler","args","Reflect","apply","makeEventHandler","EventEmitterMixin","Superclass","MixedEventEmitter","protoProps","getOwnPropertyDescriptors"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"gBAAkBA,MAAkB,sBAE7BC,MAAkB,gBAClBC,MAAa,4BACXC,MAAe,OAExB,MAAMC,qBACJA,EAAoBC,oBACpBA,EAAmBC,iBACnBA,EAAgBC,iBAChBA,GACEP,EAEEQ,EAAiBC,OAAOC,IAAI,uBAC5BC,EAAqBF,OAAO,uBAE5BG,yBACJA,EAAwBC,+BACxBA,GACEZ,SAEG,MAAMa,QAAUL,OAAO,WAC9B,MAAMM,EAAqBN,OAAO,sBAC5BO,EAAQP,OAAO,SACfQ,EAAUR,OAAO,WACjBS,EAAYT,OAAO,oBAClB,MAAMU,aAAeV,OAAO,SAEnC,MAAMW,EAAkBX,OAAOC,IAAI,0CAC5B,MAAMW,aAAeZ,OAAO,uBAC5B,MAAMa,aAAeb,OAAO,uBAC5B,MAAMc,gBAAkBd,OAAO,mBACtC,MAAMe,EAAuBf,OAAO,+BAC7B,MAAMgB,YAAchB,OAAO,eAKlC,MAAMiB,EAAQjB,OAAO,QACfkB,EAAoBlB,OAAO,oBAC3BmB,EAAcnB,OAAO,cACrBoB,EAAapB,OAAO,aACpBqB,EAAWrB,OAAO,WAClBsB,EAAYtB,OAAO,YACnBuB,EAAsBvB,OAAO,sBAE7BwB,EAAe,IAAIC,QACnBC,EAAYC,OAAOC,yBAAyB,CAC5CF,gBACF,OAAOF,EAAaK,IAAIC,KAC1B,GACC,aAAaC,IAEhB,SAASC,EAAQC,GACf,MAAiC,iBAAnBA,IAAQhB,EACxB,QAEO,MAAMiB,MACXC,YAAYC,EAAMC,EAAU,MAC1B,GAAyB,IAArBC,UAAUC,OACZ,MAAM,IAAI1C,EAAiB,QAC7B,MAAM2C,WAAEA,EAAUC,QAAEA,EAAOC,SAAEA,GAAa,IAAKL,GAC/CP,KAAKX,KAAiBqB,EACtBV,KAAKT,KAAcoB,EACnBX,KAAKR,KAAeoB,EACpBZ,KAAKb,GAAS,GAAGmB,IACjBN,KAAKZ,IAAqB,EAC1BY,KAAKV,GAAcuB,KAAKC,MACxBd,KAAKP,IAAuB,EACxBc,IAAUrB,cACZQ,EAAaqB,IAAIf,MAInBH,OAAOmB,eAAehB,KAAM,YAAa,CACvCC,IAAKL,EACLqB,YAAY,EACZC,cAAc,IAEhBlB,KAAKtB,GAAW,KAChBsB,KAAKxB,IAAsB,CAC7B,CAEA,CAACZ,EAAQuD,QAAQC,EAAOb,GACtB,IAAKL,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,MAAMqD,EAAOrB,KAAKK,YAAYgB,KAC9B,GAAID,EAAQ,EACV,OAAOC,EAET,MAAMC,EAAOzB,OAAO0B,OAAO,CAAC,EAAGhB,EAAS,CACtCa,MAAOI,OAAOC,UAAUlB,EAAQa,OAASb,EAAQa,MAAQ,EAAIb,EAAQa,QAGvE,MAAO,GAAGC,KAAQzD,EAAQ,CACxB0C,KAAMN,KAAKb,GACXuC,iBAAkB1B,KAAKZ,GACvBsB,WAAYV,KAAKX,GACjBsC,UAAW3B,KAAKV,IACfgC,IACL,CAEAM,2BACE,IAAK1B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7BgC,KAAKvB,IAAS,CAChB,CAEAoD,iBACE,IAAK3B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7BgC,KAAKZ,IAAqB,CAC5B,CAEI0C,aACF,IAAK5B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKtB,EACd,CAEIqD,oBACF,IAAK7B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKtB,EACd,CAEIsD,iBACF,IAAK9B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKtB,EACd,CAEI4B,WACF,IAAKJ,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKb,EACd,CAEIuB,iBACF,IAAKR,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKX,EACd,CAEIqC,uBACF,IAAKxB,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKX,IAAgBW,KAAKZ,EACnC,CAEIuC,gBACF,IAAKzB,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKV,EACd,CAOA2C,eACE,IAAK/B,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKxB,GAAsB,CAACwB,KAAKtB,IAAY,EACtD,CAEIwD,kBACF,IAAKhC,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAQgC,KAAK0B,gBACf,CAEIf,cACF,IAAKT,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKT,EACd,CAEIqB,eACF,IAAKV,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKR,EACd,CAEI2C,iBACF,IAAKjC,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKxB,GAAsB4B,MAAMgC,UAAYhC,MAAMiC,IAC5D,CAEIC,mBACF,IAAKpC,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7B,OAAOgC,KAAKP,EACd,CAEI6C,iBAAanC,GACf,IAAKD,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SACzBmC,GACFH,KAAKuC,iBAET,CAEAA,kBACE,IAAKrC,EAAQF,MACX,MAAM,IAAIhC,EAAiB,SAC7BgC,KAAKP,IAAuB,CAC9B,CAEA+C,YAAc,EACdA,uBAAyB,EACzBA,iBAAmB,EACnBA,sBAAwB,EAG1B,MAAMC,EAAsB5C,OAAO6C,OAAO,MAC1CD,EAAoBxB,YAAa,EAEjCpB,OAAO8C,iBACLvC,MAAMwC,UAAW,CACf,CAAC1E,OAAO2E,aAAc,CACpBC,UAAU,EACV7B,YAAY,EACZC,cAAc,EACdf,MAAO,SAETyB,yBAA0Ba,EAC1BZ,eAAgBY,EAChBX,OAAQW,EACRV,cAAeU,EACfT,WAAYS,EACZnC,KAAMmC,EACN/B,WAAY+B,EACZf,iBAAkBe,EAClBd,UAAWc,EACXR,aAAcQ,EACdP,YAAaO,EACb9B,QAAS8B,EACT7B,SAAU6B,EACVN,WAAYM,EACZH,aAAcG,EACdF,gBAAiBE,IAGrB,MAAMM,UAAwB3C,MAC5BC,YAAYC,EAAMC,GAChByC,MAAM1C,EAAMC,GACRA,GAAS0C,SACXjD,KAAKiD,OAAS1C,EAAQ0C,OAE1B,EAKF,IAAIC,EAAqB,KAGrBC,EAA0B,KAC9B,SAASC,IASP,OAR2B,OAAvBF,IACFA,EAAqB,IAAIG,sBACtBC,GAAaA,EAASC,YAGK,OAA5BJ,IACFA,EAA0B,IAAIK,SAEzB,CAAEC,SAAUP,EAAoBQ,IAAKP,EAC9C,CASA,MAAMQ,EACJtD,YAAYuD,EAAUN,EAAUO,EAAMC,EAASC,EACnCC,EAAqBC,GAC/BjE,KAAKkE,UAAOC,OACKA,IAAbP,IACFA,EAASM,KAAOlE,MAClBA,KAAK4D,SAAWA,EAChB5D,KAAKsD,SAAWA,EAEhBtD,KAAK6D,KAAOA,EACZ7D,KAAK8D,QAAUA,EACf9D,KAAK+D,QAAUA,EACf/D,KAAKgE,oBAAsBA,EAC3BhE,KAAKoE,SAAU,EACfpE,KAAKiE,KAAOI,QAAQJ,GAEhBjE,KAAKiE,MACPjE,KAAKsE,SAAW,IAAIC,QAAQjB,GAC5BF,IAAgBK,SAASe,SAASlB,EAAUtD,KAAMA,MAElDoD,IAAgBM,IAAIe,IAAIR,EAAMX,GAC9BtD,KAAKsD,SAAWtD,KAAKsE,UACQ,mBAAbhB,GAChBtD,KAAKsE,SAAWhB,EAChBtD,KAAKsD,SAAWA,IAEhBtD,KAAKsE,SAAWhB,EAASoB,YAAYC,KAAKrB,GAC1CtD,KAAKsD,SAAWA,EAEpB,CAEAsB,KAAKtB,EAAUQ,GAEb,OADmB9D,KAAKiE,KAAOjE,KAAKsD,SAASuB,QAAU7E,KAAKsD,YACtCA,GAAYtD,KAAK8D,UAAYA,CACrD,CAEAP,cACwBY,IAAlBnE,KAAK4D,WACP5D,KAAK4D,SAASM,KAAOlE,KAAKkE,WACVC,IAAdnE,KAAKkE,OACPlE,KAAKkE,KAAKN,SAAW5D,KAAK4D,UAC5B5D,KAAKoE,SAAU,EACXpE,KAAKiE,MACPb,IAAgBK,SAASqB,WAAW9E,KACxC,SAGK,SAAS+E,gBAAgBC,GAC9BA,EAAKzG,SAAW,IAAI0G,IACpBD,EAAK3G,GAA4BX,EAAawH,oBAC9CF,EAAK1G,IAAkC,CACzC,QAEO,MAAM6G,YAIX3C,OAAQvE,IAAkB,EAE1BoC,cACE0E,gBAAgB/E,KAClB,CAEAjB,CAACA,cAAcqG,EAAM9E,EAAMgD,EAAUO,EAAMC,EAASC,GAClD,GAAI/D,KAAK3B,GAA4B,GACjC+G,EAAOpF,KAAK3B,KACX2B,KAAK1B,GAAiC,CACzC0B,KAAK1B,IAAkC,EAGvC,MAAM+G,EAAI,IAAIC,MACM,8CAAGF,KAAQ9E,wBACC1C,EAAQoC,KAAM,CAAEoB,OAAQ,uDAExDiE,EAAEhE,KAAO,8BACTgE,EAAEvD,OAAS9B,KACXqF,EAAE/E,KAAOA,EACT+E,EAAEE,MAAQH,EACVzH,EAAQ6H,YAAYH,EACtB,CACF,CACArG,CAACA,iBAAiBoG,EAAM9E,EAAMgD,EAAUQ,GAAU,CAElD2B,iBAAiBnF,EAAMgD,EAAU/C,EAAU,CAAC,GAC1C,IAAKmF,cAAc1F,MACjB,MAAM,IAAIhC,EAAiB,eAC7B,GAAIwC,UAAUC,OAAS,EACrB,MAAM,IAAI1C,EAAiB,OAAQ,YAIrC,MAAM8F,KACJA,EAAIC,QACJA,EAAOC,QACPA,EAAO4B,OACPA,EAAM3B,oBACNA,EAAmBC,KACnBA,GAoUN,SAAsC1D,GACpC,MAAuB,kBAAZA,EACF,CAAEuD,QAASvD,GAEJ,OAAZA,EACK,CAAC,EACH,CACLsD,KAAMQ,QAAQ9D,EAAQsD,MACtBC,QAASO,QAAQ9D,EAAQuD,SACzBC,QAASM,QAAQ9D,EAAQwD,SACzB4B,OAAQpF,EAAQoF,OAChB1B,KAAM1D,EAAQ3B,cACdoF,oBAAqBK,QAAQ9D,EAAQtB,IAEzC,CAjVQ2G,CAA6BrF,GAEjC,IAAKsF,EAAkBvC,GAAW,CAIhC,MAAM+B,EAAI,IAAIC,MAAM,gCAAgChC,0BAMpD,OAJA+B,EAAEhE,KAAO,sCACTgE,EAAEvD,OAAS9B,KACXqF,EAAE/E,KAAOA,OACT3C,EAAQ6H,YAAYH,EAEtB,CAGA,GAFA/E,EAAOwF,OAAOxF,GAEVqF,EAAQ,CACV,GAAIA,EAAOI,QACT,OAIFJ,EAAOF,iBAAiB,SAAS,KAC/BzF,KAAKgG,oBAAoB1F,EAAMgD,EAAU/C,EAAQ,GAChD,CAAEsD,MAAM,EAAMjF,CAACA,cAAeoB,MACnC,CAEA,IAAIiG,EAAOjG,KAAKzB,SAAS0B,IAAIK,GAE7B,QAAa6D,IAAT8B,EAOF,OANAA,EAAO,CAAEb,KAAM,EAAGlB,UAAMC,GAExB,IAAIR,EAASsC,EAAM3C,EAAUO,EAAMC,EAASC,EAC/BC,EAAqBC,GAClCjE,KAAKjB,cAAckH,EAAKb,KAAM9E,EAAMgD,EAAUO,EAAMC,EAASC,QAC7D/D,KAAKzB,SAASkG,IAAInE,EAAM2F,GAI1B,IAAIC,EAAUD,EAAK/B,KACfN,EAAWqC,EAGf,UAAmB9B,IAAZ+B,IAA0BA,EAAQtB,KAAKtB,EAAUQ,IACtDF,EAAWsC,EACXA,EAAUA,EAAQhC,UAGJC,IAAZ+B,IAIJ,IAAIvC,EAASC,EAAUN,EAAUO,EAAMC,EAASC,EACnCC,EAAqBC,GAClCgC,EAAKb,OACLpF,KAAKjB,cAAckH,EAAKb,KAAM9E,EAAMgD,EAAUO,EAAMC,EAASC,GAC/D,CAEAiC,oBAAoB1F,EAAMgD,EAAU/C,EAAU,CAAC,GAC7C,IAAKmF,cAAc1F,MACjB,MAAM,IAAIhC,EAAiB,eAC7B,IAAK6H,EAAkBvC,GACrB,OAEFhD,EAAOwF,OAAOxF,GACd,MAAMwD,GAA+B,IAArBvD,GAASuD,QAEnBmC,EAAOjG,KAAKzB,SAAS0B,IAAIK,GAC/B,QAAa6D,IAAT8B,QAAoC9B,IAAd8B,EAAK/B,KAC7B,OAEF,IAAIgC,EAAUD,EAAK/B,KACnB,UAAmBC,IAAZ+B,GAAuB,CAC5B,GAAIA,EAAQtB,KAAKtB,EAAUQ,GAAU,CACnCoC,EAAQ3C,SACR0C,EAAKb,OACa,IAAda,EAAKb,MACPpF,KAAKzB,SAAS4H,OAAO7F,GACvBN,KAAKhB,iBAAiBiH,EAAKb,KAAM9E,EAAMgD,EAAUQ,GACjD,KACF,CACAoC,EAAUA,EAAQhC,IACpB,CACF,CAEAkC,cAAcC,GACZ,IAAKX,cAAc1F,MACjB,MAAM,IAAIhC,EAAiB,eAE7B,KAAMqI,aAAiBjG,OACrB,MAAM,IAAIvC,EAAqB,QAAS,QAASwI,GAEnD,GAAIA,EAAM7H,GACR,MAAM,IAAIV,EAAoBuI,EAAM/F,MAItC,OAFAN,KAAKnB,GAAiBwH,EAAOA,EAAM/F,KAAM+F,IAEP,IAA3BA,EAAM3E,gBACf,CAEA7C,CAACA,GAAiByH,EAAWhG,EAAM+F,GACjC,MAAME,EAAc,UACJpC,IAAVkC,KACFA,EAAQrG,KAAKlB,cAAcwH,EAAWhG,IAChC5B,GAAWsB,KACjBqG,EAAM7H,IAAsB,GAEvB6H,QAEKlC,IAAVkC,IACFA,EAAM3H,GAAWsB,KACjBqG,EAAM7H,IAAsB,GAG9B,MAAMyH,EAAOjG,KAAKzB,SAAS0B,IAAIK,GAC/B,QAAa6D,IAAT8B,QAAoC9B,IAAd8B,EAAK/B,KAG7B,YAFcC,IAAVkC,IACFA,EAAM7H,IAAsB,IACvB,EAGT,IACI0F,EADAgC,EAAUD,EAAK/B,KAGnB,UAAmBC,IAAZ+B,IACCA,EAAQnC,UAA8B,IAAnBsC,IAAQ5H,KAIjC,GAFAyF,EAAOgC,EAAQhC,KAEXgC,EAAQ9B,QAGV8B,EAAUhC,MAHZ,CAMA,GAAIgC,EAAQrC,KAAM,CAChBqC,EAAQ3C,SACR0C,EAAKb,OACL,MAAM9B,SAAEA,EAAQQ,QAAEA,GAAYoC,EAC9BlG,KAAKhB,iBAAiBiH,EAAKb,KAAM9E,EAAMgD,EAAUQ,EACnD,CAEA,IACE,IAAI0C,EAEFA,EADEN,EAAQlC,oBACJsC,EAEAC,IAER,MAAMjC,EAAW4B,EAAQjC,KACvBiC,EAAQ5B,SAASO,QAAUqB,EAAQ5B,SACrC,IAAImC,EACAnC,IACFmC,EAASnC,EAASoC,KAAK1G,KAAMwG,GACxBN,EAAQlC,sBACXwC,EAAIhI,IAAsB,IAG1BiI,SACFE,EAASF,EAGb,CAFE,MAAOG,GACPC,EAAsBD,EACxB,CAEAV,EAAUhC,CA9BV,MAiCYC,IAAVkC,IACFA,EAAM7H,IAAsB,EAChC,CAEAM,CAACA,cAAcwH,EAAWhG,GACxB,OAAO,IAAIyC,EAAgBzC,EAAM,CAAE2C,OAAQqD,GAC7C,CACA,CAAC1I,EAAQuD,QAAQC,EAAOb,GACtB,IAAKmF,cAAc1F,MACjB,MAAM,IAAIhC,EAAiB,eAC7B,MAAMqD,EAAOrB,KAAKK,YAAYgB,KAC9B,GAAID,EAAQ,EACV,OAAOC,EAET,MAAMC,EAAOzB,OAAO0B,OAAO,CAAC,EAAGhB,EAAS,CACtCa,MAAOI,OAAOC,UAAUlB,EAAQa,OAASb,EAAQa,MAAQ,EAAIb,EAAQa,QAGvE,MAAO,GAAGC,KAAQzD,EAAQ,CAAC,EAAG0D,IAChC,EAGFzB,OAAO8C,iBAAiBwC,YAAYvC,UAAW,CAC7C6C,iBAAkBhD,EAClBuD,oBAAqBvD,EACrB2D,cAAe3D,EACf,CAACvE,OAAO2E,aAAc,CACpBC,UAAU,EACV7B,YAAY,EACZC,cAAc,EACdf,MAAO,wBAIJ,SAAS2G,oBAAoB9B,GAClCD,gBAAgBC,EAClB,QAEO,MAAM+B,wBAAwB5B,YACnC3C,OAAQpE,IAAsB,EAC9BoE,2BAA6B,GAE7BnC,cACE2C,QACA8D,oBAAoB9G,KACtB,CAEAgH,gBAAgBC,GACd,IAAKC,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAC7BN,EAAasJ,gBAAgBC,EAAGjH,KAClC,CAEAmH,kBACE,IAAKD,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAC7B,OAAOgC,KAAK3B,EACd,CAEA+I,aACE,IAAKF,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAC7B,OAAOqJ,MAAMC,KAAKtH,KAAKzB,SAASgJ,OAClC,CAEAC,cAAclH,GACZ,IAAK4G,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAC7B,MAAMiI,EAAOjG,KAAKzB,SAAS0B,IAAI6F,OAAOxF,IACtC,YAAgB6D,IAAT8B,EAAqBA,EAAKb,KAAO,CAC1C,CAEAqC,IAAInH,EAAMgD,EAAU/C,GAClB,IAAK2G,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAE7B,OADAgC,KAAKgG,oBAAoB1F,EAAMgD,EAAU/C,GAClCP,IACT,CAEA0H,eAAepH,EAAMgD,EAAU/C,GAC7B,IAAK2G,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAE7B,OADAgC,KAAKgG,oBAAoB1F,EAAMgD,EAAU/C,GAClCP,IACT,CAEA2H,GAAGrH,EAAMgD,GACP,IAAK4D,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAE7B,OADAgC,KAAKyF,iBAAiBnF,EAAMgD,EAAU,CAAErE,CAACA,IAAuB,IACzDe,IACT,CAEA4H,YAAYtH,EAAMgD,GAChB,IAAK4D,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAE7B,OADAgC,KAAKyF,iBAAiBnF,EAAMgD,EAAU,CAAErE,CAACA,IAAuB,IACzDe,IACT,CACA6H,KAAKvH,EAAMkG,GACT,IAAKU,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAC7B,MAAM8J,EAAe9H,KAAKwH,cAAclH,GAAQ,EAEhD,OADAN,KAAKnB,GAAiB2H,EAAKlG,GACpBwH,CACT,CAEAjE,KAAKvD,EAAMgD,GACT,IAAK4D,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAG7B,OAFAgC,KAAKyF,iBAAiBnF,EAAMgD,EACN,CAAEO,MAAM,EAAM5E,CAACA,IAAuB,IACrDe,IACT,CAEA+H,mBAAmBzH,GACjB,IAAK4G,EAAkBlH,MACrB,MAAM,IAAIhC,EAAiB,mBAO7B,YANamG,IAAT7D,EACFN,KAAKzB,SAAS4H,OAAOL,OAAOxF,IAE5BN,KAAKzB,SAASyJ,QAGThI,IACT,EAmBF,SAAS6F,EAAkBvC,GACzB,GAAwB,mBAAbA,GAC0B,mBAA1BA,GAAUoB,YACnB,OAAO,EAGT,GAAgB,MAAZpB,EACF,OAAO,EAET,MAAM,IAAIzF,EAAqB,WAAY,gBAAiByF,EAC9D,CA1BAzD,OAAO8C,iBAAiBoE,gBAAgBnE,UAAW,CACjDoE,gBAAiBvE,EACjB0E,gBAAiB1E,EACjB2E,WAAY3E,EACZ+E,cAAe/E,EACfgF,IAAKhF,EACLiF,eAAgBjF,EAChBkF,GAAIlF,EACJmF,YAAanF,EACboB,KAAMpB,EACNoF,KAAMpF,EACNsF,mBAAoBtF,WAsCf,SAASiD,cAAcuC,GAC5B,OAAOA,GAAK5H,cAAcpC,EAC5B,CAEA,SAASiJ,EAAkBe,GACzB,OAAOA,GAAK5H,cAAcjC,EAC5B,CAEA,SAASuI,EAASuB,GAChB,MAAMC,EAAOD,EAAQC,KACD,mBAATA,GACTA,EAAKzB,KAAKwB,OAAS/D,GAAW,SAASyC,GAGrCC,EAAsBD,EACxB,GAEJ,CAEA,SAASC,EAAsBD,GAC7BjJ,EAAQyK,UAAS,KAAQ,MAAMxB,CAAG,GACpC,QAeO,SAASyB,mBAAmBC,EAASjH,GAE1CxB,OAAOmB,eAAesH,EAAS,KAAKjH,IAAQ,CAC1CpB,MACE,OAAOD,KAAKrB,IAAYsB,IAAIoB,IAAO6E,OACrC,EACAzB,IAAItE,GACGH,KAAKrB,KACRqB,KAAKrB,GAAa,IAAIsG,KAExB,IAAIsD,EAAiBvI,KAAKrB,IAAYsB,IAAIoB,GAC1C,GAAIkH,EAAgB,CAClB,GAAsC,mBAA3BA,EAAerC,QAAwB,CAChDlG,KAAKzB,SAAS0B,IAAIoB,GAAM+D,OACxB,MAAMA,EAAOpF,KAAKzB,SAAS0B,IAAIoB,GAAM+D,KACrCpF,KAAKhB,iBAAiBoG,EAAM/D,EAAMkH,EAAerC,SAAS,EAC5D,CAEA,GADAqC,EAAerC,QAAU/F,EACa,mBAA3BoI,EAAerC,QAAwB,CAChDlG,KAAKzB,SAAS0B,IAAIoB,GAAM+D,OACxB,MAAMA,EAAOpF,KAAKzB,SAAS0B,IAAIoB,GAAM+D,KACrCpF,KAAKjB,cAAcqG,EAAM/D,EAAMlB,GAAO,GAAO,GAAO,EACtD,CACF,MACEoI,EArCR,SAA0BrC,GAGxB,SAASsC,KAAgBC,GACvB,GAAoC,mBAAzBD,EAAatC,QAGxB,OAAOwC,QAAQC,MAAMH,EAAatC,QAASlG,KAAMyI,EACnD,CAEA,OADAD,EAAatC,QAAUA,EAChBsC,CACT,CA0ByBI,CAAiBzI,GAClCH,KAAKyF,iBAAiBpE,EAAMkH,GAE9BvI,KAAKrB,GAAW8F,IAAIpD,EAAMkH,EAC5B,EACArH,cAAc,EACdD,YAAY,GAEhB,QAEO,MAAM4H,kBAAqBC,IAChC,MAAMC,UAA0BD,EAC9BzI,eAAeoI,GACbzF,SAASyF,GACT/K,EAAagJ,KAAK1G,KACpB,EAEF,MAAMgJ,EAAanJ,OAAOoJ,0BAA0BvL,EAAakF,WAGjE,cAFOoG,EAAW3I,YAClBR,OAAO8C,iBAAiBoG,EAAkBnG,UAAWoG,GAC9CD,CAAiB"}
✄
import{codes as e}from"../errors.js";import t from"events";import n from"process";import{inspect as i}from"util";const{ERR_INVALID_ARG_TYPE:r,ERR_EVENT_RECURSION:s,ERR_MISSING_ARGS:o,ERR_INVALID_THIS:a}=e,h=Symbol.for("nodejs.event_target"),l=Symbol("kIsNodeEventTarget"),{kMaxEventTargetListeners:c,kMaxEventTargetListenersWarned:v}=t;export const kEvents=Symbol("kEvents");const d=Symbol("kIsBeingDispatched"),u=Symbol("kStop"),f=Symbol("kTarget"),p=Symbol("khandlers");export const kWeakHandler=Symbol("kWeak");const E=Symbol.for("nodejs.internal.kHybridDispatch");export const kCreateEvent=Symbol("kCreateEvent");export const kNewListener=Symbol("kNewListener");export const kRemoveListener=Symbol("kRemoveListener");const g=Symbol("kIsNodeStyleListener");export const kTrustEvent=Symbol("kTrustEvent");const w=Symbol("type"),m=Symbol("defaultPrevented"),b=Symbol("cancelable"),k=Symbol("timestamp"),y=Symbol("bubbles"),T=Symbol("composed"),L=Symbol("propagationStopped"),S=new WeakSet,x=Object.getOwnPropertyDescriptor({get isTrusted(){return S.has(this)}},"isTrusted").get;function N(e){return"string"==typeof e?.[w]}export class Event{constructor(e,t=null){if(0===arguments.length)throw new o("type");const{cancelable:n,bubbles:i,composed:r}={...t};this[b]=!!n,this[y]=!!i,this[T]=!!r,this[w]=`${e}`,this[m]=!1,this[k]=Date.now(),this[L]=!1,t?.[kTrustEvent]&&S.add(this),Object.defineProperty(this,"isTrusted",{get:x,enumerable:!0,configurable:!1}),this[f]=null,this[d]=!1}[i.custom](e,t){if(!N(this))throw new a("Event");const n=this.constructor.name;if(e<0)return n;const r=Object.assign({},t,{depth:Number.isInteger(t.depth)?t.depth-1:t.depth});return`${n} ${i({type:this[w],defaultPrevented:this[m],cancelable:this[b],timeStamp:this[k]},r)}`}stopImmediatePropagation(){if(!N(this))throw new a("Event");this[u]=!0}preventDefault(){if(!N(this))throw new a("Event");this[m]=!0}get target(){if(!N(this))throw new a("Event");return this[f]}get currentTarget(){if(!N(this))throw new a("Event");return this[f]}get srcElement(){if(!N(this))throw new a("Event");return this[f]}get type(){if(!N(this))throw new a("Event");return this[w]}get cancelable(){if(!N(this))throw new a("Event");return this[b]}get defaultPrevented(){if(!N(this))throw new a("Event");return this[b]&&this[m]}get timeStamp(){if(!N(this))throw new a("Event");return this[k]}composedPath(){if(!N(this))throw new a("Event");return this[d]?[this[f]]:[]}get returnValue(){if(!N(this))throw new a("Event");return!this.defaultPrevented}get bubbles(){if(!N(this))throw new a("Event");return this[y]}get composed(){if(!N(this))throw new a("Event");return this[T]}get eventPhase(){if(!N(this))throw new a("Event");return this[d]?Event.AT_TARGET:Event.NONE}get cancelBubble(){if(!N(this))throw new a("Event");return this[L]}set cancelBubble(e){if(!N(this))throw new a("Event");e&&this.stopPropagation()}stopPropagation(){if(!N(this))throw new a("Event");this[L]=!0}static NONE=0;static CAPTURING_PHASE=1;static AT_TARGET=2;static BUBBLING_PHASE=3}const P=Object.create(null);P.enumerable=!0,Object.defineProperties(Event.prototype,{[Symbol.toStringTag]:{writable:!1,enumerable:!1,configurable:!0,value:"Event"},stopImmediatePropagation:P,preventDefault:P,target:P,currentTarget:P,srcElement:P,type:P,cancelable:P,defaultPrevented:P,timeStamp:P,composedPath:P,returnValue:P,bubbles:P,composed:P,eventPhase:P,cancelBubble:P,stopPropagation:P});class R extends Event{constructor(e,t){super(e,t),t?.detail&&(this.detail=t.detail)}}let I=null,A=null;function M(){return null===I&&(I=new FinalizationRegistry((e=>e.remove()))),null===A&&(A=new WeakMap),{registry:I,map:A}}class O{constructor(e,t,n,i,r,s,o){this.next=void 0,void 0!==e&&(e.next=this),this.previous=e,this.listener=t,this.once=n,this.capture=i,this.passive=r,this.isNodeStyleListener=s,this.removed=!1,this.weak=Boolean(o),this.weak?(this.callback=new WeakRef(t),M().registry.register(t,this,this),M().map.set(o,t),this.listener=this.callback):"function"==typeof t?(this.callback=t,this.listener=t):(this.callback=t.handleEvent.bind(t),this.listener=t)}same(e,t){return(this.weak?this.listener.deref():this.listener)===e&&this.capture===t}remove(){void 0!==this.previous&&(this.previous.next=this.next),void 0!==this.next&&(this.next.previous=this.previous),this.removed=!0,this.weak&&M().registry.unregister(this)}}export function initEventTarget(e){e[kEvents]=new Map,e[c]=t.defaultMaxListeners,e[v]=!1}export class EventTarget{static[h]=!0;constructor(){initEventTarget(this)}[kNewListener](e,t,r,s,o,a){if(this[c]>0&&e>this[c]&&!this[v]){this[v]=!0;const r=new Error(`Possible EventTarget memory leak detected. ${e} ${t} listeners added to ${i(this,{depth:-1})}. Use events.setMaxListeners() to increase limit`);r.name="MaxListenersExceededWarning",r.target=this,r.type=t,r.count=e,n.emitWarning(r)}}[kRemoveListener](e,t,n,i){}addEventListener(e,t,i={}){if(!isEventTarget(this))throw new a("EventTarget");if(arguments.length<2)throw new o("type","listener");const{once:r,capture:s,passive:h,signal:l,isNodeStyleListener:c,weak:v}=function(e){return"boolean"==typeof e?{capture:e}:null===e?{}:{once:Boolean(e.once),capture:Boolean(e.capture),passive:Boolean(e.passive),signal:e.signal,weak:e[kWeakHandler],isNodeStyleListener:Boolean(e[g])}}(i);if(!z(t)){const i=new Error(`addEventListener called with ${t} which has no effect.`);return i.name="AddEventListenerArgumentTypeWarning",i.target=this,i.type=e,void n.emitWarning(i)}if(e=String(e),l){if(l.aborted)return;l.addEventListener("abort",(()=>{this.removeEventListener(e,t,i)}),{once:!0,[kWeakHandler]:this})}let d=this[kEvents].get(e);if(void 0===d)return d={size:1,next:void 0},new O(d,t,r,s,h,c,v),this[kNewListener](d.size,e,t,r,s,h),void this[kEvents].set(e,d);let u=d.next,f=d;for(;void 0!==u&&!u.same(t,s);)f=u,u=u.next;void 0===u&&(new O(f,t,r,s,h,c,v),d.size++,this[kNewListener](d.size,e,t,r,s,h))}removeEventListener(e,t,n={}){if(!isEventTarget(this))throw new a("EventTarget");if(!z(t))return;e=String(e);const i=!0===n?.capture,r=this[kEvents].get(e);if(void 0===r||void 0===r.next)return;let s=r.next;for(;void 0!==s;){if(s.same(t,i)){s.remove(),r.size--,0===r.size&&this[kEvents].delete(e),this[kRemoveListener](r.size,e,t,i);break}s=s.next}}dispatchEvent(e){if(!isEventTarget(this))throw new a("EventTarget");if(!(e instanceof Event))throw new r("event","Event",e);if(e[d])throw new s(e.type);return this[E](e,e.type,e),!0!==e.defaultPrevented}[E](e,t,n){const i=()=>(void 0===n&&((n=this[kCreateEvent](e,t))[f]=this,n[d]=!0),n);void 0!==n&&(n[f]=this,n[d]=!0);const r=this[kEvents].get(t);if(void 0===r||void 0===r.next)return void 0!==n&&(n[d]=!1),!0;let s,o=r.next;for(;void 0!==o&&(o.passive||!0!==n?.[u]);)if(s=o.next,o.removed)o=s;else{if(o.once){o.remove(),r.size--;const{listener:e,capture:n}=o;this[kRemoveListener](r.size,t,e,n)}try{let t;t=o.isNodeStyleListener?e:i();const n=o.weak?o.callback.deref():o.callback;let r;n&&(r=n.call(this,t),o.isNodeStyleListener||(t[d]=!1)),null!=r&&_(r)}catch(e){B(e)}o=s}void 0!==n&&(n[d]=!1)}[kCreateEvent](e,t){return new R(t,{detail:e})}[i.custom](e,t){if(!isEventTarget(this))throw new a("EventTarget");const n=this.constructor.name;if(e<0)return n;const r=Object.assign({},t,{depth:Number.isInteger(t.depth)?t.depth-1:t.depth});return`${n} ${i({},r)}`}}Object.defineProperties(EventTarget.prototype,{addEventListener:P,removeEventListener:P,dispatchEvent:P,[Symbol.toStringTag]:{writable:!1,enumerable:!1,configurable:!0,value:"EventTarget"}});export function initNodeEventTarget(e){initEventTarget(e)}export class NodeEventTarget extends EventTarget{static[l]=!0;static defaultMaxListeners=10;constructor(){super(),initNodeEventTarget(this)}setMaxListeners(e){if(!j(this))throw new a("NodeEventTarget");t.setMaxListeners(e,this)}getMaxListeners(){if(!j(this))throw new a("NodeEventTarget");return this[c]}eventNames(){if(!j(this))throw new a("NodeEventTarget");return Array.from(this[kEvents].keys())}listenerCount(e){if(!j(this))throw new a("NodeEventTarget");const t=this[kEvents].get(String(e));return void 0!==t?t.size:0}off(e,t,n){if(!j(this))throw new a("NodeEventTarget");return this.removeEventListener(e,t,n),this}removeListener(e,t,n){if(!j(this))throw new a("NodeEventTarget");return this.removeEventListener(e,t,n),this}on(e,t){if(!j(this))throw new a("NodeEventTarget");return this.addEventListener(e,t,{[g]:!0}),this}addListener(e,t){if(!j(this))throw new a("NodeEventTarget");return this.addEventListener(e,t,{[g]:!0}),this}emit(e,t){if(!j(this))throw new a("NodeEventTarget");const n=this.listenerCount(e)>0;return this[E](t,e),n}once(e,t){if(!j(this))throw new a("NodeEventTarget");return this.addEventListener(e,t,{once:!0,[g]:!0}),this}removeAllListeners(e){if(!j(this))throw new a("NodeEventTarget");return void 0!==e?this[kEvents].delete(String(e)):this[kEvents].clear(),this}}function z(e){if("function"==typeof e||"function"==typeof e?.handleEvent)return!0;if(null==e)return!1;throw new r("listener","EventListener",e)}Object.defineProperties(NodeEventTarget.prototype,{setMaxListeners:P,getMaxListeners:P,eventNames:P,listenerCount:P,off:P,removeListener:P,on:P,addListener:P,once:P,emit:P,removeAllListeners:P});export function isEventTarget(e){return e?.constructor?.[h]}function j(e){return e?.constructor?.[l]}function _(e){const t=e.then;"function"==typeof t&&t.call(e,void 0,(function(e){B(e)}))}function B(e){n.nextTick((()=>{throw e}))}export function defineEventHandler(e,t){Object.defineProperty(e,`on${t}`,{get(){return this[p]?.get(t)?.handler},set(e){this[p]||(this[p]=new Map);let n=this[p]?.get(t);if(n){if("function"==typeof n.handler){this[kEvents].get(t).size--;const e=this[kEvents].get(t).size;this[kRemoveListener](e,t,n.handler,!1)}if(n.handler=e,"function"==typeof n.handler){this[kEvents].get(t).size++;const n=this[kEvents].get(t).size;this[kNewListener](n,t,e,!1,!1,!1)}}else n=function(e){function t(...e){if("function"==typeof t.handler)return Reflect.apply(t.handler,this,e)}return t.handler=e,t}(e),this.addEventListener(t,n);this[p].set(t,n)},configurable:!0,enumerable:!0})}export const EventEmitterMixin=e=>{class n extends e{constructor(...e){super(...e),t.call(this)}}const i=Object.getOwnPropertyDescriptors(t.prototype);return delete i.constructor,Object.defineProperties(n.prototype,i),n};
✄
{"version":3,"file":"from.js","names":["errorCodes","Buffer","process","ERR_INVALID_ARG_TYPE","ERR_STREAM_NULL_VALUES","from","Readable","iterable","opts","iterator","isAsync","objectMode","read","this","push","Symbol","asyncIterator","readable","highWaterMark","reading","_read","async","value","done","next","res","then","err","destroy","_destroy","error","cb","hadError","hasThrow","throw","return","close","nextTick","e"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"gBAAkBA,MAAkB,gCAE3BC,MAAc,gBAChBC,MAAa,UAEpB,MAAMC,qBACJA,EAAoBC,uBACpBA,GACEJ,iBAEW,SAASK,EAAKC,EAAUC,EAAUC,GAC/C,IAAIC,EAYAC,EAXJ,GAAwB,iBAAbH,GAAyBA,aAAoBN,EACtD,OAAO,IAAIK,EAAS,CAClBK,YAAY,KACTH,EACHI,OACEC,KAAKC,KAAKP,GACVM,KAAKC,KAAK,KACZ,IAKJ,GAAIP,GAAYA,EAASQ,OAAOC,eAC9BN,GAAU,EACVD,EAAWF,EAASQ,OAAOC,qBACtB,KAAIT,IAAYA,EAASQ,OAAON,UAIrC,MAAM,IAAIN,EAAqB,WAAY,CAAC,YAAaI,GAHzDG,GAAU,EACVD,EAAWF,EAASQ,OAAON,WAG7B,CAEA,MAAMQ,EAAW,IAAIX,EAAS,CAC5BK,YAAY,EACZO,cAAe,KAEZV,IAKL,IAAIW,GAAU,EA6Dd,OA3DAF,EAASG,MAAQ,WACVD,IACHA,GAAU,EA4BdE,iBACE,OAAS,CACP,IACE,MAAMC,MAAEA,EAAKC,KAAEA,GAASb,QAChBD,EAASe,OACff,EAASe,OAEX,GAAID,EACFN,EAASH,KAAK,UACT,CACL,MAAMW,EAAOH,GACW,mBAAfA,EAAMI,WACPJ,EACNA,EACF,GAAY,OAARG,EAEF,MADAN,GAAU,EACJ,IAAIf,EACL,GAAIa,EAASH,KAAKW,GACvB,SAEAN,GAAU,CAEd,CAGF,CAFE,MAAOQ,GACPV,EAASW,QAAQD,EACnB,CACA,KACF,CACF,CAvDIH,GAEJ,EAEAP,EAASY,SAAW,SAASC,EAAOC,IAOpCV,eAAqBS,GACnB,MAAME,EAAW,MAACF,EACZG,EAAqC,mBAAnBxB,EAASyB,MACjC,GAAIF,GAAYC,EAAU,CACxB,MAAMX,MAAEA,EAAKC,KAAEA,SAAed,EAASyB,MAAMJ,GAE7C,SADMR,EACFC,EACF,MAEJ,CACA,GAA+B,mBAApBd,EAAS0B,OAAuB,CACzC,MAAMb,MAAEA,SAAgBb,EAAS0B,eAC3Bb,CACR,CACF,EApBEc,CAAMN,GAAOJ,MACX,IAAMxB,EAAQmC,SAASN,EAAID,KAC1BQ,GAAMpC,EAAQmC,SAASN,EAAIO,GAAKR,IAErC,EA+COb,CACT"}
✄
import{codes as t}from"../errors.js";import{Buffer as e}from"buffer";import n from"process";const{ERR_INVALID_ARG_TYPE:o,ERR_STREAM_NULL_VALUES:r}=t;export default function i(t,i,a){let c,f;if("string"==typeof i||i instanceof e)return new t({objectMode:!0,...a,read(){this.push(i),this.push(null)}});if(i&&i[Symbol.asyncIterator])f=!0,c=i[Symbol.asyncIterator]();else{if(!i||!i[Symbol.iterator])throw new o("iterable",["Iterable"],i);f=!1,c=i[Symbol.iterator]()}const s=new t({objectMode:!0,highWaterMark:1,...a});let u=!1;return s._read=function(){u||(u=!0,async function(){for(;;){try{const{value:t,done:e}=f?await c.next():c.next();if(e)s.push(null);else{const e=t&&"function"==typeof t.then?await t:t;if(null===e)throw u=!1,new r;if(s.push(e))continue;u=!1}}catch(t){s.destroy(t)}break}}())},s._destroy=function(t,e){(async function(t){const e=null!=t,n="function"==typeof c.throw;if(e&&n){const{value:e,done:n}=await c.throw(t);if(await e,n)return}if("function"==typeof c.return){const{value:t}=await c.return();await t}})(t).then((()=>n.nextTick(e,t)),(o=>n.nextTick(e,o||t)))},s}
✄
{"version":3,"file":"legacy.js","names":["EE","Stream","opts","call","this","Object","setPrototypeOf","prototype","pipe","dest","options","source","ondata","chunk","writable","write","pause","ondrain","readable","resume","on","_isStdio","end","onend","onclose","didOnEnd","destroy","onerror","er","cleanup","listenerCount","emit","removeListener","prependListener","emitter","event","fn","_events","Array","isArray","unshift"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OAAOA,MAAQ,gBAER,SAASC,OAAOC,GACrBF,EAAGG,KAAKC,KAAMF,EAChB,CACAG,OAAOC,eAAeL,OAAOM,UAAWP,EAAGO,WAC3CF,OAAOC,eAAeL,OAAQD,GAE9BC,OAAOM,UAAUC,KAAO,SAASC,EAAMC,GACrC,MAAMC,EAASP,KAEf,SAASQ,EAAOC,GACVJ,EAAKK,WAAkC,IAAtBL,EAAKM,MAAMF,IAAoBF,EAAOK,OACzDL,EAAOK,OAEX,CAIA,SAASC,IACHN,EAAOO,UAAYP,EAAOQ,QAC5BR,EAAOQ,QAEX,CANAR,EAAOS,GAAG,OAAQR,GAQlBH,EAAKW,GAAG,QAASH,GAIZR,EAAKY,UAAcX,IAA2B,IAAhBA,EAAQY,MACzCX,EAAOS,GAAG,MAAOG,GACjBZ,EAAOS,GAAG,QAASI,IAGrB,IAAIC,GAAW,EACf,SAASF,IACHE,IACJA,GAAW,EAEXhB,EAAKa,MACP,CAGA,SAASE,IACHC,IACJA,GAAW,EAEiB,mBAAjBhB,EAAKiB,SAAwBjB,EAAKiB,UAC/C,CAGA,SAASC,EAAQC,GACfC,IACwC,IAApC7B,EAAG8B,cAAc1B,KAAM,UACzBA,KAAK2B,KAAK,QAASH,EAEvB,CAMA,SAASC,IACPlB,EAAOqB,eAAe,OAAQpB,GAC9BH,EAAKuB,eAAe,QAASf,GAE7BN,EAAOqB,eAAe,MAAOT,GAC7BZ,EAAOqB,eAAe,QAASR,GAE/Bb,EAAOqB,eAAe,QAASL,GAC/BlB,EAAKuB,eAAe,QAASL,GAE7BhB,EAAOqB,eAAe,MAAOH,GAC7BlB,EAAOqB,eAAe,QAASH,GAE/BpB,EAAKuB,eAAe,QAASH,EAC/B,CASA,OA3BAI,gBAAgBtB,EAAQ,QAASgB,GACjCM,gBAAgBxB,EAAM,QAASkB,GAmB/BhB,EAAOS,GAAG,MAAOS,GACjBlB,EAAOS,GAAG,QAASS,GAEnBpB,EAAKW,GAAG,QAASS,GACjBpB,EAAKsB,KAAK,OAAQpB,GAGXF,CACT,SAEO,SAASwB,gBAAgBC,EAASC,EAAOC,GAG9C,GAAuC,mBAA5BF,EAAQD,gBACjB,OAAOC,EAAQD,gBAAgBE,EAAOC,GAMnCF,EAAQG,SAAYH,EAAQG,QAAQF,GAEhCG,MAAMC,QAAQL,EAAQG,QAAQF,IACrCD,EAAQG,QAAQF,GAAOK,QAAQJ,GAE/BF,EAAQG,QAAQF,GAAS,CAACC,EAAIF,EAAQG,QAAQF,IAJ9CD,EAAQd,GAAGe,EAAOC,EAKtB"}
✄
import e from"events";export function Stream(r){e.call(this,r)}Object.setPrototypeOf(Stream.prototype,e.prototype),Object.setPrototypeOf(Stream,e),Stream.prototype.pipe=function(r,t){const n=this;function o(e){r.writable&&!1===r.write(e)&&n.pause&&n.pause()}function i(){n.readable&&n.resume&&n.resume()}n.on("data",o),r.on("drain",i),r._isStdio||t&&!1===t.end||(n.on("end",p),n.on("close",c));let s=!1;function p(){s||(s=!0,r.end())}function c(){s||(s=!0,"function"==typeof r.destroy&&r.destroy())}function d(r){m(),0===e.listenerCount(this,"error")&&this.emit("error",r)}function m(){n.removeListener("data",o),r.removeListener("drain",i),n.removeListener("end",p),n.removeListener("close",c),n.removeListener("error",d),r.removeListener("error",d),n.removeListener("end",m),n.removeListener("close",m),r.removeListener("close",m)}return prependListener(n,"error",d),prependListener(r,"error",d),n.on("end",m),n.on("close",m),r.on("close",m),r.emit("pipe",n),r};export function prependListener(e,r,t){if("function"==typeof e.prependListener)return e.prependListener(r,t);e._events&&e._events[r]?Array.isArray(e._events[r])?e._events[r].unshift(t):e._events[r]=[t,e._events[r]]:e.on(r,t)}
✄
{"version":3,"file":"once.js","names":["once","callback","called","args","Reflect","apply","this"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"eAqBe,SAASA,EAAKC,GAC3B,IAAIC,GAAS,EACb,OAAO,YAAYC,GACbD,IACJA,GAAS,EACTE,QAAQC,MAAMJ,EAAUK,KAAMH,GAChC,CACF"}
✄
export default function t(t){let e=!1;return function(...n){e||(e=!0,Reflect.apply(t,this,n))}}
✄
{"version":3,"file":"passthrough.js","names":["Transform","Object","setPrototypeOf","PassThrough","prototype","options","this","call","_transform","chunk","encoding","cb"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OAyBOA,MAAe,iBACtBC,OAAOC,eAAeC,EAAYC,UAAWJ,EAAUI,WACvDH,OAAOC,eAAeC,EAAaH,kBAEpB,SAASG,EAAYE,GAClC,KAAMC,gBAAgBH,GACpB,OAAO,IAAIA,EAAYE,GAEzBL,EAAUO,KAAKD,KAAMD,EACvB,CAEAF,EAAYC,UAAUI,WAAa,SAASC,EAAOC,EAAUC,GAC3DA,EAAG,KAAMF,EACX"}
✄
import t from"./transform.js";Object.setPrototypeOf(o.prototype,t.prototype),Object.setPrototypeOf(o,t);export default function o(e){if(!(this instanceof o))return new o(e);t.call(this,e)}o.prototype._transform=function(t,o,e){e(null,t)};
✄
{"version":3,"file":"pipeline.js","names":["AbortController","destroyImpl","Duplex","eos","aggregateTwoErrors","errorCodes","AbortError","once","PassThrough","Readable","isIterable","isReadableNodeStream","isNodeStream","process","ERR_INVALID_ARG_TYPE","ERR_INVALID_RETURN_VALUE","ERR_MISSING_ARGS","ERR_STREAM_DESTROYED","destroyer","stream","reading","writing","callback","finished","on","readable","writable","err","rState","_readableState","code","ended","errored","errorEmitted","makeAsyncIterable","val","async","prototype","Symbol","asyncIterator","call","fromReadable","pump","iterable","finish","error","onresolve","resume","wait","Promise","resolve","reject","cleanup","writableNeedDrain","chunk","write","end","off","pipeline","streams","pop","popCallback","Array","isArray","length","pipelineImpl","opts","ac","signal","outerSignal","abort","finishImpl","value","addEventListener","destroys","ret","finishCount","final","shift","removeEventListener","i","push","from","pt","objectMode","then","destroy","pipe","stdout","stderr","aborted","nextTick"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"0BAGSA,MAAuB,kCACpBC,MAAiB,sBACtBC,MAAY,qBACZC,MAAS,kDAEdC,WACSC,gBACTC,MACK,sBACAC,MAAU,mBACVC,MAAiB,0BACjBC,MAAc,qCAEnBC,0BACAC,kBACAC,MACK,oBAEAC,MAAa,UAEpB,MAAMC,qBACJA,EAAoBC,yBACpBA,EAAwBC,iBACxBA,EAAgBC,qBAChBA,GACEZ,EAEJ,SAASa,EAAUC,EAAQC,EAASC,EAASC,GAC3CA,EAAWf,EAAKe,GAEhB,IAAIC,GAAW,EA+Bf,OA9BAJ,EAAOK,GAAG,SAAS,KACjBD,GAAW,CAAI,IAGjBpB,EAAIgB,EAAQ,CAAEM,SAAUL,EAASM,SAAUL,IAAYM,IACrDJ,GAAYI,EAEZ,MAAMC,EAAST,EAAOU,eAEpBF,GACa,+BAAbA,EAAIG,MACJV,GACCQ,GAAUA,EAAOG,QAAUH,EAAOI,UAAYJ,EAAOK,aAUtDd,EACGZ,KAAK,MAAOe,GACZf,KAAK,QAASe,GAEjBA,EAASK,EACX,IAGMA,IACFJ,IACJA,GAAW,EACXtB,EAAYiB,UAAUC,EAAQQ,GAC9BL,EAASK,GAAO,IAAIV,EAAqB,SAAQ,CAErD,CASA,SAASiB,EAAkBC,GACzB,GAAIzB,EAAWyB,GACb,OAAOA,EACF,GAAIxB,EAAqBwB,GAE9B,OAMJC,gBAA6BD,SACpB1B,EAAS4B,UAAUC,OAAOC,eAAeC,KAAKL,EACvD,CARWM,CAAaN,GAEtB,MAAM,IAAIrB,EACR,MAAO,CAAC,WAAY,WAAY,iBAAkBqB,EACtD,CAMAC,eAAeM,EAAKC,EAAUjB,EAAUkB,GACtC,IAAIC,EACAC,EAAY,KAEhB,MAAMC,EAAUpB,IAKd,GAJIA,IACFkB,EAAQlB,GAGNmB,EAAW,CACb,MAAMxB,EAAWwB,EACjBA,EAAY,KACZxB,GACF,GAGI0B,EAAO,IAAM,IAAIC,SAAQ,CAACC,EAASC,KACnCN,EACFM,EAAON,GAEPC,EAAY,KACND,EACFM,EAAON,GAEPK,GACF,CAEJ,IAGFxB,EAASF,GAAG,QAASuB,GACrB,MAAMK,EAAUjD,EAAIuB,EAAU,CAAED,UAAU,GAASsB,GAEnD,IACMrB,EAAS2B,yBACLL,IAGR,UAAW,MAAMM,KAASX,EACnBjB,EAAS6B,MAAMD,UACZN,IAIVtB,EAAS8B,YAEHR,IAENJ,GAMF,CALE,MAAOjB,GACPiB,EAAOC,IAAUlB,EAAMvB,EAAmByC,EAAOlB,GAAOA,EAC1D,CAAE,QACAyB,IACA1B,EAAS+B,IAAI,QAASV,EACxB,CACF,gBAEeW,gBAER,SAASA,YAAYC,GAC1B,MAAMrC,EAAWf,EAlFnB,SAAqBoD,GAInB,OAAOA,EAAQC,KACjB,CA6EwBC,CAAYF,IAOlC,OAJIG,MAAMC,QAAQJ,EAAQ,KAA0B,IAAnBA,EAAQK,SACvCL,EAAUA,EAAQ,IAGbM,aAAaN,EAASrC,EAC/B,QAEO,SAAS2C,aAAaN,EAASrC,EAAU4C,GAC9C,GAAIP,EAAQK,OAAS,EACnB,MAAM,IAAIhD,EAAiB,WAG7B,MAAMmD,EAAK,IAAInE,EACToE,EAASD,EAAGC,OACZC,EAAcH,GAAME,OAE1B,SAASE,IACPC,EAAW,IAAIjE,EACjB,CAIA,IAAIuC,EACA2B,EAHJH,GAAaI,iBAAiB,QAASH,GAIvC,MAAMI,EAAW,GAEjB,IA2BIC,EA3BAC,EAAc,EAElB,SAAShC,EAAOjB,GACd4C,EAAW5C,EAAuB,KAAhBiD,EACpB,CAEA,SAASL,EAAW5C,EAAKkD,GAKvB,IAJIlD,GAASkB,GAAwB,+BAAfA,EAAMf,OAC1Be,EAAQlB,GAGLkB,GAAUgC,EAAf,CAIA,KAAOH,EAASV,QACdU,EAASI,OAATJ,CAAiB7B,GAGnBwB,GAAaU,oBAAoB,QAAST,GAC1CH,EAAGG,QAECO,GACFvD,EAASuB,EAAO2B,EAVlB,CAYF,CAGA,IAAK,IAAIQ,EAAI,EAAGA,EAAIrB,EAAQK,OAAQgB,IAAK,CACvC,MAAM7D,EAASwC,EAAQqB,GACjB5D,EAAU4D,EAAIrB,EAAQK,OAAS,EAC/B3C,EAAU2D,EAAI,EAOpB,GALIpE,EAAaO,KACfyD,IACAF,EAASO,KAAK/D,EAAUC,EAAQC,EAASC,EAASuB,KAG1C,IAANoC,EACF,GAAsB,mBAAX7D,GAET,GADAwD,EAAMxD,EAAO,CAAEiD,YACV1D,EAAWiE,GACd,MAAM,IAAI5D,EACR,oCAAqC,SAAU4D,QAGnDA,EADSjE,EAAWS,IAAWR,EAAqBQ,GAC9CA,EAEAjB,EAAOgF,KAAK/D,QAEf,GAAsB,mBAAXA,EAIhB,GAHAwD,EAAMzC,EAAkByC,GACxBA,EAAMxD,EAAOwD,EAAK,CAAEP,WAEhBhD,GACF,IAAKV,EAAWiE,GAAK,GACnB,MAAM,IAAI5D,EACR,gBAAiB,aAAaiE,EAAI,KAAML,OAEvC,CASL,MAAMQ,EAAK,IAAI3E,EAAY,CACzB4E,YAAY,IAKRC,EAAOV,GAAKU,KAClB,GAAoB,mBAATA,EACTA,EAAK7C,KAAKmC,GACCxC,IACCqC,EAAQrC,EACRgD,EAAG3B,IAAIrB,EAAI,IACTR,IACFwD,EAAGG,QAAQ3D,EAAI,QAGtB,KAAIjB,EAAWiE,GAAK,GAIzB,MAAM,IAAI5D,EACR,2BAA4B,cAAe4D,GAJ7CC,IACAlC,EAAKiC,EAAKQ,EAAIvC,EAIhB,CAEA+B,EAAMQ,EAENP,IACAF,EAASO,KAAK/D,EAAUyD,GAAK,GAAO,EAAM/B,GAC5C,MACShC,EAAaO,IAClBR,EAAqBgE,IACvBA,EAAIY,KAAKpE,GAKLA,IAAWN,EAAQ2E,QAAUrE,IAAWN,EAAQ4E,QAClDd,EAAInD,GAAG,OAAO,IAAML,EAAOqC,UAG7BmB,EAAMzC,EAAkByC,GAExBC,IACAlC,EAAKiC,EAAKxD,EAAQyB,IAEpB+B,EAAMxD,GAENwD,EAAMzE,EAAOgF,KAAK/D,EAEtB,CAMA,OAJIiD,GAAQsB,SAAWrB,GAAaqB,UAClC7E,EAAQ8E,SAASrB,GAGZK,CACT"}
✄
import{AbortController as e}from"./abort_controller.js";import*as t from"./destroy.js";import r from"./duplex.js";import o from"./end-of-stream.js";import{aggregateTwoErrors as n,codes as i,AbortError as s}from"../errors.js";import a from"./once.js";import l from"./passthrough.js";import f from"./readable.js";import{isIterable as c,isReadableNodeStream as p,isNodeStream as d}from"./utils.js";import m from"process";const{ERR_INVALID_ARG_TYPE:u,ERR_INVALID_RETURN_VALUE:R,ERR_MISSING_ARGS:E,ERR_STREAM_DESTROYED:w}=i;function b(e,r,n,i){i=a(i);let s=!1;return e.on("close",(()=>{s=!0})),o(e,{readable:r,writable:n},(t=>{s=!t;const o=e._readableState;t&&"ERR_STREAM_PREMATURE_CLOSE"===t.code&&r&&o&&o.ended&&!o.errored&&!o.errorEmitted?e.once("end",i).once("error",i):i(t)})),r=>{s||(s=!0,t.destroyer(e,r),i(r||new w("pipe")))}}function y(e){if(c(e))return e;if(p(e))return async function*(e){yield*f.prototype[Symbol.asyncIterator].call(e)}(e);throw new u("val",["Readable","Iterable","AsyncIterable"],e)}async function _(e,t,r){let i,s=null;const a=e=>{if(e&&(i=e),s){const e=s;s=null,e()}},l=()=>new Promise(((e,t)=>{i?t(i):s=()=>{i?t(i):e()}}));t.on("drain",a);const f=o(t,{readable:!1},a);try{t.writableNeedDrain&&await l();for await(const r of e)t.write(r)||await l();t.end(),await l(),r()}catch(e){r(i!==e?n(i,e):e)}finally{f(),t.off("drain",a)}}export default pipeline;export function pipeline(...e){const t=a(function(e){return e.pop()}(e));return Array.isArray(e[0])&&1===e.length&&(e=e[0]),pipelineImpl(e,t)}export function pipelineImpl(t,o,n){if(t.length<2)throw new E("streams");const i=new e,a=i.signal,f=n?.signal;function u(){j(new s)}let w,h;f?.addEventListener("abort",u);const A=[];let I,S=0;function g(e){j(e,0==--S)}function j(e,t){if(!e||w&&"ERR_STREAM_PREMATURE_CLOSE"!==w.code||(w=e),w||t){for(;A.length;)A.shift()(w);f?.removeEventListener("abort",u),i.abort(),t&&o(w,h)}}for(let e=0;e<t.length;e++){const o=t[e],n=e<t.length-1,i=e>0;if(d(o)&&(S++,A.push(b(o,n,i,g))),0===e)if("function"==typeof o){if(I=o({signal:a}),!c(I))throw new R("Iterable, AsyncIterable or Stream","source",I)}else I=c(o)||p(o)?o:r.from(o);else if("function"==typeof o)if(I=y(I),I=o(I,{signal:a}),n){if(!c(I,!0))throw new R("AsyncIterable",`transform[${e-1}]`,I)}else{const e=new l({objectMode:!0}),t=I?.then;if("function"==typeof t)t.call(I,(t=>{h=t,e.end(t)}),(t=>{e.destroy(t)}));else{if(!c(I,!0))throw new R("AsyncIterable or Promise","destination",I);S++,_(I,e,g)}I=e,S++,A.push(b(I,!1,!0,g))}else d(o)?(p(I)?(I.pipe(o),o!==m.stdout&&o!==m.stderr||I.on("end",(()=>o.end()))):(I=y(I),S++,_(I,o,g)),I=o):I=r.from(o)}return(a?.aborted||f?.aborted)&&m.nextTick(u),I}
✄
{"version":3,"file":"promises.js","names":["eos","pl","isIterable","isNodeStream","pipeline","streams","Promise","resolve","reject","signal","lastArg","length","pop","err","value","finished","stream","opts"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OAAOA,MAAS,4CACSC,MAAU,qCAEjCC,kBACAC,MACK,oBAEA,SAASC,YAAYC,GAC1B,OAAO,IAAIC,SAAQ,CAACC,EAASC,KAC3B,IAAIC,EACJ,MAAMC,EAAUL,EAAQA,EAAQM,OAAS,GACzC,GAAID,GAA8B,iBAAZA,IACjBP,EAAaO,KAAaR,EAAWQ,GAAU,CAElDD,EADgBJ,EAAQO,MACPH,MACnB,CAEAR,EAAGI,GAAS,CAACQ,EAAKC,KACZD,EACFL,EAAOK,GAEPN,EAAQO,EACV,GACC,CAAEL,UAAS,GAElB,QAEO,SAASM,SAASC,EAAQC,GAC/B,OAAO,IAAIX,SAAQ,CAACC,EAASC,KAC3BR,EAAIgB,EAAQC,GAAOJ,IACbA,EACFL,EAAOK,GAEPN,GACF,GACA,GAEN"}
✄
import e from"./end-of-stream.js";import{pipelineImpl as i}from"./pipeline.js";import{isIterable as n,isNodeStream as o}from"./utils.js";export function pipeline(...e){return new Promise(((t,p)=>{let r;const s=e[e.length-1];if(s&&"object"==typeof s&&!o(s)&&!n(s)){r=e.pop().signal}i(e,((e,i)=>{e?p(e):t(i)}),{signal:r})}))}export function finished(i,n){return new Promise(((o,t)=>{e(i,n,(e=>{e?t(e):o()}))}))}
✄
{"version":3,"file":"readable.js","names":["addAbortSignal","BufferList","destroyImpl","eos","aggregateTwoErrors","errorCodes","from","Stream","prependListener","getHighWaterMark","getDefaultHighWaterMark","Buffer","EE","process","StringDecoder","Readable","ERR_INVALID_ARG_TYPE","ERR_METHOD_NOT_IMPLEMENTED","ERR_OUT_OF_RANGE","ERR_STREAM_PUSH_AFTER_EOF","ERR_STREAM_UNSHIFT_AFTER_END_EVENT","kPaused","Symbol","Object","setPrototypeOf","prototype","nop","errorOrDestroy","ReadableState","options","stream","isDuplex","Duplex","this","objectMode","readableObjectMode","highWaterMark","buffer","length","pipes","flowing","ended","endEmitted","reading","constructed","sync","needReadable","emittedReadable","readableListening","resumeScheduled","errorEmitted","emitClose","autoDestroy","destroyed","errored","closed","closeEmitted","defaultEncoding","awaitDrainWriters","multiAwaitDrain","readingMore","dataEmitted","decoder","encoding","_readableState","read","_read","destroy","_destroy","construct","_construct","signal","call","maybeReadMore","readableAddChunk","chunk","addToFront","state","err","toString","_isUint8Array","_uint8ArrayToBuffer","end","push","emitReadable","emitReadable_","onEofChunk","addChunk","write","listenerCount","clear","emit","unshift","_undestroy","undestroy","cb","captureRejectionSymbol","isPaused","setEncoding","enc","content","data","howMuchToRead","n","Number","isNaN","first","nextTick","flow","maybeReadMore_","len","updateReadableListening","self","resume","nReadingNextTick","resume_","streamToAsyncIterator","wrap","iter","async","error","callback","next","resolve","on","writable","Promise","destroyOnReturn","undefined","destroyer","createAsyncIterator","fromList","ret","shift","join","concat","consume","endReadable","endReadableNT","allowHalfOpen","endWritableNT","wState","_writableState","finished","writableEnded","NaN","isInteger","parseInt","nOrig","computeNewHighWaterMark","doRead","result","then","pipe","dest","pipeOpts","src","Set","endFn","stdout","stderr","onend","unpipe","onunpipe","readable","unpipeInfo","hasUnpiped","removeListener","onclose","onfinish","ondrain","onerror","ondata","cleanedUp","needDrain","cleanup","once","pause","includes","add","delete","size","pipeOnDrain","er","s","writableNeedDrain","dests","i","index","indexOf","splice","ev","fn","res","addListener","off","removeAllListeners","apply","arguments","paused","streamKeys","keys","j","bind","asyncIterator","iterator","defineProperties","get","r","set","val","readableDidRead","enumerable","readableAborted","readableHighWaterMark","readableBuffer","readableFlowing","readableLength","readableEncoding","value","readableEnded","pipesCount","_fromList","iterable","opts"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"yBAqBSA,MAAsB,+BACxBC,MAAgB,6BACXC,MAAiB,sBACtBC,MAAS,kDAEdC,WACSC,MACJ,sBACAC,MAAU,6BACRC,qBAAQC,MAAuB,yCAEtCC,6BACAC,MACK,8BAEEC,MAAc,gBAChBC,MAAQ,gBACRC,MAAa,kCACXC,MAAqB,gCAEfC,SAEf,MAAMC,qBACJA,EAAoBC,2BACpBA,EAA0BC,iBAC1BA,EAAgBC,0BAChBA,EAAyBC,mCACzBA,GACEf,EAEEgB,EAAUC,OAAO,WAEvBC,OAAOC,eAAeT,SAASU,UAAWlB,EAAOkB,WACjDF,OAAOC,eAAeT,SAAUR,GAChC,MAAMmB,EAAM,QAENC,eAAEA,GAAmBzB,SAEpB,SAAS0B,cAAcC,EAASC,EAAQC,GAMrB,kBAAbA,IACTA,EAAWD,aAAkBvB,EAAOyB,QAItCC,KAAKC,cAAgBL,IAAWA,EAAQK,YAEpCH,IACFE,KAAKC,WAAaD,KAAKC,eAClBL,IAAWA,EAAQM,qBAI1BF,KAAKG,cAAgBP,EACnBpB,EAAiBwB,KAAMJ,EAAS,wBAAyBE,GACzDrB,GAAwB,GAK1BuB,KAAKI,OAAS,IAAIpC,EAClBgC,KAAKK,OAAS,EACdL,KAAKM,MAAQ,GACbN,KAAKO,QAAU,KACfP,KAAKQ,OAAQ,EACbR,KAAKS,YAAa,EAClBT,KAAKU,SAAU,EAMfV,KAAKW,aAAc,EAMnBX,KAAKY,MAAO,EAIZZ,KAAKa,cAAe,EACpBb,KAAKc,iBAAkB,EACvBd,KAAKe,mBAAoB,EACzBf,KAAKgB,iBAAkB,EACvBhB,KAAKZ,GAAW,KAGhBY,KAAKiB,cAAe,EAGpBjB,KAAKkB,WAAatB,IAAiC,IAAtBA,EAAQsB,UAGrClB,KAAKmB,aAAevB,IAAmC,IAAxBA,EAAQuB,YAGvCnB,KAAKoB,WAAY,EAMjBpB,KAAKqB,QAAU,KAGfrB,KAAKsB,QAAS,EAIdtB,KAAKuB,cAAe,EAKpBvB,KAAKwB,gBAAmB5B,GAAWA,EAAQ4B,iBAAoB,OAI/DxB,KAAKyB,kBAAoB,KACzBzB,KAAK0B,iBAAkB,EAGvB1B,KAAK2B,aAAc,EAEnB3B,KAAK4B,aAAc,EAEnB5B,KAAK6B,QAAU,KACf7B,KAAK8B,SAAW,KACZlC,GAAWA,EAAQkC,WACrB9B,KAAK6B,QAAU,IAAIhD,EAAce,EAAQkC,UACzC9B,KAAK8B,SAAWlC,EAAQkC,SAE5B,QAGO,SAAShD,SAASc,GACvB,KAAMI,gBAAgBlB,UACpB,OAAO,IAAIA,SAASc,GAItB,MAAME,EAAWE,gBAAgB1B,EAAOyB,OAExCC,KAAK+B,eAAiB,IAAIpC,cAAcC,EAASI,KAAMF,GAEnDF,IAC0B,mBAAjBA,EAAQoC,OACjBhC,KAAKiC,MAAQrC,EAAQoC,MAEQ,mBAApBpC,EAAQsC,UACjBlC,KAAKmC,SAAWvC,EAAQsC,SAEO,mBAAtBtC,EAAQwC,YACjBpC,KAAKqC,WAAazC,EAAQwC,WAExBxC,EAAQ0C,SAAWxC,GACrB/B,EAAe6B,EAAQ0C,OAAQtC,OAGnC1B,EAAOiE,KAAKvC,KAAMJ,GAElB3B,EAAYmE,UAAUpC,MAAM,KACtBA,KAAK+B,eAAelB,cACtB2B,EAAcxC,KAAMA,KAAK+B,eAC3B,GAEJ,CAyBA,SAASU,EAAiB5C,EAAQ6C,EAAOZ,EAAUa,GACjD,MAAMC,EAAQ/C,EAAOkC,eAErB,IAAIc,EAyBJ,GAxBKD,EAAM3C,aACY,iBAAVyC,GACTZ,EAAWA,GAAYc,EAAMpB,gBACzBoB,EAAMd,WAAaA,IACjBa,GAAcC,EAAMd,SAGtBY,EAAQhE,EAAOL,KAAKqE,EAAOZ,GAAUgB,SAASF,EAAMd,WAEpDY,EAAQhE,EAAOL,KAAKqE,EAAOZ,GAC3BA,EAAW,MAGNY,aAAiBhE,EAC1BoD,EAAW,GACFxD,EAAOyE,cAAcL,IAC9BA,EAAQpE,EAAO0E,oBAAoBN,GACnCZ,EAAW,IACO,MAATY,IACTG,EAAM,IAAI9D,EACR,QAAS,CAAC,SAAU,SAAU,cAAe2D,KAI/CG,EACFnD,EAAeG,EAAQgD,QAClB,GAAc,OAAVH,EACTE,EAAMlC,SAAU,EAoRpB,SAAoBb,EAAQ+C,GAC1B,GAAIA,EAAMpC,MAAO,OACjB,GAAIoC,EAAMf,QAAS,CACjB,MAAMa,EAAQE,EAAMf,QAAQoB,MACxBP,GAASA,EAAMrC,SACjBuC,EAAMxC,OAAO8C,KAAKR,GAClBE,EAAMvC,QAAUuC,EAAM3C,WAAa,EAAIyC,EAAMrC,OAEjD,CACAuC,EAAMpC,OAAQ,EAEVoC,EAAMhC,KAIRuC,EAAatD,IAGb+C,EAAM/B,cAAe,EACrB+B,EAAM9B,iBAAkB,EAGxBsC,EAAcvD,GAElB,CA3SIwD,CAAWxD,EAAQ+C,QACd,GAAIA,EAAM3C,YAAeyC,GAASA,EAAMrC,OAAS,EACtD,GAAIsC,EACF,GAAIC,EAAMnC,WACRf,EAAeG,EAAQ,IAAIV,OACxB,IAAIyD,EAAMxB,WAAawB,EAAMvB,QAChC,OAAO,EAEPiC,EAASzD,EAAQ+C,EAAOF,GAAO,EAAK,MACjC,GAAIE,EAAMpC,MACfd,EAAeG,EAAQ,IAAIX,OACtB,IAAI0D,EAAMxB,WAAawB,EAAMvB,QAClC,OAAO,EAEPuB,EAAMlC,SAAU,EACZkC,EAAMf,UAAYC,GACpBY,EAAQE,EAAMf,QAAQ0B,MAAMb,GACxBE,EAAM3C,YAA+B,IAAjByC,EAAMrC,OAC5BiD,EAASzD,EAAQ+C,EAAOF,GAAO,GAE/BF,EAAc3C,EAAQ+C,IAExBU,EAASzD,EAAQ+C,EAAOF,GAAO,EAEnC,MACUC,IACVC,EAAMlC,SAAU,EAChB8B,EAAc3C,EAAQ+C,IAMxB,OAAQA,EAAMpC,QACXoC,EAAMvC,OAASuC,EAAMzC,eAAkC,IAAjByC,EAAMvC,OACjD,CAEA,SAASiD,EAASzD,EAAQ+C,EAAOF,EAAOC,GAClCC,EAAMrC,SAA4B,IAAjBqC,EAAMvC,SAAiBuC,EAAMhC,MAC9Cf,EAAO2D,cAAc,QAAU,GAG7BZ,EAAMlB,gBACRkB,EAAMnB,kBAAkBgC,QAExBb,EAAMnB,kBAAoB,KAG5BmB,EAAMhB,aAAc,EACpB/B,EAAO6D,KAAK,OAAQhB,KAGpBE,EAAMvC,QAAUuC,EAAM3C,WAAa,EAAIyC,EAAMrC,OACzCsC,EACFC,EAAMxC,OAAOuD,QAAQjB,GAErBE,EAAMxC,OAAO8C,KAAKR,GAEhBE,EAAM/B,cACRsC,EAAatD,IAEjB2C,EAAc3C,EAAQ+C,EACxB,CArHA9D,SAASU,UAAU0C,QAAUjE,EAAYiE,QACzCpD,SAASU,UAAUoE,WAAa3F,EAAY4F,UAC5C/E,SAASU,UAAU2C,SAAW,SAASU,EAAKiB,GAC1CA,EAAGjB,EACL,EAEA/D,SAASU,UAAUb,EAAGoF,wBAA0B,SAASlB,GACvD7C,KAAKkC,QAAQW,EACf,EAMA/D,SAASU,UAAU0D,KAAO,SAASR,EAAOZ,GACxC,OAAOW,EAAiBzC,KAAM0C,EAAOZ,GAAU,EACjD,EAGAhD,SAASU,UAAUmE,QAAU,SAASjB,EAAOZ,GAC3C,OAAOW,EAAiBzC,KAAM0C,EAAOZ,GAAU,EACjD,EAkGAhD,SAASU,UAAUwE,SAAW,WAC5B,MAAMpB,EAAQ5C,KAAK+B,eACnB,OAA0B,IAAnBa,EAAMxD,KAAuC,IAAlBwD,EAAMrC,OAC1C,EAGAzB,SAASU,UAAUyE,YAAc,SAASC,GACxC,MAAMrC,EAAU,IAAIhD,EAAcqF,GAClClE,KAAK+B,eAAeF,QAAUA,EAE9B7B,KAAK+B,eAAeD,SAAW9B,KAAK+B,eAAeF,QAAQC,SAE3D,MAAM1B,EAASJ,KAAK+B,eAAe3B,OAEnC,IAAI+D,EAAU,GACd,IAAK,MAAMC,KAAQhE,EACjB+D,GAAWtC,EAAQ0B,MAAMa,GAM3B,OAJAhE,EAAOqD,QACS,KAAZU,GACF/D,EAAO8C,KAAKiB,GACdnE,KAAK+B,eAAe1B,OAAS8D,EAAQ9D,OAC9BL,IACT,EAuBA,SAASqE,EAAcC,EAAG1B,GACxB,OAAI0B,GAAK,GAAuB,IAAjB1B,EAAMvC,QAAgBuC,EAAMpC,MAClC,EACLoC,EAAM3C,WACD,EACLsE,OAAOC,MAAMF,GAEX1B,EAAMrC,SAAWqC,EAAMvC,OAClBuC,EAAMxC,OAAOqE,QAAQpE,OACvBuC,EAAMvC,OAEXiE,GAAK1B,EAAMvC,OACNiE,EACF1B,EAAMpC,MAAQoC,EAAMvC,OAAS,CACtC,CAoLA,SAAS8C,EAAatD,GACpB,MAAM+C,EAAQ/C,EAAOkC,eACrBa,EAAM/B,cAAe,EAChB+B,EAAM9B,kBACT8B,EAAM9B,iBAAkB,EACxBlC,EAAQ8F,SAAStB,EAAevD,GAEpC,CAEA,SAASuD,EAAcvD,GACrB,MAAM+C,EAAQ/C,EAAOkC,eAChBa,EAAMxB,WAAcwB,EAAMvB,UAAYuB,EAAMvC,SAAUuC,EAAMpC,QAC/DX,EAAO6D,KAAK,YACZd,EAAM9B,iBAAkB,GAS1B8B,EAAM/B,cACH+B,EAAMrC,UACNqC,EAAMpC,OACPoC,EAAMvC,QAAUuC,EAAMzC,cACxBwE,EAAK9E,EACP,CASA,SAAS2C,EAAc3C,EAAQ+C,IACxBA,EAAMjB,aAAeiB,EAAMjC,cAC9BiC,EAAMjB,aAAc,EACpB/C,EAAQ8F,SAASE,EAAgB/E,EAAQ+C,GAE7C,CAEA,SAASgC,EAAe/E,EAAQ+C,GAwB9B,MAAQA,EAAMlC,UAAYkC,EAAMpC,QACxBoC,EAAMvC,OAASuC,EAAMzC,eACpByC,EAAMrC,SAA4B,IAAjBqC,EAAMvC,SAAgB,CAC9C,MAAMwE,EAAMjC,EAAMvC,OAElB,GADAR,EAAOmC,KAAK,GACR6C,IAAQjC,EAAMvC,OAEhB,KACJ,CACAuC,EAAMjB,aAAc,CACtB,CAyRA,SAASmD,EAAwBC,GAC/B,MAAMnC,EAAQmC,EAAKhD,eACnBa,EAAM7B,kBAAoBgE,EAAKvB,cAAc,YAAc,EAEvDZ,EAAM5B,kBAAsC,IAAnB4B,EAAMxD,GAGjCwD,EAAMrC,SAAU,EAGPwE,EAAKvB,cAAc,QAAU,EACtCuB,EAAKC,SACKpC,EAAM7B,oBAChB6B,EAAMrC,QAAU,KAEpB,CAEA,SAAS0E,EAAiBF,GACxBA,EAAK/C,KAAK,EACZ,CAwBA,SAASkD,EAAQrF,EAAQ+C,GAClBA,EAAMlC,SACTb,EAAOmC,KAAK,GAGdY,EAAM5B,iBAAkB,EACxBnB,EAAO6D,KAAK,UACZiB,EAAK9E,GACD+C,EAAMrC,UAAYqC,EAAMlC,SAC1Bb,EAAOmC,KAAK,EAChB,CAWA,SAAS2C,EAAK9E,GACZ,MAAM+C,EAAQ/C,EAAOkC,eACrB,KAAOa,EAAMrC,SAA6B,OAAlBV,EAAOmC,SACjC,CA8DA,SAASmD,EAAsBtF,EAAQD,GACV,mBAAhBC,EAAOmC,OAChBnC,EAASf,SAASsG,KAAKvF,EAAQ,CAAEI,YAAY,KAG/C,MAAMoF,EAKRC,gBAAoCzF,EAAQD,GAC1C,IAaI2F,EAbAC,EAAW/F,EAEf,SAASgG,EAAKC,GACR1F,OAASH,GACX2F,IACAA,EAAW/F,GAEX+F,EAAWE,CAEf,CAEA7F,EAAO8F,GAAG,WAAYF,GAGtBvH,EAAI2B,EAAQ,CAAE+F,UAAU,IAAU/C,IAChC0C,EAAQ1C,EAAM1E,EAAmBoH,EAAO1C,GAAO,KAC/C2C,IACAA,EAAW/F,CAAG,IAGhB,IACE,OAAa,CACX,MAAMiD,EAAQ7C,EAAOuB,UAAY,KAAOvB,EAAOmC,OAC/C,GAAc,OAAVU,QACIA,MACD,IAAI6C,EACT,MAAMA,EACD,GAAc,OAAVA,EACT,aAEM,IAAIM,QAAQJ,EACpB,CACF,CAWF,CAVE,MAAO5C,GAEP,MADA0C,EAAQpH,EAAmBoH,EAAO1C,GAC5B0C,CACR,CAAE,SAEGA,IAAsC,IAA7B3F,GAASkG,sBACRC,IAAVR,IAAuB1F,EAAOkC,eAAeZ,aAE9ClD,EAAY+H,UAAUnG,EAAQ,KAElC,CACF,CAlDeoG,CAAoBpG,EAAQD,GAEzC,OADAyF,EAAKxF,OAASA,EACPwF,CACT,CA6LA,SAASa,EAAS5B,EAAG1B,GAEnB,GAAqB,IAAjBA,EAAMvC,OACR,OAAO,KAET,IAAI8F,EAiBJ,OAhBIvD,EAAM3C,WACRkG,EAAMvD,EAAMxC,OAAOgG,SACX9B,GAAKA,GAAK1B,EAAMvC,QAGtB8F,EADEvD,EAAMf,QACFe,EAAMxC,OAAOiG,KAAK,IACO,IAAxBzD,EAAMxC,OAAOC,OACduC,EAAMxC,OAAOqE,QAEb7B,EAAMxC,OAAOkG,OAAO1D,EAAMvC,QAClCuC,EAAMxC,OAAOqD,SAGb0C,EAAMvD,EAAMxC,OAAOmG,QAAQjC,EAAG1B,EAAMf,SAG/BsE,CACT,CAEA,SAASK,EAAY3G,GACnB,MAAM+C,EAAQ/C,EAAOkC,eAEhBa,EAAMnC,aACTmC,EAAMpC,OAAQ,EACd5B,EAAQ8F,SAAS+B,EAAe7D,EAAO/C,GAE3C,CAEA,SAAS4G,EAAc7D,EAAO/C,GAE5B,IAAK+C,EAAMvB,UAAYuB,EAAMrB,eACxBqB,EAAMnC,YAA+B,IAAjBmC,EAAMvC,OAI7B,GAHAuC,EAAMnC,YAAa,EACnBZ,EAAO6D,KAAK,OAER7D,EAAO+F,WAAqC,IAAzB/F,EAAO6G,cAC5B9H,EAAQ8F,SAASiC,EAAe9G,QAC3B,GAAI+C,EAAMzB,YAAa,CAG5B,MAAMyF,EAAS/G,EAAOgH,iBACDD,GACnBA,EAAOzF,cAGNyF,EAAOE,WAAgC,IAApBF,EAAOhB,YAI3B/F,EAAOqC,SAEX,CAEJ,CAEA,SAASyE,EAAc9G,GACJA,EAAO+F,WAAa/F,EAAOkH,gBACzClH,EAAOuB,WAERvB,EAAOoD,KAEX,CAh6BAnE,SAASU,UAAUwC,KAAO,SAASsC,QAGvByB,IAANzB,EACFA,EAAI0C,IACMzC,OAAO0C,UAAU3C,KAC3BA,EAAIC,OAAO2C,SAAS5C,EAAG,KAEzB,MAAM1B,EAAQ5C,KAAK+B,eACboF,EAAQ7C,EAYd,GATIA,EAAI1B,EAAMzC,gBACZyC,EAAMzC,cAjDV,SAAiCmE,GAC/B,GAAIA,EAFU,WAGZ,MAAM,IAAIrF,EAAiB,OAAQ,UAAWqF,GAYhD,OAREA,IACAA,GAAKA,IAAM,EACXA,GAAKA,IAAM,EACXA,GAAKA,IAAM,EACXA,GAAKA,IAAM,EACXA,GAAKA,IAAM,KACXA,CAGJ,CAkC0B8C,CAAwB9C,IAEtC,IAANA,IACF1B,EAAM9B,iBAAkB,GAKhB,IAANwD,GACA1B,EAAM/B,gBACoB,IAAxB+B,EAAMzC,cACNyC,EAAMvC,QAAUuC,EAAMzC,cACtByC,EAAMvC,OAAS,IAChBuC,EAAMpC,OAKT,OAJqB,IAAjBoC,EAAMvC,QAAgBuC,EAAMpC,MAC9BgG,EAAYxG,MAEZmD,EAAanD,MACR,KAMT,GAAU,KAHVsE,EAAID,EAAcC,EAAG1B,KAGNA,EAAMpC,MAGnB,OAFqB,IAAjBoC,EAAMvC,QACRmG,EAAYxG,MACP,KA0BT,IA6CImG,EA7CAkB,EAASzE,EAAM/B,aAUnB,IAPqB,IAAjB+B,EAAMvC,QAAgBuC,EAAMvC,OAASiE,EAAI1B,EAAMzC,iBACjDkH,GAAS,GAMPzE,EAAMpC,OAASoC,EAAMlC,SAAWkC,EAAMxB,WAAawB,EAAMvB,UACxDuB,EAAMjC,YACT0G,GAAS,OACJ,GAAIA,EAAQ,CACjBzE,EAAMlC,SAAU,EAChBkC,EAAMhC,MAAO,EAEQ,IAAjBgC,EAAMvC,SACRuC,EAAM/B,cAAe,GAGvB,IACE,MAAMyG,EAAStH,KAAKiC,MAAMW,EAAMzC,eAChC,GAAc,MAAVmH,EAAgB,CAClB,MAAMC,EAAOD,EAAOC,KACA,mBAATA,GACTA,EAAKhF,KACH+E,EACA7H,GACA,SAASoD,GACPnD,EAAeM,KAAM6C,EACvB,GAEN,CAGF,CAFE,MAAOA,GACPnD,EAAeM,KAAM6C,EACvB,CAEAD,EAAMhC,MAAO,EAGRgC,EAAMlC,UACT4D,EAAID,EAAc8C,EAAOvE,GAC7B,CAoCA,OAhCEuD,EADE7B,EAAI,EACA4B,EAAS5B,EAAG1B,GAEZ,KAEI,OAARuD,GACFvD,EAAM/B,aAAe+B,EAAMvC,QAAUuC,EAAMzC,cAC3CmE,EAAI,IAEJ1B,EAAMvC,QAAUiE,EACZ1B,EAAMlB,gBACRkB,EAAMnB,kBAAkBgC,QAExBb,EAAMnB,kBAAoB,MAIT,IAAjBmB,EAAMvC,SAGHuC,EAAMpC,QACToC,EAAM/B,cAAe,GAGnBsG,IAAU7C,GAAK1B,EAAMpC,OACvBgG,EAAYxG,OAGJ,OAARmG,GAAiBvD,EAAM3B,cAAiB2B,EAAMrB,eAChDqB,EAAMhB,aAAc,EACpB5B,KAAK0D,KAAK,OAAQyC,IAGbA,CACT,EAkHArH,SAASU,UAAUyC,MAAQ,SAASqC,GAClC,MAAM,IAAItF,EAA2B,UACvC,EAEAF,SAASU,UAAUgI,KAAO,SAASC,EAAMC,GACvC,MAAMC,EAAM3H,KACN4C,EAAQ5C,KAAK+B,eAEQ,IAAvBa,EAAMtC,MAAMD,SACTuC,EAAMlB,kBACTkB,EAAMlB,iBAAkB,EACxBkB,EAAMnB,kBAAoB,IAAImG,IAC5BhF,EAAMnB,kBAAoB,CAACmB,EAAMnB,mBAAqB,MAK5DmB,EAAMtC,MAAM4C,KAAKuE,GAEjB,MAIMI,IAJUH,IAA6B,IAAjBA,EAASzE,MACzBwE,IAAS7I,EAAQkJ,QACjBL,IAAS7I,EAAQmJ,OAEPC,EAAQC,EAO9B,SAASC,EAASC,EAAUC,GACtBD,IAAaR,GACXS,IAAwC,IAA1BA,EAAWC,aAC3BD,EAAWC,YAAa,EAa9B,WAEEZ,EAAKa,eAAe,QAASC,GAC7Bd,EAAKa,eAAe,SAAUE,GAC1BC,GACFhB,EAAKa,eAAe,QAASG,GAE/BhB,EAAKa,eAAe,QAASI,GAC7BjB,EAAKa,eAAe,SAAUJ,GAC9BP,EAAIW,eAAe,MAAON,GAC1BL,EAAIW,eAAe,MAAOL,GAC1BN,EAAIW,eAAe,OAAQK,GAE3BC,GAAY,EAORH,GAAW7F,EAAMnB,qBACfgG,EAAKZ,gBAAkBY,EAAKZ,eAAegC,YAC/CJ,GACJ,CAnCMK,GAGN,CAEA,SAASd,IACPP,EAAKxE,KACP,CAEA,IAAIwF,EAnBA7F,EAAMnC,WACR7B,EAAQ8F,SAASmD,GAEjBF,EAAIoB,KAAK,MAAOlB,GAElBJ,EAAK9B,GAAG,SAAUuC,GAgBlB,IAAIU,GAAY,EA0BhB,SAASI,IAKFJ,IACwB,IAAvBhG,EAAMtC,MAAMD,QAAgBuC,EAAMtC,MAAM,KAAOmH,GACjD7E,EAAMnB,kBAAoBgG,EAC1B7E,EAAMlB,iBAAkB,GACfkB,EAAMtC,MAAMD,OAAS,GAAKuC,EAAMtC,MAAM2I,SAASxB,IACxD7E,EAAMnB,kBAAkByH,IAAIzB,GAE9BE,EAAIqB,SAEDP,IAKHA,EAgEN,SAAqBd,EAAKF,GACxB,OAAO,WACL,MAAM7E,EAAQ+E,EAAI5F,eAKda,EAAMnB,oBAAsBgG,EAC9B7E,EAAMnB,kBAAoB,KACjBmB,EAAMlB,iBACfkB,EAAMnB,kBAAkB0H,OAAO1B,GAG3B7E,EAAMnB,mBAAsD,IAAjCmB,EAAMnB,kBAAkB2H,OACvDzK,EAAG6E,cAAcmE,EAAK,UACtB/E,EAAMrC,SAAU,EAChBoE,EAAKgD,GAET,CACF,CAnFgB0B,CAAY1B,EAAKF,GAC3BA,EAAK9B,GAAG,QAAS8C,GAErB,CAGA,SAASE,EAAOjG,IAEF,IADA+E,EAAKlE,MAAMb,IAErBsG,GAEJ,CAIA,SAASN,EAAQY,GAGf,GAFArB,IACAR,EAAKa,eAAe,QAASI,GACW,IAApC/J,EAAG6E,cAAciE,EAAM,SAAgB,CACzC,MAAM8B,EAAI9B,EAAKZ,gBAAkBY,EAAK1F,eAClCwH,IAAMA,EAAEtI,aAEVvB,EAAe+H,EAAM6B,GAErB7B,EAAK/D,KAAK,QAAS4F,EAEvB,CACF,CAMA,SAASf,IACPd,EAAKa,eAAe,SAAUE,GAC9BP,GACF,CAEA,SAASO,IACPf,EAAKa,eAAe,QAASC,GAC7BN,GACF,CAGA,SAASA,IACPN,EAAIM,OAAOR,EACb,CAeA,OAxDAE,EAAIhC,GAAG,OAAQgD,GAyBfpK,EAAgBkJ,EAAM,QAASiB,GAO/BjB,EAAKsB,KAAK,QAASR,GAKnBd,EAAKsB,KAAK,SAAUP,GAOpBf,EAAK/D,KAAK,OAAQiE,IAIa,IAA3BF,EAAK+B,kBACH5G,EAAMrC,SACRyI,IAEQpG,EAAMrC,SAChBoH,EAAI3C,SAGCyC,CACT,EAwBA3I,SAASU,UAAUyI,OAAS,SAASR,GACnC,MAAM7E,EAAQ5C,KAAK+B,eAInB,GAA2B,IAAvBa,EAAMtC,MAAMD,OACd,OAAOL,KAET,IAAKyH,EAAM,CAET,MAAMgC,EAAQ7G,EAAMtC,MACpBsC,EAAMtC,MAAQ,GACdN,KAAKgJ,QAEL,IAAK,IAAIU,EAAI,EAAGA,EAAID,EAAMpJ,OAAQqJ,IAChCD,EAAMC,GAAGhG,KAAK,SAAU1D,KAAM,CAAEqI,YAAY,IAC9C,OAAOrI,IACT,CAGA,MAAM2J,EAAQ/G,EAAMtC,MAAMsJ,QAAQnC,GAClC,OAAe,IAAXkC,IAGJ/G,EAAMtC,MAAMuJ,OAAOF,EAAO,GACC,IAAvB/G,EAAMtC,MAAMD,QACdL,KAAKgJ,QAEPvB,EAAK/D,KAAK,SAAU1D,KA1BD,CAAEqI,YAAY,KAoBxBrI,IASX,EAIAlB,SAASU,UAAUmG,GAAK,SAASmE,EAAIC,GACnC,MAAMC,EAAM1L,EAAOkB,UAAUmG,GAAGpD,KAAKvC,KAAM8J,EAAIC,GACzCnH,EAAQ5C,KAAK+B,eAuBnB,MArBW,SAAP+H,GAGFlH,EAAM7B,kBAAoBf,KAAKwD,cAAc,YAAc,GAGrC,IAAlBZ,EAAMrC,SACRP,KAAKgF,UACS,aAAP8E,IACJlH,EAAMnC,YAAemC,EAAM7B,oBAC9B6B,EAAM7B,kBAAoB6B,EAAM/B,cAAe,EAC/C+B,EAAMrC,SAAU,EAChBqC,EAAM9B,iBAAkB,EACpB8B,EAAMvC,OACR8C,EAAanD,MACH4C,EAAMlC,SAChB9B,EAAQ8F,SAASO,EAAkBjF,QAKlCgK,CACT,EACAlL,SAASU,UAAUyK,YAAcnL,SAASU,UAAUmG,GAEpD7G,SAASU,UAAU8I,eAAiB,SAASwB,EAAIC,GAC/C,MAAMC,EAAM1L,EAAOkB,UAAU8I,eAAe/F,KAAKvC,KACA8J,EAAIC,GAYrD,MAVW,aAAPD,GAOFlL,EAAQ8F,SAASI,EAAyB9E,MAGrCgK,CACT,EACAlL,SAASU,UAAU0K,IAAMpL,SAASU,UAAU8I,eAE5CxJ,SAASU,UAAU2K,mBAAqB,SAASL,GAC/C,MAAME,EAAM1L,EAAOkB,UAAU2K,mBAAmBC,MAAMpK,KACAqK,WAYtD,MAVW,aAAPP,QAA4B/D,IAAP+D,GAOvBlL,EAAQ8F,SAASI,EAAyB9E,MAGrCgK,CACT,EAyBAlL,SAASU,UAAUwF,OAAS,WAC1B,MAAMpC,EAAQ5C,KAAK+B,eASnB,OARKa,EAAMrC,UAITqC,EAAMrC,SAAWqC,EAAM7B,kBAO3B,SAAgBlB,EAAQ+C,GACjBA,EAAM5B,kBACT4B,EAAM5B,iBAAkB,EACxBpC,EAAQ8F,SAASQ,EAASrF,EAAQ+C,GAEtC,CAXIoC,CAAOhF,KAAM4C,IAEfA,EAAMxD,IAAW,EACVY,IACT,EAqBAlB,SAASU,UAAUwJ,MAAQ,WAMzB,OALoC,IAAhChJ,KAAK+B,eAAexB,UACtBP,KAAK+B,eAAexB,SAAU,EAC9BP,KAAK0D,KAAK,UAEZ1D,KAAK+B,eAAe3C,IAAW,EACxBY,IACT,EAUAlB,SAASU,UAAU4F,KAAO,SAASvF,GACjC,IAAIyK,GAAS,EAMbzK,EAAO8F,GAAG,QAASjD,KACZ1C,KAAKkD,KAAKR,IAAU7C,EAAOmJ,QAC9BsB,GAAS,EACTzK,EAAOmJ,QACT,IAGFnJ,EAAO8F,GAAG,OAAO,KACf3F,KAAKkD,KAAK,KAAK,IAGjBrD,EAAO8F,GAAG,SAAU9C,IAClBnD,EAAeM,KAAM6C,EAAI,IAG3BhD,EAAO8F,GAAG,SAAS,KACjB3F,KAAKkC,SAAS,IAGhBrC,EAAO8F,GAAG,WAAW,KACnB3F,KAAKkC,SAAS,IAGhBlC,KAAKiC,MAAQ,KACPqI,GAAUzK,EAAOmF,SACnBsF,GAAS,EACTzK,EAAOmF,SACT,EAIF,MAAMuF,EAAajL,OAAOkL,KAAK3K,GAC/B,IAAK,IAAI4K,EAAI,EAAGA,EAAIF,EAAWlK,OAAQoK,IAAK,CAC1C,MAAMf,EAAIa,EAAWE,QACL1E,IAAZ/F,KAAK0J,IAAyC,mBAAd7J,EAAO6J,KACzC1J,KAAK0J,GAAK7J,EAAO6J,GAAGgB,KAAK7K,GAE7B,CAEA,OAAOG,IACT,EAEAlB,SAASU,UAAUH,OAAOsL,eAAiB,WACzC,OAAOxF,EAAsBnF,KAC/B,EAEAlB,SAASU,UAAUoL,SAAW,SAAShL,GACrC,OAAOuF,EAAsBnF,KAAMJ,EACrC,EA8DAN,OAAOuL,iBAAiB/L,SAASU,UAAW,CAC1C2I,SAAU,CACR2C,MACE,MAAMC,EAAI/K,KAAK+B,eAKf,SAASgJ,IAAoB,IAAfA,EAAE5C,UAAuB4C,EAAE3J,WAAc2J,EAAE9J,cACtD8J,EAAEtK,WACP,EACAuK,IAAIC,GAEEjL,KAAK+B,iBACP/B,KAAK+B,eAAeoG,WAAa8C,EAErC,GAGFC,gBAAiB,CACfC,YAAY,EACZL,IAAK,WACH,OAAO9K,KAAK+B,eAAeH,WAC7B,GAGFwJ,gBAAiB,CACfD,YAAY,EACZL,IAAK,WACH,SAAU9K,KAAK+B,eAAeX,YAAapB,KAAK+B,eAAeV,SAC5DrB,KAAK+B,eAAetB,WACzB,GAGF4K,sBAAuB,CACrBF,YAAY,EACZL,IAAK,WACH,OAAO9K,KAAK+B,eAAe5B,aAC7B,GAGFmL,eAAgB,CACdH,YAAY,EACZL,IAAK,WACH,OAAO9K,KAAK+B,gBAAkB/B,KAAK+B,eAAe3B,MACpD,GAGFmL,gBAAiB,CACfJ,YAAY,EACZL,IAAK,WACH,OAAO9K,KAAK+B,eAAexB,OAC7B,EACAyK,IAAK,SAASpI,GACR5C,KAAK+B,iBACP/B,KAAK+B,eAAexB,QAAUqC,EAElC,GAGF4I,eAAgB,CACdL,YAAY,EACZL,MACE,OAAO9K,KAAK+B,eAAe1B,MAC7B,GAGFH,mBAAoB,CAClBiL,YAAY,EACZL,MACE,QAAO9K,KAAK+B,gBAAiB/B,KAAK+B,eAAe9B,UACnD,GAGFwL,iBAAkB,CAChBN,YAAY,EACZL,MACE,OAAO9K,KAAK+B,eAAiB/B,KAAK+B,eAAeD,SAAW,IAC9D,GAGFV,UAAW,CACT+J,YAAY,EACZL,MACE,YAA4B/E,IAAxB/F,KAAK+B,gBAGF/B,KAAK+B,eAAeX,SAC7B,EACA4J,IAAIU,GAGG1L,KAAK+B,iBAMV/B,KAAK+B,eAAeX,UAAYsK,EAClC,GAGFC,cAAe,CACbR,YAAY,EACZL,MACE,QAAO9K,KAAK+B,gBAAiB/B,KAAK+B,eAAetB,UACnD,KAKJnB,OAAOuL,iBAAiBlL,cAAcH,UAAW,CAE/CoM,WAAY,CACVd,MACE,OAAO9K,KAAKM,MAAMD,MACpB,GAIFiK,OAAQ,CACNQ,MACE,OAAyB,IAAlB9K,KAAKZ,EACd,EACA4L,IAAIU,GACF1L,KAAKZ,KAAasM,CACpB,KAKJ5M,SAAS+M,UAAY3F,EA2ErBpH,SAAST,KAAO,SAASyN,EAAUC,GACjC,OAAO1N,EAAKS,SAAUgN,EAAUC,EAClC,EAEAjN,SAASsG,KAAO,SAASuC,EAAK/H,GAC5B,OAAO,IAAId,SAAS,CAClBmB,WAAY0H,EAAIzH,oBAAsByH,EAAI1H,aAAc,KACrDL,EACHsC,QAAQW,EAAK2C,GACXvH,EAAY+H,UAAU2B,EAAK9E,GAC3B2C,EAAS3C,EACX,IACCuC,KAAKuC,EACV"}
✄
import{addAbortSignal as e}from"./add-abort-signal.js";import t from"./buffer_list.js";import*as r from"./destroy.js";import n from"./end-of-stream.js";import{aggregateTwoErrors as a,codes as i}from"../errors.js";import o from"./from.js";import{Stream as d,prependListener as s}from"./legacy.js";import{getHighWaterMark as l,getDefaultHighWaterMark as u}from"./state.js";import{Buffer as h}from"buffer";import c from"events";import f from"process";import{StringDecoder as b}from"string_decoder";export default Readable;const{ERR_INVALID_ARG_TYPE:p,ERR_METHOD_NOT_IMPLEMENTED:g,ERR_OUT_OF_RANGE:m,ERR_STREAM_PUSH_AFTER_EOF:y,ERR_STREAM_UNSHIFT_AFTER_END_EVENT:_}=i,w=Symbol("kPaused");Object.setPrototypeOf(Readable.prototype,d.prototype),Object.setPrototypeOf(Readable,d);const R=()=>{},{errorOrDestroy:S}=r;export function ReadableState(e,r,n){"boolean"!=typeof n&&(n=r instanceof d.Duplex),this.objectMode=!(!e||!e.objectMode),n&&(this.objectMode=this.objectMode||!(!e||!e.readableObjectMode)),this.highWaterMark=e?l(this,e,"readableHighWaterMark",n):u(!1),this.buffer=new t,this.length=0,this.pipes=[],this.flowing=null,this.ended=!1,this.endEmitted=!1,this.reading=!1,this.constructed=!0,this.sync=!0,this.needReadable=!1,this.emittedReadable=!1,this.readableListening=!1,this.resumeScheduled=!1,this[w]=null,this.errorEmitted=!1,this.emitClose=!e||!1!==e.emitClose,this.autoDestroy=!e||!1!==e.autoDestroy,this.destroyed=!1,this.errored=null,this.closed=!1,this.closeEmitted=!1,this.defaultEncoding=e&&e.defaultEncoding||"utf8",this.awaitDrainWriters=null,this.multiAwaitDrain=!1,this.readingMore=!1,this.dataEmitted=!1,this.decoder=null,this.encoding=null,e&&e.encoding&&(this.decoder=new b(e.encoding),this.encoding=e.encoding)}export function Readable(t){if(!(this instanceof Readable))return new Readable(t);const n=this instanceof d.Duplex;this._readableState=new ReadableState(t,this,n),t&&("function"==typeof t.read&&(this._read=t.read),"function"==typeof t.destroy&&(this._destroy=t.destroy),"function"==typeof t.construct&&(this._construct=t.construct),t.signal&&!n&&e(t.signal,this)),d.call(this,t),r.construct(this,(()=>{this._readableState.needReadable&&L(this,this._readableState)}))}function E(e,t,r,n){const a=e._readableState;let i;if(a.objectMode||("string"==typeof t?(r=r||a.defaultEncoding,a.encoding!==r&&(n&&a.encoding?t=h.from(t,r).toString(a.encoding):(t=h.from(t,r),r=""))):t instanceof h?r="":d._isUint8Array(t)?(t=d._uint8ArrayToBuffer(t),r=""):null!=t&&(i=new p("chunk",["string","Buffer","Uint8Array"],t))),i)S(e,i);else if(null===t)a.reading=!1,function(e,t){if(t.ended)return;if(t.decoder){const e=t.decoder.end();e&&e.length&&(t.buffer.push(e),t.length+=t.objectMode?1:e.length)}t.ended=!0,t.sync?j(e):(t.needReadable=!1,t.emittedReadable=!0,W(e))}(e,a);else if(a.objectMode||t&&t.length>0)if(n)if(a.endEmitted)S(e,new _);else{if(a.destroyed||a.errored)return!1;M(e,a,t,!0)}else if(a.ended)S(e,new y);else{if(a.destroyed||a.errored)return!1;a.reading=!1,a.decoder&&!r?(t=a.decoder.write(t),a.objectMode||0!==t.length?M(e,a,t,!1):L(e,a)):M(e,a,t,!1)}else n||(a.reading=!1,L(e,a));return!a.ended&&(a.length<a.highWaterMark||0===a.length)}function M(e,t,r,n){t.flowing&&0===t.length&&!t.sync&&e.listenerCount("data")>0?(t.multiAwaitDrain?t.awaitDrainWriters.clear():t.awaitDrainWriters=null,t.dataEmitted=!0,e.emit("data",r)):(t.length+=t.objectMode?1:r.length,n?t.buffer.unshift(r):t.buffer.push(r),t.needReadable&&j(e)),L(e,t)}Readable.prototype.destroy=r.destroy,Readable.prototype._undestroy=r.undestroy,Readable.prototype._destroy=function(e,t){t(e)},Readable.prototype[c.captureRejectionSymbol]=function(e){this.destroy(e)},Readable.prototype.push=function(e,t){return E(this,e,t,!1)},Readable.prototype.unshift=function(e,t){return E(this,e,t,!0)},Readable.prototype.isPaused=function(){const e=this._readableState;return!0===e[w]||!1===e.flowing},Readable.prototype.setEncoding=function(e){const t=new b(e);this._readableState.decoder=t,this._readableState.encoding=this._readableState.decoder.encoding;const r=this._readableState.buffer;let n="";for(const e of r)n+=t.write(e);return r.clear(),""!==n&&r.push(n),this._readableState.length=n.length,this};function D(e,t){return e<=0||0===t.length&&t.ended?0:t.objectMode?1:Number.isNaN(e)?t.flowing&&t.length?t.buffer.first().length:t.length:e<=t.length?e:t.ended?t.length:0}function j(e){const t=e._readableState;t.needReadable=!1,t.emittedReadable||(t.emittedReadable=!0,f.nextTick(W,e))}function W(e){const t=e._readableState;t.destroyed||t.errored||!t.length&&!t.ended||(e.emit("readable"),t.emittedReadable=!1),t.needReadable=!t.flowing&&!t.ended&&t.length<=t.highWaterMark,O(e)}function L(e,t){!t.readingMore&&t.constructed&&(t.readingMore=!0,f.nextTick(k,e,t))}function k(e,t){for(;!t.reading&&!t.ended&&(t.length<t.highWaterMark||t.flowing&&0===t.length);){const r=t.length;if(e.read(0),r===t.length)break}t.readingMore=!1}function v(e){const t=e._readableState;t.readableListening=e.listenerCount("readable")>0,t.resumeScheduled&&!1===t[w]?t.flowing=!0:e.listenerCount("data")>0?e.resume():t.readableListening||(t.flowing=null)}function T(e){e.read(0)}function A(e,t){t.reading||e.read(0),t.resumeScheduled=!1,e.emit("resume"),O(e),t.flowing&&!t.reading&&e.read(0)}function O(e){const t=e._readableState;for(;t.flowing&&null!==e.read(););}function x(e,t){"function"!=typeof e.read&&(e=Readable.wrap(e,{objectMode:!0}));const i=async function*(e,t){let i,o=R;function d(t){this===e?(o(),o=R):o=t}e.on("readable",d),n(e,{writable:!1},(e=>{i=e?a(i,e):null,o(),o=R}));try{for(;;){const t=e.destroyed?null:e.read();if(null!==t)yield t;else{if(i)throw i;if(null===i)return;await new Promise(d)}}}catch(e){throw i=a(i,e),i}finally{!i&&!1===t?.destroyOnReturn||void 0!==i&&!e._readableState.autoDestroy||r.destroyer(e,null)}}(e,t);return i.stream=e,i}function N(e,t){if(0===t.length)return null;let r;return t.objectMode?r=t.buffer.shift():!e||e>=t.length?(r=t.decoder?t.buffer.join(""):1===t.buffer.length?t.buffer.first():t.buffer.concat(t.length),t.buffer.clear()):r=t.buffer.consume(e,t.decoder),r}function P(e){const t=e._readableState;t.endEmitted||(t.ended=!0,f.nextTick(C,t,e))}function C(e,t){if(!e.errored&&!e.closeEmitted&&!e.endEmitted&&0===e.length)if(e.endEmitted=!0,t.emit("end"),t.writable&&!1===t.allowHalfOpen)f.nextTick(U,t);else if(e.autoDestroy){const e=t._writableState;(!e||e.autoDestroy&&(e.finished||!1===e.writable))&&t.destroy()}}function U(e){e.writable&&!e.writableEnded&&!e.destroyed&&e.end()}Readable.prototype.read=function(e){void 0===e?e=NaN:Number.isInteger(e)||(e=Number.parseInt(e,10));const t=this._readableState,r=e;if(e>t.highWaterMark&&(t.highWaterMark=function(e){if(e>1073741824)throw new m("size","<= 1GiB",e);return e--,e|=e>>>1,e|=e>>>2,e|=e>>>4,e|=e>>>8,e|=e>>>16,++e}(e)),0!==e&&(t.emittedReadable=!1),0===e&&t.needReadable&&((0!==t.highWaterMark?t.length>=t.highWaterMark:t.length>0)||t.ended))return 0===t.length&&t.ended?P(this):j(this),null;if(0===(e=D(e,t))&&t.ended)return 0===t.length&&P(this),null;let n,a=t.needReadable;if((0===t.length||t.length-e<t.highWaterMark)&&(a=!0),t.ended||t.reading||t.destroyed||t.errored||!t.constructed)a=!1;else if(a){t.reading=!0,t.sync=!0,0===t.length&&(t.needReadable=!0);try{const e=this._read(t.highWaterMark);if(null!=e){const t=e.then;"function"==typeof t&&t.call(e,R,(function(e){S(this,e)}))}}catch(e){S(this,e)}t.sync=!1,t.reading||(e=D(r,t))}return n=e>0?N(e,t):null,null===n?(t.needReadable=t.length<=t.highWaterMark,e=0):(t.length-=e,t.multiAwaitDrain?t.awaitDrainWriters.clear():t.awaitDrainWriters=null),0===t.length&&(t.ended||(t.needReadable=!0),r!==e&&t.ended&&P(this)),null===n||t.errorEmitted||t.closeEmitted||(t.dataEmitted=!0,this.emit("data",n)),n},Readable.prototype._read=function(e){throw new g("_read()")},Readable.prototype.pipe=function(e,t){const r=this,n=this._readableState;1===n.pipes.length&&(n.multiAwaitDrain||(n.multiAwaitDrain=!0,n.awaitDrainWriters=new Set(n.awaitDrainWriters?[n.awaitDrainWriters]:[]))),n.pipes.push(e);const a=(!t||!1!==t.end)&&e!==f.stdout&&e!==f.stderr?o:m;function i(t,a){t===r&&a&&!1===a.hasUnpiped&&(a.hasUnpiped=!0,function(){e.removeListener("close",p),e.removeListener("finish",g),d&&e.removeListener("drain",d);e.removeListener("error",b),e.removeListener("unpipe",i),r.removeListener("end",o),r.removeListener("end",m),r.removeListener("data",h),l=!0,d&&n.awaitDrainWriters&&(!e._writableState||e._writableState.needDrain)&&d()}())}function o(){e.end()}let d;n.endEmitted?f.nextTick(a):r.once("end",a),e.on("unpipe",i);let l=!1;function u(){l||(1===n.pipes.length&&n.pipes[0]===e?(n.awaitDrainWriters=e,n.multiAwaitDrain=!1):n.pipes.length>1&&n.pipes.includes(e)&&n.awaitDrainWriters.add(e),r.pause()),d||(d=function(e,t){return function(){const r=e._readableState;r.awaitDrainWriters===t?r.awaitDrainWriters=null:r.multiAwaitDrain&&r.awaitDrainWriters.delete(t),r.awaitDrainWriters&&0!==r.awaitDrainWriters.size||!c.listenerCount(e,"data")||(r.flowing=!0,O(e))}}(r,e),e.on("drain",d))}function h(t){!1===e.write(t)&&u()}function b(t){if(m(),e.removeListener("error",b),0===c.listenerCount(e,"error")){const r=e._writableState||e._readableState;r&&!r.errorEmitted?S(e,t):e.emit("error",t)}}function p(){e.removeListener("finish",g),m()}function g(){e.removeListener("close",p),m()}function m(){r.unpipe(e)}return r.on("data",h),s(e,"error",b),e.once("close",p),e.once("finish",g),e.emit("pipe",r),!0===e.writableNeedDrain?n.flowing&&u():n.flowing||r.resume(),e},Readable.prototype.unpipe=function(e){const t=this._readableState;if(0===t.pipes.length)return this;if(!e){const e=t.pipes;t.pipes=[],this.pause();for(let t=0;t<e.length;t++)e[t].emit("unpipe",this,{hasUnpiped:!1});return this}const r=t.pipes.indexOf(e);return-1===r||(t.pipes.splice(r,1),0===t.pipes.length&&this.pause(),e.emit("unpipe",this,{hasUnpiped:!1})),this},Readable.prototype.on=function(e,t){const r=d.prototype.on.call(this,e,t),n=this._readableState;return"data"===e?(n.readableListening=this.listenerCount("readable")>0,!1!==n.flowing&&this.resume()):"readable"===e&&(n.endEmitted||n.readableListening||(n.readableListening=n.needReadable=!0,n.flowing=!1,n.emittedReadable=!1,n.length?j(this):n.reading||f.nextTick(T,this))),r},Readable.prototype.addListener=Readable.prototype.on,Readable.prototype.removeListener=function(e,t){const r=d.prototype.removeListener.call(this,e,t);return"readable"===e&&f.nextTick(v,this),r},Readable.prototype.off=Readable.prototype.removeListener,Readable.prototype.removeAllListeners=function(e){const t=d.prototype.removeAllListeners.apply(this,arguments);return"readable"!==e&&void 0!==e||f.nextTick(v,this),t},Readable.prototype.resume=function(){const e=this._readableState;return e.flowing||(e.flowing=!e.readableListening,function(e,t){t.resumeScheduled||(t.resumeScheduled=!0,f.nextTick(A,e,t))}(this,e)),e[w]=!1,this},Readable.prototype.pause=function(){return!1!==this._readableState.flowing&&(this._readableState.flowing=!1,this.emit("pause")),this._readableState[w]=!0,this},Readable.prototype.wrap=function(e){let t=!1;e.on("data",(r=>{!this.push(r)&&e.pause&&(t=!0,e.pause())})),e.on("end",(()=>{this.push(null)})),e.on("error",(e=>{S(this,e)})),e.on("close",(()=>{this.destroy()})),e.on("destroy",(()=>{this.destroy()})),this._read=()=>{t&&e.resume&&(t=!1,e.resume())};const r=Object.keys(e);for(let t=1;t<r.length;t++){const n=r[t];void 0===this[n]&&"function"==typeof e[n]&&(this[n]=e[n].bind(e))}return this},Readable.prototype[Symbol.asyncIterator]=function(){return x(this)},Readable.prototype.iterator=function(e){return x(this,e)},Object.defineProperties(Readable.prototype,{readable:{get(){const e=this._readableState;return!(!e||!1===e.readable||e.destroyed||e.errorEmitted||e.endEmitted)},set(e){this._readableState&&(this._readableState.readable=!!e)}},readableDidRead:{enumerable:!1,get:function(){return this._readableState.dataEmitted}},readableAborted:{enumerable:!1,get:function(){return!(!this._readableState.destroyed&&!this._readableState.errored||this._readableState.endEmitted)}},readableHighWaterMark:{enumerable:!1,get:function(){return this._readableState.highWaterMark}},readableBuffer:{enumerable:!1,get:function(){return this._readableState&&this._readableState.buffer}},readableFlowing:{enumerable:!1,get:function(){return this._readableState.flowing},set:function(e){this._readableState&&(this._readableState.flowing=e)}},readableLength:{enumerable:!1,get(){return this._readableState.length}},readableObjectMode:{enumerable:!1,get(){return!!this._readableState&&this._readableState.objectMode}},readableEncoding:{enumerable:!1,get(){return this._readableState?this._readableState.encoding:null}},destroyed:{enumerable:!1,get(){return void 0!==this._readableState&&this._readableState.destroyed},set(e){this._readableState&&(this._readableState.destroyed=e)}},readableEnded:{enumerable:!1,get(){return!!this._readableState&&this._readableState.endEmitted}}}),Object.defineProperties(ReadableState.prototype,{pipesCount:{get(){return this.pipes.length}},paused:{get(){return!1!==this[w]},set(e){this[w]=!!e}}}),Readable._fromList=N,Readable.from=function(e,t){return o(Readable,e,t)},Readable.wrap=function(e,t){return new Readable({objectMode:e.readableObjectMode??e.objectMode??!0,...t,destroy(t,n){r.destroyer(e,t),n(t)}}).wrap(e)};
✄
{"version":3,"file":"state.js","names":["errorCodes","ERR_INVALID_ARG_VALUE","getDefaultHighWaterMark","objectMode","getHighWaterMark","state","options","duplexKey","isDuplex","hwm","highWaterMark","highWaterMarkFrom","Number","isInteger","Math","floor"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"gBAAkBA,MAAkB,eAEpC,MAAMC,sBAAEA,GAA0BD,SAO3B,SAASE,wBAAwBC,GACtC,OAAOA,EAAa,GAAK,KAC3B,QAEO,SAASC,iBAAiBC,EAAOC,EAASC,EAAWC,GAC1D,MAAMC,EAVR,SAA2BH,EAASE,EAAUD,GAC5C,OAAgC,MAAzBD,EAAQI,cAAwBJ,EAAQI,cAC7CF,EAAWF,EAAQC,GAAa,IACpC,CAOcI,CAAkBL,EAASE,EAAUD,GACjD,GAAW,MAAPE,EAAa,CACf,IAAKG,OAAOC,UAAUJ,IAAQA,EAAM,EAAG,CAErC,MAAM,IAAIR,EADGO,EAAW,WAAWD,IAAc,wBACXE,EACxC,CACA,OAAOK,KAAKC,MAAMN,EACpB,CAGA,OAAOP,wBAAwBG,EAAMF,WACvC"}
✄
import{codes as r}from"../errors.js";const{ERR_INVALID_ARG_VALUE:t}=r;export function getDefaultHighWaterMark(r){return r?16:16384}export function getHighWaterMark(r,e,o,n){const a=function(r,t,e){return null!=r.highWaterMark?r.highWaterMark:t?r[e]:null}(e,n,o);if(null!=a){if(!Number.isInteger(a)||a<0){throw new t(n?`options.${o}`:"options.highWaterMark",a)}return Math.floor(a)}return getDefaultHighWaterMark(r.objectMode)}
✄
{"version":3,"file":"transform.js","names":["Duplex","errorCodes","process","ERR_METHOD_NOT_IMPLEMENTED","Object","setPrototypeOf","Transform","prototype","kCallback","Symbol","options","this","call","_readableState","sync","transform","_transform","flush","_flush","on","prefinish","final","cb","called","destroyed","push","result","er","data","destroy","then","nextTick","err","_final","chunk","encoding","callback","_write","rState","wState","_writableState","length","val","ended","highWaterMark","undefined","_read"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OA+DOA,MAAY,8BACDC,MAAkB,sBAE7BC,MAAa,UAEpB,MAAMC,2BACJA,GACEF,EAEJG,OAAOC,eAAeC,EAAUC,UAAWP,EAAOO,WAClDH,OAAOC,eAAeC,EAAWN,GAEjC,MAAMQ,EAAYC,OAAO,4BAEV,SAASH,EAAUI,GAChC,KAAMC,gBAAgBL,GACpB,OAAO,IAAIA,EAAUI,GAEvBV,EAAOY,KAAKD,KAAMD,GAKlBC,KAAKE,eAAeC,MAAO,EAE3BH,KAAKH,GAAa,KAEdE,IAC+B,mBAAtBA,EAAQK,YACjBJ,KAAKK,WAAaN,EAAQK,WAEC,mBAAlBL,EAAQO,QACjBN,KAAKO,OAASR,EAAQO,QAO1BN,KAAKQ,GAAG,YAAaC,EACvB,CAEA,SAASC,EAAMC,GACb,IAAIC,GAAS,EACb,GAA2B,mBAAhBZ,KAAKO,QAA0BP,KAAKa,UAgD7Cb,KAAKc,KAAK,MACNH,GACFA,QAlDsD,CACxD,MAAMI,EAASf,KAAKO,QAAO,CAACS,EAAIC,KAC9BL,GAAS,EACLI,EACEL,EACFA,EAAGK,GAEHhB,KAAKkB,QAAQF,IAKL,MAARC,GACFjB,KAAKc,KAAKG,GAEZjB,KAAKc,KAAK,MACNH,GACFA,IACF,IAEF,GAAII,QACF,IACE,MAAMI,EAAOJ,EAAOI,KACA,mBAATA,GACTA,EAAKlB,KACHc,GACCE,IACKL,IAEQ,MAARK,GACFjB,KAAKc,KAAKG,GACZjB,KAAKc,KAAK,MACNH,GACFpB,EAAQ6B,SAAST,GAAG,IAEvBU,IACKV,EACFpB,EAAQ6B,SAAST,EAAIU,GAErB9B,EAAQ6B,UAAS,IAAMpB,KAAKkB,QAAQG,IACtC,GAKR,CAFE,MAAOA,GACP9B,EAAQ6B,UAAS,IAAMpB,KAAKkB,QAAQG,IACtC,CAEJ,CAMF,CAEA,SAASZ,IACHT,KAAKsB,SAAWZ,GAClBA,EAAMT,KAAKD,KAEf,CAEAL,EAAUC,UAAU0B,OAASZ,EAE7Bf,EAAUC,UAAUS,WAAa,SAASkB,EAAOC,EAAUC,GACzD,MAAM,IAAIjC,EAA2B,eACvC,EAEAG,EAAUC,UAAU8B,OAAS,SAASH,EAAOC,EAAUC,GACrD,MAAME,EAAS3B,KAAKE,eACd0B,EAAS5B,KAAK6B,eACdC,EAASH,EAAOG,OAEtB,IAAIlB,GAAS,EACb,MAAMG,EAASf,KAAKK,WAAWkB,EAAOC,GAAU,CAACH,EAAKU,KACpDnB,GAAS,EACLS,EACFI,EAASJ,IAIA,MAAPU,GACF/B,KAAKc,KAAKiB,GAIVH,EAAOI,OACPF,IAAWH,EAAOG,QAClBH,EAAOG,OAASH,EAAOM,eACL,IAAlBN,EAAOG,OAEPL,IAEAzB,KAAKH,GAAa4B,EACpB,IAEF,QAAeS,IAAXnB,GAAkC,MAAVA,EAC1B,IACE,MAAMI,EAAOJ,EAAOI,KACA,mBAATA,GACTA,EAAKlB,KACHc,GACCgB,IACKnB,IAGO,MAAPmB,GACF/B,KAAKc,KAAKiB,GAIVH,EAAOI,OACPF,IAAWH,EAAOG,QAClBH,EAAOG,OAASH,EAAOM,eACL,IAAlBN,EAAOG,OACPvC,EAAQ6B,SAASK,GAEjBzB,KAAKH,GAAa4B,EACpB,IAEDJ,IACC9B,EAAQ6B,SAASK,EAAUJ,EAAI,GAKvC,CAFE,MAAOA,GACP9B,EAAQ6B,SAASK,EAAUJ,EAC7B,CAEJ,EAEA1B,EAAUC,UAAUuC,MAAQ,WAC1B,GAAInC,KAAKH,GAAY,CACnB,MAAM4B,EAAWzB,KAAKH,GACtBG,KAAKH,GAAa,KAClB4B,GACF,CACF"}
✄
import t from"./duplex.js";import{codes as n}from"../errors.js";import s from"process";const{ERR_METHOD_NOT_IMPLEMENTED:e}=n;Object.setPrototypeOf(o.prototype,t.prototype),Object.setPrototypeOf(o,t);const i=Symbol("kCallback");export default function o(n){if(!(this instanceof o))return new o(n);t.call(this,n),this._readableState.sync=!1,this[i]=null,n&&("function"==typeof n.transform&&(this._transform=n.transform),"function"==typeof n.flush&&(this._flush=n.flush)),this.on("prefinish",l)}function h(t){let n=!1;if("function"!=typeof this._flush||this.destroyed)this.push(null),t&&t();else{const e=this._flush(((s,e)=>{n=!0,s?t?t(s):this.destroy(s):(null!=e&&this.push(e),this.push(null),t&&t())}));if(null!=e)try{const i=e.then;"function"==typeof i&&i.call(e,(e=>{n||(null!=e&&this.push(e),this.push(null),t&&s.nextTick(t))}),(n=>{t?s.nextTick(t,n):s.nextTick((()=>this.destroy(n)))}))}catch(t){s.nextTick((()=>this.destroy(t)))}}}function l(){this._final!==h&&h.call(this)}o.prototype._final=h,o.prototype._transform=function(t,n,s){throw new e("_transform()")},o.prototype._write=function(t,n,e){const o=this._readableState,h=this._writableState,l=o.length;let r=!1;const f=this._transform(t,n,((t,n)=>{r=!0,t?e(t):(null!=n&&this.push(n),h.ended||l===o.length||o.length<o.highWaterMark||0===o.length?e():this[i]=e)}));if(void 0!==f&&null!=f)try{const t=f.then;"function"==typeof t&&t.call(f,(t=>{r||(null!=t&&this.push(t),h.ended||l===o.length||o.length<o.highWaterMark||0===o.length?s.nextTick(e):this[i]=e)}),(t=>{s.nextTick(e,t)}))}catch(t){s.nextTick(e,t)}},o.prototype._read=function(){if(this[i]){const t=this[i];this[i]=null,t()}};
✄
{"version":3,"file":"utils.js","names":["kDestroyed","Symbol","kIsDisturbed","isReadableNodeStream","obj","pipe","on","_writableState","_readableState","readable","isWritableNodeStream","write","writable","isDuplexNodeStream","isNodeStream","isIterable","isAsync","asyncIterator","iterator","isDestroyed","stream","wState","rState","state","destroyed","isWritableEnded","writableEnded","errored","ended","isWritableFinished","strict","writableFinished","finished","length","isReadableEnded","readableEnded","isReadableFinished","endEmitted","isReadable","r","isWritable","isFinished","opts","isClosed","closed","_closed","isOutgoingMessage","_defaultKeepAlive","_removedConnection","_removedContLen","isServerResponse","_sent100","isServerRequest","_consuming","_dumped","undefined","req","upgradeOrConnect","willEmitClose","autoDestroy","emitClose","isDisturbed","readableDidRead","readableAborted"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"OAAO,MAAMA,WAAaC,OAAO,qBAC1B,MAAMC,aAAeD,OAAO,uBAE5B,SAASE,qBAAqBC,GACnC,SACEA,GACoB,mBAAbA,EAAIC,MACO,mBAAXD,EAAIE,IACTF,EAAIG,iBAAmD,IAAjCH,EAAII,gBAAgBC,UAC1CL,EAAIG,iBAAkBH,EAAII,eAEhC,QAEO,SAASE,qBAAqBN,GACnC,SACEA,GACqB,mBAAdA,EAAIO,OACO,mBAAXP,EAAIE,IACTF,EAAII,iBAAmD,IAAjCJ,EAAIG,gBAAgBK,SAEhD,QAEO,SAASC,mBAAmBT,GACjC,SACEA,GACqB,mBAAbA,EAAIC,OAAuBD,EAAII,gBACrB,mBAAXJ,EAAIE,IACU,mBAAdF,EAAIO,MAEf,QAEO,SAASG,aAAaV,GAC3B,OACEA,IAEEA,EAAII,gBACJJ,EAAIG,gBACkB,mBAAdH,EAAIO,OAA0C,mBAAXP,EAAIE,IAC1B,mBAAbF,EAAIC,MAAyC,mBAAXD,EAAIE,GAGpD,QAEO,SAASS,WAAWX,EAAKY,GAC9B,OAAW,MAAPZ,KACY,IAAZY,EAA8D,mBAA9BZ,EAAIH,OAAOgB,gBAC/B,IAAZD,EAA0D,mBAAzBZ,EAAIH,OAAOiB,UACJ,mBAA9Bd,EAAIH,OAAOgB,gBACS,mBAAzBb,EAAIH,OAAOiB,UACtB,QAEO,SAASC,YAAYC,GAC1B,IAAKN,aAAaM,GAAS,OAAO,KAClC,MAAMC,EAASD,EAAOb,eAChBe,EAASF,EAAOZ,eAChBe,EAAQF,GAAUC,EACxB,SAAUF,EAAOI,WAAaJ,EAAOpB,aAAeuB,GAAOC,UAC7D,QAGO,SAASC,gBAAgBL,GAC9B,IAAKV,qBAAqBU,GAAS,OAAO,KAC1C,IAA6B,IAAzBA,EAAOM,cAAwB,OAAO,EAC1C,MAAML,EAASD,EAAOb,eACtB,OAAIc,GAAQM,UACiB,kBAAlBN,GAAQO,MAA4B,KACxCP,EAAOO,MAChB,QAGO,SAASC,mBAAmBT,EAAQU,GACzC,IAAKpB,qBAAqBU,GAAS,OAAO,KAC1C,IAAgC,IAA5BA,EAAOW,iBAA2B,OAAO,EAC7C,MAAMV,EAASD,EAAOb,eACtB,OAAIc,GAAQM,UACoB,kBAArBN,GAAQW,SAA+B,QAEhDX,EAAOW,WACK,IAAXF,IAAqC,IAAjBT,EAAOO,OAAoC,IAAlBP,EAAOY,QAEzD,QAGO,SAASC,gBAAgBd,GAC9B,IAAKjB,qBAAqBiB,GAAS,OAAO,KAC1C,IAA6B,IAAzBA,EAAOe,cAAwB,OAAO,EAC1C,MAAMb,EAASF,EAAOZ,eACtB,SAAKc,GAAUA,EAAOK,WACO,kBAAlBL,GAAQM,MAA4B,KACxCN,EAAOM,MAChB,QAGO,SAASQ,mBAAmBhB,EAAQU,GACzC,IAAK3B,qBAAqBiB,GAAS,OAAO,KAC1C,MAAME,EAASF,EAAOZ,eACtB,OAAIc,GAAQK,UACsB,kBAAvBL,GAAQe,WAAiC,QAElDf,EAAOe,aACK,IAAXP,IAAqC,IAAjBR,EAAOM,OAAoC,IAAlBN,EAAOW,QAEzD,QAEO,SAASK,WAAWlB,GACzB,MAAMmB,EAAIpC,qBAAqBiB,GAC/B,OAAU,OAANmB,GAA0C,kBAArBnB,GAAQX,SAA+B,MAC5DU,YAAYC,KACTmB,GAAKnB,EAAOX,WAAa2B,mBAAmBhB,GACrD,QAEO,SAASoB,WAAWpB,GACzB,MAAMmB,EAAI7B,qBAAqBU,GAC/B,OAAU,OAANmB,GAA0C,kBAArBnB,GAAQR,SAA+B,MAC5DO,YAAYC,KACTmB,GAAKnB,EAAOR,WAAaa,gBAAgBL,GAClD,QAEO,SAASqB,WAAWrB,EAAQsB,GACjC,OAAK5B,aAAaM,KAIdD,YAAYC,MAIO,IAAnBsB,GAAMjC,WAAsB6B,WAAWlB,OAIpB,IAAnBsB,GAAM9B,WAAsB4B,WAAWpB,IAXlC,IAgBX,QAEO,SAASuB,SAASvB,GACvB,IAAKN,aAAaM,GAChB,OAAO,KAGT,MAAMC,EAASD,EAAOb,eAChBe,EAASF,EAAOZ,eAEtB,MAC4B,kBAAnBa,GAAQuB,QACW,kBAAnBtB,GAAQsB,OAERvB,GAAQuB,QAAUtB,GAAQsB,OAGL,kBAAnBxB,EAAOyB,SAAyBC,EAAkB1B,GACpDA,EAAOyB,QAGT,IACT,CAEA,SAASC,EAAkB1B,GACzB,MAC4B,kBAAnBA,EAAOyB,SACsB,kBAA7BzB,EAAO2B,mBACuB,kBAA9B3B,EAAO4B,oBACoB,kBAA3B5B,EAAO6B,eAElB,QAEO,SAASC,iBAAiB9B,GAC/B,MAC6B,kBAApBA,EAAO+B,UACdL,EAAkB1B,EAEtB,QAEO,SAASgC,gBAAgBhC,GAC9B,MAC+B,kBAAtBA,EAAOiC,YACY,kBAAnBjC,EAAOkC,cACmBC,IAAjCnC,EAAOoC,KAAKC,gBAEhB,QAEO,SAASC,cAActC,GAC5B,IAAKN,aAAaM,GAAS,OAAO,KAElC,MAAMC,EAASD,EAAOb,eAChBe,EAASF,EAAOZ,eAChBe,EAAQF,GAAUC,EAExB,OAASC,GAAS2B,iBAAiB9B,OACjCG,GACAA,EAAMoC,aACNpC,EAAMqC,YACW,IAAjBrC,EAAMqB,OAEV,QAEO,SAASiB,YAAYzC,GAC1B,SAAUA,KACRA,EAAO0C,iBACP1C,EAAO2C,iBACP3C,EAAOlB,eAEX"}
✄
export const kDestroyed=Symbol("kDestroyed");export const kIsDisturbed=Symbol("kIsDisturbed");export function isReadableNodeStream(e){return!(!e||"function"!=typeof e.pipe||"function"!=typeof e.on||e._writableState&&!1===e._readableState?.readable||e._writableState&&!e._readableState)}export function isWritableNodeStream(e){return!(!e||"function"!=typeof e.write||"function"!=typeof e.on||e._readableState&&!1===e._writableState?.writable)}export function isDuplexNodeStream(e){return!(!e||"function"!=typeof e.pipe||!e._readableState||"function"!=typeof e.on||"function"!=typeof e.write)}export function isNodeStream(e){return e&&(e._readableState||e._writableState||"function"==typeof e.write&&"function"==typeof e.on||"function"==typeof e.pipe&&"function"==typeof e.on)}export function isIterable(e,t){return null!=e&&(!0===t?"function"==typeof e[Symbol.asyncIterator]:!1===t?"function"==typeof e[Symbol.iterator]:"function"==typeof e[Symbol.asyncIterator]||"function"==typeof e[Symbol.iterator])}export function isDestroyed(e){if(!isNodeStream(e))return null;const t=e._writableState,o=e._readableState,n=t||o;return!!(e.destroyed||e[kDestroyed]||n?.destroyed)}export function isWritableEnded(e){if(!isWritableNodeStream(e))return null;if(!0===e.writableEnded)return!0;const t=e._writableState;return!t?.errored&&("boolean"!=typeof t?.ended?null:t.ended)}export function isWritableFinished(e,t){if(!isWritableNodeStream(e))return null;if(!0===e.writableFinished)return!0;const o=e._writableState;return!o?.errored&&("boolean"!=typeof o?.finished?null:!!(o.finished||!1===t&&!0===o.ended&&0===o.length))}export function isReadableEnded(e){if(!isReadableNodeStream(e))return null;if(!0===e.readableEnded)return!0;const t=e._readableState;return!(!t||t.errored)&&("boolean"!=typeof t?.ended?null:t.ended)}export function isReadableFinished(e,t){if(!isReadableNodeStream(e))return null;const o=e._readableState;return!o?.errored&&("boolean"!=typeof o?.endEmitted?null:!!(o.endEmitted||!1===t&&!0===o.ended&&0===o.length))}export function isReadable(e){const t=isReadableNodeStream(e);return null===t||"boolean"!=typeof e?.readable?null:!isDestroyed(e)&&(t&&e.readable&&!isReadableFinished(e))}export function isWritable(e){const t=isWritableNodeStream(e);return null===t||"boolean"!=typeof e?.writable?null:!isDestroyed(e)&&(t&&e.writable&&!isWritableEnded(e))}export function isFinished(e,t){return isNodeStream(e)?!!isDestroyed(e)||(!1===t?.readable||!isReadable(e))&&(!1===t?.writable||!isWritable(e)):null}export function isClosed(t){if(!isNodeStream(t))return null;const o=t._writableState,n=t._readableState;return"boolean"==typeof o?.closed||"boolean"==typeof n?.closed?o?.closed||n?.closed:"boolean"==typeof t._closed&&e(t)?t._closed:null}function e(e){return"boolean"==typeof e._closed&&"boolean"==typeof e._defaultKeepAlive&&"boolean"==typeof e._removedConnection&&"boolean"==typeof e._removedContLen}export function isServerResponse(t){return"boolean"==typeof t._sent100&&e(t)}export function isServerRequest(e){return"boolean"==typeof e._consuming&&"boolean"==typeof e._dumped&&void 0===e.req?.upgradeOrConnect}export function willEmitClose(e){if(!isNodeStream(e))return null;const t=e._writableState,o=e._readableState,n=t||o;return!n&&isServerResponse(e)||!!(n&&n.autoDestroy&&n.emitClose&&!1===n.closed)}export function isDisturbed(e){return!(!e||!(e.readableDidRead||e.readableAborted||e[kIsDisturbed]))}
✄
{"version":3,"file":"writable.js","names":["addAbortSignal","destroyImpl","errorCodes","Stream","getHighWaterMark","getDefaultHighWaterMark","Buffer","EE","process","Writable","ERR_INVALID_ARG_TYPE","ERR_METHOD_NOT_IMPLEMENTED","ERR_MULTIPLE_CALLBACK","ERR_STREAM_CANNOT_PIPE","ERR_STREAM_DESTROYED","ERR_STREAM_ALREADY_FINISHED","ERR_STREAM_NULL_VALUES","ERR_STREAM_WRITE_AFTER_END","ERR_UNKNOWN_ENCODING","errorOrDestroy","nop","Object","setPrototypeOf","prototype","kOnFinished","Symbol","WritableState","options","stream","isDuplex","Duplex","this","objectMode","writableObjectMode","highWaterMark","finalCalled","needDrain","ending","ended","finished","destroyed","noDecode","decodeStrings","defaultEncoding","length","writing","corked","sync","bufferProcessing","onwrite","bind","undefined","writecb","writelen","afterWriteTickInfo","resetBuffer","pendingcb","constructed","prefinished","errorEmitted","emitClose","autoDestroy","errored","closed","closeEmitted","state","buffered","bufferedIndex","allBuffers","allNoop","getBuffer","slice","defineProperty","get","realHasInstance","Function","hasInstance","call","_writableState","write","_write","writev","_writev","destroy","_destroy","final","_final","construct","_construct","signal","clearBuffer","finishMaybe","chunk","encoding","cb","isEncoding","from","_isUint8Array","_uint8ArrayToBuffer","err","nextTick","callback","len","ret","push","writeOrBuffer","doWrite","onwriteError","er","errorBuffer","stack","_readableState","count","afterWriteTick","afterWrite","emit","n","onfinishCallbacks","splice","i","bufferedLength","chunks","needFinish","prefinish","called","onFinish","finish","result","then","callFinal","rState","endEmitted","readable","value","object","pipe","cork","uncork","setDefaultEncoding","toLowerCase","end","Error","defineProperties","set","writable","w","val","writableFinished","writableBuffer","writableEnded","writableNeedDrain","wState","writableHighWaterMark","writableCorked","writableLength","_undestroy","undestroy","captureRejectionSymbol"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/lib/","sources":[""],"mappings":"yBAyBSA,MAAsB,kCACnBC,MAAiB,+BACXC,MAAkB,gCAC3BC,MAAc,yCAErBC,6BACAC,MACK,8BAEEC,MAAc,gBAChBC,MAAQ,gBACRC,MAAa,yBAELC,SAEf,MAAMC,qBACJA,EAAoBC,2BACpBA,EAA0BC,sBAC1BA,EAAqBC,uBACrBA,EAAsBC,qBACtBA,EAAoBC,4BACpBA,EAA2BC,uBAC3BA,EAAsBC,2BACtBA,EAA0BC,qBAC1BA,GACEhB,GAEEiB,eAAEA,GAAmBlB,EAK3B,SAASmB,IAAO,CAHhBC,OAAOC,eAAeb,SAASc,UAAWpB,EAAOoB,WACjDF,OAAOC,eAAeb,SAAUN,GAIhC,MAAMqB,EAAcC,OAAO,sBAEpB,SAASC,cAAcC,EAASC,EAAQC,GAMrB,kBAAbA,IACTA,EAAWD,aAAkBzB,EAAO2B,QAItCC,KAAKC,cAAgBL,IAAWA,EAAQK,YAEpCH,IACFE,KAAKC,WAAaD,KAAKC,eAClBL,IAAWA,EAAQM,qBAK1BF,KAAKG,cAAgBP,EACnBvB,EAAiB2B,KAAMJ,EAAS,wBAAyBE,GACzDxB,GAAwB,GAG1B0B,KAAKI,aAAc,EAGnBJ,KAAKK,WAAY,EAEjBL,KAAKM,QAAS,EAEdN,KAAKO,OAAQ,EAEbP,KAAKQ,UAAW,EAGhBR,KAAKS,WAAY,EAKjB,MAAMC,KAAcd,IAAqC,IAA1BA,EAAQe,eACvCX,KAAKW,eAAiBD,EAKtBV,KAAKY,gBAAmBhB,GAAWA,EAAQgB,iBAAoB,OAK/DZ,KAAKa,OAAS,EAGdb,KAAKc,SAAU,EAGfd,KAAKe,OAAS,EAMdf,KAAKgB,MAAO,EAKZhB,KAAKiB,kBAAmB,EAGxBjB,KAAKkB,QAAUA,EAAQC,UAAKC,EAAWvB,GAGvCG,KAAKqB,QAAU,KAGfrB,KAAKsB,SAAW,EAIhBtB,KAAKuB,mBAAqB,KAE1BC,EAAYxB,MAIZA,KAAKyB,UAAY,EAMjBzB,KAAK0B,aAAc,EAInB1B,KAAK2B,aAAc,EAGnB3B,KAAK4B,cAAe,EAGpB5B,KAAK6B,WAAajC,IAAiC,IAAtBA,EAAQiC,UAGrC7B,KAAK8B,aAAelC,IAAmC,IAAxBA,EAAQkC,YAKvC9B,KAAK+B,QAAU,KAGf/B,KAAKgC,QAAS,EAIdhC,KAAKiC,cAAe,EAEpBjC,KAAKP,GAAe,EACtB,CAEA,SAAS+B,EAAYU,GACnBA,EAAMC,SAAW,GACjBD,EAAME,cAAgB,EACtBF,EAAMG,YAAa,EACnBH,EAAMI,SAAU,CAClB,CAEA3C,cAAcH,UAAU+C,UAAY,WAClC,OAAOvC,KAAKmC,SAASK,MAAMxC,KAAKoC,cAClC,EAEA9C,OAAOmD,eAAe9C,cAAcH,UAAW,uBAAwB,CACrEkD,MACE,OAAO1C,KAAKmC,SAAStB,OAASb,KAAKoC,aACrC,IAGF,MAAMO,EAAkBC,SAASpD,UAAUE,OAAOmD,oBAE3C,SAASnE,SAASkB,GAWvB,MAAME,EAAYE,gBAAgB5B,EAAO2B,OAEzC,IAAKD,IAAa6C,EAAgBG,KAAKpE,SAAUsB,MAC/C,OAAO,IAAItB,SAASkB,GAEtBI,KAAK+C,eAAiB,IAAIpD,cAAcC,EAASI,KAAMF,GAEnDF,IAC2B,mBAAlBA,EAAQoD,QACjBhD,KAAKiD,OAASrD,EAAQoD,OAEM,mBAAnBpD,EAAQsD,SACjBlD,KAAKmD,QAAUvD,EAAQsD,QAEM,mBAApBtD,EAAQwD,UACjBpD,KAAKqD,SAAWzD,EAAQwD,SAEG,mBAAlBxD,EAAQ0D,QACjBtD,KAAKuD,OAAS3D,EAAQ0D,OAES,mBAAtB1D,EAAQ4D,YACjBxD,KAAKyD,WAAa7D,EAAQ4D,WAExB5D,EAAQ8D,QACVzF,EAAe2B,EAAQ8D,OAAQ1D,OAGnC5B,EAAO0E,KAAK9C,KAAMJ,GAElB1B,EAAYsF,UAAUxD,MAAM,KAC1B,MAAMkC,EAAQlC,KAAK+C,eAEdb,EAAMpB,SACT6C,EAAY3D,KAAMkC,GAGpB0B,EAAY5D,KAAMkC,EAAM,GAE5B,CAgBA,SAASe,EAAOpD,EAAQgE,EAAOC,EAAUC,GACvC,MAAM7B,EAAQrC,EAAOkD,eAErB,GAAwB,mBAAbe,EACTC,EAAKD,EACLA,EAAW5B,EAAMtB,oBACZ,CACL,GAAKkD,GAEA,GAAiB,WAAbA,IAA0BvF,EAAOyF,WAAWF,GACnD,MAAM,IAAI3E,EAAqB2E,QAF/BA,EAAW5B,EAAMtB,gBAGD,mBAAPmD,IACTA,EAAK1E,EACT,CAEA,GAAc,OAAVwE,EACF,MAAM,IAAI5E,EACL,IAAKiD,EAAMjC,WAChB,GAAqB,iBAAV4D,GACmB,IAAxB3B,EAAMvB,gBACRkD,EAAQtF,EAAO0F,KAAKJ,EAAOC,GAC3BA,EAAW,eAER,GAAID,aAAiBtF,EAC1BuF,EAAW,aACN,KAAI1F,EAAO8F,cAAcL,GAI9B,MAAM,IAAIlF,EACR,QAAS,CAAC,SAAU,SAAU,cAAekF,GAJ/CA,EAAQzF,EAAO+F,oBAAoBN,GACnCC,EAAW,QAIb,CAGF,IAAIM,EAOJ,OANIlC,EAAM5B,OACR8D,EAAM,IAAIlF,EACDgD,EAAMzB,YACf2D,EAAM,IAAIrF,EAAqB,UAG7BqF,GACF3F,EAAQ4F,SAASN,EAAIK,GACrBhF,EAAeS,EAAQuE,GAAK,GACrBA,IAETlC,EAAMT,YAoCR,SAAuB5B,EAAQqC,EAAO2B,EAAOC,EAAUQ,GACrD,MAAMC,EAAMrC,EAAMjC,WAAa,EAAI4D,EAAMhD,OAEzCqB,EAAMrB,QAAU0D,EAGhB,MAAMC,EAAMtC,EAAMrB,OAASqB,EAAM/B,cAE5BqE,IACHtC,EAAM7B,WAAY,GAEhB6B,EAAMpB,SAAWoB,EAAMnB,QAAUmB,EAAMH,UAAYG,EAAMR,aAC3DQ,EAAMC,SAASsC,KAAK,CAAEZ,QAAOC,WAAUQ,aACnCpC,EAAMG,YAA2B,WAAbyB,IACtB5B,EAAMG,YAAa,GAEjBH,EAAMI,SAAWgC,IAAajF,IAChC6C,EAAMI,SAAU,KAGlBJ,EAAMZ,SAAWiD,EACjBrC,EAAMb,QAAUiD,EAChBpC,EAAMpB,SAAU,EAChBoB,EAAMlB,MAAO,EACbnB,EAAOoD,OAAOY,EAAOC,EAAU5B,EAAMhB,SACrCgB,EAAMlB,MAAO,GAKf,OAAOwD,IAAQtC,EAAMH,UAAYG,EAAMzB,SACzC,CAlESiE,CAAc7E,EAAQqC,EAAO2B,EAAOC,EAAUC,GACvD,CAmEA,SAASY,EAAQ9E,EAAQqC,EAAOgB,EAAQqB,EAAKV,EAAOC,EAAUC,GAC5D7B,EAAMZ,SAAWiD,EACjBrC,EAAMb,QAAU0C,EAChB7B,EAAMpB,SAAU,EAChBoB,EAAMlB,MAAO,EACTkB,EAAMzB,UACRyB,EAAMhB,QAAQ,IAAInC,EAAqB,UAChCmE,EACPrD,EAAOsD,QAAQU,EAAO3B,EAAMhB,SAE5BrB,EAAOoD,OAAOY,EAAOC,EAAU5B,EAAMhB,SACvCgB,EAAMlB,MAAO,CACf,CAEA,SAAS4D,EAAa/E,EAAQqC,EAAO2C,EAAId,KACrC7B,EAAMT,UAERsC,EAAGc,GAKHC,EAAY5C,GAEZ9C,EAAeS,EAAQgF,EACzB,CAEA,SAAS3D,EAAQrB,EAAQgF,GACvB,MAAM3C,EAAQrC,EAAOkD,eACf/B,EAAOkB,EAAMlB,KACb+C,EAAK7B,EAAMb,QAEC,mBAAP0C,GAKX7B,EAAMpB,SAAU,EAChBoB,EAAMb,QAAU,KAChBa,EAAMrB,QAAUqB,EAAMZ,SACtBY,EAAMZ,SAAW,EAEbuD,GAEFA,EAAGE,MAEE7C,EAAMH,UACTG,EAAMH,QAAU8C,GAKdhF,EAAOmF,iBAAmBnF,EAAOmF,eAAejD,UAClDlC,EAAOmF,eAAejD,QAAU8C,GAG9B7D,EACFvC,EAAQ4F,SAASO,EAAc/E,EAAQqC,EAAO2C,EAAId,GAElDa,EAAa/E,EAAQqC,EAAO2C,EAAId,KAG9B7B,EAAMC,SAAStB,OAASqB,EAAME,eAChCuB,EAAY9D,EAAQqC,GAGlBlB,EAK+B,OAA7BkB,EAAMX,oBACNW,EAAMX,mBAAmBwC,KAAOA,EAClC7B,EAAMX,mBAAmB0D,SAEzB/C,EAAMX,mBAAqB,CAAE0D,MAAO,EAAGlB,KAAIlE,SAAQqC,SACnDzD,EAAQ4F,SAASa,EAAgBhD,EAAMX,qBAGzC4D,EAAWtF,EAAQqC,EAAO,EAAG6B,KA9C/B3E,EAAeS,EAAQ,IAAIhB,EAiD/B,CAEA,SAASqG,GAAerF,OAAEA,EAAMqC,MAAEA,EAAK+C,MAAEA,EAAKlB,GAAEA,IAE9C,OADA7B,EAAMX,mBAAqB,KACpB4D,EAAWtF,EAAQqC,EAAO+C,EAAOlB,EAC1C,CAEA,SAASoB,EAAWtF,EAAQqC,EAAO+C,EAAOlB,GAQxC,KAPmB7B,EAAM5B,SAAWT,EAAOY,WAA8B,IAAjByB,EAAMrB,QAC5DqB,EAAM7B,YAEN6B,EAAM7B,WAAY,EAClBR,EAAOuF,KAAK,UAGPH,KAAU,GACf/C,EAAMT,YACNsC,IAGE7B,EAAMzB,WACRqE,EAAY5C,GAGd0B,EAAY/D,EAAQqC,EACtB,CAGA,SAAS4C,EAAY5C,GACnB,GAAIA,EAAMpB,QACR,OAGF,IAAK,IAAIuE,EAAInD,EAAME,cAAeiD,EAAInD,EAAMC,SAAStB,SAAUwE,EAAG,CAChE,MAAMxB,MAAEA,EAAKS,SAAEA,GAAapC,EAAMC,SAASkD,GACrCd,EAAMrC,EAAMjC,WAAa,EAAI4D,EAAMhD,OACzCqB,EAAMrB,QAAU0D,EAChBD,EAASpC,EAAMH,SAAW,IAAIhD,EAAqB,SACrD,CAEA,MAAMuG,EAAoBpD,EAAMzC,GAAa8F,OAAO,GACpD,IAAK,IAAIC,EAAI,EAAGA,EAAIF,EAAkBzE,OAAQ2E,IAC5CF,EAAkBE,GAAGtD,EAAMH,SAAW,IAAIhD,EAAqB,QAGjEyC,EAAYU,EACd,CAGA,SAASyB,EAAY9D,EAAQqC,GAC3B,GAAIA,EAAMnB,QACNmB,EAAMjB,kBACNiB,EAAMzB,YACLyB,EAAMR,YACT,OAGF,MAAMS,SAAEA,EAAQC,cAAEA,EAAanC,WAAEA,GAAeiC,EAC1CuD,EAAiBtD,EAAStB,OAASuB,EAEzC,IAAKqD,EACH,OAGF,IAAID,EAAIpD,EAGR,GADAF,EAAMjB,kBAAmB,EACrBwE,EAAiB,GAAK5F,EAAOsD,QAAS,CACxCjB,EAAMT,WAAagE,EAAiB,EAEpC,MAAMnB,EAAWpC,EAAMI,QAAUjD,EAAO+E,IACtC,IAAK,IAAIiB,EAAIG,EAAGH,EAAIlD,EAAStB,SAAUwE,EACrClD,EAASkD,GAAGf,SAASF,EACvB,EAIIsB,EAASxD,EAAMI,SAAiB,IAANkD,EAAUrD,EAAWA,EAASK,MAAMgD,GACpEE,EAAOrD,WAAaH,EAAMG,WAE1BsC,EAAQ9E,EAAQqC,GAAO,EAAMA,EAAMrB,OAAQ6E,EAAQ,GAAIpB,GAEvD9C,EAAYU,EACd,KAAO,CACL,EAAG,CACD,MAAM2B,MAAEA,EAAKC,SAAEA,EAAQQ,SAAEA,GAAanC,EAASqD,GAC/CrD,EAASqD,KAAO,KAEhBb,EAAQ9E,EAAQqC,GAAO,EADXjC,EAAa,EAAI4D,EAAMhD,OACAgD,EAAOC,EAAUQ,EACtD,OAASkB,EAAIrD,EAAStB,SAAWqB,EAAMpB,SAEnC0E,IAAMrD,EAAStB,OACjBW,EAAYU,GACHsD,EAAI,KACbrD,EAASoD,OAAO,EAAGC,GACnBtD,EAAME,cAAgB,GAEtBF,EAAME,cAAgBoD,CAE1B,CACAtD,EAAMjB,kBAAmB,CAC3B,CAoEA,SAAS0E,EAAWzD,GAClB,OAAQA,EAAM5B,QACN4B,EAAMR,aACW,IAAjBQ,EAAMrB,SACLqB,EAAMH,SACmB,IAA1BG,EAAMC,SAAStB,SACdqB,EAAM1B,WACN0B,EAAMpB,UACNoB,EAAMN,eACNM,EAAMD,YACjB,CAuDA,SAAS2D,EAAU/F,EAAQqC,GACpBA,EAAMP,aAAgBO,EAAM9B,cACF,mBAAlBP,EAAO0D,QAA0BrB,EAAMzB,WAIhDyB,EAAMP,aAAc,EACpB9B,EAAOuF,KAAK,eAJZlD,EAAM9B,aAAc,EAxD1B,SAAmBP,EAAQqC,GACzB,IAAI2D,GAAS,EAEb,SAASC,EAAS1B,GAChB,GAAIyB,EACFzG,EAAeS,EAAQuE,GAAOvF,UAMhC,GAHAgH,GAAS,EAET3D,EAAMT,YACF2C,EAAK,CACP,MAAMkB,EAAoBpD,EAAMzC,GAAa8F,OAAO,GACpD,IAAK,IAAIC,EAAI,EAAGA,EAAIF,EAAkBzE,OAAQ2E,IAC5CF,EAAkBE,GAAGpB,GAEvBhF,EAAeS,EAAQuE,EAAKlC,EAAMlB,KACpC,MAAW2E,EAAWzD,KACpBA,EAAMP,aAAc,EACpB9B,EAAOuF,KAAK,aAIZlD,EAAMT,YACNhD,EAAQ4F,SAAS0B,EAAQlG,EAAQqC,GAErC,CAEAA,EAAMlB,MAAO,EACbkB,EAAMT,YAEN,IACE,MAAMuE,EAASnG,EAAO0D,OAAOuC,GAC7B,GAAc,MAAVE,EAAgB,CAClB,MAAMC,EAAOD,EAAOC,KACA,mBAATA,GACTA,EAAKnD,KACHkD,GACA,WACEvH,EAAQ4F,SAASyB,EAAU,KAC7B,IACA,SAAS1B,GACP3F,EAAQ4F,SAASyB,EAAU1B,EAC7B,GAEN,CAGF,CAFE,MAAOA,GACP0B,EAASjG,EACX,CAEAqC,EAAMlB,MAAO,CACf,CAMMkF,CAAUrG,EAAQqC,IAMxB,CAEA,SAAS0B,EAAY/D,EAAQqC,EAAOlB,GAC9B2E,EAAWzD,KACb0D,EAAU/F,EAAQqC,GACM,IAApBA,EAAMT,WAAmBkE,EAAWzD,KACtCA,EAAMT,YACFT,EACFvC,EAAQ4F,SAAS0B,EAAQlG,EAAQqC,GAEjC6D,EAAOlG,EAAQqC,IAIvB,CAEA,SAAS6D,EAAOlG,EAAQqC,GACtBA,EAAMT,YACNS,EAAM1B,UAAW,EAEjB,MAAM8E,EAAoBpD,EAAMzC,GAAa8F,OAAO,GACpD,IAAK,IAAIC,EAAI,EAAGA,EAAIF,EAAkBzE,OAAQ2E,IAC5CF,EAAkBE,KAKpB,GAFA3F,EAAOuF,KAAK,UAERlD,EAAMJ,YAAa,CAGrB,MAAMqE,EAAStG,EAAOmF,iBACDmB,GACnBA,EAAOrE,cAGNqE,EAAOC,aAAkC,IAApBD,EAAOE,YAG7BxG,EAAOuD,SAEX,CACF,CAhfA9D,OAAOmD,eAAe/D,SAAUgB,OAAOmD,YAAa,CAClDyD,MAAO,SAASC,GACd,QAAI5D,EAAgBG,KAAK9C,KAAMuG,IAC3BvG,OAAStB,WAEN6H,GAAUA,EAAOxD,0BAA0BpD,cACpD,IAIFjB,SAASc,UAAUgH,KAAO,WACxBpH,EAAeY,KAAM,IAAIlB,EAC3B,EAoDAJ,SAASc,UAAUwD,MAAQ,SAASa,EAAOC,EAAUC,GACnD,OAA6C,IAAtCd,EAAOjD,KAAM6D,EAAOC,EAAUC,EACvC,EAEArF,SAASc,UAAUiH,KAAO,WACxBzG,KAAK+C,eAAehC,QACtB,EAEArC,SAASc,UAAUkH,OAAS,WAC1B,MAAMxE,EAAQlC,KAAK+C,eAEfb,EAAMnB,SACRmB,EAAMnB,SAEDmB,EAAMpB,SACT6C,EAAY3D,KAAMkC,GAExB,EAEAxD,SAASc,UAAUmH,mBAAqB,SAA4B7C,GAIlE,GAFwB,iBAAbA,IACTA,EAAWA,EAAS8C,gBACjBrI,EAAOyF,WAAWF,GACrB,MAAM,IAAI3E,EAAqB2E,GAEjC,OADA9D,KAAK+C,eAAenC,gBAAkBkD,EAC/B9D,IACT,EA+NAtB,SAASc,UAAUyD,OAAS,SAASY,EAAOC,EAAUC,GACpD,IAAI/D,KAAKmD,QAGP,MAAM,IAAIvE,EAA2B,YAFrCoB,KAAKmD,QAAQ,CAAC,CAAEU,QAAOC,aAAaC,EAIxC,EAEArF,SAASc,UAAU2D,QAAU,KAE7BzE,SAASc,UAAUqH,IAAM,SAAShD,EAAOC,EAAUC,GACjD,MAAM7B,EAAQlC,KAAK+C,eAWnB,IAAIqB,EAEJ,GAXqB,mBAAVP,GACTE,EAAKF,EACLA,EAAQ,KACRC,EAAW,MACkB,mBAAbA,IAChBC,EAAKD,EACLA,EAAW,MAKTD,QAAuC,CACzC,MAAMW,EAAMvB,EAAOjD,KAAM6D,EAAOC,GAC5BU,aAAesC,QACjB1C,EAAMI,EAEV,CAkCA,OA/BItC,EAAMnB,SACRmB,EAAMnB,OAAS,EACff,KAAK0G,UAGHtC,IAEQlC,EAAMH,SAAYG,EAAM5B,OAUzB4B,EAAM1B,SACf4D,EAAM,IAAIpF,EAA4B,OAC7BkD,EAAMzB,YACf2D,EAAM,IAAIrF,EAAqB,SAN/BmD,EAAM5B,QAAS,EACfsD,EAAY5D,KAAMkC,GAAO,GACzBA,EAAM3B,OAAQ,IAOE,mBAAPwD,IACLK,GAAOlC,EAAM1B,SACf/B,EAAQ4F,SAASN,EAAIK,GAErBlC,EAAMzC,GAAagF,KAAKV,IAIrB/D,IACT,EAwHAV,OAAOyH,iBAAiBrI,SAASc,UAAW,CAE1CiB,UAAW,CACTiC,MACE,QAAO1C,KAAK+C,gBAAiB/C,KAAK+C,eAAetC,SACnD,EACAuG,IAAIV,GAEEtG,KAAK+C,iBACP/C,KAAK+C,eAAetC,UAAY6F,EAEpC,GAGFW,SAAU,CACRvE,MACE,MAAMwE,EAAIlH,KAAK+C,eAKf,SAASmE,IAAoB,IAAfA,EAAED,UAAuBC,EAAEzG,WAAcyG,EAAEnF,SACtDmF,EAAE5G,QAAW4G,EAAE3G,MACpB,EACAyG,IAAIG,GAEEnH,KAAK+C,iBACP/C,KAAK+C,eAAekE,WAAaE,EAErC,GAGFC,iBAAkB,CAChB1E,MACE,QAAO1C,KAAK+C,gBAAiB/C,KAAK+C,eAAevC,QACnD,GAGFN,mBAAoB,CAClBwC,MACE,QAAO1C,KAAK+C,gBAAiB/C,KAAK+C,eAAe9C,UACnD,GAGFoH,eAAgB,CACd3E,MACE,OAAO1C,KAAK+C,gBAAkB/C,KAAK+C,eAAeR,WACpD,GAGF+E,cAAe,CACb5E,MACE,QAAO1C,KAAK+C,gBAAiB/C,KAAK+C,eAAezC,MACnD,GAGFiH,kBAAmB,CACjB7E,MACE,MAAM8E,EAASxH,KAAK+C,eACpB,QAAKyE,KACGA,EAAO/G,YAAc+G,EAAOlH,QAAUkH,EAAOnH,UACvD,GAGFoH,sBAAuB,CACrB/E,MACE,OAAO1C,KAAK+C,gBAAkB/C,KAAK+C,eAAe5C,aACpD,GAGFuH,eAAgB,CACdhF,MACE,OAAO1C,KAAK+C,eAAiB/C,KAAK+C,eAAehC,OAAS,CAC5D,GAGF4G,eAAgB,CACdjF,MACE,OAAO1C,KAAK+C,gBAAkB/C,KAAK+C,eAAelC,MACpD,KAIJ,MAAMuC,EAAUlF,EAAYkF,QAC5B1E,SAASc,UAAU4D,QAAU,SAASgB,EAAKL,GACzC,MAAM7B,EAAQlC,KAAK+C,eAUnB,OAPKb,EAAMzB,YACRyB,EAAME,cAAgBF,EAAMC,SAAStB,QACpCqB,EAAMzC,GAAaoB,SACrBpC,EAAQ4F,SAASS,EAAa5C,GAGhCkB,EAAQN,KAAK9C,KAAMoE,EAAKL,GACjB/D,IACT,EAEAtB,SAASc,UAAUoI,WAAa1J,EAAY2J,UAC5CnJ,SAASc,UAAU6D,SAAW,SAASe,EAAKL,GAC1CA,EAAGK,EACL,EAEA1F,SAASc,UAAUhB,EAAGsJ,wBAA0B,SAAS1D,GACvDpE,KAAKoD,QAAQgB,EACf"}
✄
import{addAbortSignal as t}from"./add-abort-signal.js";import*as e from"./destroy.js";import{codes as i}from"../errors.js";import{Stream as r}from"./legacy.js";import{getHighWaterMark as n,getDefaultHighWaterMark as o}from"./state.js";import{Buffer as s}from"buffer";import l from"events";import f from"process";export default Writable;const{ERR_INVALID_ARG_TYPE:c,ERR_METHOD_NOT_IMPLEMENTED:a,ERR_MULTIPLE_CALLBACK:d,ERR_STREAM_CANNOT_PIPE:u,ERR_STREAM_DESTROYED:h,ERR_STREAM_ALREADY_FINISHED:b,ERR_STREAM_NULL_VALUES:w,ERR_STREAM_WRITE_AFTER_END:p,ERR_UNKNOWN_ENCODING:g}=i,{errorOrDestroy:y}=e;function _(){}Object.setPrototypeOf(Writable.prototype,r.prototype),Object.setPrototypeOf(Writable,r);const S=Symbol("kOnFinished");export function WritableState(t,e,i){"boolean"!=typeof i&&(i=e instanceof r.Duplex),this.objectMode=!(!t||!t.objectMode),i&&(this.objectMode=this.objectMode||!(!t||!t.writableObjectMode)),this.highWaterMark=t?n(this,t,"writableHighWaterMark",i):o(!1),this.finalCalled=!1,this.needDrain=!1,this.ending=!1,this.ended=!1,this.finished=!1,this.destroyed=!1;const s=!(!t||!1!==t.decodeStrings);this.decodeStrings=!s,this.defaultEncoding=t&&t.defaultEncoding||"utf8",this.length=0,this.writing=!1,this.corked=0,this.sync=!0,this.bufferProcessing=!1,this.onwrite=R.bind(void 0,e),this.writecb=null,this.writelen=0,this.afterWriteTickInfo=null,k(this),this.pendingcb=0,this.constructed=!0,this.prefinished=!1,this.errorEmitted=!1,this.emitClose=!t||!1!==t.emitClose,this.autoDestroy=!t||!1!==t.autoDestroy,this.errored=null,this.closed=!1,this.closeEmitted=!1,this[S]=[]}function k(t){t.buffered=[],t.bufferedIndex=0,t.allBuffers=!0,t.allNoop=!0}WritableState.prototype.getBuffer=function(){return this.buffered.slice(this.bufferedIndex)},Object.defineProperty(WritableState.prototype,"bufferedRequestCount",{get(){return this.buffered.length-this.bufferedIndex}});const E=Function.prototype[Symbol.hasInstance];export function Writable(i){const n=this instanceof r.Duplex;if(!n&&!E.call(Writable,this))return new Writable(i);this._writableState=new WritableState(i,this,n),i&&("function"==typeof i.write&&(this._write=i.write),"function"==typeof i.writev&&(this._writev=i.writev),"function"==typeof i.destroy&&(this._destroy=i.destroy),"function"==typeof i.final&&(this._final=i.final),"function"==typeof i.construct&&(this._construct=i.construct),i.signal&&t(i.signal,this)),r.call(this,i),e.construct(this,(()=>{const t=this._writableState;t.writing||D(this,t),A(this,t)}))}function W(t,e,i,n){const o=t._writableState;if("function"==typeof i)n=i,i=o.defaultEncoding;else{if(i){if("buffer"!==i&&!s.isEncoding(i))throw new g(i)}else i=o.defaultEncoding;"function"!=typeof n&&(n=_)}if(null===e)throw new w;if(!o.objectMode)if("string"==typeof e)!1!==o.decodeStrings&&(e=s.from(e,i),i="buffer");else if(e instanceof s)i="buffer";else{if(!r._isUint8Array(e))throw new c("chunk",["string","Buffer","Uint8Array"],e);e=r._uint8ArrayToBuffer(e),i="buffer"}let l;return o.ending?l=new p:o.destroyed&&(l=new h("write")),l?(f.nextTick(n,l),y(t,l,!0),l):(o.pendingcb++,function(t,e,i,r,n){const o=e.objectMode?1:i.length;e.length+=o;const s=e.length<e.highWaterMark;s||(e.needDrain=!0);e.writing||e.corked||e.errored||!e.constructed?(e.buffered.push({chunk:i,encoding:r,callback:n}),e.allBuffers&&"buffer"!==r&&(e.allBuffers=!1),e.allNoop&&n!==_&&(e.allNoop=!1)):(e.writelen=o,e.writecb=n,e.writing=!0,e.sync=!0,t._write(i,r,e.onwrite),e.sync=!1);return s&&!e.errored&&!e.destroyed}(t,o,e,i,n))}function m(t,e,i,r,n,o,s){e.writelen=r,e.writecb=s,e.writing=!0,e.sync=!0,e.destroyed?e.onwrite(new h("write")):i?t._writev(n,e.onwrite):t._write(n,o,e.onwrite),e.sync=!1}function T(t,e,i,r){--e.pendingcb,r(i),x(e),y(t,i)}function R(t,e){const i=t._writableState,r=i.sync,n=i.writecb;"function"==typeof n?(i.writing=!1,i.writecb=null,i.length-=i.writelen,i.writelen=0,e?(e.stack,i.errored||(i.errored=e),t._readableState&&!t._readableState.errored&&(t._readableState.errored=e),r?f.nextTick(T,t,i,e,n):T(t,i,e,n)):(i.buffered.length>i.bufferedIndex&&D(t,i),r?null!==i.afterWriteTickInfo&&i.afterWriteTickInfo.cb===n?i.afterWriteTickInfo.count++:(i.afterWriteTickInfo={count:1,cb:n,stream:t,state:i},f.nextTick(I,i.afterWriteTickInfo)):M(t,i,1,n))):y(t,new d)}function I({stream:t,state:e,count:i,cb:r}){return e.afterWriteTickInfo=null,M(t,e,i,r)}function M(t,e,i,r){for(!e.ending&&!t.destroyed&&0===e.length&&e.needDrain&&(e.needDrain=!1,t.emit("drain"));i-- >0;)e.pendingcb--,r();e.destroyed&&x(e),A(t,e)}function x(t){if(t.writing)return;for(let e=t.bufferedIndex;e<t.buffered.length;++e){const{chunk:i,callback:r}=t.buffered[e],n=t.objectMode?1:i.length;t.length-=n,r(t.errored??new h("write"))}const e=t[S].splice(0);for(let i=0;i<e.length;i++)e[i](t.errored??new h("end"));k(t)}function D(t,e){if(e.corked||e.bufferProcessing||e.destroyed||!e.constructed)return;const{buffered:i,bufferedIndex:r,objectMode:n}=e,o=i.length-r;if(!o)return;let s=r;if(e.bufferProcessing=!0,o>1&&t._writev){e.pendingcb-=o-1;const r=e.allNoop?_:t=>{for(let e=s;e<i.length;++e)i[e].callback(t)},n=e.allNoop&&0===s?i:i.slice(s);n.allBuffers=e.allBuffers,m(t,e,!0,e.length,n,"",r),k(e)}else{do{const{chunk:r,encoding:o,callback:l}=i[s];i[s++]=null;m(t,e,!1,n?1:r.length,r,o,l)}while(s<i.length&&!e.writing);s===i.length?k(e):s>256?(i.splice(0,s),e.bufferedIndex=0):e.bufferedIndex=s}e.bufferProcessing=!1}function j(t){return t.ending&&t.constructed&&0===t.length&&!t.errored&&0===t.buffered.length&&!t.finished&&!t.writing&&!t.errorEmitted&&!t.closeEmitted}function N(t,e){e.prefinished||e.finalCalled||("function"!=typeof t._final||e.destroyed?(e.prefinished=!0,t.emit("prefinish")):(e.finalCalled=!0,function(t,e){let i=!1;function r(r){if(i)y(t,r??d());else if(i=!0,e.pendingcb--,r){const i=e[S].splice(0);for(let t=0;t<i.length;t++)i[t](r);y(t,r,e.sync)}else j(e)&&(e.prefinished=!0,t.emit("prefinish"),e.pendingcb++,f.nextTick(O,t,e))}e.sync=!0,e.pendingcb++;try{const e=t._final(r);if(null!=e){const t=e.then;"function"==typeof t&&t.call(e,(function(){f.nextTick(r,null)}),(function(t){f.nextTick(r,t)}))}}catch(e){r(t)}e.sync=!1}(t,e)))}function A(t,e,i){j(e)&&(N(t,e),0===e.pendingcb&&j(e)&&(e.pendingcb++,i?f.nextTick(O,t,e):O(t,e)))}function O(t,e){e.pendingcb--,e.finished=!0;const i=e[S].splice(0);for(let t=0;t<i.length;t++)i[t]();if(t.emit("finish"),e.autoDestroy){const e=t._readableState;(!e||e.autoDestroy&&(e.endEmitted||!1===e.readable))&&t.destroy()}}Object.defineProperty(Writable,Symbol.hasInstance,{value:function(t){return!!E.call(this,t)||this===Writable&&(t&&t._writableState instanceof WritableState)}}),Writable.prototype.pipe=function(){y(this,new u)},Writable.prototype.write=function(t,e,i){return!0===W(this,t,e,i)},Writable.prototype.cork=function(){this._writableState.corked++},Writable.prototype.uncork=function(){const t=this._writableState;t.corked&&(t.corked--,t.writing||D(this,t))},Writable.prototype.setDefaultEncoding=function(t){if("string"==typeof t&&(t=t.toLowerCase()),!s.isEncoding(t))throw new g(t);return this._writableState.defaultEncoding=t,this},Writable.prototype._write=function(t,e,i){if(!this._writev)throw new a("_write()");this._writev([{chunk:t,encoding:e}],i)},Writable.prototype._writev=null,Writable.prototype.end=function(t,e,i){const r=this._writableState;let n;if("function"==typeof t?(i=t,t=null,e=null):"function"==typeof e&&(i=e,e=null),null!=t){const i=W(this,t,e);i instanceof Error&&(n=i)}return r.corked&&(r.corked=1,this.uncork()),n||(r.errored||r.ending?r.finished?n=new b("end"):r.destroyed&&(n=new h("end")):(r.ending=!0,A(this,r,!0),r.ended=!0)),"function"==typeof i&&(n||r.finished?f.nextTick(i,n):r[S].push(i)),this},Object.defineProperties(Writable.prototype,{destroyed:{get(){return!!this._writableState&&this._writableState.destroyed},set(t){this._writableState&&(this._writableState.destroyed=t)}},writable:{get(){const t=this._writableState;return!(!t||!1===t.writable||t.destroyed||t.errored||t.ending||t.ended)},set(t){this._writableState&&(this._writableState.writable=!!t)}},writableFinished:{get(){return!!this._writableState&&this._writableState.finished}},writableObjectMode:{get(){return!!this._writableState&&this._writableState.objectMode}},writableBuffer:{get(){return this._writableState&&this._writableState.getBuffer()}},writableEnded:{get(){return!!this._writableState&&this._writableState.ending}},writableNeedDrain:{get(){const t=this._writableState;return!!t&&(!t.destroyed&&!t.ending&&t.needDrain)}},writableHighWaterMark:{get(){return this._writableState&&this._writableState.highWaterMark}},writableCorked:{get(){return this._writableState?this._writableState.corked:0}},writableLength:{get(){return this._writableState&&this._writableState.length}}});const P=e.destroy;Writable.prototype.destroy=function(t,e){const i=this._writableState;return!i.destroyed&&(i.bufferedIndex<i.buffered.length||i[S].length)&&f.nextTick(x,i),P.call(this,t,e),this},Writable.prototype._undestroy=e.undestroy,Writable.prototype._destroy=function(t,e){e(t)},Writable.prototype[l.captureRejectionSymbol]=function(t){this.destroy(t)};
✄
{"version":3,"file":"readable.js","names":["addAbortSignal","compose","destroyer","Duplex","eos","LegacyStream","PassThrough","pipeline","promises","Readable","Transform","isDisturbed","Writable","Buffer","promisify","types","finished","destroy","Object","defineProperty","configurable","enumerable","get","custom","Stream","_isUint8Array","isUint8Array","_uint8ArrayToBuffer","from"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/readable-stream/","sources":[""],"mappings":"yBAASA,MAAsB,mCACxBC,MAAa,uCACXC,MAAiB,0BACnBC,MAAY,yBACZC,MAAS,0CACGC,MAAoB,yBAChCC,MAAiB,8BACjBC,MAAc,8BACTC,MAAc,2BACnBC,MAAc,2BACdC,MAAe,2CACbC,MAAmB,wBACrBC,MAAc,qCAEZC,MAAc,6BACdC,WAAWC,MAAa,sBAElBN,SAEbE,iBACAF,YACAJ,kBACAI,cACAG,cACAT,YACAO,eACAJ,iBACAC,cACAP,oBACAI,cACAF,aACAD,aACAO,eAGFH,EAAaM,YAAcA,EAC3BN,EAAaI,SAAWA,EACxBJ,EAAaO,SAAWA,EACxBP,EAAaF,OAASA,EACtBE,EAAaK,UAAYA,EACzBL,EAAaC,YAAcA,EAC3BD,EAAaE,SAAWA,EACxBF,EAAaL,eAAiBA,EAC9BK,EAAaW,SAAWZ,EACxBC,EAAaY,QAAUf,EACvBG,EAAaJ,QAAUA,EAEvBiB,OAAOC,eAAed,EAAc,WAAY,CAC9Ce,cAAc,EACdC,YAAY,EACZC,IAAG,IACMd,IAIXU,OAAOC,eAAeZ,EAAUO,EAAUS,OAAQ,CAChDF,YAAY,EACZC,IAAG,IACMd,EAASD,WAIpBW,OAAOC,eAAef,EAAKU,EAAUS,OAAQ,CAC3CF,YAAY,EACZC,IAAG,IACMd,EAASQ,WAKpBX,EAAamB,OAASnB,EAEtBA,EAAaoB,cAAgBV,EAAMW,aACnCrB,EAAasB,oBAAsBd,EAAOe"}
✄
import{addAbortSignal as r}from"./lib/add-abort-signal.js";import e from"./lib/compose.js";import{destroyer as i}from"./lib/destroy.js";import o from"./lib/duplex.js";import t from"./lib/end-of-stream.js";import{Stream as m}from"./lib/legacy.js";import s from"./lib/passthrough.js";import a from"./lib/pipeline.js";import*as p from"./lib/promises.js";import l from"./lib/readable.js";import f from"./lib/transform.js";import{isDisturbed as b}from"./lib/utils.js";import n from"./lib/writable.js";import{Buffer as d}from"buffer";import{promisify as u,types as j}from"util";export default l;export{b as isDisturbed,l as Stream,m as LegacyStream,l as Readable,n as Writable,o as Duplex,f as Transform,s as PassThrough,a as pipeline,r as addAbortSignal,t as finished,i as destroy,e as compose,p as promises};m.isDisturbed=b,m.Readable=l,m.Writable=n,m.Duplex=o,m.Transform=f,m.PassThrough=s,m.pipeline=a,m.addAbortSignal=r,m.finished=t,m.destroy=i,m.compose=e,Object.defineProperty(m,"promises",{configurable:!0,enumerable:!0,get:()=>p}),Object.defineProperty(a,u.custom,{enumerable:!0,get:()=>p.pipeline}),Object.defineProperty(t,u.custom,{enumerable:!0,get:()=>p.finished}),m.Stream=m,m._isUint8Array=j.isUint8Array,m._uint8ArrayToBuffer=d.from;
✄
{"version":3,"file":"index.js","names":["isDisturbed","LegacyStream","Readable","Writable","Duplex","Transform","PassThrough","pipeline","addAbortSignal","finished","destroy","compose","promises"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/stream/","sources":[""],"mappings":"sBACEA,kBACAC,cACAC,cACAC,YACAC,eACAC,iBACAC,cACAC,oBACAC,cACAC,aACAC,aACAC,cACAC,MACK,iCAEQX,SAEbD,iBACAC,YACAC,cACAC,cACAC,YACAC,eACAC,iBACAC,cACAC,oBACAC,cACAC,aACAC,aACAC"}
✄
import{isDisturbed as e,LegacyStream as r,Readable as a,Writable as t,Duplex as m,Transform as o,PassThrough as p,pipeline as d,addAbortSignal as f,finished as l,destroy as s,compose as x,promises as b}from"readable-stream";export default r;export{e as isDisturbed,r as Stream,a as Readable,t as Writable,m as Duplex,o as Transform,p as PassThrough,d as pipeline,f as addAbortSignal,l as finished,s as destroy,x as compose,b as promises};
✄
{"version":3,"file":"string_decoder.js","names":["Buffer","StringDecoder","isEncoding","encoding","nb","this","enc","nenc","retried","toLowerCase","_normalizeEncoding","undefined","Error","normalizeEncoding","text","utf16Text","end","utf16End","fillLast","utf8FillLast","base64Text","base64End","write","simpleWrite","simpleEnd","lastNeed","lastTotal","lastChar","allocUnsafe","utf8CheckByte","byte","buf","p","r","self","length","utf8CheckExtraBytes","copy","toString","i","c","charCodeAt","slice","n","prototype","total","j","utf8CheckIncomplete"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/string_decoder/lib/","sources":[""],"mappings":"iBAqBSA,MAAc,uBAER,CAAEC,eAEjB,MAAMC,EAAaF,EAAOE,kBAyCnB,SAASD,cAAcE,GAE5B,IAAIC,EACJ,OAFAC,KAAKF,SAVP,SAA2BG,GACzB,MAAMC,EA/BR,SAA4BD,GAC1B,IAAKA,EAAK,MAAO,OACjB,IAAIE,GAAU,EACd,OACE,OAAQF,GACN,IAAK,OACL,IAAK,QACH,MAAO,OACT,IAAK,OACL,IAAK,QACL,IAAK,UACL,IAAK,WACH,MAAO,UACT,IAAK,SACL,IAAK,SACH,MAAO,SACT,IAAK,SACL,IAAK,QACL,IAAK,MACH,OAAOA,EACT,QACE,GAAIE,EAAS,OACbF,GAAO,GAAKA,GAAKG,cACjBD,GAAU,EAGlB,CAKeE,CAAmBJ,GAChC,QAAaK,IAATJ,IAAuBP,EAAOE,aAAeA,IAAeA,EAAWI,IAAO,MAAM,IAAIM,MAAM,qBAAuBN,GACzH,OAAOC,GAAQD,CACjB,CAMkBO,CAAkBV,GAE1BE,KAAKF,UACX,IAAK,UACHE,KAAKS,KAAOC,EACZV,KAAKW,IAAMC,EACXb,EAAK,EACL,MACF,IAAK,OACHC,KAAKa,SAAWC,EAChBf,EAAK,EACL,MACF,IAAK,SACHC,KAAKS,KAAOM,EACZf,KAAKW,IAAMK,EACXjB,EAAK,EACL,MACF,QAGE,OAFAC,KAAKiB,MAAQC,OACblB,KAAKW,IAAMQ,GAGfnB,KAAKoB,SAAW,EAChBpB,KAAKqB,UAAY,EACjBrB,KAAKsB,SAAW3B,EAAO4B,YAAYxB,EACrC,CAmCA,SAASyB,EAAcC,GACrB,OAAIA,GAAQ,IAAa,EAAWA,GAAQ,GAAM,EAAa,EAAWA,GAAQ,GAAM,GAAa,EAAWA,GAAQ,GAAM,GAAa,EACpIA,GAAQ,GAAM,GAAQ,GAAK,CACpC,CA0DA,SAASX,EAAaY,GACpB,MAAMC,EAAI3B,KAAKqB,UAAYrB,KAAKoB,SAC1BQ,EAtBR,SAA6BC,EAAMH,EAAKC,GACtC,GAAwB,MAAV,IAATD,EAAI,IAEP,OADAG,EAAKT,SAAW,EACT,IAET,GAAIS,EAAKT,SAAW,GAAKM,EAAII,OAAS,EAAG,CACvC,GAAwB,MAAV,IAATJ,EAAI,IAEP,OADAG,EAAKT,SAAW,EACT,IAET,GAAIS,EAAKT,SAAW,GAAKM,EAAII,OAAS,GACZ,MAAV,IAATJ,EAAI,IAEP,OADAG,EAAKT,SAAW,EACT,GAGb,CACF,CAKYW,CAAoB/B,KAAM0B,GACpC,YAAUpB,IAANsB,EAAwBA,EACxB5B,KAAKoB,UAAYM,EAAII,QACvBJ,EAAIM,KAAKhC,KAAKsB,SAAUK,EAAG,EAAG3B,KAAKoB,UAC5BpB,KAAKsB,SAASW,SAASjC,KAAKF,SAAU,EAAGE,KAAKqB,aAEvDK,EAAIM,KAAKhC,KAAKsB,SAAUK,EAAG,EAAGD,EAAII,aAClC9B,KAAKoB,UAAYM,EAAII,QACvB,CA0BA,SAASpB,EAAUgB,EAAKQ,GACtB,IAAKR,EAAII,OAASI,GAAK,GAAM,EAAG,CAC9B,MAAMN,EAAIF,EAAIO,SAAS,UAAWC,GAClC,GAAIN,EAAG,CACL,MAAMO,EAAIP,EAAEQ,WAAWR,EAAEE,OAAS,GAClC,GAAIK,GAAK,OAAUA,GAAK,MAKtB,OAJAnC,KAAKoB,SAAW,EAChBpB,KAAKqB,UAAY,EACjBrB,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,GACpC9B,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,GAC7BF,EAAES,MAAM,GAAI,EAEvB,CACA,OAAOT,CACT,CAIA,OAHA5B,KAAKoB,SAAW,EAChBpB,KAAKqB,UAAY,EACjBrB,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,GAC7BJ,EAAIO,SAAS,UAAWC,EAAGR,EAAII,OAAS,EACjD,CAIA,SAASlB,EAASc,GAChB,MAAME,EAAIF,GAAOA,EAAII,OAAS9B,KAAKiB,MAAMS,GAAO,GAChD,GAAI1B,KAAKoB,SAAU,CACjB,MAAMT,EAAMX,KAAKqB,UAAYrB,KAAKoB,SAClC,OAAOQ,EAAI5B,KAAKsB,SAASW,SAAS,UAAW,EAAGtB,EAClD,CACA,OAAOiB,CACT,CAEA,SAASb,EAAWW,EAAKQ,GACvB,MAAMI,GAAKZ,EAAII,OAASI,GAAK,EAC7B,OAAU,IAANI,EAAgBZ,EAAIO,SAAS,SAAUC,IAC3ClC,KAAKoB,SAAW,EAAIkB,EACpBtC,KAAKqB,UAAY,EACP,IAANiB,EACFtC,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,IAEpC9B,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,GACpC9B,KAAKsB,SAAS,GAAKI,EAAIA,EAAII,OAAS,IAE/BJ,EAAIO,SAAS,SAAUC,EAAGR,EAAII,OAASQ,GAChD,CAEA,SAAStB,EAAUU,GACjB,MAAME,EAAIF,GAAOA,EAAII,OAAS9B,KAAKiB,MAAMS,GAAO,GAChD,OAAI1B,KAAKoB,SAAiBQ,EAAI5B,KAAKsB,SAASW,SAAS,SAAU,EAAG,EAAIjC,KAAKoB,UACpEQ,CACT,CAGA,SAASV,EAAYQ,GACnB,OAAOA,EAAIO,SAASjC,KAAKF,SAC3B,CAEA,SAASqB,EAAUO,GACjB,OAAOA,GAAOA,EAAII,OAAS9B,KAAKiB,MAAMS,GAAO,EAC/C,CA7LA9B,cAAc2C,UAAUtB,MAAQ,SAAUS,GACxC,GAAmB,IAAfA,EAAII,OAAc,MAAO,GAC7B,IAAIF,EACAM,EACJ,GAAIlC,KAAKoB,SAAU,CAEjB,GADAQ,EAAI5B,KAAKa,SAASa,QACRpB,IAANsB,EAAiB,MAAO,GAC5BM,EAAIlC,KAAKoB,SACTpB,KAAKoB,SAAW,CAClB,MACEc,EAAI,EAEN,OAAIA,EAAIR,EAAII,OAAeF,EAAIA,EAAI5B,KAAKS,KAAKiB,EAAKQ,GAAKlC,KAAKS,KAAKiB,EAAKQ,GAC/DN,GAAK,EACd,EAEAhC,cAAc2C,UAAU5B,IAwGxB,SAAiBe,GACf,MAAME,EAAIF,GAAOA,EAAII,OAAS9B,KAAKiB,MAAMS,GAAO,GAChD,OAAI1B,KAAKoB,SAAiBQ,EAAI,IACvBA,CACT,EAzGAhC,cAAc2C,UAAU9B,KA0FxB,SAAkBiB,EAAKQ,GACrB,MAAMM,EArER,SAA6BX,EAAMH,EAAKQ,GACtC,IAAIO,EAAIf,EAAII,OAAS,EACrB,GAAIW,EAAIP,EAAG,OAAO,EAClB,IAAInC,EAAKyB,EAAcE,EAAIe,IAC3B,GAAI1C,GAAM,EAER,OADIA,EAAK,IAAG8B,EAAKT,SAAWrB,EAAK,GAC1BA,EAET,KAAM0C,EAAIP,IAAa,IAARnC,EAAW,OAAO,EAEjC,GADAA,EAAKyB,EAAcE,EAAIe,IACnB1C,GAAM,EAER,OADIA,EAAK,IAAG8B,EAAKT,SAAWrB,EAAK,GAC1BA,EAET,KAAM0C,EAAIP,IAAa,IAARnC,EAAW,OAAO,EAEjC,GADAA,EAAKyB,EAAcE,EAAIe,IACnB1C,GAAM,EAIR,OAHIA,EAAK,IACI,IAAPA,EAAUA,EAAK,EAAO8B,EAAKT,SAAWrB,EAAK,GAE1CA,EAET,OAAO,CACT,CA8CgB2C,CAAoB1C,KAAM0B,EAAKQ,GAC7C,IAAKlC,KAAKoB,SAAU,OAAOM,EAAIO,SAAS,OAAQC,GAChDlC,KAAKqB,UAAYmB,EACjB,MAAM7B,EAAMe,EAAII,QAAUU,EAAQxC,KAAKoB,UAEvC,OADAM,EAAIM,KAAKhC,KAAKsB,SAAU,EAAGX,GACpBe,EAAIO,SAAS,OAAQC,EAAGvB,EACjC,EA9FAf,cAAc2C,UAAU1B,SAAW,SAAUa,GAC3C,GAAI1B,KAAKoB,UAAYM,EAAII,OAEvB,OADAJ,EAAIM,KAAKhC,KAAKsB,SAAUtB,KAAKqB,UAAYrB,KAAKoB,SAAU,EAAGpB,KAAKoB,UACzDpB,KAAKsB,SAASW,SAASjC,KAAKF,SAAU,EAAGE,KAAKqB,WAEvDK,EAAIM,KAAKhC,KAAKsB,SAAUtB,KAAKqB,UAAYrB,KAAKoB,SAAU,EAAGM,EAAII,QAC/D9B,KAAKoB,UAAYM,EAAII,MACvB"}
✄
import{Buffer as t}from"buffer";export default{StringDecoder};const e=t.isEncoding;export function StringDecoder(s){let u;switch(this.encoding=function(s){const i=function(t){if(!t)return"utf8";let e=!1;for(;;)switch(t){case"utf8":case"utf-8":return"utf8";case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return"utf16le";case"latin1":case"binary":return"latin1";case"base64":case"ascii":case"hex":return t;default:if(e)return;t=(""+t).toLowerCase(),e=!0}}(s);if(void 0===i&&(t.isEncoding===e||!e(s)))throw new Error("Unknown encoding: "+s);return i||s}(s),this.encoding){case"utf16le":this.text=n,this.end=r,u=4;break;case"utf8":this.fillLast=i,u=4;break;case"base64":this.text=l,this.end=a,u=3;break;default:return this.write=h,void(this.end=o)}this.lastNeed=0,this.lastTotal=0,this.lastChar=t.allocUnsafe(u)}function s(t){return t<=127?0:t>>5==6?2:t>>4==14?3:t>>3==30?4:t>>6==2?-1:-2}function i(t){const e=this.lastTotal-this.lastNeed,s=function(t,e,s){if(128!=(192&e[0]))return t.lastNeed=0,"�";if(t.lastNeed>1&&e.length>1){if(128!=(192&e[1]))return t.lastNeed=1,"�";if(t.lastNeed>2&&e.length>2&&128!=(192&e[2]))return t.lastNeed=2,"�"}}(this,t);return void 0!==s?s:this.lastNeed<=t.length?(t.copy(this.lastChar,e,0,this.lastNeed),this.lastChar.toString(this.encoding,0,this.lastTotal)):(t.copy(this.lastChar,e,0,t.length),void(this.lastNeed-=t.length))}function n(t,e){if((t.length-e)%2==0){const s=t.toString("utf16le",e);if(s){const e=s.charCodeAt(s.length-1);if(e>=55296&&e<=56319)return this.lastNeed=2,this.lastTotal=4,this.lastChar[0]=t[t.length-2],this.lastChar[1]=t[t.length-1],s.slice(0,-1)}return s}return this.lastNeed=1,this.lastTotal=2,this.lastChar[0]=t[t.length-1],t.toString("utf16le",e,t.length-1)}function r(t){const e=t&&t.length?this.write(t):"";if(this.lastNeed){const t=this.lastTotal-this.lastNeed;return e+this.lastChar.toString("utf16le",0,t)}return e}function l(t,e){const s=(t.length-e)%3;return 0===s?t.toString("base64",e):(this.lastNeed=3-s,this.lastTotal=3,1===s?this.lastChar[0]=t[t.length-1]:(this.lastChar[0]=t[t.length-2],this.lastChar[1]=t[t.length-1]),t.toString("base64",e,t.length-s))}function a(t){const e=t&&t.length?this.write(t):"";return this.lastNeed?e+this.lastChar.toString("base64",0,3-this.lastNeed):e}function h(t){return t.toString(this.encoding)}function o(t){return t&&t.length?this.write(t):""}StringDecoder.prototype.write=function(t){if(0===t.length)return"";let e,s;if(this.lastNeed){if(e=this.fillLast(t),void 0===e)return"";s=this.lastNeed,this.lastNeed=0}else s=0;return s<t.length?e?e+this.text(t,s):this.text(t,s):e||""},StringDecoder.prototype.end=function(t){const e=t&&t.length?this.write(t):"";return this.lastNeed?e+"�":e},StringDecoder.prototype.text=function(t,e){const i=function(t,e,i){let n=e.length-1;if(n<i)return 0;let r=s(e[n]);if(r>=0)return r>0&&(t.lastNeed=r-1),r;if(--n<i||-2===r)return 0;if(r=s(e[n]),r>=0)return r>0&&(t.lastNeed=r-2),r;if(--n<i||-2===r)return 0;if(r=s(e[n]),r>=0)return r>0&&(2===r?r=0:t.lastNeed=r-3),r;return 0}(this,t,e);if(!this.lastNeed)return t.toString("utf8",e);this.lastTotal=i;const n=t.length-(i-this.lastNeed);return t.copy(this.lastChar,0,n),t.toString("utf8",e,n)},StringDecoder.prototype.fillLast=function(t){if(this.lastNeed<=t.length)return t.copy(this.lastChar,this.lastTotal-this.lastNeed,0,this.lastNeed),this.lastChar.toString(this.encoding,0,this.lastTotal);t.copy(this.lastChar,this.lastTotal-this.lastNeed,0,t.length),this.lastNeed-=t.length};
✄
{"version":3,"file":"types.js","names":["ObjectToString","uncurryThis","Object","prototype","toString","numberValue","Number","valueOf","stringValue","String","booleanValue","Boolean","bigIntValue","BigInt","symbolValue","Symbol","generatorPrototype","getPrototypeOf","typedArrayPrototype","Int8Array","isArgumentsObject","value","toStringTag","isGeneratorFunction","isTypedArray","isPromise","input","Promise","isArrayBufferView","ArrayBuffer","isView","isUint8Array","Uint8Array","isUint8ClampedArray","Uint8ClampedArray","isUint16Array","Uint16Array","isUint32Array","Uint32Array","isInt8Array","isInt16Array","Int16Array","isInt32Array","Int32Array","isFloat32Array","Float32Array","isFloat64Array","Float64Array","isBigInt64Array","BigInt64Array","isBigUint64Array","BigUint64Array","isMap","isSet","isWeakMap","isWeakSet","isArrayBuffer","isDataView","isSharedArrayBuffer","isAsyncFunction","isMapIterator","isSetIterator","isGeneratorObject","isWebAssemblyCompiledModule","isNumberObject","checkBoxedPrimitive","isStringObject","isBooleanObject","isBigIntObject","isSymbolObject","prototypeValueOf","e","isBoxedPrimitive","isAnyArrayBuffer","isProxy","throwNotSupported","isExternal","isModuleNamespaceObject","method","Error","f","call","bind"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/util/support/","sources":[""],"mappings":"AAGA,MAAMA,EAAiBC,EAAYC,OAAOC,UAAUC,UAE9CC,EAAcJ,EAAYK,OAAOH,UAAUI,SAC3CC,EAAcP,EAAYQ,OAAON,UAAUI,SAC3CG,EAAeT,EAAYU,QAAQR,UAAUI,SAE7CK,EAAcX,EAAYY,OAAOV,UAAUI,SAE3CO,EAAcb,EAAYc,OAAOZ,UAAUI,SAE3CS,EAAqBd,OAAOe,gBAAe,YAAc,IACzDC,EAAsBhB,OAAOe,eAAeE,kBAE3C,SAASC,kBAAkBC,GAChC,OAAc,OAAVA,GAAmC,iBAAVA,KAAsBN,OAAOO,eAAeD,KAGxC,uBAA1BrB,EAAeqB,EACxB,QAEO,SAASE,oBAAoBF,GAClC,OAAOnB,OAAOe,eAAeI,KAAWL,CAC1C,QAEO,SAASQ,aAAaH,GAC3B,OAAOA,aAAiBH,CAC1B,QAEO,SAASO,UAAUC,GACxB,OAAOA,aAAiBC,OAC1B,QAEO,SAASC,kBAAkBP,GAChC,OAAOQ,YAAYC,OAAOT,EAC5B,QAEO,SAASU,aAAaV,GAC3B,OAAOA,aAAiBW,UAC1B,QAEO,SAASC,oBAAoBZ,GAClC,OAAOA,aAAiBa,iBAC1B,QAEO,SAASC,cAAcd,GAC5B,OAAOA,aAAiBe,WAC1B,QAEO,SAASC,cAAchB,GAC5B,OAAOA,aAAiBiB,WAC1B,QAEO,SAASC,YAAYlB,GAC1B,OAAOA,aAAiBF,SAC1B,QAEO,SAASqB,aAAanB,GAC3B,OAAOA,aAAiBoB,UAC1B,QAEO,SAASC,aAAarB,GAC3B,OAAOA,aAAiBsB,UAC1B,QAEO,SAASC,eAAevB,GAC7B,OAAOA,aAAiBwB,YAC1B,QAEO,SAASC,eAAezB,GAC7B,OAAOA,aAAiB0B,YAC1B,QAEO,SAASC,gBAAgB3B,GAC9B,OAAOA,aAAiB4B,aAC1B,QAEO,SAASC,iBAAiB7B,GAC/B,OAAOA,aAAiB8B,cAC1B,QAEO,SAASC,MAAM/B,GACpB,MAAiC,iBAA1BrB,EAAeqB,EACxB,QAEO,SAASgC,MAAMhC,GACpB,MAAiC,iBAA1BrB,EAAeqB,EACxB,QAEO,SAASiC,UAAUjC,GACxB,MAAiC,qBAA1BrB,EAAeqB,EACxB,QAEO,SAASkC,UAAUlC,GACxB,MAAiC,qBAA1BrB,EAAeqB,EACxB,QAEO,SAASmC,cAAcnC,GAC5B,MAAiC,yBAA1BrB,EAAeqB,EACxB,QAEO,SAASoC,WAAWpC,GACzB,MAAiC,sBAA1BrB,EAAeqB,EACxB,QAEO,SAASqC,oBAAoBrC,GAClC,MAAiC,+BAA1BrB,EAAeqB,EACxB,QAEO,SAASsC,gBAAgBtC,GAC9B,MAAiC,2BAA1BrB,EAAeqB,EACxB,QAEO,SAASuC,cAAcvC,GAC5B,MAAiC,0BAA1BrB,EAAeqB,EACxB,QAEO,SAASwC,cAAcxC,GAC5B,MAAiC,0BAA1BrB,EAAeqB,EACxB,QAEO,SAASyC,kBAAkBzC,GAChC,MAAiC,uBAA1BrB,EAAeqB,EACxB,QAEO,SAAS0C,4BAA4B1C,GAC1C,MAAiC,gCAA1BrB,EAAeqB,EACxB,QAEO,SAAS2C,eAAe3C,GAC7B,OAAO4C,EAAoB5C,EAAOhB,EACpC,QAEO,SAAS6D,eAAe7C,GAC7B,OAAO4C,EAAoB5C,EAAOb,EACpC,QAEO,SAAS2D,gBAAgB9C,GAC9B,OAAO4C,EAAoB5C,EAAOX,EACpC,QAEO,SAAS0D,eAAe/C,GAC7B,OAAO4C,EAAoB5C,EAAOT,EACpC,QAEO,SAASyD,eAAehD,GAC7B,OAAO4C,EAAoB5C,EAAOP,EACpC,CAEA,SAASmD,EAAoB5C,EAAOiD,GAClC,GAAqB,iBAAVjD,EACT,OAAO,EAET,IAEE,OADAiD,EAAiBjD,IACV,CAGT,CAFE,MAAMkD,GACN,OAAO,CACT,CACF,QAEO,SAASC,iBAAiBnD,GAC/B,OACE2C,eAAe3C,IACf6C,eAAe7C,IACf8C,gBAAgB9C,IAChB+C,eAAe/C,IACfgD,eAAehD,EAEnB,QAEO,SAASoD,iBAAiBpD,GAC/B,OAAOmC,cAAcnC,IAAUqC,oBAAoBrC,EACrD,QAEO,SAASqD,QAAQrD,GACtBsD,EAAkB,UACpB,QAEO,SAASC,WAAWvD,GACzBsD,EAAkB,aACpB,QAEO,SAASE,wBAAwBxD,GACtCsD,EAAkB,0BACpB,CAEA,SAASA,EAAkBG,GACzB,MAAM,IAAIC,MAAM,GAAGD,iCACrB,CAEA,SAAS7E,EAAY+E,GACnB,OAAOA,EAAEC,KAAKC,KAAKF,EACrB"}
✄
const t=s(Object.prototype.toString),r=s(Number.prototype.valueOf),e=s(String.prototype.valueOf),n=s(Boolean.prototype.valueOf),o=s(BigInt.prototype.valueOf),i=s(Symbol.prototype.valueOf),u=Object.getPrototypeOf((function*(){})),c=Object.getPrototypeOf(Int8Array);export function isArgumentsObject(r){return(null===r||"object"!=typeof r||!(Symbol.toStringTag in r))&&"[object Arguments]"===t(r)}export function isGeneratorFunction(t){return Object.getPrototypeOf(t)===u}export function isTypedArray(t){return t instanceof c}export function isPromise(t){return t instanceof Promise}export function isArrayBufferView(t){return ArrayBuffer.isView(t)}export function isUint8Array(t){return t instanceof Uint8Array}export function isUint8ClampedArray(t){return t instanceof Uint8ClampedArray}export function isUint16Array(t){return t instanceof Uint16Array}export function isUint32Array(t){return t instanceof Uint32Array}export function isInt8Array(t){return t instanceof Int8Array}export function isInt16Array(t){return t instanceof Int16Array}export function isInt32Array(t){return t instanceof Int32Array}export function isFloat32Array(t){return t instanceof Float32Array}export function isFloat64Array(t){return t instanceof Float64Array}export function isBigInt64Array(t){return t instanceof BigInt64Array}export function isBigUint64Array(t){return t instanceof BigUint64Array}export function isMap(r){return"[object Map]"===t(r)}export function isSet(r){return"[object Set]"===t(r)}export function isWeakMap(r){return"[object WeakMap]"===t(r)}export function isWeakSet(r){return"[object WeakSet]"===t(r)}export function isArrayBuffer(r){return"[object ArrayBuffer]"===t(r)}export function isDataView(r){return"[object DataView]"===t(r)}export function isSharedArrayBuffer(r){return"[object SharedArrayBuffer]"===t(r)}export function isAsyncFunction(r){return"[object AsyncFunction]"===t(r)}export function isMapIterator(r){return"[object Map Iterator]"===t(r)}export function isSetIterator(r){return"[object Set Iterator]"===t(r)}export function isGeneratorObject(r){return"[object Generator]"===t(r)}export function isWebAssemblyCompiledModule(r){return"[object WebAssembly.Module]"===t(r)}export function isNumberObject(t){return a(t,r)}export function isStringObject(t){return a(t,e)}export function isBooleanObject(t){return a(t,n)}export function isBigIntObject(t){return a(t,o)}export function isSymbolObject(t){return a(t,i)}function a(t,r){if("object"!=typeof t)return!1;try{return r(t),!0}catch(t){return!1}}export function isBoxedPrimitive(t){return isNumberObject(t)||isStringObject(t)||isBooleanObject(t)||isBigIntObject(t)||isSymbolObject(t)}export function isAnyArrayBuffer(t){return isArrayBuffer(t)||isSharedArrayBuffer(t)}export function isProxy(t){f("isProxy")}export function isExternal(t){f("isExternal")}export function isModuleNamespaceObject(t){f("isModuleNamespaceObject")}function f(t){throw new Error(`${t} is not supported in userland`)}function s(t){return t.call.bind(t)}
✄
{"version":3,"file":"util.js","names":["_types","process","types","isRegExp","isDate","isNativeError","isError","format","deprecate","debuglog","inspect","isArray","isBoolean","isNull","isNullOrUndefined","isNumber","isString","isSymbol","isUndefined","isObject","isFunction","isPrimitive","isBuffer","log","inherits","_extend","promisify","callbackify","formatRegExp","f","objects","i","arguments","length","push","join","args","len","str","String","replace","x","Number","JSON","stringify","_","fn","msg","noDeprecation","warned","throwDeprecation","Error","traceDeprecation","console","trace","error","apply","this","debugs","debugEnvRegex","env","NODE_DEBUG","debugEnv","toUpperCase","RegExp","set","test","pid","obj","opts","ctx","seen","stylize","stylizeNoColor","depth","colors","showHidden","customInspect","stylizeWithColor","formatValue","styleType","style","styles","value","recurseTimes","constructor","prototype","ret","primitive","simple","formatPrimitive","keys","Object","visibleKeys","array","hash","forEach","val","idx","arrayToHash","getOwnPropertyNames","indexOf","formatError","name","toString","call","Date","output","base","braces","toUTCString","l","hasOwnProperty","formatProperty","key","match","formatArray","map","pop","reduce","prev","cur","numLinesEst","reduceToSingleString","desc","getOwnPropertyDescriptor","get","split","line","substr","custom","Symbol","for","bold","italic","underline","inverse","white","grey","black","blue","cyan","green","magenta","red","yellow","special","number","boolean","undefined","null","string","date","regexp","ar","Array","arg","re","objectToString","d","e","Buffer","o","pad","n","months","timestamp","time","getHours","getMinutes","getSeconds","getDate","getMonth","ctor","superCtor","defineProperty","writable","configurable","setPrototypeOf","origin","add","prop","kCustomPromisifiedSymbol","original","TypeError","enumerable","promiseResolve","promiseReject","promise","Promise","resolve","reject","err","getPrototypeOf","defineProperties","getOwnPropertyDescriptors","callbackifyOnRejected","reason","cb","newReason","callbackified","maybeCb","self","then","nextTick","bind","rej"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/@frida/util/","sources":[""],"mappings":"UAqBYA,MAAY,4BAEjBC,MAAa,iBAEb,MAAMC,MAAQ,IAChBF,EACHG,SACAC,OACAC,cAAeC,uBAGF,CACbC,OACAC,UACAC,SACAC,QACAR,MACAS,QACAC,UACAC,OACAC,kBACAC,SACAC,SACAC,SACAC,YACAf,SACAgB,SACAf,OACAE,QACAc,WACAC,YACAC,SACAC,IACAC,SACAC,QACAC,UACAC,aAGF,MAAMC,EAAe,kBAEd,SAASrB,OAAOsB,GACrB,IAAKb,SAASa,GAAI,CAChB,MAAMC,EAAU,GAChB,IAAK,IAAIC,EAAI,EAAGA,EAAIC,UAAUC,OAAQF,IACpCD,EAAQI,KAAKxB,QAAQsB,UAAUD,KAEjC,OAAOD,EAAQK,KAAK,IACtB,CAEA,IAAIJ,EAAI,EACR,MAAMK,EAAOJ,UACPK,EAAMD,EAAKH,OACjB,IAAIK,EAAMC,OAAOV,GAAGW,QAAQZ,GAAc,SAASa,GACjD,GAAU,OAANA,EAAY,MAAO,IACvB,GAAIV,GAAKM,EAAK,OAAOI,EACrB,OAAQA,GACN,IAAK,KAAM,OAAOF,OAAOH,EAAKL,MAC9B,IAAK,KAAM,OAAOW,OAAON,EAAKL,MAC9B,IAAK,KACH,IACE,OAAOY,KAAKC,UAAUR,EAAKL,KAG7B,CAFE,MAAOc,GACP,MAAO,YACT,CACF,QACE,OAAOJ,EAEb,IACA,IAAK,IAAIA,EAAIL,EAAKL,GAAIA,EAAIM,EAAKI,EAAIL,IAAOL,GACpClB,OAAO4B,KAAOtB,SAASsB,GACzBH,GAAO,IAAMG,EAEbH,GAAO,IAAM5B,QAAQ+B,GAGzB,OAAOH,CACT,QAMO,SAAS9B,UAAUsC,EAAIC,GAC5B,IAA8B,IAA1B9C,EAAQ+C,cACV,OAAOF,EAGT,IAAIG,GAAS,EAeb,OAdA,WACE,IAAKA,EAAQ,CACX,GAAIhD,EAAQiD,iBACV,MAAM,IAAIC,MAAMJ,GACP9C,EAAQmD,iBACjBC,QAAQC,MAAMP,GAEdM,QAAQE,MAAMR,GAEhBE,GAAS,CACX,CACA,OAAOH,EAAGU,MAAMC,KAAMzB,UACxB,CAGF,CAGA,MAAM0B,EAAS,CAAC,EAChB,IAAIC,EAAgB,KAEpB,GAAI1D,EAAQ2D,IAAIC,WAAY,CAC1B,IAAIC,EAAW7D,EAAQ2D,IAAIC,WAC3BC,EAAWA,EAAStB,QAAQ,qBAAsB,QAC/CA,QAAQ,MAAO,MACfA,QAAQ,KAAM,OACduB,cACHJ,EAAgB,IAAIK,OAAO,IAAMF,EAAW,IAAK,IACnD,QAEO,SAASrD,SAASwD,GAEvB,GADAA,EAAMA,EAAIF,eACLL,EAAOO,GACV,GAAIN,EAAcO,KAAKD,GAAM,CAC3B,MAAME,EAAMlE,EAAQkE,IACpBT,EAAOO,GAAO,WACZ,MAAMlB,EAAMxC,OAAOiD,MAAM,KAAMxB,WAC/BqB,QAAQE,MAAM,YAAaU,EAAKE,EAAKpB,EACvC,CACF,MACEW,EAAOO,GAAO,WAAY,EAG9B,OAAOP,EAAOO,EAChB,QAWO,SAASvD,QAAQ0D,EAAKC,GAE3B,MAAMC,EAAM,CACVC,KAAM,GACNC,QAASC,GAkBX,OAfIzC,UAAUC,QAAU,IAAGqC,EAAII,MAAQ1C,UAAU,IAC7CA,UAAUC,QAAU,IAAGqC,EAAIK,OAAS3C,UAAU,IAC9CpB,UAAUyD,GAEZC,EAAIM,WAAaP,EACRA,GAET5C,QAAQ6C,EAAKD,GAGXnD,YAAYoD,EAAIM,cAAaN,EAAIM,YAAa,GAC9C1D,YAAYoD,EAAII,SAAQJ,EAAII,MAAQ,GACpCxD,YAAYoD,EAAIK,UAASL,EAAIK,QAAS,GACtCzD,YAAYoD,EAAIO,iBAAgBP,EAAIO,eAAgB,GACpDP,EAAIK,SAAQL,EAAIE,QAAUM,GACvBC,EAAYT,EAAKF,EAAKE,EAAII,MACnC,CAmCA,SAASI,EAAiBxC,EAAK0C,GAC7B,MAAMC,EAAQvE,QAAQwE,OAAOF,GAE7B,OAAIC,EACK,KAAYvE,QAAQiE,OAAOM,GAAO,GAAK,IAAM3C,EAC7C,KAAY5B,QAAQiE,OAAOM,GAAO,GAAK,IAEvC3C,CAEX,CAGA,SAASmC,EAAenC,EAAK0C,GAC3B,OAAO1C,CACT,CAcA,SAASyC,EAAYT,EAAKa,EAAOC,GAG/B,GAAId,EAAIO,eACJM,GACA/D,WAAW+D,EAAMzE,UAEjByE,EAAMzE,UAAYA,WAEhByE,EAAME,aAAeF,EAAME,YAAYC,YAAcH,GAAQ,CACjE,IAAII,EAAMJ,EAAMzE,QAAQ0E,EAAcd,GAItC,OAHKtD,SAASuE,KACZA,EAAMR,EAAYT,EAAKiB,EAAKH,IAEvBG,CACT,CAGA,MAAMC,EA+FR,SAAyBlB,EAAKa,GAC5B,GAAIjE,YAAYiE,GACd,OAAOb,EAAIE,QAAQ,YAAa,aAClC,GAAIxD,SAASmE,GAAQ,CACnB,MAAMM,EAAS,IAAO9C,KAAKC,UAAUuC,GAAO3C,QAAQ,SAAU,IAClBA,QAAQ,KAAM,OACdA,QAAQ,OAAQ,KAAO,IACnE,OAAO8B,EAAIE,QAAQiB,EAAQ,SAC7B,CACA,GAAI1E,SAASoE,GACX,OAAOb,EAAIE,QAAQ,GAAKW,EAAO,UACjC,GAAIvE,UAAUuE,GACZ,OAAOb,EAAIE,QAAQ,GAAKW,EAAO,WAEjC,GAAItE,OAAOsE,GACT,OAAOb,EAAIE,QAAQ,OAAQ,OAC/B,CA/GoBkB,CAAgBpB,EAAKa,GACvC,GAAIK,EACF,OAAOA,EAIT,IAAIG,EAAOC,OAAOD,KAAKR,GACvB,MAAMU,EApCR,SAAqBC,GACnB,MAAMC,EAAO,CAAC,EAMd,OAJAD,EAAME,SAAQ,SAASC,EAAKC,GAC1BH,EAAKE,IAAO,CACd,IAEOF,CACT,CA4BsBI,CAAYR,GAQhC,GANIrB,EAAIM,aACNe,EAAOC,OAAOQ,oBAAoBjB,IAKhC7E,QAAQ6E,KACJQ,EAAKU,QAAQ,YAAc,GAAKV,EAAKU,QAAQ,gBAAkB,GACrE,OAAOC,EAAYnB,GAIrB,GAAoB,IAAhBQ,EAAK1D,OAAc,CACrB,GAAIb,WAAW+D,GAAQ,CACrB,MAAMoB,EAAOpB,EAAMoB,KAAO,KAAOpB,EAAMoB,KAAO,GAC9C,OAAOjC,EAAIE,QAAQ,YAAc+B,EAAO,IAAK,UAC/C,CACA,GAAIpG,SAASgF,GACX,OAAOb,EAAIE,QAAQR,OAAOsB,UAAUkB,SAASC,KAAKtB,GAAQ,UAE5D,GAAI/E,OAAO+E,GACT,OAAOb,EAAIE,QAAQkC,KAAKpB,UAAUkB,SAASC,KAAKtB,GAAQ,QAE1D,GAAI7E,QAAQ6E,GACV,OAAOmB,EAAYnB,EAEvB,CAEA,IA2CIwB,EA3CAC,EAAO,GAAId,GAAQ,EAAOe,EAAS,CAAC,IAAK,KAS7C,GANIlG,QAAQwE,KACVW,GAAQ,EACRe,EAAS,CAAC,IAAK,MAIbzF,WAAW+D,GAAQ,CAErByB,EAAO,cADGzB,EAAMoB,KAAO,KAAOpB,EAAMoB,KAAO,IACjB,GAC5B,CAiBA,OAdIpG,SAASgF,KACXyB,EAAO,IAAM5C,OAAOsB,UAAUkB,SAASC,KAAKtB,IAI1C/E,OAAO+E,KACTyB,EAAO,IAAMF,KAAKpB,UAAUwB,YAAYL,KAAKtB,IAI3C7E,QAAQ6E,KACVyB,EAAO,IAAMN,EAAYnB,IAGP,IAAhBQ,EAAK1D,QAAkB6D,GAAyB,GAAhBX,EAAMlD,OAItCmD,EAAe,EACbjF,SAASgF,GACJb,EAAIE,QAAQR,OAAOsB,UAAUkB,SAASC,KAAKtB,GAAQ,UAEnDb,EAAIE,QAAQ,WAAY,YAInCF,EAAIC,KAAKrC,KAAKiD,GAIZwB,EADEb,EAsCN,SAAqBxB,EAAKa,EAAOC,EAAcS,EAAaF,GAC1D,MAAMgB,EAAS,GACf,IAAK,IAAI5E,EAAI,EAAGgF,EAAI5B,EAAMlD,OAAQF,EAAIgF,IAAKhF,EACrCiF,EAAe7B,EAAO5C,OAAOR,IAC/B4E,EAAOzE,KAAK+E,EAAe3C,EAAKa,EAAOC,EAAcS,EACjDtD,OAAOR,IAAI,IAEf4E,EAAOzE,KAAK,IAShB,OANAyD,EAAKK,SAAQ,SAASkB,GACfA,EAAIC,MAAM,UACbR,EAAOzE,KAAK+E,EAAe3C,EAAKa,EAAOC,EAAcS,EACjDqB,GAAK,GAEb,IACOP,CACT,CAtDaS,CAAY9C,EAAKa,EAAOC,EAAcS,EAAaF,GAEnDA,EAAK0B,KAAI,SAASH,GACzB,OAAOD,EAAe3C,EAAKa,EAAOC,EAAcS,EAAaqB,EAAKpB,EACpE,IAGFxB,EAAIC,KAAK+C,MA6GX,SAA8BX,EAAQC,EAAMC,GAQ1C,GANeF,EAAOY,QAAO,SAASC,EAAMC,GAG1C,OADIA,EAAIpB,QAAQ,OAAS,GAAGqB,EACrBF,EAAOC,EAAIjF,QAAQ,kBAAmB,IAAIP,OAAS,CAC5D,GAAG,GAEU,GACX,OAAO4E,EAAO,IACG,KAATD,EAAc,GAAKA,EAAO,OAC3B,IACAD,EAAOxE,KAAK,SACZ,IACA0E,EAAO,GAGhB,OAAOA,EAAO,GAAKD,EAAO,IAAMD,EAAOxE,KAAK,MAAQ,IAAM0E,EAAO,EACnE,CA7HSc,CAAqBhB,EAAQC,EAAMC,IAxBjCA,EAAO,GAAKD,EAAOC,EAAO,EAyBrC,CAsBA,SAASP,EAAYnB,GACnB,MAAO,IAAMhC,MAAMmC,UAAUkB,SAASC,KAAKtB,GAAS,GACtD,CAuBA,SAAS8B,EAAe3C,EAAKa,EAAOC,EAAcS,EAAaqB,EAAKpB,GAClE,IAAIS,EAAMjE,EAAKsF,EAsCf,GArCAA,EAAOhC,OAAOiC,yBAAyB1C,EAAO+B,IAAQ,CAAE/B,MAAOA,EAAM+B,IACjEU,EAAKE,IAELxF,EADEsF,EAAK3D,IACDK,EAAIE,QAAQ,kBAAmB,WAE/BF,EAAIE,QAAQ,WAAY,WAG5BoD,EAAK3D,MACP3B,EAAMgC,EAAIE,QAAQ,WAAY,YAG7BwC,EAAenB,EAAaqB,KAC/BX,EAAO,IAAMW,EAAM,KAEhB5E,IACCgC,EAAIC,KAAK8B,QAAQuB,EAAKzC,OAAS,GAE/B7C,EADEzB,OAAOuE,GACHL,EAAYT,EAAKsD,EAAKzC,MAAO,MAE7BJ,EAAYT,EAAKsD,EAAKzC,MAAOC,EAAe,GAEhD9C,EAAI+D,QAAQ,OAAS,IAErB/D,EADEwD,EACIxD,EAAIyF,MAAM,MAAMV,KAAI,SAASW,GACjC,MAAO,KAAOA,CAChB,IAAG7F,KAAK,MAAM8F,OAAO,GAEf,KAAO3F,EAAIyF,MAAM,MAAMV,KAAI,SAASW,GACxC,MAAO,MAAQA,CACjB,IAAG7F,KAAK,QAIZG,EAAMgC,EAAIE,QAAQ,aAAc,YAGhCtD,YAAYqF,GAAO,CACrB,GAAIT,GAASoB,EAAIC,MAAM,SACrB,OAAO7E,EAETiE,EAAO5D,KAAKC,UAAU,GAAKsE,GACvBX,EAAKY,MAAM,iCACbZ,EAAOA,EAAK0B,OAAO,EAAG1B,EAAKtE,OAAS,GACpCsE,EAAOjC,EAAIE,QAAQ+B,EAAM,UAEzBA,EAAOA,EAAK/D,QAAQ,KAAM,OACdA,QAAQ,OAAQ,KAChBA,QAAQ,WAAY,KAChC+D,EAAOjC,EAAIE,QAAQ+B,EAAM,UAE7B,CAEA,OAAOA,EAAO,KAAOjE,CACvB,CAnRA5B,QAAQwH,OAASC,OAAOC,IAAI,8BAI5B1H,QAAQiE,OAAS,CACf0D,KAAS,CAAC,EAAG,IACbC,OAAW,CAAC,EAAG,IACfC,UAAc,CAAC,EAAG,IAClBC,QAAY,CAAC,EAAG,IAChBC,MAAU,CAAC,GAAI,IACfC,KAAS,CAAC,GAAI,IACdC,MAAU,CAAC,GAAI,IACfC,KAAS,CAAC,GAAI,IACdC,KAAS,CAAC,GAAI,IACdC,MAAU,CAAC,GAAI,IACfC,QAAY,CAAC,GAAI,IACjBC,IAAQ,CAAC,GAAI,IACbC,OAAW,CAAC,GAAI,KAIlBvI,QAAQwE,OAAS,CACfgE,QAAW,OACXC,OAAU,SACVC,QAAW,SACXC,UAAa,OACbC,KAAQ,OACRC,OAAU,QACVC,KAAQ,UAERC,OAAU,cA4QL,SAAS9I,QAAQ+I,GACtB,OAAOC,MAAMhJ,QAAQ+I,EACvB,QAEO,SAAS9I,UAAUgJ,GACxB,MAAsB,kBAARA,CAChB,QAEO,SAAS/I,OAAO+I,GACrB,OAAe,OAARA,CACT,QAEO,SAAS9I,kBAAkB8I,GAChC,OAAc,MAAPA,CACT,QAEO,SAAS7I,SAAS6I,GACvB,MAAsB,iBAARA,CAChB,QAEO,SAAS5I,SAAS4I,GACvB,MAAsB,iBAARA,CAChB,QAEO,SAAS3I,SAAS2I,GACvB,MAAsB,iBAARA,CAChB,QAEO,SAAS1I,YAAY0I,GAC1B,YAAe,IAARA,CACT,QAEO,SAASzJ,SAAS0J,GACvB,OAAO1I,SAAS0I,IAA8B,oBAAvBC,EAAeD,EACxC,QAEO,SAAS1I,SAASyI,GACvB,MAAsB,iBAARA,GAA4B,OAARA,CACpC,QAEO,SAASxJ,OAAO2J,GACrB,OAAO5I,SAAS4I,IAA4B,kBAAtBD,EAAeC,EACvC,QAEO,SAASzJ,QAAQ0J,GACtB,OAAO7I,SAAS6I,KACW,mBAAtBF,EAAeE,IAA2BA,aAAa7G,MAC9D,QAEO,SAAS/B,WAAWwI,GACzB,MAAsB,mBAARA,CAChB,QAEO,SAASvI,YAAYuI,GAC1B,OAAe,OAARA,GACe,kBAARA,GACQ,iBAARA,GACQ,iBAARA,GACQ,iBAARA,QACQ,IAARA,CAChB,QAEO,SAAStI,SAASsI,GACvB,OAAOA,aAAeK,MACxB,CAEA,SAASH,EAAeI,GACtB,OAAOtE,OAAON,UAAUkB,SAASC,KAAKyD,EACxC,CAGA,SAASC,EAAIC,GACX,OAAOA,EAAI,GAAK,IAAMA,EAAE5D,SAAS,IAAM4D,EAAE5D,SAAS,GACpD,CAGA,MAAM6D,EAAS,CAAC,MAAO,MAAO,MAAO,MAAO,MAAO,MAAO,MAAO,MAAO,MACxD,MAAO,MAAO,OAG9B,SAASC,IACP,MAAMP,EAAI,IAAIrD,KACR6D,EAAO,CAACJ,EAAIJ,EAAES,YACNL,EAAIJ,EAAEU,cACNN,EAAIJ,EAAEW,eAAevI,KAAK,KACxC,MAAO,CAAC4H,EAAEY,UAAWN,EAAON,EAAEa,YAAaL,GAAMpI,KAAK,IACxD,QAIO,SAASZ,MACd8B,QAAQ9B,IAAI,UAAW+I,IAAa/J,OAAOiD,MAAM,KAAMxB,WACzD,QAgBO,SAASR,SAASqJ,EAAMC,GAC7BlF,OAAOmF,eAAeF,EAAM,SAAU,CACpC1F,MAAO2F,EACPE,UAAU,EACVC,cAAc,IAEhBrF,OAAOsF,eAAeL,EAAKvF,UAAWwF,EAAUxF,UAClD,QAEO,SAAS7D,QAAQ0J,EAAQC,GAE9B,IAAKA,IAAQjK,SAASiK,GAAM,OAAOD,EAEnC,MAAMxF,EAAOC,OAAOD,KAAKyF,GACzB,IAAIrJ,EAAI4D,EAAK1D,OACb,KAAOF,KACLoJ,EAAOxF,EAAK5D,IAAMqJ,EAAIzF,EAAK5D,IAE7B,OAAOoJ,CACT,CAEA,SAASnE,EAAe5C,EAAKiH,GAC3B,OAAOzF,OAAON,UAAU0B,eAAeP,KAAKrC,EAAKiH,EACnD,CAEA,MAAMC,EAA2BnD,OAAO,gCAEjC,SAASzG,UAAU6J,GACxB,GAAwB,mBAAbA,EACT,MAAM,IAAIC,UAAU,oDAEtB,GAAIF,GAA4BC,EAASD,GAA2B,CAClE,MAAMxI,EAAKyI,EAASD,GACpB,GAAkB,mBAAPxI,EACT,MAAM,IAAI0I,UAAU,iEAKtB,OAHA5F,OAAOmF,eAAejI,EAAIwI,EAA0B,CAClDnG,MAAOrC,EAAI2I,YAAY,EAAOT,UAAU,EAAOC,cAAc,IAExDnI,CACT,CAEA,SAASA,IACP,IAAI4I,EAAgBC,EACpB,MAAMC,EAAU,IAAIC,SAAQ,SAAUC,EAASC,GAC7CL,EAAiBI,EACjBH,EAAgBI,CAClB,IAEM3J,EAAO,GACb,IAAK,IAAIL,EAAI,EAAGA,EAAIC,UAAUC,OAAQF,IACpCK,EAAKF,KAAKF,UAAUD,IAEtBK,EAAKF,MAAK,SAAU8J,EAAK7G,GACnB6G,EACFL,EAAcK,GAEdN,EAAevG,EAEnB,IAEA,IACEoG,EAAS/H,MAAMC,KAAMrB,EAGvB,CAFE,MAAO4J,GACPL,EAAcK,EAChB,CAEA,OAAOJ,CACT,CAOA,OALAhG,OAAOsF,eAAepI,EAAI8C,OAAOqG,eAAeV,IAE5CD,GAA0B1F,OAAOmF,eAAejI,EAAIwI,EAA0B,CAChFnG,MAAOrC,EAAI2I,YAAY,EAAOT,UAAU,EAAOC,cAAc,IAExDrF,OAAOsG,iBACZpJ,EACA8C,OAAOuG,0BAA0BZ,GAErC,CAIA,SAASa,EAAsBC,EAAQC,GAKrC,IAAKD,EAAQ,CACX,MAAME,EAAY,IAAIpJ,MAAM,2CAC5BoJ,EAAUF,OAASA,EACnBA,EAASE,CACX,CACA,OAAOD,EAAGD,EACZ,CAbA3K,UAAUwG,OAASoD,SAeZ,SAAS3J,YAAY4J,GAC1B,GAAwB,mBAAbA,EACT,MAAM,IAAIC,UAAU,oDAMtB,SAASgB,IACP,MAAMpK,EAAO,GACb,IAAK,IAAIL,EAAI,EAAGA,EAAIC,UAAUC,OAAQF,IACpCK,EAAKF,KAAKF,UAAUD,IAGtB,MAAM0K,EAAUrK,EAAKkF,MACrB,GAAuB,mBAAZmF,EACT,MAAM,IAAIjB,UAAU,8CAEtB,MAAMkB,EAAOjJ,KACP6I,EAAK,WACT,OAAOG,EAAQjJ,MAAMkJ,EAAM1K,UAC7B,EAGAuJ,EAAS/H,MAAMC,KAAMrB,GAClBuK,MAAK,SAASpH,GAAOtF,EAAQ2M,SAASN,EAAGO,KAAK,KAAM,KAAMtH,GAAM,IAC3D,SAASuH,GAAO7M,EAAQ2M,SAASR,EAAsBS,KAAK,KAAMC,EAAKR,GAAK,GACtF,CAKA,OAHA1G,OAAOsF,eAAesB,EAAe5G,OAAOqG,eAAeV,IAC3D3F,OAAOsG,iBAAiBM,EACA5G,OAAOuG,0BAA0BZ,IAClDiB,CACT"}
✄
import*as e from"./support/types.js";import t from"process";export const types={...e,isRegExp,isDate,isNativeError:isError};export default{format,deprecate,debuglog,inspect,types,isArray,isBoolean,isNull,isNullOrUndefined,isNumber,isString,isSymbol,isUndefined,isRegExp,isObject,isDate,isError,isFunction,isPrimitive,isBuffer,log,inherits,_extend,promisify,callbackify};const n=/%[sdj%]/g;export function format(e){if(!isString(e)){const e=[];for(let t=0;t<arguments.length;t++)e.push(inspect(arguments[t]));return e.join(" ")}let t=1;const r=arguments,i=r.length;let o=String(e).replace(n,(function(e){if("%%"===e)return"%";if(t>=i)return e;switch(e){case"%s":return String(r[t++]);case"%d":return Number(r[t++]);case"%j":try{return JSON.stringify(r[t++])}catch(e){return"[Circular]"}default:return e}}));for(let e=r[t];t<i;e=r[++t])isNull(e)||!isObject(e)?o+=" "+e:o+=" "+inspect(e);return o}export function deprecate(e,n){if(!0===t.noDeprecation)return e;let r=!1;return function(){if(!r){if(t.throwDeprecation)throw new Error(n);t.traceDeprecation?console.trace(n):console.error(n),r=!0}return e.apply(this,arguments)}}const r={};let i=/^$/;if(t.env.NODE_DEBUG){let m=t.env.NODE_DEBUG;m=m.replace(/[|\\{}()[\]^$+?.]/g,"\\$&").replace(/\*/g,".*").replace(/,/g,"$|^").toUpperCase(),i=new RegExp("^"+m+"$","i")}export function debuglog(e){if(e=e.toUpperCase(),!r[e])if(i.test(e)){const n=t.pid;r[e]=function(){const t=format.apply(null,arguments);console.error("%s %d: %s",e,n,t)}}else r[e]=function(){};return r[e]}export function inspect(e,t){const n={seen:[],stylize:s};return arguments.length>=3&&(n.depth=arguments[2]),arguments.length>=4&&(n.colors=arguments[3]),isBoolean(t)?n.showHidden=t:t&&_extend(n,t),isUndefined(n.showHidden)&&(n.showHidden=!1),isUndefined(n.depth)&&(n.depth=2),isUndefined(n.colors)&&(n.colors=!1),isUndefined(n.customInspect)&&(n.customInspect=!0),n.colors&&(n.stylize=o),c(n,e,n.depth)}function o(e,t){const n=inspect.styles[t];return n?"["+inspect.colors[n][0]+"m"+e+"["+inspect.colors[n][1]+"m":e}function s(e,t){return e}function c(e,t,n){if(e.customInspect&&t&&isFunction(t.inspect)&&t.inspect!==inspect&&(!t.constructor||t.constructor.prototype!==t)){let r=t.inspect(n,e);return isString(r)||(r=c(e,r,n)),r}const r=function(e,t){if(isUndefined(t))return e.stylize("undefined","undefined");if(isString(t)){const n="'"+JSON.stringify(t).replace(/^"|"$/g,"").replace(/'/g,"\\'").replace(/\\"/g,'"')+"'";return e.stylize(n,"string")}if(isNumber(t))return e.stylize(""+t,"number");if(isBoolean(t))return e.stylize(""+t,"boolean");if(isNull(t))return e.stylize("null","null")}(e,t);if(r)return r;let i=Object.keys(t);const o=function(e){const t={};return e.forEach((function(e,n){t[e]=!0})),t}(i);if(e.showHidden&&(i=Object.getOwnPropertyNames(t)),isError(t)&&(i.indexOf("message")>=0||i.indexOf("description")>=0))return u(t);if(0===i.length){if(isFunction(t)){const n=t.name?": "+t.name:"";return e.stylize("[Function"+n+"]","special")}if(isRegExp(t))return e.stylize(RegExp.prototype.toString.call(t),"regexp");if(isDate(t))return e.stylize(Date.prototype.toString.call(t),"date");if(isError(t))return u(t)}let s,p="",f=!1,a=["{","}"];if(isArray(t)&&(f=!0,a=["[","]"]),isFunction(t)){p=" [Function"+(t.name?": "+t.name:"")+"]"}return isRegExp(t)&&(p=" "+RegExp.prototype.toString.call(t)),isDate(t)&&(p=" "+Date.prototype.toUTCString.call(t)),isError(t)&&(p=" "+u(t)),0!==i.length||f&&0!=t.length?n<0?isRegExp(t)?e.stylize(RegExp.prototype.toString.call(t),"regexp"):e.stylize("[Object]","special"):(e.seen.push(t),s=f?function(e,t,n,r,i){const o=[];for(let i=0,s=t.length;i<s;++i)g(t,String(i))?o.push(l(e,t,n,r,String(i),!0)):o.push("");return i.forEach((function(i){i.match(/^\d+$/)||o.push(l(e,t,n,r,i,!0))})),o}(e,t,n,o,i):i.map((function(r){return l(e,t,n,o,r,f)})),e.seen.pop(),function(e,t,n){if(e.reduce((function(e,t){return t.indexOf("\n")>=0&&0,e+t.replace(/\u001b\[\d\d?m/g,"").length+1}),0)>60)return n[0]+(""===t?"":t+"\n ")+" "+e.join(",\n  ")+" "+n[1];return n[0]+t+" "+e.join(", ")+" "+n[1]}(s,p,a)):a[0]+p+a[1]}function u(e){return"["+Error.prototype.toString.call(e)+"]"}function l(e,t,n,r,i,o){let s,u,l;if(l=Object.getOwnPropertyDescriptor(t,i)||{value:t[i]},l.get?u=l.set?e.stylize("[Getter/Setter]","special"):e.stylize("[Getter]","special"):l.set&&(u=e.stylize("[Setter]","special")),g(r,i)||(s="["+i+"]"),u||(e.seen.indexOf(l.value)<0?(u=isNull(n)?c(e,l.value,null):c(e,l.value,n-1),u.indexOf("\n")>-1&&(u=o?u.split("\n").map((function(e){return"  "+e})).join("\n").substr(2):"\n"+u.split("\n").map((function(e){return"   "+e})).join("\n"))):u=e.stylize("[Circular]","special")),isUndefined(s)){if(o&&i.match(/^\d+$/))return u;s=JSON.stringify(""+i),s.match(/^"([a-zA-Z_][a-zA-Z_0-9]*)"$/)?(s=s.substr(1,s.length-2),s=e.stylize(s,"name")):(s=s.replace(/'/g,"\\'").replace(/\\"/g,'"').replace(/(^"|"$)/g,"'"),s=e.stylize(s,"string"))}return s+": "+u}inspect.custom=Symbol.for("nodejs.util.inspect.custom"),inspect.colors={bold:[1,22],italic:[3,23],underline:[4,24],inverse:[7,27],white:[37,39],grey:[90,39],black:[30,39],blue:[34,39],cyan:[36,39],green:[32,39],magenta:[35,39],red:[31,39],yellow:[33,39]},inspect.styles={special:"cyan",number:"yellow",boolean:"yellow",undefined:"grey",null:"bold",string:"green",date:"magenta",regexp:"red"};export function isArray(e){return Array.isArray(e)}export function isBoolean(e){return"boolean"==typeof e}export function isNull(e){return null===e}export function isNullOrUndefined(e){return null==e}export function isNumber(e){return"number"==typeof e}export function isString(e){return"string"==typeof e}export function isSymbol(e){return"symbol"==typeof e}export function isUndefined(e){return void 0===e}export function isRegExp(e){return isObject(e)&&"[object RegExp]"===p(e)}export function isObject(e){return"object"==typeof e&&null!==e}export function isDate(e){return isObject(e)&&"[object Date]"===p(e)}export function isError(e){return isObject(e)&&("[object Error]"===p(e)||e instanceof Error)}export function isFunction(e){return"function"==typeof e}export function isPrimitive(e){return null===e||"boolean"==typeof e||"number"==typeof e||"string"==typeof e||"symbol"==typeof e||void 0===e}export function isBuffer(e){return e instanceof Buffer}function p(e){return Object.prototype.toString.call(e)}function f(e){return e<10?"0"+e.toString(10):e.toString(10)}const a=["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"];function y(){const e=new Date,t=[f(e.getHours()),f(e.getMinutes()),f(e.getSeconds())].join(":");return[e.getDate(),a[e.getMonth()],t].join(" ")}export function log(){console.log("%s - %s",y(),format.apply(null,arguments))}export function inherits(e,t){Object.defineProperty(e,"super_",{value:t,writable:!0,configurable:!0}),Object.setPrototypeOf(e.prototype,t.prototype)}export function _extend(e,t){if(!t||!isObject(t))return e;const n=Object.keys(t);let r=n.length;for(;r--;)e[n[r]]=t[n[r]];return e}function g(e,t){return Object.prototype.hasOwnProperty.call(e,t)}const d=Symbol("util.promisify.custom");export function promisify(e){if("function"!=typeof e)throw new TypeError('The "original" argument must be of type Function');if(d&&e[d]){const t=e[d];if("function"!=typeof t)throw new TypeError('The "util.promisify.custom" argument must be of type Function');return Object.defineProperty(t,d,{value:t,enumerable:!1,writable:!1,configurable:!0}),t}function t(){let t,n;const r=new Promise((function(e,r){t=e,n=r})),i=[];for(let e=0;e<arguments.length;e++)i.push(arguments[e]);i.push((function(e,r){e?n(e):t(r)}));try{e.apply(this,i)}catch(e){n(e)}return r}return Object.setPrototypeOf(t,Object.getPrototypeOf(e)),d&&Object.defineProperty(t,d,{value:t,enumerable:!1,writable:!1,configurable:!0}),Object.defineProperties(t,Object.getOwnPropertyDescriptors(e))}function b(e,t){if(!e){const t=new Error("Promise was rejected with a falsy value");t.reason=e,e=t}return t(e)}promisify.custom=d;export function callbackify(e){if("function"!=typeof e)throw new TypeError('The "original" argument must be of type Function');function n(){const n=[];for(let e=0;e<arguments.length;e++)n.push(arguments[e]);const r=n.pop();if("function"!=typeof r)throw new TypeError("The last argument must be of type Function");const i=this,o=function(){return r.apply(i,arguments)};e.apply(this,n).then((function(e){t.nextTick(o.bind(null,null,e))}),(function(e){t.nextTick(b.bind(null,e,o))}))}return Object.setPrototypeOf(n,Object.getPrototypeOf(e)),Object.defineProperties(n,Object.getOwnPropertyDescriptors(e)),n}
✄
{"version":3,"file":"index.js","names":["_ReadStream_input","_ReadStream_readRequest","_WriteStream_output","_WriteStream_writeRequest","__classPrivateFieldSet","this","receiver","state","value","kind","f","TypeError","has","call","set","__classPrivateFieldGet","get","Buffer","fsPath","process","stream","getWindowsApi","memoize","makeApi","SF","NF","getPosixApi","offsetType","ssizeType","sizeType","invokeXstat","platform","pointerSize","Process","isWindows","constants","S_IFMT","S_IFREG","S_IFDIR","S_IFCHR","S_IFBLK","S_IFIFO","S_IFLNK","S_IFSOCK","S_IRWXU","S_IRUSR","S_IWUSR","S_IXUSR","S_IRWXG","S_IRGRP","S_IWGRP","S_IXGRP","S_IRWXO","S_IROTH","S_IWOTH","S_IXOTH","DT_UNKNOWN","DT_FIFO","DT_CHR","DT_DIR","DT_BLK","DT_REG","DT_LNK","DT_SOCK","DT_WHT","darwin","O_RDONLY","O_WRONLY","O_RDWR","O_CREAT","O_EXCL","O_NOCTTY","O_TRUNC","O_APPEND","O_DIRECTORY","O_NOFOLLOW","O_SYNC","O_DSYNC","O_SYMLINK","O_NONBLOCK","linux","O_NOATIME","O_DIRECT","INVALID_HANDLE_VALUE","ptr","ReadStream","Readable","constructor","path","super","highWaterMark","result","CreateFileW","Memory","allocUtf16String","NULL","handle","equals","nextTick","destroy","makeWindowsError","lastError","Win32InputStream","autoClose","open","allocUtf8String","fd","makePosixError","errno","UnixInputStream","_destroy","error","callback","close","_read","size","read","then","buffer","byteLength","push","from","catch","WeakMap","WriteStream","Writable","FILE_ATTRIBUTE_NORMAL","Win32OutputStream","api","pathStr","flags","mode","UnixOutputStream","_write","chunk","encoding","writeAll","windowsBackend","enumerateDirectoryEntries","enumerateWindowsDirectoryEntriesMatching","readFileSync","options","GetFileSizeEx","ReadFile","CloseHandle","createRes","throwWindowsError","scratchBuf","alloc","fileSizeBuf","getRes","fileSize","readU64","valueOf","buf","numBytesReadBuf","readRes","readU32","Error","parseReadFileResult","readlinkSync","GetFinalPathNameByHandleW","FILE_SHARE_READ","maxLength","readUtf16String","substring","rmdirSync","RemoveDirectoryW","unlinkSync","DeleteFileW","statSync","s","lstatSync","isSymbolicLink","target","GetFileAttributesExW","fileAttrData","data","dup","makeStatsProxy","filename","FindFirstFileW","FindNextFileW","FindClose","posixBackend","opendir","opendir$INODE64","closedir","readdir","readdir$INODE64","readdirImpl","dir","dirHandle","isNull","throwPosixError","entry","lseek","openResult","readResult","n","readFailed","linkSize","readlink","readUtf8String","rmdir","unlink","performStatPosix","getStatSpec","_stat","_lstat","writeFileSync","rawData","encodingIsUtf8","file","File","write","impl","statBufSize","readByteArray","toString","backend","direntSpecs","windows","d_name","d_type","readWindowsFileAttributes","atime","readWindowsFileTime","mtime","ctime","readWindowsFileSize","direntSpec","readdirSync","entries","name","readDirentField","list","extraFieldNames","Object","keys","filter","k","startsWith","type","join","extras","args","fieldSpec","offset","NativePointer","prototype","add","Int64","UInt64","statFields","Set","statSpecs","fields","dev","returnZero","nlink","returnOne","ino","uid","gid","rdev","birthtime","blocks","blksize","readTimespec32","readTimespec64","cachedStatSpec","statSpec","stat64Impl","stat64","__xstat64","platformId","undefined","stat","lstat64","__lxstat64","lstat","Stats","isFile","isDirectory","isCharacterDevice","isBlockDevice","isFIFO","isSocket","Proxy","property","statsHasField","statsReadField","ownKeys","Array","getOwnPropertyDescriptor","writable","configurable","enumerable","field","msPos","lastIndexOf","length","getTime","attributes","isLink","reserved0","isDir","fileTime","BigInt","Date","parseInt","high","low","uint64","shl","or","sec","nsec","FormatMessageW","FORMAT_MESSAGE_FROM_SYSTEM","message","strerror","callbackify","original","numArgsMinusOne","implArgs","slice","e","SystemFunction","NativeFunction","spec","reduce","defineProperty","Ctor","retType","argTypes","wrapper","address","Module","findExportByName","nativeOpts","bind","addApiPlaceholder","abi","createReadStream","createWriteStream","readFile","writeFile","compute","computed"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/frida-fs/dist/","sources":[""],"mappings":"AAAA,IAWIA,EAAmBC,EAAyBC,EAAqBC,EAXjEC,EAA0BC,MAAQA,KAAKD,wBAA2B,SAAUE,EAAUC,EAAOC,EAAOC,EAAMC,GAC1G,GAAa,MAATD,EAAc,MAAM,IAAIE,UAAU,kCACtC,GAAa,MAATF,IAAiBC,EAAG,MAAM,IAAIC,UAAU,iDAC5C,GAAqB,mBAAVJ,EAAuBD,IAAaC,IAAUG,GAAKH,EAAMK,IAAIN,GAAW,MAAM,IAAIK,UAAU,2EACvG,MAAiB,MAATF,EAAeC,EAAEG,KAAKP,EAAUE,GAASE,EAAIA,EAAEF,MAAQA,EAAQD,EAAMO,IAAIR,EAAUE,GAASA,CACxG,EACIO,EAA0BV,MAAQA,KAAKU,wBAA2B,SAAUT,EAAUC,EAAOE,EAAMC,GACnG,GAAa,MAATD,IAAiBC,EAAG,MAAM,IAAIC,UAAU,iDAC5C,GAAqB,mBAAVJ,EAAuBD,IAAaC,IAAUG,GAAKH,EAAMK,IAAIN,GAAW,MAAM,IAAIK,UAAU,4EACvG,MAAgB,MAATF,EAAeC,EAAa,MAATD,EAAeC,EAAEG,KAAKP,GAAYI,EAAIA,EAAEF,MAAQD,EAAMS,IAAIV,EACxF,mBAESW,MAAc,gBAChBC,MAAY,cACZC,MAAa,iBACbC,MAAY,SACnB,MAAMC,EAAgBC,IAw0BtB,WACI,OAAOC,GAAQ,CACX,CAAC,cAAeC,GAAI,UAAW,CAAC,UAAW,OAAQ,OAAQ,UAAW,OAAQ,OAAQ,YACtF,CAAC,cAAeA,GAAI,OAAQ,CAAC,YAC7B,CAAC,gBAAiBA,GAAI,OAAQ,CAAC,UAAW,YAC1C,CAAC,WAAYA,GAAI,OAAQ,CAAC,UAAW,UAAW,OAAQ,UAAW,YACnE,CAAC,mBAAoBA,GAAI,OAAQ,CAAC,YAClC,CAAC,cAAeC,GAAI,OAAQ,CAAC,YAC7B,CAAC,iBAAkBD,GAAI,UAAW,CAAC,UAAW,YAC9C,CAAC,gBAAiBC,GAAI,OAAQ,CAAC,UAAW,YAC1C,CAAC,YAAaA,GAAI,OAAQ,CAAC,YAC3B,CAAC,uBAAwBD,GAAI,OAAQ,CAAC,UAAW,OAAQ,YACzD,CAAC,4BAA6BA,GAAI,OAAQ,CAAC,UAAW,UAAW,OAAQ,SACzE,CAAC,iBAAkBC,GAAI,OAAQ,CAAC,OAAQ,UAAW,OAAQ,OAAQ,UAAW,OAAQ,aAE9F,IAt1BMC,EAAcJ,IAu1BpB,WACI,OAAOC,GAAQ,CACX,CAAC,OAAQC,GAAI,MAAO,CAAC,UAAW,MAAO,MAAO,QAC9C,CAAC,QAASC,GAAI,MAAO,CAAC,QACtB,CAAC,QAASA,GAAIE,GAAY,CAAC,MAAOA,GAAY,QAC9C,CAAC,OAAQH,GAAII,GAAW,CAAC,MAAO,UAAWC,KAC3C,CAAC,UAAWL,GAAI,UAAW,CAAC,YAC5B,CAAC,kBAAmBA,GAAI,UAAW,CAAC,YACpC,CAAC,WAAYC,GAAI,MAAO,CAAC,YACzB,CAAC,UAAWA,GAAI,UAAW,CAAC,YAC5B,CAAC,kBAAmBA,GAAI,UAAW,CAAC,YACpC,CAAC,WAAYD,GAAII,GAAW,CAAC,UAAW,UAAWC,KACnD,CAAC,QAASL,GAAI,MAAO,CAAC,YACtB,CAAC,SAAUA,GAAI,MAAO,CAAC,YACvB,CAAC,OAAQA,GAAI,MAAO,CAAC,UAAW,YAChC,CAAC,SAAUA,GAAI,MAAO,CAAC,UAAW,YAClC,CAAC,YAAaA,GAAI,MAAO,CAAC,MAAO,UAAW,WAAYM,IACxD,CAAC,QAASN,GAAI,MAAO,CAAC,UAAW,YACjC,CAAC,UAAWA,GAAI,MAAO,CAAC,UAAW,YACnC,CAAC,aAAcA,GAAI,MAAO,CAAC,MAAO,UAAW,WAAYM,IACzD,CAAC,WAAYL,GAAI,UAAW,CAAC,SAErC,KA52BMM,SAAEA,EAAQC,YAAEA,GAAgBC,QAC5BC,EAAyB,YAAbH,EA2EZI,EAAY,IAlES,CACvBC,OATW,MAUXC,QATY,MAUZC,QATY,MAUZC,QATY,KAUZC,QATY,MAUZC,QATY,KAUZC,QATY,MAUZC,SATa,MAUbC,QAAS,IACTC,QAAS,IACTC,QAAS,IACTC,QAAS,GACTC,QAAS,GACTC,QAAS,GACTC,QAAS,GACTC,QAAS,EACTC,QAAS,EACTC,QAAS,EACTC,QAAS,EACTC,QAAS,EACTC,WAAY,EACZC,QAAS,EACTC,OAAQ,EACRC,OAAQ,EACRC,OAAQ,EACRC,OAAQ,EACRC,OAAQ,GACRC,QAAS,GACTC,OAAQ,OAEc,CACtBC,OAAQ,CACJC,SAAU,EACVC,SAAU,EACVC,OAAQ,EACRC,QAAS,IACTC,OAAQ,KACRC,SAAU,OACVC,QAAS,KACTC,SAAU,EACVC,YAAa,QACbC,WAAY,IACZC,OAAQ,IACRC,QAAS,QACTC,UAAW,QACXC,WAAY,GAEhBC,MAAO,CACHd,SAAU,EACVC,SAAU,EACVC,OAAQ,EACRC,QAAS,GACTC,OAAQ,IACRC,SAAU,IACVC,QAAS,IACTC,SAAU,KACVC,YAAa,MACbO,UAAW,OACXN,WAAY,OACZC,OAAQ,QACRC,QAAS,KACTK,SAAU,MACVH,WAAY,OAKKhD,IAEnBoD,EAAuBC,KAAK,GAqBlC,MAAMC,UAAmBjE,EAAOkE,SAC5BC,YAAYC,GAMR,GALAC,MAAM,CACFC,cAAe,UAEnB1F,EAAkBc,IAAIT,KAAM,MAC5BJ,EAAwBa,IAAIT,KAAM,MAC9B6B,EAAW,CACX,MACMyD,EADMtE,IACOuE,YAAYC,OAAOC,iBAAiBN,GA7B9C,WAEG,EA2BiFO,KAvBnF,EAMO,WAiBuHA,MAClIC,EAASL,EAAOnF,MACtB,GAAIwF,EAAOC,OAAOd,GAId,YAHAhE,EAAQ+E,UAAS,KACb7F,KAAK8F,QAAQC,GAAiBT,EAAOU,WAAW,IAIxDjG,EAAuBC,KAAML,EAAmB,IAAIsG,iBAAiBN,EAAQ,CAAEO,WAAW,IAAS,IACvG,KACK,CACD,MACMZ,EADMjE,IACO8E,KAAKX,OAAOY,gBAAgBjB,GAAOrD,EAAU+B,SAAU,GACpEwC,EAAKf,EAAOnF,MAClB,IAAY,IAARkG,EAIA,YAHAvF,EAAQ+E,UAAS,KACb7F,KAAK8F,QAAQQ,GAAehB,EAAOiB,OAAO,IAIlDxG,EAAuBC,KAAML,EAAmB,IAAI6G,gBAAgBH,EAAI,CAAEH,WAAW,IAAS,IAClG,CACJ,CACAO,SAASC,EAAOC,GACZjG,EAAuBV,KAAML,EAAmB,MAAMiH,QACtD7G,EAAuBC,KAAML,EAAmB,KAAM,KACtDgH,EAASD,EACb,CACAG,MAAMC,GACiE,OAA/DpG,EAAuBV,KAAMJ,EAAyB,MAE1DG,EAAuBC,KAAMJ,EAAyBc,EAAuBV,KAAML,EAAmB,KAAKoH,KAAKD,GAC3GE,MAAKC,IACNlH,EAAuBC,KAAMJ,EAAyB,KAAM,KAClC,IAAtBqH,EAAOC,WAIPlH,KAAKmH,KAAKvG,EAAOwG,KAAKH,KACtBjH,KAAK6G,MAAMC,GAJX9G,KAAKmH,KAAK,KAIM,IAEnBE,OAAMX,IACP3G,EAAuBC,KAAMJ,EAAyB,KAAM,KAC5DI,KAAK8F,QAAQY,EAAM,IACnB,IACR,EAEJ/G,EAAoB,IAAI2H,QAAW1H,EAA0B,IAAI0H,QACjE,MAAMC,UAAoBxG,EAAOyG,SAC7BtC,YAAYC,GAMR,GALAC,MAAM,CACFC,cAAe,UAEnBxF,EAAoBY,IAAIT,KAAM,MAC9BF,EAA0BW,IAAIT,KAAM,MAChC6B,EAAW,CACX,MACMyD,EADMtE,IACOuE,YAAYC,OAAOC,iBAAiBN,GArF7C,WAqFmE,EAAGO,KAjFtE,EAiF2F+B,WAA8C/B,MAC7IC,EAASL,EAAOnF,MACtB,GAAIwF,EAAOC,OAAOd,GAId,YAHAhE,EAAQ+E,UAAS,KACb7F,KAAK8F,QAAQC,GAAiBT,EAAOU,WAAW,IAIxDjG,EAAuBC,KAAMH,EAAqB,IAAI6H,kBAAkB/B,EAAQ,CAAEO,WAAW,IAAS,IAC1G,KACK,CACD,MAAMyB,EAAMtG,IACNuG,EAAUpC,OAAOY,gBAAgBjB,GACjC0C,EAAQ/F,EAAUgC,SAAWhC,EAAUkC,QAAUlC,EAAUqC,QAC3D2D,EAAOhG,EAAUU,QAAUV,EAAUW,QAAUX,EAAUc,QAAUd,EAAUkB,QAC7EsC,EAASqC,EAAIxB,KAAKyB,EAASC,EAAOC,GAClCzB,EAAKf,EAAOnF,MAClB,IAAY,IAARkG,EAIA,YAHAvF,EAAQ+E,UAAS,KACb7F,KAAK8F,QAAQQ,GAAehB,EAAOiB,OAAO,IAIlDxG,EAAuBC,KAAMH,EAAqB,IAAIkI,iBAAiB1B,EAAI,CAAEH,WAAW,IAAS,IACrG,CACJ,CACAO,SAASC,EAAOC,GACZjG,EAAuBV,KAAMH,EAAqB,MAAM+G,QACxD7G,EAAuBC,KAAMH,EAAqB,KAAM,KACxD8G,EAASD,EACb,CACAsB,OAAOC,EAAOC,EAAUvB,GACiD,OAAjEjG,EAAuBV,KAAMF,EAA2B,MAE5DC,EAAuBC,KAAMF,EAA2BY,EAAuBV,KAAMH,EAAqB,KAAKsI,SAASF,GACnHjB,MAAKF,IACN/G,EAAuBC,KAAMF,EAA2B,KAAM,KAC9D6G,GAAU,IAETU,OAAMX,IACP3G,EAAuBC,KAAMF,EAA2B,KAAM,KAC9D6G,EAASD,EAAM,IACf,IACR,EAEJ7G,EAAsB,IAAIyH,QAAWxH,EAA4B,IAAIwH,QACrE,MAAMc,EAAiB,CACnBC,0BAA0BlD,EAAMwB,GAC5B2B,EAAyCnD,EAAO,MAAOwB,EAC3D,EACA4B,aAAapD,EAAMqD,EAAU,CAAC,GACH,iBAAZA,IACPA,EAAU,CAAEN,SAAUM,IAC1B,MAAMN,SAAEA,EAAW,MAASM,GACtBjD,YAAEA,EAAWkD,cAAEA,EAAaC,SAAEA,EAAQC,YAAEA,GAAgB3H,IACxD4H,EAAYrD,EAAYC,OAAOC,iBAAiBN,GA7IzC,WAEG,EA2I4EO,KAvI9E,EAuImG,EAAGA,MAC9GC,EAASiD,EAAUzI,MACrBwF,EAAOC,OAAOd,IACd+D,GAAkBD,EAAU5C,WAChC,IACI,MAAM8C,EAAatD,OAAOuD,MAAM,GAC1BC,EAAcF,EACdG,EAASR,EAAc9C,EAAQqD,GAChB,IAAjBC,EAAO9I,OACP0I,GAAkBI,EAAOjD,WAC7B,MAAMkD,EAAWF,EAAYG,UAAUC,UACjCC,EAAM7D,OAAOuD,MAAMG,GACnBI,EAAkBR,EAClBS,EAAUb,EAAS/C,EAAQ0D,EAAKH,EAAUI,EAAiB5D,MAC3C,IAAlB6D,EAAQpJ,OACR0I,GAAkBU,EAAQvD,WAE9B,GADUsD,EAAgBE,YAChBN,EACN,MAAM,IAAIO,MAAM,cACpB,OAAOC,EAAoBL,EAAKH,EAAUhB,EAI9C,CAFA,QACIS,EAAYhD,EAChB,CACJ,EACAgE,aAAaxE,GACT,MAAMI,YAAEA,EAAWqE,0BAAEA,EAAyBjB,YAAEA,GAAgB3H,IAC1D4H,EAAYrD,EAAYC,OAAOC,iBAAiBN,GAAO,EAAG0E,EAAwDnE,KAlK1G,EAOa,SA2J8IA,MACnKC,EAASiD,EAAUzI,MACrBwF,EAAOC,OAAOd,IACd+D,GAAkBD,EAAU5C,WAChC,IACI,IAAI8D,EAAY,IAChB,OAAa,CACT,MAAMT,EAAM7D,OAAOuD,MAAkB,EAAZe,IACnB3J,MAAEA,EAAK6F,UAAEA,GAAc4D,EAA0BjE,EAAQ0D,EAAKS,EAAW,GAG/E,GAFc,IAAV3J,GACA0I,GAAkB7C,GApKN,IAqKZA,EAIJ,OAAOqD,EAAIU,kBAAkBC,UAAU,GAHnCF,GAAa,CAIrB,CAIJ,CAFA,QACInB,EAAYhD,EAChB,CACJ,EACAsE,UAAU9E,GACN,MAAMG,EAAStE,IAAgBkJ,iBAAiB1E,OAAOC,iBAAiBN,IACnD,IAAjBG,EAAOnF,OACP0I,GAAkBvD,EAAOU,UACjC,EACAmE,WAAWhF,GACP,MAAMG,EAAStE,IAAgBoJ,YAAY5E,OAAOC,iBAAiBN,IAC9C,IAAjBG,EAAOnF,OACP0I,GAAkBvD,EAAOU,UACjC,EACAqE,SAASlF,GACL,MAAMmF,EAAIlC,EAAemC,UAAUpF,GACnC,IAAKmF,EAAEE,iBACH,OAAOF,EACX,MAAMG,EAASrC,EAAeuB,aAAaxE,GAC3C,OAAOiD,EAAemC,UAAUE,EACpC,EACAF,UAAUpF,GACN,MACMkE,EAAM7D,OAAOuD,MAAM,IACnBzD,EAAStE,IAAgB0J,qBAAqBlF,OAAOC,iBAAiBN,GAF9C,EAE4EkE,GAC1G,GAAqB,IAAjB/D,EAAOnF,MAAa,CACpB,GArMoB,KAqMhBmF,EAAOU,UAAuC,CAC9C,IAAI2E,EAKJ,OAJArC,EAAyCnD,GAAMyF,IAE3CD,EAAenF,OAAOqF,IAAID,EAAM,GAAG,IAEhCE,EAAe3F,EAAMwF,EAChC,CACA9B,GAAkBvD,EAAOU,UAC7B,CACA,OAAO8E,EAAe3F,EAAMkE,EAChC,GAEJ,SAASf,EAAyCyC,EAAUpE,GACxD,MAAMqE,eAAEA,EAAcC,cAAEA,EAAaC,UAAEA,GAAclK,IAC/C4J,EAAOpF,OAAOuD,MAAM,KACpBzD,EAAS0F,EAAexF,OAAOC,iBAAiBsF,GAAWH,GAC3DjF,EAASL,EAAOnF,MAClBwF,EAAOC,OAAOd,IACd+D,GAAkBvD,EAAOU,WAC7B,IACI,GACIW,EAASiE,SAC4B,IAAhCK,EAActF,EAAQiF,GAInC,CAFA,QACIM,EAAUvF,EACd,CACJ,CACA,MAAMwF,EAAe,CACjB9C,0BAA0BlD,EAAMwB,GAC5B,MAAMyE,QAAEA,EAAOC,gBAAEA,EAAeC,SAAEA,EAAQC,QAAEA,EAAOC,gBAAEA,GAAoBnK,IAEnEoK,EAAcD,GAAmBD,EACjCG,GAFcL,GAAmBD,GAEf5F,OAAOY,gBAAgBjB,IACzCwG,EAAYD,EAAIvL,MAClBwL,EAAUC,UACVC,GAAgBH,EAAInF,OACxB,IACI,IAAIuF,EACJ,OAAUA,EAAQL,EAAYE,IAAYC,UACtCjF,EAASmF,EAKjB,CAFA,QACIR,EAASK,EACb,CACJ,EACApD,aAAapD,EAAMqD,EAAU,CAAC,GACH,iBAAZA,IACPA,EAAU,CAAEN,SAAUM,IAC1B,MAAMN,SAAEA,EAAW,MAASM,GACtBrC,KAAEA,EAAIS,MAAEA,EAAKmF,MAAEA,EAAKhF,KAAEA,GAAS1F,IAC/B2K,EAAa7F,EAAKX,OAAOY,gBAAgBjB,GAAOrD,EAAU+B,SAAU,GACpEwC,EAAK2F,EAAW7L,OACV,IAARkG,GACAwF,GAAgBG,EAAWzF,OAC/B,IACI,MAAM2C,EAAW6C,EAAM1F,EAAI,EA5PtB,GA4PmC+C,UACxC2C,EAAM1F,EAAI,EA/PL,GAgQL,MAAMgD,EAAM7D,OAAOuD,MAAMG,GACzB,IAAI+C,EAAYC,EAAGC,EACnB,GACIF,EAAalF,EAAKV,EAAIgD,EAAKH,GAC3BgD,EAAID,EAAW9L,MAAMiJ,UACrB+C,GAAoB,IAAPD,QACRC,GAnQP,IAmQqBF,EAAW1F,OAGlC,GAFI4F,GACAN,GAAgBI,EAAW1F,OAC3B2F,IAAMhD,EAASE,UACf,MAAM,IAAIK,MAAM,cACpB,OAAOC,EAAoBL,EAAKH,EAAUhB,EAI9C,CAFA,QACItB,EAAMP,EACV,CACJ,EACAsD,aAAaxE,GACT,MAAMyC,EAAUpC,OAAOY,gBAAgBjB,GACjCiH,EAAWjB,EAAaZ,UAAUpF,GAAM2B,KAAKsC,UAC7CC,EAAM7D,OAAOuD,MAAMqD,GACnB9G,EAASjE,IAAcgL,SAASzE,EAASyB,EAAK+C,GAC9CF,EAAI5G,EAAOnF,MAAMiJ,UAGvB,OAFW,IAAP8C,GACAL,GAAgBvG,EAAOiB,OACpB8C,EAAIiD,eAAeJ,EAC9B,EACAjC,UAAU9E,GACN,MAAMG,EAASjE,IAAckL,MAAM/G,OAAOY,gBAAgBjB,KACpC,IAAlBG,EAAOnF,OACP0L,GAAgBvG,EAAOiB,MAC/B,EACA4D,WAAWhF,GACP,MAAMG,EAASjE,IAAcmL,OAAOhH,OAAOY,gBAAgBjB,KACrC,IAAlBG,EAAOnF,OACP0L,GAAgBvG,EAAOiB,MAC/B,EACA8D,SAASlF,GACEsH,EAAiBC,IAAcC,MAAOxH,GAEjDoF,UAAUpF,GACCsH,EAAiBC,IAAcE,OAAQzH,IAGtD,SAAS0H,EAAc1H,EAAMyF,EAAMpC,EAAU,CAAC,GACnB,iBAAZA,IACPA,EAAU,CAAEN,SAAUM,IAC1B,MAAMN,SAAEA,EAAW,MAASM,EAC5B,IAAIsE,EAKIA,EAJY,iBAATlC,EACU,OAAb1C,GAAsB6E,EAAe7E,GAG3B0C,EAFAhK,EAAOwG,KAAKwD,EAAM1C,GAAUjB,OAKhC2D,EAAK3D,OAEnB,MAAM+F,EAAO,IAAIC,KAAK9H,EAAM,MAC5B,IACI6H,EAAKE,MAAMJ,EAIf,CAFA,QACIE,EAAKpG,OACT,CACJ,CACA,SAAS6F,EAAiBU,EAAMhI,GAC5B,MAAMkE,EAAM7D,OAAOuD,MAAMqE,GACnB9H,EAAS6H,EAAK3H,OAAOY,gBAAgBjB,GAAOkE,GAGlD,OAFqB,IAAjB/D,EAAOnF,OACP0L,GAAgBvG,EAAOiB,OACpBuE,EAAe3F,EAAMkE,EAChC,CACA,SAASK,EAAoBL,EAAKH,EAAUhB,GACxC,GAAI6E,EAAe7E,GACf,OAAOmB,EAAIiD,eAAepD,GAC9B,MAAM/I,EAAQS,EAAOwG,KAAKiC,EAAIgE,cAAcnE,IAC5C,OAAiB,OAAbhB,EACO/H,EAAMmN,SAASpF,GACnB/H,CACX,CACA,SAAS4M,EAAe7E,GACpB,MAAoB,SAAbA,GAAoC,UAAbA,CAClC,CACA,MAAMqF,EAAU1L,EAAYuG,EAAiB+C,GACvC9C,0BAAEA,EAAyBE,aAAEA,EAAYoB,aAAEA,EAAYM,UAAEA,EAASE,WAAEA,EAAUE,SAAEA,EAAQE,UAAEA,GAAegD,EACzGC,EAAc,CAChBC,QAAW,CACPC,OAAU,CAAC,GAAI,eACfC,OAAU,CAAC,EAAGC,GACdC,MAAS,CAAC,GAAIC,GACdC,MAAS,CAAC,GAAID,GACdE,MAAS,CAAC,EAAGF,GACbhH,KAAQ,CAAC,GAAImH,IAEjB,WAAY,CACRP,OAAU,CAAC,GAAI,cACfC,OAAU,CAAC,GAAI,OAEnB,WAAY,CACRD,OAAU,CAAC,GAAI,cACfC,OAAU,CAAC,GAAI,OAEnB,YAAa,CACTD,OAAU,CAAC,GAAI,cACfC,OAAU,CAAC,GAAI,OAEnB,YAAa,CACTD,OAAU,CAAC,GAAI,cACfC,OAAU,CAAC,GAAI,QAGjBO,EAAarM,EAAY2L,EAAYC,QAAUD,EAAY,GAAG9L,KAA0B,EAAdC,KAChF,SAASwM,EAAYhJ,GACjB,MAAMiJ,EAAU,GAKhB,OAJA/F,EAA0BlD,GAAM2G,IAC5B,MAAMuC,EAAOC,EAAgBxC,EAAO,UACpCsC,EAAQjH,KAAKkH,EAAK,IAEfD,CACX,CACA,SAASG,EAAKpJ,GACV,MAAMqJ,EAAkBC,OAAOC,KAAKR,GAAYS,QAAOC,IAAMA,EAAEC,WAAW,QACpET,EAAU,GAahB,OAZA/F,EAA0BlD,GAAM2G,IAC5B,MAAMuC,EAAOC,EAAgBxC,EAAO,UAC9BgD,EAAOR,EAAgBxC,EAAO,SAAUjL,EAAOkO,KAAK5J,EAAMkJ,IAC1DW,EAAS,CAAC,EAChB,IAAK,MAAM3O,KAAKmO,EACZQ,EAAO3O,GAAKiO,EAAgBxC,EAAOzL,GACvC+N,EAAQjH,KAAK,CACTkH,OACAS,UACGE,GACL,IAECZ,CACX,CACA,SAASE,EAAgBxC,EAAOuC,KAASY,GACrC,MAAMC,EAAYhB,EAAWG,IACtBc,EAAQL,GAAQI,EAEjB/O,GADwB,iBAAT2O,EAAqBM,cAAcC,UAAU,OAASP,GAAQA,GAChEtO,KAAKsL,EAAMwD,IAAIH,MAAYF,GAC9C,OAAI9O,aAAiBoP,OAASpP,aAAiBqP,OACpCrP,EAAMiJ,UACVjJ,CACX,CACA,MAAMsP,EAAa,IAAIC,IAAI,CACvB,MACA,OACA,QACA,MACA,MACA,OACA,UACA,MACA,OACA,SACA,UACA,UACA,UACA,cACA,QACA,QACA,QACA,cAEEC,EAAY,CACdlC,QAAW,CACP3G,KAAM,GACN8I,OAAQ,CACJC,IAAO,CAAC,EAAGC,GACXhI,KAAQ,CAAC,EAAG8F,GACZmC,MAAS,CAAC,EAAGC,IACbC,IAAO,CAAC,EAAGH,GACXI,IAAO,CAAC,EAAGJ,GACXK,IAAO,CAAC,EAAGL,GACXM,KAAQ,CAAC,EAAGN,GACZjC,MAAS,CAAC,GAAIC,GACdC,MAAS,CAAC,GAAID,GACdE,MAAS,CAAC,GAAIF,GACduC,UAAa,CAAC,EAAGvC,GACjBhH,KAAQ,CAAC,GAAImH,GACbqC,OAAU,CAAC,GAAIrC,GACfsC,QAAW,CAAC,EAAGP,MAGvB,YAAa,CACTlJ,KAAM,IACN8I,OAAQ,CACJC,IAAO,CAAC,EAAG,OACX/H,KAAQ,CAAC,EAAG,OACZiI,MAAS,CAAC,EAAG,OACbE,IAAO,CAAC,EAAG,OACXC,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,KAAQ,CAAC,GAAI,OACbvC,MAAS,CAAC,GAAI2C,GACdzC,MAAS,CAAC,GAAIyC,GACdxC,MAAS,CAAC,GAAIwC,GACdH,UAAa,CAAC,GAAIG,GAClB1J,KAAQ,CAAC,GAAI,OACbwJ,OAAU,CAAC,GAAI,OACfC,QAAW,CAAC,GAAI,SAGxB,YAAa,CACTzJ,KAAM,IACN8I,OAAQ,CACJC,IAAO,CAAC,EAAG,OACX/H,KAAQ,CAAC,EAAG,OACZiI,MAAS,CAAC,EAAG,OACbE,IAAO,CAAC,EAAG,OACXC,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,KAAQ,CAAC,GAAI,OACbvC,MAAS,CAAC,GAAI4C,GACd1C,MAAS,CAAC,GAAI0C,GACdzC,MAAS,CAAC,GAAIyC,GACdJ,UAAa,CAAC,GAAII,GAClB3J,KAAQ,CAAC,GAAI,OACbwJ,OAAU,CAAC,IAAK,OAChBC,QAAW,CAAC,IAAK,SAGzB,WAAY,CACRzJ,KAAM,GACN8I,OAAQ,CACJC,IAAO,CAAC,EAAG,OACX/H,KAAQ,CAAC,GAAI,OACbiI,MAAS,CAAC,GAAI,OACdE,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,KAAQ,CAAC,GAAI,OACbvC,MAAS,CAAC,GAAI2C,GACdzC,MAAS,CAAC,GAAIyC,GACdxC,MAAS,CAAC,GAAIwC,GACd1J,KAAQ,CAAC,GAAI,OACbwJ,OAAU,CAAC,GAAI,OACfC,QAAW,CAAC,GAAI,SAGxB,kBAAmB,CACfzJ,KAAM,IACN8I,OAAQ,CACJC,IAAO,CAAC,EAAG,OACX/H,KAAQ,CAAC,GAAI,OACbiI,MAAS,CAAC,GAAI,OACdE,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,KAAQ,CAAC,GAAI,OACbvC,MAAS,CAAC,GAAI2C,GACdzC,MAAS,CAAC,GAAIyC,GACdxC,MAAS,CAAC,GAAIwC,GACd1J,KAAQ,CAAC,GAAI,OACbwJ,OAAU,CAAC,GAAI,OACfC,QAAW,CAAC,GAAI,SAGxB,WAAY,CACRzJ,KAAM,IACN8I,OAAQ,CACJC,IAAO,CAAC,EAAG,OACX/H,KAAQ,CAAC,GAAI,OACbiI,MAAS,CAAC,GAAI,OACdE,IAAO,CAAC,EAAG,OACXC,IAAO,CAAC,GAAI,OACZC,IAAO,CAAC,GAAI,OACZC,KAAQ,CAAC,GAAI,OACbvC,MAAS,CAAC,GAAI4C,GACd1C,MAAS,CAAC,GAAI0C,GACdzC,MAAS,CAAC,IAAKyC,GACf3J,KAAQ,CAAC,GAAI,OACbwJ,OAAU,CAAC,GAAI,OACfC,QAAW,CAAC,GAAI,UAI5B,IAAIG,EAAiB,KACrB,MAAMtD,EAAc,IACpB,SAASV,IACL,GAAuB,OAAnBgE,EACA,OAAOA,EACX,IAAIC,EACJ,GAAI9O,EACA8O,EAAWhB,EAAUlC,YAEpB,CACD,MAAM9F,EAAMtG,IACNuP,EAAajJ,EAAIkJ,QAAUlJ,EAAImJ,UACrC,IAAIC,EAAa,GAAGrP,KAA0B,EAAdC,IAMhC,GALmB,aAAfoP,QACmBC,IAAfJ,IACAG,GAAc,WAEtBJ,EAAWhB,EAAUoB,QACJC,IAAbL,EACA,MAAM,IAAIlH,MAAM,qDACpBkH,EAAShE,MAAQiE,GAAcjJ,EAAIsJ,KACnCN,EAAS/D,OAASjF,EAAIuJ,SAAWvJ,EAAIwJ,YAAcxJ,EAAIyJ,KAC3D,CAEA,OADAV,EAAiBC,EACVA,CACX,CACA,MAAMU,EACFC,SACI,OAlpBQ,QADD,MAmpBCtR,KAAK8H,KACjB,CACAyJ,cACI,OAppBQ,QAFD,MAspBCvR,KAAK8H,KACjB,CACA0J,oBACI,OAtpBQ,OAHD,MAypBCxR,KAAK8H,KACjB,CACA2J,gBACI,OAxpBQ,QAJD,MA4pBCzR,KAAK8H,KACjB,CACA4J,SACI,OA1pBQ,OALD,MA+pBC1R,KAAK8H,KACjB,CACA0C,iBACI,OA5pBQ,QAND,MAkqBCxK,KAAK8H,KACjB,CACA6J,WACI,OA9pBS,QAPF,MAqqBC3R,KAAK8H,KACjB,EAEJ,SAASgD,EAAe3F,EAAMkE,GAC1B,OAAO,IAAIuI,MAAM,IAAIP,EAAS,CAC1B9Q,IAAG,CAACkK,EAAQoH,IACgB,iBAAbA,EACAA,KAAYpH,EAChBqH,EAAcD,GAEzBlR,IAAI8J,EAAQoH,EAAU5R,GAClB,OAAQ4R,GACJ,IAAK,YACD,OACJ,IAAK,cACL,IAAK,WACD,OAAOpH,EAAOoH,GAClB,IAAK,iBACD,OAAOC,EACX,IAAK,UACD,OAAO7R,EACX,IAAK,SACD,OAAOoJ,EACX,QACI,MAAwB,iBAAbwI,GAAyBA,KAAYpH,EACrCA,EAAOoH,GACXE,EAAevR,KAAKP,EAAU4R,EAAU1M,GAE3D,EACA1E,IAAG,CAACgK,EAAQoH,EAAU1R,EAAOF,KAClB,EAEX+R,QAAQvH,GACGwH,MAAM7K,KAAKqI,GAEtByC,yBAAwB,CAACzH,EAAQoH,KACtB,CACHM,UAAU,EACVC,cAAc,EACdC,YAAY,KAI5B,CACA,SAASP,EAAczD,GACnB,OAAOoB,EAAWlP,IAAI8N,EAC1B,CACA,SAAS0D,EAAe1D,EAAMlJ,GAC1B,IAAImN,EAAQ5F,IAAckD,OAAOvB,GACjC,QAAc2C,IAAVsB,EAAqB,CACrB,GAAa,cAATjE,EACA,OAAO0D,EAAevR,KAAKR,KAAM,QAASmF,GAE9C,MAAMoN,EAAQlE,EAAKmE,YAAY,MAC/B,OAAID,IAAUlE,EAAKoE,OAAS,EACjBV,EAAevR,KAAKR,KAAMqO,EAAKrE,UAAU,EAAGuI,GAAQpN,GAAMuN,eAErE,CACJ,CACA,MAAOvD,EAAQL,GAAQwD,EAEjBnS,GADwB,iBAAT2O,EAAqBM,cAAcC,UAAU,OAASP,GAAQA,GAChEtO,KAAKR,KAAKiH,OAAOqI,IAAIH,GAAShK,GACjD,OAAIhF,aAAiBoP,OAASpP,aAAiBqP,OACpCrP,EAAMiJ,UACVjJ,CACX,CACA,SAASyN,EAA0BzI,GAC/B,MAAMwN,EAAa3S,KAAKwJ,UACxB,IAAIoJ,GAAS,EACuC,IAlpBnB,KAkpB5BD,IACDrK,EAAyCnD,GAAMyF,IAC3C,MAAMiI,EAAYjI,EAAK0E,IAAI,IAAI9F,UAC/BoJ,EAppBuB,aAopBbC,GAnpBS,aAmpBmCA,CAAqC,IAGnG,MAAMC,EAAoD,IAzpB7B,GAypBdH,GACf,IAAI7K,EAWJ,OATIA,EADA8K,EA5uBQ,MA8uBHE,EAlvBG,MADA,MAwvBRhL,GADAgL,EACQ,IAEA,IACLhL,CACX,CACA,SAASgG,IACL,MAAMiF,EAAWC,OAAOhT,KAAKmJ,UAAUmE,YAAYlE,UAInD,OAAO,IAAI6J,KAAKC,UADEH,EAFG,OACG,iBAEUzF,YACtC,CACA,SAASW,IACL,MAAMkF,EAAOnT,KAAKwJ,UACZ4J,EAAMpT,KAAKsP,IAAI,GAAG9F,UACxB,OAAO6J,OAAOF,GAAMG,IAAI,IAAIC,GAAGH,EACnC,CACA,SAAS5C,IACL,MAAMgD,EAAMxT,KAAKwJ,UACXiK,EAAOzT,KAAKsP,IAAI,GAAG9F,UAEzB,OAAO,IAAIyJ,KAAY,IAANO,EADJC,EAAO,IAExB,CACA,SAAShD,IAEL,MAAM+C,EAAMxT,KAAKmJ,UAAUC,UACrBqK,EAAOzT,KAAKsP,IAAI,GAAGnG,UAAUC,UAEnC,OAAO,IAAI6J,KAAY,IAANO,EADJC,EAAO,IAExB,CACA,SAAS3D,IACL,OAAO,CACX,CACA,SAASE,KACL,OAAO,CACX,CACA,SAASnH,GAAkB7C,GACvB,MAAMD,GAAiBC,EAC3B,CACA,SAAS6F,GAAgBtF,GACrB,MAAMD,GAAeC,EACzB,CACA,SAASR,GAAiBC,GACtB,MAGMqD,EAAM7D,OAAOuD,MAAMe,KAEzB,OADA9I,IAAgB0S,eAAeC,KAA4DjO,KAAMM,EAAW,EAAGqD,EAJ7F,IAI6G3D,MACxH,IAAI+D,MAAMJ,EAAIU,kBACzB,CACA,SAASzD,GAAeC,GACpB,MAAMqN,EAAUvS,IAAcwS,SAAStN,GAAO+F,iBAC9C,OAAO,IAAI7C,MAAMmK,EACrB,CACA,SAASE,GAAYC,GACjB,OAAO,YAAa9E,GAChB,MAAM+E,EAAkB/E,EAAKwD,OAAS,EAChCwB,EAAWhF,EAAKiF,MAAM,EAAGF,GACzBrN,EAAWsI,EAAK+E,GACtBlT,EAAQ+E,UAAS,WACb,IACI,MAAMP,EAASyO,KAAYE,GAC3BtN,EAAS,KAAMrB,EAInB,CAFA,MAAO6O,GACHxN,EAASwN,EACb,CACJ,GACJ,CACJ,CACA,MAAMhT,GAAKiT,eACLhT,GAAKiT,eACL9S,GAA6B,IAAhBI,EAAqB,QAAU,QAC5CH,GAAW,IAAMD,GACjBD,GAA2B,WAAbI,GAAyC,IAAhBC,EAAqB,QAAU,QAwC5E,SAASF,GAAY0L,EAAMhI,EAAMkE,GAE7B,OAAO8D,EADgB,EACKhI,EAAMkE,EACtC,CACA,SAASnI,GAAQoT,GACb,OAAOA,EAAKC,QAAO,CAAC5M,EAAKmE,KAM7B,SAA2BnE,EAAKmE,GAC5B,MAAOuC,GAAQvC,EACf2C,OAAO+F,eAAe7M,EAAK0G,EAAM,CAC7B+D,cAAc,EACdzR,MACI,MAAO,CAAE8T,EAAMC,EAASC,EAAUC,GAAW9I,EAC7C,IAAIqB,EAAO,KACX,MAAM0H,EAAUhT,EACViT,OAAOC,iBAAiB,eAAgB1G,GACxCyG,OAAOC,iBAAiB,KAAM1G,GAMpC,OALgB,OAAZwG,IACA1H,EAAO,IAAIsH,EAAKI,EAASH,EAASC,EAAUK,UAChChE,IAAZ4D,IACAzH,EAAOyH,EAAQK,KAAK,KAAM9H,IAC9BsB,OAAO+F,eAAe7M,EAAK0G,EAAM,CAAElO,MAAOgN,IACnCA,CACX,GAER,CAvBQ+H,CAAkBvN,EAAKmE,GAChBnE,IACR,CAAC,EACR,CACA,MAAMqN,GAAcnT,GAA6B,IAAhBF,EAAqB,CAAEwT,IAAK,WAAc,CAAC,SAoBrE,SAASC,iBAAiBjQ,GAC7B,OAAO,IAAIH,EAAWG,EAC1B,QACO,SAASkQ,kBAAkBlQ,GAC9B,OAAO,IAAIoC,EAAYpC,EAC3B,QACO,MAAMoG,QAAUuI,GAAY3F,UAC5B,MAAMmH,SAAWxB,GAAYvL,UAC7B,MAAMgN,UAAYzB,GAAYjH,UAC9B,MAAMR,SAAWyH,GAAYnK,UAC7B,MAAM4C,MAAQuH,GAAY7J,UAC1B,MAAMuC,OAASsH,GAAY3J,UAC3B,MAAM8G,KAAO6C,GAAYzJ,UACzB,MAAM+G,MAAQ0C,GAAYvJ,GACjC,SAAStJ,GAAQuU,GACb,IAAIrV,EACAsV,GAAW,EACf,OAAO,YAAaxG,GAKhB,OAJKwG,IACDtV,EAAQqV,KAAWvG,GACnBwG,GAAW,GAERtV,CACX,CACJ,QACS2B,eAAWqM,iBAAaI,UAAMhG,kBAAcsE,mBAAelD,kBAAcM,eAAWE,gBAAYE,cAAUE,eAAW8G,0BAC/G,CACXvP,YACAsT,iBACAC,kBACA9J,QACA4C,cACAI,OACA+G,SACA/M,eACAgN,UACA1I,gBACAR,SACA1C,eACA4C,MACAtC,YACAuC,OACArC,aACA8G,KACA5G,WACA+G,MACA7G,YACA8G"}
✄
var t,e,n,r,i=this&&this.__classPrivateFieldSet||function(t,e,n,r,i){if("m"===r)throw new TypeError("Private method is not writable");if("a"===r&&!i)throw new TypeError("Private accessor was defined without a setter");if("function"==typeof e?t!==e||!i:!e.has(t))throw new TypeError("Cannot write private member to an object whose class did not declare it");return"a"===r?i.call(t,n):i?i.value=n:e.set(t,n),n},o=this&&this.__classPrivateFieldGet||function(t,e,n,r){if("a"===n&&!r)throw new TypeError("Private accessor was defined without a getter");if("function"==typeof e?t!==e||!r:!e.has(t))throw new TypeError("Cannot read private member from an object whose class did not declare it");return"m"===n?r:"a"===n?r.call(t):r?r.value:e.get(t)};import{Buffer as l}from"buffer";import s from"path";import a from"process";import c from"stream";const u=pt((function(){return ft([["CreateFileW",lt,"pointer",["pointer","uint","uint","pointer","uint","uint","pointer"]],["DeleteFileW",lt,"uint",["pointer"]],["GetFileSizeEx",lt,"uint",["pointer","pointer"]],["ReadFile",lt,"uint",["pointer","pointer","uint","pointer","pointer"]],["RemoveDirectoryW",lt,"uint",["pointer"]],["CloseHandle",st,"uint",["pointer"]],["FindFirstFileW",lt,"pointer",["pointer","pointer"]],["FindNextFileW",st,"uint",["pointer","pointer"]],["FindClose",st,"uint",["pointer"]],["GetFileAttributesExW",lt,"uint",["pointer","uint","pointer"]],["GetFinalPathNameByHandleW",lt,"uint",["pointer","pointer","uint","uint"]],["FormatMessageW",st,"uint",["uint","pointer","uint","uint","pointer","uint","pointer"]]])})),d=pt((function(){return ft([["open",lt,"int",["pointer","int","...","int"]],["close",st,"int",["int"]],["lseek",st,ut,["int",ut,"int"]],["read",lt,at,["int","pointer",ct]],["opendir",lt,"pointer",["pointer"]],["opendir$INODE64",lt,"pointer",["pointer"]],["closedir",st,"int",["pointer"]],["readdir",st,"pointer",["pointer"]],["readdir$INODE64",st,"pointer",["pointer"]],["readlink",lt,at,["pointer","pointer",ct]],["rmdir",lt,"int",["pointer"]],["unlink",lt,"int",["pointer"]],["stat",lt,"int",["pointer","pointer"]],["stat64",lt,"int",["pointer","pointer"]],["__xstat64",lt,"int",["int","pointer","pointer"],dt],["lstat",lt,"int",["pointer","pointer"]],["lstat64",lt,"int",["pointer","pointer"]],["__lxstat64",lt,"int",["int","pointer","pointer"],dt],["strerror",st,"pointer",["int"]]])})),{platform:f,pointerSize:m}=Process,p="windows"===f,S={...{S_IFMT:61440,S_IFREG:32768,S_IFDIR:16384,S_IFCHR:8192,S_IFBLK:24576,S_IFIFO:4096,S_IFLNK:40960,S_IFSOCK:49152,S_IRWXU:448,S_IRUSR:256,S_IWUSR:128,S_IXUSR:64,S_IRWXG:56,S_IRGRP:32,S_IWGRP:16,S_IXGRP:8,S_IRWXO:7,S_IROTH:4,S_IWOTH:2,S_IXOTH:1,DT_UNKNOWN:0,DT_FIFO:1,DT_CHR:2,DT_DIR:4,DT_BLK:6,DT_REG:8,DT_LNK:10,DT_SOCK:12,DT_WHT:14},...{darwin:{O_RDONLY:0,O_WRONLY:1,O_RDWR:2,O_CREAT:512,O_EXCL:2048,O_NOCTTY:131072,O_TRUNC:1024,O_APPEND:8,O_DIRECTORY:1048576,O_NOFOLLOW:256,O_SYNC:128,O_DSYNC:4194304,O_SYMLINK:2097152,O_NONBLOCK:4},linux:{O_RDONLY:0,O_WRONLY:1,O_RDWR:2,O_CREAT:64,O_EXCL:128,O_NOCTTY:256,O_TRUNC:512,O_APPEND:1024,O_DIRECTORY:65536,O_NOATIME:262144,O_NOFOLLOW:131072,O_SYNC:1052672,O_DSYNC:4096,O_DIRECT:16384,O_NONBLOCK:2048}}[f]},y=ptr(-1);class h extends c.Readable{constructor(n){if(super({highWaterMark:4194304}),t.set(this,null),e.set(this,null),p){const e=u().CreateFileW(Memory.allocUtf16String(n),2147483648,1,NULL,3,1073741824,NULL),r=e.value;if(r.equals(y))return void a.nextTick((()=>{this.destroy(rt(e.lastError))}));i(this,t,new Win32InputStream(r,{autoClose:!0}),"f")}else{const e=d().open(Memory.allocUtf8String(n),S.O_RDONLY,0),r=e.value;if(-1===r)return void a.nextTick((()=>{this.destroy(it(e.errno))}));i(this,t,new UnixInputStream(r,{autoClose:!0}),"f")}}_destroy(e,n){o(this,t,"f")?.close(),i(this,t,null,"f"),n(e)}_read(n){null===o(this,e,"f")&&i(this,e,o(this,t,"f").read(n).then((t=>{i(this,e,null,"f"),0!==t.byteLength?this.push(l.from(t))&&this._read(n):this.push(null)})).catch((t=>{i(this,e,null,"f"),this.destroy(t)})),"f")}}t=new WeakMap,e=new WeakMap;class _ extends c.Writable{constructor(t){if(super({highWaterMark:4194304}),n.set(this,null),r.set(this,null),p){const e=u().CreateFileW(Memory.allocUtf16String(t),1073741824,0,NULL,2,1073741952,NULL),r=e.value;if(r.equals(y))return void a.nextTick((()=>{this.destroy(rt(e.lastError))}));i(this,n,new Win32OutputStream(r,{autoClose:!0}),"f")}else{const e=d(),r=Memory.allocUtf8String(t),o=S.O_WRONLY|S.O_CREAT|S.O_TRUNC,l=S.S_IRUSR|S.S_IWUSR|S.S_IRGRP|S.S_IROTH,s=e.open(r,o,l),c=s.value;if(-1===c)return void a.nextTick((()=>{this.destroy(it(s.errno))}));i(this,n,new UnixOutputStream(c,{autoClose:!0}),"f")}}_destroy(t,e){o(this,n,"f")?.close(),i(this,n,null,"f"),e(t)}_write(t,e,l){null===o(this,r,"f")&&i(this,r,o(this,n,"f").writeAll(t).then((t=>{i(this,r,null,"f"),l()})).catch((t=>{i(this,r,null,"f"),l(t)})),"f")}}n=new WeakMap,r=new WeakMap;const U={enumerateDirectoryEntries(t,e){O(t+"\\*",e)},readFileSync(t,e={}){"string"==typeof e&&(e={encoding:e});const{encoding:n=null}=e,{CreateFileW:r,GetFileSizeEx:i,ReadFile:o,CloseHandle:l}=u(),s=r(Memory.allocUtf16String(t),2147483648,1,NULL,3,0,NULL),a=s.value;a.equals(y)&&et(s.lastError);try{const t=Memory.alloc(8),e=t,r=i(a,e);0===r.value&&et(r.lastError);const s=e.readU64().valueOf(),c=Memory.alloc(s),u=t,d=o(a,c,s,u,NULL);0===d.value&&et(d.lastError);if(u.readU32()!==s)throw new Error("Short read");return F(c,s,n)}finally{l(a)}},readlinkSync(t){const{CreateFileW:e,GetFinalPathNameByHandleW:n,CloseHandle:r}=u(),i=e(Memory.allocUtf16String(t),0,7,NULL,3,33554432,NULL),o=i.value;o.equals(y)&&et(i.lastError);try{let t=256;for(;;){const e=Memory.alloc(2*t),{value:r,lastError:i}=n(o,e,t,0);if(0===r&&et(i),8!==i)return e.readUtf16String().substring(4);t*=2}}finally{r(o)}},rmdirSync(t){const e=u().RemoveDirectoryW(Memory.allocUtf16String(t));0===e.value&&et(e.lastError)},unlinkSync(t){const e=u().DeleteFileW(Memory.allocUtf16String(t));0===e.value&&et(e.lastError)},statSync(t){const e=U.lstatSync(t);if(!e.isSymbolicLink())return e;const n=U.readlinkSync(t);return U.lstatSync(n)},lstatSync(t){const e=Memory.alloc(36),n=u().GetFileAttributesExW(Memory.allocUtf16String(t),0,e);if(0===n.value){if(32===n.lastError){let e;return O(t,(t=>{e=Memory.dup(t,36)})),K(t,e)}et(n.lastError)}return K(t,e)}};function O(t,e){const{FindFirstFileW:n,FindNextFileW:r,FindClose:i}=u(),o=Memory.alloc(592),l=n(Memory.allocUtf16String(t),o),s=l.value;s.equals(y)&&et(l.lastError);try{do{e(o)}while(0!==r(s,o))}finally{i(s)}}const w={enumerateDirectoryEntries(t,e){const{opendir:n,opendir$INODE64:r,closedir:i,readdir:o,readdir$INODE64:l}=d(),s=l||o,a=(r||n)(Memory.allocUtf8String(t)),c=a.value;c.isNull()&&nt(a.errno);try{let t;for(;!(t=s(c)).isNull();)e(t)}finally{i(c)}},readFileSync(t,e={}){"string"==typeof e&&(e={encoding:e});const{encoding:n=null}=e,{open:r,close:i,lseek:o,read:l}=d(),s=r(Memory.allocUtf8String(t),S.O_RDONLY,0),a=s.value;-1===a&&nt(s.errno);try{const t=o(a,0,2).valueOf();o(a,0,0);const e=Memory.alloc(t);let r,s,c;do{r=l(a,e,t),s=r.value.valueOf(),c=-1===s}while(c&&4===r.errno);if(c&&nt(r.errno),s!==t.valueOf())throw new Error("Short read");return F(e,t,n)}finally{i(a)}},readlinkSync(t){const e=Memory.allocUtf8String(t),n=w.lstatSync(t).size.valueOf(),r=Memory.alloc(n),i=d().readlink(e,r,n),o=i.value.valueOf();return-1===o&&nt(i.errno),r.readUtf8String(o)},rmdirSync(t){const e=d().rmdir(Memory.allocUtf8String(t));-1===e.value&&nt(e.errno)},unlinkSync(t){const e=d().unlink(Memory.allocUtf8String(t));-1===e.value&&nt(e.errno)},statSync:t=>g(A()._stat,t),lstatSync:t=>g(A()._lstat,t)};function v(t,e,n={}){"string"==typeof n&&(n={encoding:n});const{encoding:r=null}=n;let i;i="string"==typeof e?null===r||k(r)?e:l.from(e,r).buffer:e.buffer;const o=new File(t,"wb");try{o.write(i)}finally{o.close()}}function g(t,e){const n=Memory.alloc(H),r=t(Memory.allocUtf8String(e),n);return 0!==r.value&&nt(r.errno),K(e,n)}function F(t,e,n){if(k(n))return t.readUtf8String(e);const r=l.from(t.readByteArray(e));return null!==n?r.toString(n):r}function k(t){return"utf8"===t||"utf-8"===t}const N=p?U:w,{enumerateDirectoryEntries:R,readFileSync:b,readlinkSync:E,rmdirSync:W,unlinkSync:I,statSync:L,lstatSync:M}=N,x={windows:{d_name:[44,"Utf16String"],d_type:[0,j],atime:[12,q],mtime:[20,q],ctime:[4,q],size:[28,J]},"linux-32":{d_name:[11,"Utf8String"],d_type:[10,"U8"]},"linux-64":{d_name:[19,"Utf8String"],d_type:[18,"U8"]},"darwin-32":{d_name:[21,"Utf8String"],d_type:[20,"U8"]},"darwin-64":{d_name:[21,"Utf8String"],d_type:[20,"U8"]}},D=p?x.windows:x[`${f}-${8*m}`];function C(t){const e=[];return R(t,(t=>{const n=z(t,"d_name");e.push(n)})),e}function T(t){const e=Object.keys(D).filter((t=>!t.startsWith("d_"))),n=[];return R(t,(r=>{const i=z(r,"d_name"),o=z(r,"d_type",s.join(t,i)),l={};for(const t of e)l[t]=z(r,t);n.push({name:i,type:o,...l})})),n}function z(t,e,...n){const r=D[e],[i,o]=r,l=("string"==typeof o?NativePointer.prototype["read"+o]:o).call(t.add(i),...n);return l instanceof Int64||l instanceof UInt64?l.valueOf():l}const P=new Set(["dev","mode","nlink","uid","gid","rdev","blksize","ino","size","blocks","atimeMs","mtimeMs","ctimeMs","birthtimeMs","atime","mtime","ctime","birthtime"]),Y={windows:{size:36,fields:{dev:[0,Z],mode:[0,j],nlink:[0,tt],ino:[0,Z],uid:[0,Z],gid:[0,Z],rdev:[0,Z],atime:[12,q],mtime:[20,q],ctime:[20,q],birthtime:[4,q],size:[28,J],blocks:[28,J],blksize:[0,tt]}},"darwin-32":{size:108,fields:{dev:[0,"S32"],mode:[4,"U16"],nlink:[6,"U16"],ino:[8,"U64"],uid:[16,"U32"],gid:[20,"U32"],rdev:[24,"S32"],atime:[28,Q],mtime:[36,Q],ctime:[44,Q],birthtime:[52,Q],size:[60,"S64"],blocks:[68,"S64"],blksize:[76,"S32"]}},"darwin-64":{size:144,fields:{dev:[0,"S32"],mode:[4,"U16"],nlink:[6,"U16"],ino:[8,"U64"],uid:[16,"U32"],gid:[20,"U32"],rdev:[24,"S32"],atime:[32,V],mtime:[48,V],ctime:[64,V],birthtime:[80,V],size:[96,"S64"],blocks:[104,"S64"],blksize:[112,"S32"]}},"linux-32":{size:88,fields:{dev:[0,"U64"],mode:[16,"U32"],nlink:[20,"U32"],ino:[12,"U32"],uid:[24,"U32"],gid:[28,"U32"],rdev:[32,"U64"],atime:[56,Q],mtime:[64,Q],ctime:[72,Q],size:[44,"S32"],blocks:[52,"S32"],blksize:[48,"S32"]}},"linux-32-stat64":{size:104,fields:{dev:[0,"U64"],mode:[16,"U32"],nlink:[20,"U32"],ino:[96,"U64"],uid:[24,"U32"],gid:[28,"U32"],rdev:[32,"U64"],atime:[72,Q],mtime:[80,Q],ctime:[88,Q],size:[48,"S64"],blocks:[64,"S64"],blksize:[56,"S32"]}},"linux-64":{size:144,fields:{dev:[0,"U64"],mode:[24,"U32"],nlink:[16,"U64"],ino:[8,"U64"],uid:[28,"U32"],gid:[32,"U32"],rdev:[40,"U64"],atime:[72,V],mtime:[88,V],ctime:[104,V],size:[48,"S64"],blocks:[64,"S64"],blksize:[56,"S64"]}}};let G=null;const H=256;function A(){if(null!==G)return G;let t;if(p)t=Y.windows;else{const e=d(),n=e.stat64??e.__xstat64;let r=`${f}-${8*m}`;if("linux-32"===r&&void 0!==n&&(r+="-stat64"),t=Y[r],void 0===t)throw new Error("Current OS is not yet supported; please open a PR");t._stat=n??e.stat,t._lstat=e.lstat64??e.__lxstat64??e.lstat}return G=t,t}class B{isFile(){return 32768==(61440&this.mode)}isDirectory(){return 16384==(61440&this.mode)}isCharacterDevice(){return 8192==(61440&this.mode)}isBlockDevice(){return 24576==(61440&this.mode)}isFIFO(){return 4096==(61440&this.mode)}isSymbolicLink(){return 40960==(61440&this.mode)}isSocket(){return 49152==(61440&this.mode)}}function K(t,e){return new Proxy(new B,{has:(t,e)=>"symbol"==typeof e?e in t:X(e),get(n,r,i){switch(r){case"prototype":return;case"constructor":case"toString":return n[r];case"hasOwnProperty":return X;case"valueOf":return i;case"buffer":return e;default:return"symbol"==typeof r||r in n?n[r]:$.call(i,r,t)}},set:(t,e,n,r)=>!1,ownKeys:t=>Array.from(P),getOwnPropertyDescriptor:(t,e)=>({writable:!1,configurable:!0,enumerable:!0})})}function X(t){return P.has(t)}function $(t,e){let n=A().fields[t];if(void 0===n){if("birthtime"===t)return $.call(this,"ctime",e);const n=t.lastIndexOf("Ms");return n===t.length-2?$.call(this,t.substring(0,n),e).getTime():void 0}const[r,i]=n,o=("string"==typeof i?NativePointer.prototype["read"+i]:i).call(this.buffer.add(r),e);return o instanceof Int64||o instanceof UInt64?o.valueOf():o}function j(t){const e=this.readU32();let n=!1;0!=(1024&e)&&O(t,(t=>{const e=t.add(36).readU32();n=2684354563===e||2684354572===e}));const r=0!=(16&e);let i;return i=n?40960:r?16384:32768,i|=r?493:420,i}function q(){const t=BigInt(this.readU64().toString()).valueOf();return new Date(parseInt((t/10000n-11644473600000n).toString()))}function J(){const t=this.readU32(),e=this.add(4).readU32();return uint64(t).shl(32).or(e)}function Q(){const t=this.readU32(),e=this.add(4).readU32();return new Date(1e3*t+e/1e6)}function V(){const t=this.readU64().valueOf(),e=this.add(8).readU64().valueOf();return new Date(1e3*t+e/1e6)}function Z(){return 0}function tt(){return 1}function et(t){throw rt(t)}function nt(t){throw it(t)}function rt(t){const e=Memory.alloc(512);return u().FormatMessageW(4608,NULL,t,0,e,256,NULL),new Error(e.readUtf16String())}function it(t){const e=d().strerror(t).readUtf8String();return new Error(e)}function ot(t){return function(...e){const n=e.length-1,r=e.slice(0,n),i=e[n];a.nextTick((function(){try{const e=t(...r);i(null,e)}catch(t){i(t)}}))}}const lt=SystemFunction,st=NativeFunction,at=8===m?"int64":"int32",ct="u"+at,ut="darwin"===f||8===m?"int64":"int32";function dt(t,e,n){return t(3,e,n)}function ft(t){return t.reduce(((t,e)=>(function(t,e){const[n]=e;Object.defineProperty(t,n,{configurable:!0,get(){const[,r,i,o,l]=e;let s=null;const a=p?Module.findExportByName("kernel32.dll",n):Module.findExportByName(null,n);return null!==a&&(s=new r(a,i,o,mt)),void 0!==l&&(s=l.bind(null,s)),Object.defineProperty(t,n,{value:s}),s}})}(t,e),t)),{})}const mt=p&&4===m?{abi:"stdcall"}:{};export function createReadStream(t){return new h(t)}export function createWriteStream(t){return new _(t)}export const readdir=ot(C);export const readFile=ot(b);export const writeFile=ot(v);export const readlink=ot(E);export const rmdir=ot(W);export const unlink=ot(I);export const stat=ot(L);export const lstat=ot(M);function pt(t){let e,n=!1;return function(...r){return n||(e=t(...r),n=!0),e}}export{S as constants,C as readdirSync,T as list,b as readFileSync,v as writeFileSync,E as readlinkSync,W as rmdirSync,I as unlinkSync,L as statSync,M as lstatSync,B as Stats};export default{constants:S,createReadStream,createWriteStream,readdir,readdirSync:C,list:T,readFile,readFileSync:b,writeFile,writeFileSync:v,readlink,readlinkSync:E,rmdir,rmdirSync:W,unlink,unlinkSync:I,stat,statSync:L,lstat,lstatSync:M,Stats:B};
✄
{"version":3,"file":"index.js","names":["EventEmitter","Readable","Writable","Controller","constructor","this","events","sources","Map","nextEndpointId","requests","nextRequestId","onCreate","payload","endpoint","source","Source","set","id","emit","onFinish","get","undefined","Error","delete","push","onWrite","data","deliver","handlers","open","label","details","Sink","receive","packet","stanza","name","type","onRequest","onNotification","_request","Promise","resolve","reject","handler","result","e","then","value","catch","error","message","toString","request","response","super","onReadComplete","delivery","_read","size","chunk","length","tryComplete","controller","once","_onFinish","bind","_write","encoding","callback","_"],"sourceRoot":"/Users/oleavr/src/frida-tools/build/agents/fs/fs_agent.js.p/node_modules/frida-remote-stream/dist/","sources":[""],"mappings":"OAAOA,MAAkB,4BAChBC,cAAUC,MAAgB,gBAC5B,MAAMC,WACTC,cACIC,KAAKC,OAAS,IAAIN,EAClBK,KAAKE,QAAU,IAAIC,IACnBH,KAAKI,eAAiB,EACtBJ,KAAKK,SAAW,IAAIF,IACpBH,KAAKM,cAAgB,EACrBN,KAAKO,SAAYC,IACb,MAAMC,EAAWD,EAAQC,SACnBC,EAAS,IAAIC,EAAOF,GAC1BT,KAAKE,QAAQU,IAAIH,EAASI,GAAIH,GAC9BV,KAAKC,OAAOa,KAAK,SAAUJ,EAAO,EAEtCV,KAAKe,SAAYP,IACb,MAAMK,EAAKL,EAAQC,SAASI,GACtBH,EAASV,KAAKE,QAAQc,IAAIH,GAChC,QAAeI,IAAXP,EACA,MAAM,IAAIQ,MAAM,uBAEpBlB,KAAKE,QAAQiB,OAAON,GACpBH,EAAOU,KAAK,KAAK,EAErBpB,KAAKqB,QAAU,CAACb,EAASc,KACrB,MAAMT,EAAKL,EAAQC,SAASI,GACtBH,EAASV,KAAKE,QAAQc,IAAIH,GAChC,QAAeI,IAAXP,EACA,MAAM,IAAIQ,MAAM,uBAEpB,GAAa,OAATI,EACA,MAAM,IAAIJ,MAAM,iCAEpB,OAAOR,EAAOa,QAAQD,EAAK,EAE/BtB,KAAKwB,SAAW,CACZ,UAAWxB,KAAKO,SAChB,UAAWP,KAAKe,SAChB,SAAUf,KAAKqB,QAEvB,CACAI,KAAKC,EAAOC,EAAU,CAAC,GACnB,MAAMlB,EAAW,CACbI,GAAIb,KAAKI,iBACTsB,QACAC,WAEJ,OAAO,IAAIC,EAAK5B,KAAMS,EAC1B,CACAoB,QAAQC,GACJ,MAAMC,EAASD,EAAOC,QAChBlB,GAAEA,EAAEmB,KAAEA,EAAIxB,QAAEA,GAAYuB,EACxBE,EAAOD,EAAK,GAClB,GAAa,MAATC,EACAjC,KAAKkC,UAAUrB,EAAImB,EAAMxB,EAASsB,EAAOR,UAExC,IAAa,MAATW,EAIL,MAAM,IAAIf,MAAM,mBAAqBc,GAHrChC,KAAKmC,eAAetB,EAAImB,EAAMxB,EAIlC,CACJ,CACA4B,SAASJ,EAAMxB,EAASc,GACpB,OAAO,IAAIe,SAAQ,CAACC,EAASC,KACzB,MAAM1B,EAAKb,KAAKM,gBAChBN,KAAKK,SAASO,IAAIC,EAAI,CAClByB,QAASA,EACTC,OAAQA,IAEZ,MAAMR,EAAS,CACXlB,KACAmB,OACAxB,WAEJR,KAAKC,OAAOa,KAAK,OAAQ,CACrBiB,SACAT,QACF,GAEV,CACAY,UAAUrB,EAAImB,EAAMxB,EAASc,GACzB,MAAMkB,EAAUxC,KAAKwB,SAASQ,GAC9B,QAAgBf,IAAZuB,EACA,MAAM,IAAItB,MAAM,oBAAoBc,KAExC,IAAIS,EACJ,IACIA,EAASD,EAAQhC,EAASc,EAK9B,CAHA,MAAOoB,GAEH,YADA1C,KAAKuC,OAAO1B,EAAI6B,EAEpB,CACID,aAAkBJ,QAClBI,EACKE,MAAKC,GAAS5C,KAAKsC,QAAQzB,EAAI+B,KAC/BC,OAAMC,GAAS9C,KAAKuC,OAAO1B,EAAIiC,KAGpC9C,KAAKsC,QAAQzB,EAAI4B,EAEzB,CACAH,QAAQzB,EAAI+B,GACR,MAAMb,EAAS,CACXlB,GAAIA,EACJmB,KAAM,UACNxB,QAASoC,GAEb5C,KAAKC,OAAOa,KAAK,OAAQ,CACrBiB,SACAT,KAAM,MAEd,CACAiB,OAAO1B,EAAIiC,GACP,MAAMf,EAAS,CACXlB,GAAIA,EACJmB,KAAM,SACNxB,QAAS,CACLuC,QAASD,EAAME,aAGvBhD,KAAKC,OAAOa,KAAK,OAAQ,CACrBiB,SACAT,KAAM,MAEd,CACAa,eAAetB,EAAImB,EAAMxB,GACrB,MAAMyC,EAAUjD,KAAKK,SAASW,IAAIH,GAClC,QAAgBI,IAAZgC,EACA,MAAM,IAAI/B,MAAM,sBAGpB,GADAlB,KAAKK,SAASc,OAAON,GACR,YAATmB,EACAiB,EAAQX,QAAQ9B,OAEf,IAAa,WAATwB,EAKL,MAAM,IAAId,MAAM,yBAA2Bc,GALnB,CACxB,MAAMkB,EAAW1C,EACjByC,EAAQV,OAAO,IAAIrB,MAAMgC,EAASH,SACtC,CAGA,CACJ,iBAEWjD,WACf,MAAMa,UAAef,EACjBG,aAAY2B,MAAEA,EAAKC,QAAEA,IACjBwB,QACAnD,KAAKoD,eAAiB,KACtBpD,KAAKqD,SAAW,KAChBrD,KAAK0B,MAAQA,EACb1B,KAAK2B,QAAUA,CACnB,CACA2B,MAAMC,GAC0B,OAAxBvD,KAAKoD,iBAGTpD,KAAKoD,eAAiBI,IAClBxD,KAAKoD,eAAiB,KACD,IAAjBI,EAAMC,QACNzD,KAAKoB,KAAK,OACH,IAEPpB,KAAKoB,KAAKoC,IACVxD,KAAKsD,MAAMC,IAER,IAEXvD,KAAK0D,cACT,CACAnC,QAAQiC,GACJ,OAAO,IAAInB,SAAQ,CAACC,EAASC,KACzB,GAAsB,OAAlBvC,KAAKqD,SACL,MAAM,IAAInC,MAAM,sBAEpBlB,KAAKqD,SAAW,CACZG,MAAOA,EACPlB,QAASA,EACTC,OAAQA,GAEZvC,KAAK0D,aAAa,GAE1B,CACAA,cACI,MAAMN,eAAEA,EAAcC,SAAEA,GAAarD,KACd,OAAnBoD,GAAwC,OAAbC,IAG/BrD,KAAKoD,eAAiB,KACtBpD,KAAKqD,SAAW,KACZD,EAAeC,EAASG,OACxBH,EAASf,UAGTe,EAASd,OAAO,IAAIrB,MAAM,kBAElC,EAEJ,MAAMU,UAAa/B,EACfE,YAAY4D,EAAYlD,GACpB0C,QACAnD,KAAK2D,WAAaA,EAClB3D,KAAKS,SAAWA,EAChBT,KAAK2D,WAAWvB,SAAS,UAAW,CAAE3B,SAAUT,KAAKS,UAAY,MACjET,KAAK4D,KAAK,SAAU5D,KAAK6D,UAAUC,KAAK9D,MAC5C,CACA+D,OAAOP,EAAOQ,EAAUC,GACpBjE,KAAK2D,WAAWvB,SAAS,SAAU,CAAE3B,SAAUT,KAAKS,UAAY+C,GAC3Db,MAAKuB,GAAKD,MACVpB,OAAMC,GAASmB,EAASnB,IACjC,CACAe,YACI7D,KAAK2D,WAAWvB,SAAS,UAAW,CAAE3B,SAAUT,KAAKS,UAAY,KACrE"}
✄
import e from"events";import{Readable as t,Writable as s}from"stream";export class Controller{constructor(){this.events=new e,this.sources=new Map,this.nextEndpointId=1,this.requests=new Map,this.nextRequestId=1,this.onCreate=e=>{const t=e.endpoint,s=new n(t);this.sources.set(t.id,s),this.events.emit("stream",s)},this.onFinish=e=>{const t=e.endpoint.id,s=this.sources.get(t);if(void 0===s)throw new Error("Invalid endpoint ID");this.sources.delete(t),s.push(null)},this.onWrite=(e,t)=>{const s=e.endpoint.id,n=this.sources.get(s);if(void 0===n)throw new Error("Invalid endpoint ID");if(null===t)throw new Error("Invalid request: missing data");return n.deliver(t)},this.handlers={".create":this.onCreate,".finish":this.onFinish,".write":this.onWrite}}open(e,t={}){const s={id:this.nextEndpointId++,label:e,details:t};return new i(this,s)}receive(e){const t=e.stanza,{id:s,name:n,payload:i}=t,r=n[0];if("."===r)this.onRequest(s,n,i,e.data);else{if("+"!==r)throw new Error("Unknown stanza: "+n);this.onNotification(s,n,i)}}_request(e,t,s){return new Promise(((n,i)=>{const r=this.nextRequestId++;this.requests.set(r,{resolve:n,reject:i});const o={id:r,name:e,payload:t};this.events.emit("send",{stanza:o,data:s})}))}onRequest(e,t,s,n){const i=this.handlers[t];if(void 0===i)throw new Error(`Invalid request: ${t}`);let r;try{r=i(s,n)}catch(t){return void this.reject(e,t)}r instanceof Promise?r.then((t=>this.resolve(e,t))).catch((t=>this.reject(e,t))):this.resolve(e,r)}resolve(e,t){const s={id:e,name:"+result",payload:t};this.events.emit("send",{stanza:s,data:null})}reject(e,t){const s={id:e,name:"+error",payload:{message:t.toString()}};this.events.emit("send",{stanza:s,data:null})}onNotification(e,t,s){const n=this.requests.get(e);if(void 0===n)throw new Error("Invalid request ID");if(this.requests.delete(e),"+result"===t)n.resolve(s);else{if("+error"!==t)throw new Error("Unknown notification: "+t);{const e=s;n.reject(new Error(e.message))}}}}export default Controller;class n extends t{constructor({label:e,details:t}){super(),this.onReadComplete=null,this.delivery=null,this.label=e,this.details=t}_read(e){null===this.onReadComplete&&(this.onReadComplete=t=>(this.onReadComplete=null,0===t.length?(this.push(null),!1):(this.push(t)&&this._read(e),!0)),this.tryComplete())}deliver(e){return new Promise(((t,s)=>{if(null!==this.delivery)throw new Error("Protocol violation");this.delivery={chunk:e,resolve:t,reject:s},this.tryComplete()}))}tryComplete(){const{onReadComplete:e,delivery:t}=this;null!==e&&null!==t&&(this.onReadComplete=null,this.delivery=null,e(t.chunk)?t.resolve():t.reject(new Error("Stream closed")))}}class i extends s{constructor(e,t){super(),this.controller=e,this.endpoint=t,this.controller._request(".create",{endpoint:this.endpoint},null),this.once("finish",this._onFinish.bind(this))}_write(e,t,s){this.controller._request(".write",{endpoint:this.endpoint},e).then((e=>s())).catch((e=>s(e)))}_onFinish(){this.controller._request(".finish",{endpoint:this.endpoint},null)}}